#include <iostream>
#include <iomanip>
#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <cublas_v2.h>
#include <cuda_runtime.h>
#include <clibgrey.h>

#define gpuErrchk(ans) { gpuAssert((ans), __FILE__, __LINE__); }
inline void gpuAssert(cudaError_t code, char *file, int line, bool abort=true){
    if (code != cudaSuccess) {
        fprintf(stderr,"GPUassert: %s %s %d\n", cudaGetErrorString(code), file, line);
        if (abort) exit(code);
    }
}

template <class T>
class devicearray{
public:
    T *data;
    int n[10]; // dimensions
    int rank; // number of dimensions
    int size; // number of elements
    bool isAllocated; // whether or not allocated
    
    // constructor (this calls init but separate in case want to free or reallocate later)
    devicearray(int n1, int n2, int n3){isAllocated=0;init(n1,n2,n3);};
    devicearray(int n1, int n2){isAllocated=0;init(n1,n2);};
    devicearray(int n1){isAllocated=0;init(n1);};
//    devicearray(){}; // could define as a scalar, but this will likely lead to incorrect usage when declaration is accidentally left out

    // destructor 
    ~devicearray(){if (isAllocated) final();};

    // initialize
    void init(int n1, int n2, int n3, int n4, int n5){
        if (!isAllocated) {
            for (int i=0; i<10; i++) {
                n[i] = 1;
            }
            rank = 5;
            n[0] = n1;
            n[1] = n2;
            n[2] = n3;
            n[3] = n4;
            n[4] = n5;
            size = n1*n2*n3*n4*n5;
            gpuErrchk(cudaMalloc((void **) &data, sizeof(T) * size));
            isAllocated = 1;
        } else {
            std::cerr<<"Error: "<<__FILE__<<" line "<<__LINE__<<" array already initialized"<<std::endl;
            std::exit(EXIT_FAILURE);
        }
    }
    void init(int n1, int n2, int n3, int n4){init(n1,n2,n3,n4,1);rank = 4;}
    void init(int n1, int n2, int n3){init(n1,n2,n3,1,1);rank = 3;}
    void init(int n1, int n2){init(n1,n2,1,1,1);rank = 2;}
    void init(int n1){init(n1,1,1,1,1);rank = 1;}

    // finalize
    void final(){
        cudaFree(data);
        isAllocated = 0;
    }

    // overload brackets for each access to device memory
    T operator[](int i){
        T x;
        cudaMemcpy(&x,&data[i],sizeof(T) * 1,cudaMemcpyDeviceToHost);
        return x;
    };

};

template <class T>
__host__ void disp(std::string str, devicearray<T> &x){

     std::cout<<str<<":"<<std::endl;

     if (x.rank>2) 
         std::cout<<" only displaying values for (:,:,1)"<<std::endl;
 
     for (int i=0; i<x.n[0]; i++) {
         for (int j=0; j<x.n[1]; j++) {
             std::cout<<std::setw(15)<<std::setprecision(6)<<x[i + j*x.n[0]]; // brackets call cudaMemcopy
         }
         std::cout<<std::endl;
     }
 
}


void mtimes(double *c, double *a, double *b, int na1, int na2, int nb1, int nb2, char ta, char tb);
void mtimes(double *c, double *a, double *b, int na1, int na2, int nb1, int nb2);
void mtimes(float *c, float *a, float *b, int na1, int na2, int nb1, int nb2, char ta, char tb);
void mtimes(float *c, float *a, float *b, int na1, int na2, int nb1, int nb2);

void testbellman();
void testcubellman();

template <class T>
__global__ void global_zeros(T *x, int n){
    int i = threadIdx.x + blockDim.x * blockIdx.x;
    if (i<n)
        x[i] = 0;
}

template <class T>
void zeros(devicearray<T> &x){
    int nThreadsPerBlock = 1024;
    int nBlocks = 1+(x.size-1)/1024;
    global_zeros<<<nBlocks,nThreadsPerBlock>>>(x.data,x.size);
}

// inplace absolute value
template <class T>
__global__ void global_abs(T *x,int n){
    int i = threadIdx.x + blockDim.x * blockIdx.x;
    if (i<n)
        x[i] = abs(x[i]);
}

template <class T>
__host__ void abs(devicearray<T> &x){
    int nThreadsPerBlock = 1024;
    int nBlocks = 1+(x.size-1)/1024;
    global_abs<<<nBlocks,nThreadsPerBlock>>>(x.data,x.size);
}

// inplace max (in Fortran sense). Sets x = max(x,c) element by element.
// This is useful for maxval, enabling it to be solved by cublas.
__global__ void global_max(float *x, float c);
__host__ void max(devicearray<float> &x, float c);



// Use binary search to find the lowerbound on a bracketing interval for 
// Given sorted ascending array, find lowerbound on bracketing interval. Note that
// ilo will always be in 1,...,size(x)-1.
// 
// This is supposed to work much like the C++ standard library where the interface comes from
template <class T>
__host__ __device__ int bsearch(T xhat,T *x,int n){
    int ilo,ihi,mid;
        
    ilo = 0;
    ihi = n-1;
    mid = (ihi+ilo)/2; // mid will have ilo<=mid<ihi as long as ihi>ilo. If ihi=ilo, then mid=ilo, so the 
                       // while loop is not entered and ilo=1 is returned 
    
    while (mid>ilo){ // mid=ilo means ihi=ilo+1.

        // Here, we must have ihi>ilo and ilo<=mid<ihi

        // Bisect. 
        // If xhat<x(1), then xhat>=x(mid) will never occur and ilo will never be modified.
        // If xhat>=x(nx), then xhat>=x(mid) will always occur (note mid<ihi). Consequently, 
        //      ilo = mid occurs until mid==ilo, which only occures if ihi==ilo+1.
        if (xhat>=x[mid]) 
            ilo = mid;
        else 
            ihi = mid;

        mid = (ihi+ilo)/2;

    }

    return ilo;

}

template <class T>
__host__ __device__ void interp1qlocweight(int &ilo, T &mlo, T xi, T *x, int n){
    ilo = bsearch(xi,x,n); 
    mlo = (x[ilo+1]-xi)/(x[ilo+1]-x[ilo]);
    return;
}
template <class T>
__host__ __device__ T interp1q(T *x,T *y,T xi, int nx){
    int ilo;
    T mlo, one = 1;

    // Get interpolation location and weights
    interp1qlocweight(ilo,mlo,xi,x,nx);

    // Perform the interpolation
    return y[ilo]*mlo + y[ilo+1]*(one - mlo);

}


template <class T>
T rand01(){
    return (T)rand() / (T)RAND_MAX;
}

template <class T>
void rand01(T *x, int n){
    int i;
    for (i=0; i<n; i++)
        x[i] = rand01<T>();
}


template <class T>
__host__ __device__ void eye(T *x, int n){
    T zero = 0, one = 1;
    int i;
    for (i=0;i<n*n;i++)
        x[i] = zero;

    for (i=0;i<n;i++)
        x[i + i*n] = one;
}

template <class T>
__global__ void global_linspace(T *x, T a, T b, int n){
    if (blockIdx.x==0) 
        linspace(x,a,b,n);

}

template <class T>
void linspace(devicearray<T> &x, T a, T b){
    global_linspace<<<1,1>>>(x.data,a,b,x.size);
}

template <class T>
__global__ void global_tauchen(T *z, T *Fzz, T cover, T ar1, T cdnalstd, int n){
    if (blockIdx.x==0) 
        tauchen(z,Fzz,cover,ar1,cdnalstd,n);
}

template <class T>
void tauchen(devicearray<T> &z, devicearray<T> &Fzz, T cover, T ar1, T cdnalstd){
    global_tauchen<<<1,1>>>(z.data,Fzz.data,cover,ar1,cdnalstd,z.size);
}
