! A module that attempts to simplify the process of using mpi w/o a significant
! computational cost (which of course would defeat the point!)
!
! Most routines, if available, are listed as myXXX where XXX is an mpi command.
! E.g., mympi_allgatherv calls MPI_Allgatherv.
! There are some exceptions, such as mympi_stop which calls MPI_Abort after
! displaying an error message.
module mod_mympi
    implicit none

    include 'mpif.h'

    logical, private :: mympi_initialized = .false.
    integer, private :: nproc,iproc

    interface mympi_allgatherv
        module procedure mympi_allgatherv_vec, mympi_allgatherv_mat,  mympi_allgatherv_3d,  mympi_allgatherv_4d, mympi_allgatherv_5d, mympi_allgatherv_6d,&
                         mympi_allgatherv_r4mat,mympi_allgatherv_r43d,mympi_allgatherv_r44d,&
                         mympi_allgatherv_imat, mympi_allgatherv_i3d, mympi_allgatherv_i4d, &
                         mympi_allgatherv_l1d, mympi_allgatherv_l4d, mympi_allgatherv_l5d
    end interface
    interface mympi_bcast
        module procedure mympi_bcast_vec, mympi_bcast_mat, mympi_bcast_3d, mympi_bcast_4d, mympi_bcast_5d, mympi_bcast_6d, &
                         mympi_bcast_logical_vec, mympi_bcast_logical_mat, mympi_bcast_logical_3d, mympi_bcast_logical_4d, mympi_bcast_logical_5d, mympi_bcast_logical_6d
    end interface
    interface mympi_irecv
        module procedure mympi_irecv_1d
    end interface
    interface mympi_recv
        module procedure mympi_recv_1d
    end interface
    interface mympi_send
        module procedure mympi_send_1d
    end interface

    private :: mympi_bcast_core
    private :: mympi_allgatherv_core
    private :: mympi_allgatherv_core_logical
    private :: mympi_allgatherv_icore
    private :: checkinit
    
contains

function get_ierr_msg(ierr) result(msg)
    implicit none
    integer, intent(in) :: ierr
    character(len=50) :: msg

    select case (ierr)
        case (MPI_SUCCESS)
            msg='Successful return code'
        case (MPI_ERR_BUFFER)
            msg='Invalid buffer pointer'
        case (MPI_ERR_COUNT)
            msg='Invalid count argument'
        case (MPI_ERR_TYPE)
            msg='Invalid datatype argument'
        case (MPI_ERR_TAG)
            msg='Invalid tag argument'
        case (MPI_ERR_COMM)
            msg='Invalid communicator'
        case (MPI_ERR_RANK)
            msg='Invalid rank'
        case (MPI_ERR_ROOT)
            msg='Invalid root'
        case (MPI_ERR_GROUP)
            msg='Null group passed to function'
        case (MPI_ERR_OP)
            msg='Invalid operation'
        case (MPI_ERR_TOPOLOGY)
            msg='Invalid topology'
        case (MPI_ERR_DIMS)
            msg='Illegal dimension argument'
        case (MPI_ERR_ARG)
            msg='Invalid argument'
        case (MPI_ERR_UNKNOWN)
            msg='Unknown error'
        case (MPI_ERR_TRUNCATE)
            msg='message truncated on receive'
        case (MPI_ERR_OTHER)
            msg='Other error; use Error_string'
        case (MPI_ERR_INTERN)
            msg='internal error code'
        case (MPI_ERR_IN_STATUS)
            msg='Look in status for error value'
        case (MPI_ERR_PENDING)
            msg='Pending request'
        case (MPI_ERR_REQUEST)
            msg='illegal mpi_request handle'
        case (MPI_ERR_LASTCODE)
            msg='Last error code -- always at end'
    end select

end function

subroutine checkinit()
    implicit none
    if (.not. mympi_initialized) STOP 'ERROR: mpi has not been initialized'
end subroutine 

integer function mympi_iproc()
    implicit none
    call checkinit()
    mympi_iproc = iproc
end function 
integer function mympi_nproc()
    implicit none
    call checkinit()
    mympi_nproc = nproc
end function 

subroutine mympi_barrier()
    implicit none
    integer :: ierr

    call checkinit()
    call MPI_Barrier(MPI_COMM_WORLD,ierr)

end subroutine 

! Aborts all mpi routines and also displays a message
subroutine mympi_stop(msg)
    implicit none
    character(len=*), intent(in) :: msg
    ! local
    integer :: errorcode
    integer :: ierr

    call checkinit()

    ! Display the error message
    print*,msg

    ! Pass error code to environment
    errorcode = 1 
    
    ! Terminate all processes
    call MPI_Abort(MPI_COMM_WORLD,errorcode,ierr)


end subroutine 


! Divides n elts as evenly as possible into ranges [a,b]
! If b<a, then b is set to -1 and a to 0. This is convenient as b-a+1 
! then returns the number of elements for each node regardless of whether 
! the range is 0 or not. Also, since b<a, "do i = a,b" is not executed. 
! NOTE: if used to assign trailing dimensions of an array, 
! as it defines contiguous array section this is very useful
subroutine mympi_assign_range(a,b,n)
    implicit none
    integer, intent(in) :: n
    integer, intent(out) :: a,b
    
    call checkinit()
    call mympi_assign_range_core(a,b,iproc,n)

end subroutine 
! Helper functions
integer function mympi_a(n)
    implicit none
    integer, intent(in) :: n
    ! local
    integer :: b
    call checkinit()
    call mympi_assign_range(mympi_a,b,n)
end function mympi_a
integer function mympi_b(n)
    implicit none
    integer, intent(in) :: n
    ! local
    integer :: a
    call checkinit()
    call mympi_assign_range(a,mympi_b,n)
end function mympi_b
! Gives the combined list
subroutine mympi_assign_range_list(a,b,n)
    implicit none
    integer, intent(in) :: n
    integer, intent(out) :: a(nproc),b(nproc)
    ! local
    integer :: ip

    call checkinit()

    do ip = 0,nproc-1
        call mympi_assign_range_core(a(ip+1),b(ip+1),ip,n)
    end do
end subroutine 

! The core routine for dividing elements
! ip = 0,...,nproc-1
subroutine mympi_assign_range_core(a,b,ip,n)
    implicit none
    integer, intent(in) :: ip,n
    integer, intent(out) :: a,b
    ! local
    integer :: rem,jproc,times

    call checkinit()
    if (n<nproc) then
        ! If here, can assign one element to each node. 
        ! Hence, ip gets the ip+1 element if ip+1<=n
        if (ip<n) then
            a = ip+1
            b = ip+1
        else
            a = 0
            b = -1
        end if
    else
        ! If here, need to assign at least n/nproc to each element and the remainder
        ! across the first elements

        rem = mod(n,nproc)
        times = n/nproc ! >=1

        if (n/=times*nproc + rem) call mympi_stop('ERROR: mympi_assign_range: arithmetic not as expected')

        b = 0
        do jproc = 0,ip
            if (jproc<rem) then
                ! Gets times+1 elements
                a = 1 + b
                b = times + 1 + b
            else
                ! Gets times elements
                a = 1 + b
                b = times + b
            end if
        end do

    end if

end subroutine


subroutine mympi_bcast_logical_6d(A,root)
    implicit none
    logical, intent(inout) :: A(:,:,:,:,:,:)
    integer, intent(in) :: root
    call mympi_bcast_logical_core(A,size(A),root)
end subroutine
subroutine mympi_bcast_logical_5d(A,root)
    implicit none
    logical, intent(inout) :: A(:,:,:,:,:)
    integer, intent(in) :: root
    call mympi_bcast_logical_core(A,size(A),root)
end subroutine
subroutine mympi_bcast_logical_4d(A,root)
    implicit none
    logical, intent(inout) :: A(:,:,:,:)
    integer, intent(in) :: root
    call mympi_bcast_logical_core(A,size(A),root)
end subroutine
subroutine mympi_bcast_logical_3d(A,root)
    implicit none
    logical, intent(inout) :: A(:,:,:)
    integer, intent(in) :: root
    call mympi_bcast_logical_core(A,size(A),root)
end subroutine
subroutine mympi_bcast_logical_mat(A,root)
    implicit none
    logical, intent(inout) :: A(:,:)
    integer, intent(in) :: root
    call mympi_bcast_logical_core(A,size(A),root)
end subroutine
subroutine mympi_bcast_logical_vec(A,root)
    implicit none
    logical, intent(inout) :: A(:)
    integer, intent(in) :: root
    call mympi_bcast_logical_core(A,size(A),root)
end subroutine

subroutine mympi_bcast_logical_core(A,n,root)
    implicit none
    integer, intent(in) :: n, root
    logical, intent(inout) :: A(*)
    ! local
    integer :: ierr

    call checkinit()

    ! Call the mpi routine
    call MPI_Bcast(A(1),n,MPI_LOGICAL,root,MPI_COMM_WORLD,ierr)

end subroutine


subroutine mympi_bcast_6d(A,root)
    implicit none
    real(8), intent(inout) :: A(:,:,:,:,:,:)
    integer, intent(in) :: root
    call mympi_bcast_core(A,size(A),root)
end subroutine
subroutine mympi_bcast_5d(A,root)
    implicit none
    real(8), intent(inout) :: A(:,:,:,:,:)
    integer, intent(in) :: root
    call mympi_bcast_core(A,size(A),root)
end subroutine
subroutine mympi_bcast_4d(A,root)
    implicit none
    real(8), intent(inout) :: A(:,:,:,:)
    integer, intent(in) :: root
    call mympi_bcast_core(A,size(A),root)
end subroutine
subroutine mympi_bcast_3d(A,root)
    implicit none
    real(8), intent(inout) :: A(:,:,:)
    integer, intent(in) :: root
    call mympi_bcast_core(A,size(A),root)
end subroutine
subroutine mympi_bcast_mat(A,root)
    implicit none
    real(8), intent(inout) :: A(:,:)
    integer, intent(in) :: root
    call mympi_bcast_core(A,size(A),root)
end subroutine
subroutine mympi_bcast_vec(A,root)
    implicit none
    real(8), intent(inout) :: A(:)
    integer, intent(in) :: root
    call mympi_bcast_core(A,size(A),root)
end subroutine

subroutine mympi_bcast_core(A,n,root)
    implicit none
    integer, intent(in) :: n,root
    real(8), intent(inout) :: A(*)
    ! local
    integer :: ierr

    call checkinit()

    ! Call the mpi routine
    call MPI_Bcast(A(1),n,MPI_DOUBLE_PRECISION,root,MPI_COMM_WORLD,ierr)

end subroutine

function mympi_test(request) result(isComplete)
    implicit none
    integer, intent(in) :: request
    logical :: isComplete
    ! local
    logical :: flag
    integer :: ierr, status(MPI_STATUS_SIZE)

    call MPI_Test(request,flag,status,ierr)

    isComplete = flag

end function

subroutine mympi_irecv_1d(request,A,source,tag)
    implicit none
    integer, intent(out) :: request
    real(8), intent(out) :: A(:)
    integer, intent(in) :: source
    integer, intent(in), optional :: tag

    if (present(tag)) then
        call mympi_irecv_core(request,A,size(A),source,tag)
    else
        call mympi_irecv_core(request,A,size(A),source,MPI_ANY_TAG)
    end if

end subroutine

subroutine mympi_irecv_core(request,A,n,source,tag)
    implicit none
    integer, intent(out) :: request
    real(8), intent(out) :: A(*)
    integer, intent(in) :: n,source,tag
    ! local
    integer :: ierr

    call MPI_Irecv(A(1), n, MPI_DOUBLE_PRECISION, source, tag, &
                  MPI_COMM_WORLD, request, ierr)

    if (ierr/=MPI_SUCCESS) then
        print*,'ERROR: mympi_irecv: ',trim(get_ierr_msg(ierr))
        call MPI_Abort(MPI_COMM_WORLD,1,ierr)
    end if

end subroutine 

subroutine mympi_waitall(array_of_requests)
    implicit none
    integer, intent(in) :: array_of_requests(:)
    ! local
    integer, allocatable :: status(:,:)
    integer :: ierr

    allocate(status(MPI_STATUS_SIZE,size(array_of_requests)))

    call MPI_Waitall(size(array_of_requests),array_of_requests(1),status,ierr)

    if (ierr/=MPI_SUCCESS) then
        print*,'ERROR: mympi_waitall: ',trim(get_ierr_msg(ierr))
        call MPI_Abort(MPI_COMM_WORLD,1,ierr)
    end if
end subroutine


subroutine mympi_recv_1d(A,source,tag)
    implicit none
    real(8), intent(out) :: A(:)
    integer, intent(in) :: source
    integer, intent(in), optional :: tag

    if (present(tag)) then
        call mympi_recv_core(A,size(A),source,tag)
    else
        call mympi_recv_core(A,size(A),source,MPI_ANY_TAG)
    end if

end subroutine

subroutine mympi_recv_core(A,n,source,tag)
    implicit none
    real(8), intent(out) :: A(*)
    integer, intent(in) :: n,source,tag
    ! local
    integer :: ierr, status(MPI_STATUS_SIZE)

    call MPI_Recv(A(1), n, MPI_DOUBLE_PRECISION, source, tag, &
                  MPI_COMM_WORLD, status, ierr)

    if (ierr/=MPI_SUCCESS) then
        print*,'ERROR: mympi_recv: ',trim(get_ierr_msg(ierr))
        call MPI_Abort(MPI_COMM_WORLD,1,ierr)
    end if

end subroutine 

subroutine mympi_send_1d(A,dest,tag)
    implicit none
    real(8), intent(in) :: A(:)
    integer, intent(in) :: dest
    integer, intent(in), optional :: tag

    if (present(tag)) then
        call mympi_send_core(A,size(A),dest,tag)
    else
        call mympi_send_core(A,size(A),dest,0)
    end if

end subroutine

subroutine mympi_send_core(A,n,dest,tag)
    implicit none
    real(8), intent(in) :: A(*)
    integer, intent(in) :: n,dest,tag
    ! local
    integer :: ierr

    call MPI_send(A(1), n, MPI_DOUBLE_PRECISION, dest, tag, &
                   MPI_COMM_WORLD, ierr)

    if (ierr/=MPI_SUCCESS) then
        print*,'ERROR: mympi_send: ',trim(get_ierr_msg(ierr))
        call MPI_Abort(MPI_COMM_WORLD,1,ierr)
    end if

end subroutine 



subroutine mympi_allgatherv_l5d(A)
    use mod_matlab
    implicit none
    logical, intent(inout) :: A(:,:,:,:,:)
    call mympi_allgatherv_core_logical(A,size(A,1)*size(A,2)*size(A,3)*size(A,4),size(A,5))
end subroutine 
subroutine mympi_allgatherv_l4d(A)
    use mod_matlab
    implicit none
    logical, intent(inout) :: A(:,:,:,:)
    call mympi_allgatherv_core_logical(A,size(A,1)*size(A,2)*size(A,3),size(A,4))
end subroutine 
subroutine mympi_allgatherv_l1d(A)
    use mod_matlab
    implicit none
    logical, intent(inout) :: A(:)
    call mympi_allgatherv_core_logical(A,1,size(A,1))
end subroutine 

! This gathers data from all tasks and delivers the combined data to all tasks.
! NOTE: it assumes the work was distributed across the trailing dimension according
!       to mympi_assign_range_list.
subroutine mympi_allgatherv_6d(A)
    use mod_matlab
    implicit none
    real(8), intent(inout) :: A(:,:,:,:,:,:)
    call mympi_allgatherv_core(A,size(A,1)*size(A,2)*size(A,3)*size(A,4)*size(A,5),size(A,6))
end subroutine 
subroutine mympi_allgatherv_5d(A)
    use mod_matlab
    implicit none
    real(8), intent(inout) :: A(:,:,:,:,:)
    call mympi_allgatherv_core(A,size(A,1)*size(A,2)*size(A,3)*size(A,4),size(A,5))
end subroutine 
subroutine mympi_allgatherv_4d(A)
    use mod_matlab
    implicit none
    real(8), intent(inout) :: A(:,:,:,:)
    call mympi_allgatherv_core(A,size(A,1)*size(A,2)*size(A,3),size(A,4))
end subroutine 
subroutine mympi_allgatherv_3d(A)
    use mod_matlab
    implicit none
    real(8), intent(inout) :: A(:,:,:)
    call mympi_allgatherv_core(A,size(A,1)*size(A,2),size(A,3))
end subroutine 
subroutine mympi_allgatherv_mat(A)
    use mod_matlab
    implicit none
    real(8), intent(inout) :: A(:,:)
    call mympi_allgatherv_core(A,size(A,1),size(A,2))
end subroutine 
subroutine mympi_allgatherv_vec(A)
    use mod_matlab
    implicit none
    real(8), intent(inout) :: A(:)
    call mympi_allgatherv_core(A,1,size(A,1))
end subroutine 
subroutine mympi_allgatherv_core(A,nrow,ncol)
    use mod_matlab
    implicit none
    integer, intent(in) :: nrow,ncol
    real(8), intent(inout) :: A(nrow,*)
    ! local
    integer :: lo(nproc),hi(nproc),nelts(nproc),offset(nproc)
    integer :: ierr,lo_iproc

    call checkinit()


    ! Get how the list was distributed
    call mympi_assign_range_list(lo,hi,ncol)

    ! Number of elements (columns)
    nelts = hi - lo + 1

    ! Offset
    offset(1) = 0
    offset(2:nproc) = cumsum(nrow*nelts(1:nproc-1))

    ! Check whether number of elements to be sent is > 0 
    if (nelts(iproc+1)>0) then
        ! If so, then first element of A to be sent occurs at lo_iproc
        lo_iproc = lo(iproc+1)
    else
        ! Otherwise, need to specify one knowing that no values are sent 
        lo_iproc = 1
    end if

    ! Call the mpi routine
    call MPI_Allgatherv(A(1,lo_iproc),& ! first element assigned to iproc
                        nrow*nelts(iproc+1),& ! total number of elements assigned to iproc --- may be zero
                        MPI_DOUBLE_PRECISION,& ! data type of send elements
                        A(1,1),& ! first element of receive 
                        nrow*nelts,& ! number of elements to be received by each process
                        offset,& ! Entry iproc specifies the displacement (relative to recvbuf) at which to place data from iproc
                        MPI_DOUBLE_PRECISION,& ! data type of receive elements
                        MPI_COMM_WORLD,&
                        ierr) 

end subroutine

! Same as mympi_allgatherv_core but with logical
subroutine mympi_allgatherv_core_logical(A,nrow,ncol)
    use mod_matlab
    implicit none
    integer, intent(in) :: nrow,ncol
    logical, intent(inout) :: A(nrow,*)
    ! local
    integer :: lo(nproc),hi(nproc),nelts(nproc),offset(nproc)
    integer :: ierr,lo_iproc

    call checkinit()


    ! Get how the list was distributed
    call mympi_assign_range_list(lo,hi,ncol)

    ! Number of elements (columns)
    nelts = hi - lo + 1

    ! Offset
    offset(1) = 0
    offset(2:nproc) = cumsum(nrow*nelts(1:nproc-1))

    ! Check whether number of elements to be sent is > 0 
    if (nelts(iproc+1)>0) then
        ! If so, then first element of A to be sent occurs at lo_iproc
        lo_iproc = lo(iproc+1)
    else
        ! Otherwise, need to specify one knowing that no values are sent 
        lo_iproc = 1
    end if

    ! Call the mpi routine
    call MPI_Allgatherv(A(1,lo_iproc),& ! first element assigned to iproc
                        nrow*nelts(iproc+1),& ! total number of elements assigned to iproc --- may be zero
                        MPI_LOGICAL,& ! data type of send elements
                        A(1,1),& ! first element of receive 
                        nrow*nelts,& ! number of elements to be received by each process
                        offset,& ! Entry iproc specifies the displacement (relative to recvbuf) at which to place data from iproc
                        MPI_LOGICAL,& ! data type of receive elements
                        MPI_COMM_WORLD,&
                        ierr) 

end subroutine

! Single precision version of the above
subroutine mympi_allgatherv_r44d(A)
    use mod_matlab
    implicit none
    real(4), intent(inout) :: A(:,:,:,:)
    call mympi_allgatherv_r4core(A,size(A,1)*size(A,2)*size(A,3),size(A,4))
end subroutine 
subroutine mympi_allgatherv_r43d(A)
    use mod_matlab
    implicit none
    real(4), intent(inout) :: A(:,:,:)
    call mympi_allgatherv_r4core(A,size(A,1)*size(A,2),size(A,3))
end subroutine 
subroutine mympi_allgatherv_r4mat(A)
    use mod_matlab
    implicit none
    real(4), intent(inout) :: A(:,:)
    call mympi_allgatherv_r4core(A,size(A,1),size(A,2))
end subroutine 
subroutine mympi_allgatherv_r4core(A,nrow,ncol)
    use mod_matlab
    implicit none
    integer, intent(in) :: nrow,ncol
    real(4), intent(inout) :: A(nrow,*)
    ! local
    integer :: lo(nproc),hi(nproc),nelts(nproc),offset(nproc)
    integer :: ierr,lo_iproc

    call checkinit()


    ! Get how the list was distributed
    call mympi_assign_range_list(lo,hi,ncol)

    ! Number of elements (columns)
    nelts = hi - lo + 1

    ! Offset
    offset(1) = 0
    offset(2:nproc) = cumsum(nrow*nelts(1:nproc-1))

    ! Check whether number of elements to be sent is > 0 
    if (nelts(iproc+1)>0) then
        ! If so, then first element of A to be sent occurs at lo_iproc
        lo_iproc = lo(iproc+1)
    else
        ! Otherwise, need to specify one knowing that no values are sent 
        lo_iproc = 1
    end if

    ! Call the mpi routine
    call MPI_Allgatherv(A(1,lo_iproc),& ! first element assigned to iproc
                        nrow*nelts(iproc+1),& ! total number of elements assigned to iproc --- may be zero
                        MPI_REAL4,& ! data type of send elements
                        A(1,1),& ! first element of receive 
                        nrow*nelts,& ! number of elements to be received by each process
                        offset,& ! Entry iproc specifies the displacement (relative to recvbuf) at which to place data from iproc
                        MPI_REAL4,& ! data type of receive elements
                        MPI_COMM_WORLD,&
                        ierr) 

end subroutine


! Integer versions of the above
subroutine mympi_allgatherv_i4d(A)
    use mod_matlab
    implicit none
    integer, intent(inout) :: A(:,:,:,:)
    call mympi_allgatherv_icore(A,size(A,1)*size(A,2)*size(A,3),size(A,4))
end subroutine 
subroutine mympi_allgatherv_i3d(A)
    use mod_matlab
    implicit none
    integer, intent(inout) :: A(:,:,:)
    call mympi_allgatherv_icore(A,size(A,1)*size(A,2),size(A,3))
end subroutine 
subroutine mympi_allgatherv_imat(A)
    use mod_matlab
    implicit none
    integer, intent(inout) :: A(:,:)
    call mympi_allgatherv_icore(A,size(A,1),size(A,2))
end subroutine 
subroutine mympi_allgatherv_icore(A,nrow,ncol)
    use mod_matlab
    implicit none
    integer, intent(in) :: nrow,ncol
    integer, intent(inout) :: A(nrow,*)
    ! local
    integer :: lo(nproc),hi(nproc),nelts(nproc),offset(nproc)
    integer :: ierr,lo_iproc

    call checkinit()


    ! Get how the list was distributed
    call mympi_assign_range_list(lo,hi,ncol)

    ! Number of elements (columns)
    nelts = hi - lo + 1

    ! Offset
    offset(1) = 0
    offset(2:nproc) = cumsum(nrow*nelts(1:nproc-1))

    ! Check whether number of elements to be sent is > 0 
    if (nelts(iproc+1)>0) then
        ! If so, then first element of A to be sent occurs at lo_iproc
        lo_iproc = lo(iproc+1)
    else
        ! Otherwise, need to specify one knowing that no values are sent 
        lo_iproc = 1
    end if

    ! Call the mpi routine
    call MPI_Allgatherv(A(1,lo_iproc),& ! first element assigned to iproc
                        nrow*nelts(iproc+1),& ! total number of elements assigned to iproc --- may be zero
                        MPI_INTEGER,& ! data type of send elements
                        A(1,1),& ! first element of receive 
                        nrow*nelts,& ! number of elements to be received by each process
                        offset,& ! Entry iproc specifies the displacement (relative to recvbuf) at which to place data from iproc
                        MPI_INTEGER,& ! data type of receive elements
                        MPI_COMM_WORLD,&
                        ierr) 

end subroutine

subroutine mympi_init()
    implicit none
    integer :: ierr

    ! Initialize mpi
    mympi_initialized = .true.
    call mpi_init(ierr)

    ! Get the number of processors
    call MPI_Comm_size(MPI_COMM_WORLD,nproc,ierr)

    ! Get the processor id
    call MPI_Comm_rank(MPI_COMM_WORLD,iproc,ierr)

end subroutine
subroutine mympi_finalize()
    implicit none
    integer :: ierr

    call checkinit()
    call MPI_Finalize(ierr)

end subroutine 

subroutine mympi_hello_world()
    implicit none
    character(len=MPI_MAX_PROCESSOR_NAME) :: procname
    integer :: lenprocname,ierr

    call checkinit()

    ! Get the processor name
    call MPI_Get_processor_name(procname,lenprocname,ierr)

    ! Display
    write(*,'(3A,i0,A,i0,A)') 'Hello from ',trim(procname),'. I am process ',iproc,' of ',nproc,'.'

end subroutine 
    
    

end module mod_mympi

module mod_mympi_test
    use mod_matlab
    use mod_mympi
    implicit none
contains
    
    ! mympi_init and mympi_finalize should be called before and after this routine
    subroutine test_mympi
        implicit none
        integer, parameter :: n = 9
        integer :: a,b,i
        real(8) :: x(17,n)
        integer :: avec(100),bvec(100)
        
        call mympi_hello_world()
        call mympi_assign_range(a,b,n)
        if (mympi_a(n)/=a) then
            call mympi_stop('ERROR: in mympi_a')
        elseif (mympi_b(n)/=b) then
            call mympi_stop('ERROR: in mympi_b')
        end if

        call mympi_barrier() ! prettier i/o
        print*,'Id ',mympi_iproc(),'has range',a,b,'#elts',b-a+1

        call mympi_barrier() ! prettier i/o


        do i = a,b
            x(:,i) = sqrt(linspace(1d0,real(size(x,1),8),size(x,1))) + mympi_iproc()
        end do

        call mympi_allgatherv_mat(x)

        avec = -1
        bvec = -1
        call mympi_assign_range_list(avec(1:mympi_nproc()),bvec(1:mympi_nproc()),n)
        
        if (mympi_iproc()==0) then
            print*,'avec',avec(1:mympi_nproc())
            print*,'bvec',bvec(1:mympi_nproc())
            print*,'list:'
            do i = 1,mympi_nproc()
                write(*,'(3(i0,A))') i,' ',avec(i),' ',bvec(i),' '
            end do

            call disp('x',x)
        end if

    end subroutine

end module mod_mympi_test
