! A collection of core routines used by many of the libgrey routines. These are designed to be basic.
module mod_core
    implicit none

    interface toc
        module procedure toc_print, toc_pr_return
    end interface toc

    real(4), private :: time_start
    integer(4), private :: count_start,countrate_start

contains

! Computes the euclidean norm of X
function norm(X) result(l2)
    implicit none
    real(8), dimension(1:), intent(in) :: X
    real(8) :: l2
    
    ! Call the minpack version of enorm
    l2 = enorm(size(X),X)

contains

DOUBLE PRECISION FUNCTION ENORM(N,X)
IMPLICIT NONE
INTEGER N
DOUBLE PRECISION X(N)
!     **********
!
!     function enorm
!
!     given an n-vector x, this function calculates the
!     euclidean norm of x.
!
!     the euclidean norm is computed by accumulating the sum of
!     squares in three different sums. the sums of squares for the
!     small and large components are scaled so that no overflows
!     occur. non-destructive underflows are permitted. underflows
!     and overflows do not occur in the computation of the unscaled
!     sum of squares for the intermediate components.
!     the definitions of small, intermediate and large components
!     depend on two constants, rdwarf and rgiant. the main
!     restrictions on these constants are that rdwarf**2 not
!     underflow and rgiant**2 not overflow. the constants
!     given here are suitable for every known computer.
!
!     the function statement is
!
!       double precision function enorm(n,x)
!
!     where
!
!       n is a positive integer input variable.
!
!       x is an input array of length n.
!
!     subprograms called
!
!       fortran-supplied ... dabs,dsqrt
!
!     argonne national laboratory. minpack project. march 1980.
!     burton s. garbow, kenneth e. hillstrom, jorge j. more
!
!     **********
INTEGER i
DOUBLE PRECISION agiant , floatn , one , rdwarf , rgiant , s1 , s2 , s3 , xabs , x1max , x3max , zero
DATA one , zero , rdwarf , rgiant/1.0D0 , 0.0D0 , 3.834D-20 ,  1.304D19/
s1 = zero
s2 = zero
s3 = zero
x1max = zero
x3max = zero
floatn = N
agiant = rgiant/floatn
DO i = 1 , N
    xabs = DABS(X(i))
    IF ( xabs.GT.rdwarf .AND. xabs.LT.agiant ) THEN
    !
    !           sum for intermediate components.
    !
        s2 = s2 + xabs**2
    ELSEIF ( xabs.LE.rdwarf ) THEN
    !
    !              sum for small components.
    !
        IF ( xabs.LE.x3max ) THEN
            IF ( xabs.NE.zero ) s3 = s3 + (xabs/x3max)**2
        ELSE
            s3 = one + s3*(x3max/xabs)**2
            x3max = xabs
        ENDIF
    !
    !              sum for large components.
    !
    ELSEIF ( xabs.LE.x1max ) THEN
        s1 = s1 + (xabs/x1max)**2
    ELSE
        s1 = one + s1*(x1max/xabs)**2
        x1max = xabs
    ENDIF
ENDDO
!
!     calculation of norm.
!
IF ( s1.NE.zero ) THEN
    ENORM = x1max*DSQRT(s1+(s2/x1max)/x1max)
ELSEIF ( s2.EQ.zero ) THEN
    ENORM = x3max*DSQRT(s3)
ELSE
    IF ( s2.GE.x3max ) ENORM = DSQRT(s2*(one+(x3max/s2)*(x3max*s3)))
    IF ( s2.LT.x3max ) ENORM = DSQRT(x3max*((s2/x3max)+(x3max*s3)))
ENDIF
!
!     last card of function enorm.
!
END function


end function 

! linspace(a,b,n) (of course) constructs a grid of n linearly spaced points between
! a and b. If n==1, then the grid is just "a", consistent with Matlab. If n<=0, the routine
! stops with an error.
function linspace(x1,x2,n) result(grid)
implicit none
real(8), intent(in) :: x1,x2
integer, intent(in) :: n
real(8), dimension(1:n) :: grid
!local
integer :: i 

if (n>1) then
grid(1) = x1
do i=2,n-1
grid(i) = x1 + (x2-x1)*(real(i,8)-1.d0)/(real(n,8)-1.d0)
end do   
grid(n) = x2
elseif (n==1) then
grid(1) = x1
else
STOP 'linspace: ERROR: grid has size <= 0'
end if
end function linspace

! Tic and toc for timing routines.
! Because of the odd behavior of the system clock, there is a known bug here that is 
! more likely to be experienced for long run times.
subroutine tic()
implicit none
call cpu_time(time_start)
call system_clock(count=count_start,count_rate=countrate_start)
end subroutine tic
subroutine toc_print()
implicit none
real(4) :: time_end
!integer :: date_and_time_value_end(8)
integer :: count_end, countrate_end
call cpu_time(time_end)
call system_clock(count=count_end,count_rate=countrate_end)
write(*,*) 'Elapsed CPU time is ', (time_end-time_start)/60., 'minutes'
write(*,*) 'Elapsed Wall time is ',(real(count_end,4)/real(countrate_end,4)- &
                            real(count_start,4)/real(countrate_start,4))/60., 'minutes'
end subroutine toc_print
subroutine toc_pr_return(elapsed)    
implicit none
real(4) :: time_end
real(8) :: elapsed
call cpu_time(time_end)
! write(*,*) 'Elapsed CPU time is ', (time_end-time_start)/60., 'minutes'
elapsed = (time_end-time_start)/60.
end subroutine toc_pr_return




end module mod_core
