module mod_deriv
    implicit none
contains
    ! Computes the jacobian of a function f:R^n->R^n via forward differencing
    ! This is done in a robust way using (many more function evaluations)
    ! by computing more 3 jacobians at various step sizes and then taking the
    ! average of them. 
    subroutine sub_jacobian_robust(x,f,fx,jac,eps_in)
        implicit none
        real(8), dimension(1:), intent(in) :: x
        real(8), dimension(1:), intent(in) :: fx ! f evaluated at x
        interface    
            function f(x) result(y)
                implicit none
                real(8), dimension(1:), intent(in) :: x
                real(8), dimension(1:size(x)) :: y
            end function f
        end interface
        real(8), dimension(size(x),size(x)), intent(out) :: jac
        real(8), optional, intent(in) :: eps_in
        ! local
        real(8), allocatable :: jactmp(:,:)
        real(8) :: eps

        eps = 1d-6; if (present(eps_in)) eps = eps_in

        allocate(jactmp(size(x),size(x)))

        call sub_jacobian(x,f,fx,jactmp,eps)
        jac = jactmp/2d0
        call sub_jacobian(x,f,fx,jactmp,eps*100d0)
        jac = jac + jactmp/2d0

        deallocate(jactmp)

    end subroutine


    ! Computes the jacobian of a function f:R^n->R^n via forward differencing
    ! The logic behind the tolerances and flow is from MINPACK's fdjac1
    ! "eps_in" should be roughly the amount of error in the calculation of f
    !
    ! The jacobian is of the form
    ! [df_1/dx_1 df_1/dx_2 ... df_1/dx_n]
    ! [df_2/dx_1 df_2/dx_2 ... df_2/dx_n]
    !  ...
    ! [df_n/dx_1 df_n/dx_2 ... df_n/dx_n]
    subroutine sub_jacobian(x,f,fx,jac,eps_in)
        implicit none
        real(8), dimension(1:), intent(in) :: x
        real(8), dimension(1:), intent(in) :: fx ! f evaluated at x
        interface    
            function f(x) result(y)
                implicit none
                real(8), dimension(1:), intent(in) :: x
                real(8), dimension(1:size(x)) :: y
            end function f
        end interface
        real(8), dimension(size(x),size(x)), intent(out) :: jac
        real(8), optional, intent(in) :: eps_in
        ! local
        real(8), allocatable :: wa(:), xlocal(:)
        real(8) :: temp, h, eps
        integer :: n, i, j

        n = size(x)

        allocate(wa(n),xlocal(n))

        xlocal = x

        eps = 1d-6; if (present(eps_in)) eps = eps_in

        ! Now this is basically fdjac1 code
        eps = sqrt(max(eps,epsilon(0d0)))
        do j = 1,n
            temp = xlocal(j)
            h = eps*abs(temp)
            if (h==0d0) h = eps
            xlocal(j) = temp + h
            wa = f(xlocal)
            xlocal(j) = temp
            do i = 1,n
                jac(i,j) = (wa(i) - fx(i))/h
            end do
        end do

        deallocate(wa,xlocal)

    end subroutine


    ! sub_numGradHess
    ! 
    ! Compute the gradient and optionally the hessian of a function through finite differencing.
    ! There is a default step size used to try to obtain a good approximation. It is fairly small.
    ! However, for non or *almost* differentiable problems, this step size can be scaled up to make
    ! the steps larger. A "scaling" of 1 is the default, but the step size increases linearly in 
    ! scaling.
    !
    subroutine sub_numGradHess(x0,f,scaling,grad,hess,f0)
        implicit none
        real(8), intent(in) :: x0(1:), scaling(1:)
        interface 
            function f(x) 
                implicit none
                real(8), dimension(1:), intent(in) :: x
                real(8) :: f
            end function f
        end interface
        real(8), intent(out) :: grad(size(x0))
        real(8), intent(out), optional :: f0, hess(size(x0),size(x0))
        ! local        
        real(8), parameter :: eps12 = epsilon(1d0)**.5d0
        real(8), parameter :: eps13 = epsilon(1d0)**.33333333333333d0
        real(8), parameter :: eps14 = epsilon(1d0)**.25d0
        real(8) :: f0val
        real(8), dimension(size(x0)) :: h,ei,ej, fxph
        integer :: i,j,n

        ! Get dim
        n = size(x0)

        h = scaling * max(abs(x0),1d0) * eps12 ! * 2d0*sqrt(epsilon(0d0))/10d0! eps14

        f0val = f(x0)
        if (present(f0)) then
            f0 = f0val
        end if

        ! Compute the gradient
        do i = 1,n
            ei = 0d0
            ei(i) = h(i)
            fxph(i) = f(x0+ei)
        end do
        grad = (fxph-f0val)/h

        ! Compute the hessian

        ! Choose the new step size for the second derivative calculation. I have not done the error
        ! analysis to find the optimal, but this is the analagous one from the equal step size
        h = scaling * max(abs(x0),1d0) * eps14

        if (present(hess)) then

            ! According to my calculations based on Taylor expansions,
            ! the formula for a cross partial d2f/didj using the 5 points f(x), f(x+-ei+-ej)
            ! puts 0 weight on f(x), 1/(4/(hi*hj)) where hi and hj are the step sizes. 
            do i = 1,n
                ei = 0d0
                ei(i) = h(i)
                ! Handle the "special" i==j case (b/c can use f0val) separately
                hess(i,i) = (  f(x0+2d0*ei) + f(x0-2d0*ei) - 2d0*f0val  )/(4d0*h(i)**2)
                do j = i+1,n
                    ej = 0d0
                    ej(j) = h(j)
                    hess(i,j) = ( f(x0+ei+ej) + f(x0-ei-ej) - f(x0+ei-ej) - f(x0-ei+ej) ) / (4d0*h(i)*h(j)) 
                end do
            end do

            ! Fill out the symmetric hessian
            do i = 1,n
                do j = i+1,n
                    hess(j,i) = hess(i,j)
                end do
            end do
            
        end if

    end subroutine sub_numGradHess

! 
!     subroutine sub_numGradHess(x0,f,scaling,grad,hess,f0)
!         implicit none
!         real(8), intent(in) :: x0(1:), scaling
!         interface 
!             function f(x) 
!                 implicit none
!                 real(8), dimension(1:), intent(in) :: x
!                 real(8) :: f
!             end function f
!         end interface
!         real(8), intent(out) :: grad(size(x0))
!         real(8), intent(out), optional :: f0, hess(size(x0),size(x0))
!         ! local        
!         real(8), parameter :: eps13 = epsilon(1d0)**.33333333333333d0
!         real(8), parameter :: eps14 = epsilon(1d0)**.25d0
!         real(8) :: f0val
!         real(8), dimension(size(x0)) :: h,ei,ej
!         integer :: i,j,n
! 
!         ! Get dim
!         n = size(x0)
! 
!         h = scaling * max(abs(x0),1d0) * eps13
! 
!         ! Compute the gradient
!         do i = 1,n
!             ei = 0d0
!             ei(i) = h(i)
!             grad(i) = (f(x0+ei)-f(x0-ei))/(2d0*h(i))
!         end do
! 
!         ! Optionally compute the value at f0 and the hessian
!         if (present(f0)) then
!             f0 = f(x0)
!             f0val = f0
!         elseif (present(hess)) then
!             f0val = f(x0)
!         end if
! 
!         ! Choose the new step size for the second derivative calculation. I have not done the error
!         ! analysis to find the optimal, but this is the analagous one from the equal step size
!         h = scaling * max(abs(x0),1d0) * eps14
! 
!         if (present(hess)) then
! 
!             ! According to my calculations based on Taylor expansions,
!             ! the formula for a cross partial d2f/didj using the 5 points f(x), f(x+-ei+-ej)
!             ! puts 0 weight on f(x), 1/(4/(hi*hj)) where hi and hj are the step sizes. 
!             do i = 1,n
!                 ei = 0d0
!                 ei(i) = h(i)
!                 ! Handle the "special" i==j case (b/c can use f0val) separately
!                 hess(i,i) = (  f(x0+2d0*ei) + f(x0-2d0*ei) - 2d0*f0val  )/(4d0*h(i)**2)
!                 do j = i+1,n
!                     ej = 0d0
!                     ej(j) = h(j)
!                     hess(i,j) = ( f(x0+ei+ej) + f(x0-ei-ej) - f(x0+ei-ej) - f(x0-ei+ej) ) / (4d0*h(i)*h(j)) 
!                 end do
!             end do
! 
!             ! Fill out the symmetric hessian
!             do i = 1,n
!                 do j = i+1,n
!                     hess(j,i) = hess(i,j)
!                 end do
!             end do
!             
!         end if
! 
!     end subroutine sub_numGradHess


end module mod_deriv
