! This module contains several ways to discretize an AR1 process
! -- sub_rouwenhorst
! -- sub_tauchen
! Additional, it provides a routine that will sample from an AR1 process.
! -- sub_drawAR1
module mod_discreteAR1
    implicit none
contains

! Draw from the markov chain given current state i
! There are two options in this routine
! option == 0, slow but easy for user, F is the mdf
! option == 1, fast but involved for user, F is the cumulative summation transpose of the markov chain
function drawIid(F,option) result(izp)
    use mod_matlab, only: cumsum,find
    implicit none
    real(8), dimension(1:), intent(in) :: F 
    integer, intent(in) :: option
    integer :: izp
    ! local
    real(8) :: randval
    real(8), allocatable :: Ftmp(:)
    
    call random_number(randval)

    if (option==0) then
        ! Slow: temporary created
        allocate(Ftmp(size(F)))
        Ftmp = cumsum(F)
        do izp = 1,size(F)
            if (randval<=Ftmp(izp)) return
        end do


    else
        ! Fast
        do izp = 1,size(F)
            if (randval<=F(izp)) return
        end do

    end if
    
    
end function drawIid

! drawAR1core(F,iz,randval) result(izp)
! Same as drawAR1(F,iz,option=1) but takes randval as input
function drawAR1core(F,iz,randval) result(izp_out)
    use mod_matlab, only: cumsum,bsearch
    implicit none
    real(8), dimension(1:,1:), intent(in) :: F ! Either the transition matrix or the cumulatively summed transpose 
    integer, intent(in) :: iz !, option
    integer :: izp_out
    real(8), intent(in) :: randval 
    ! local
    integer :: izp
    
    ! Find the izp such that randval \in (F(izp-1,iZ),F(izp,iZ)]
    ! If randval<F(1), return 1
    ! If randval>F(n), then issue an error
    ! If randval\in (F(1),F(2)], return 2 
    ! If randval\in (F(i),F(i+1)], return i+1
    ! Note: bsearch finds the i such that randval \in (F(i),F(i+1)), so need to add 1
    if (randval<=F(1,iZ)) then
        izp = 1
    elseif (randval>=F(size(F,1),iZ)) then
        izp = size(F,1)
        print*,'WARNING: drawAR1: F must not sum to 1, simulation will be wrong'
    else
        call bsearch(izp,randval,F(:,iZ))
        izp = izp + 1
    end if

    izp_out = izp

end function
! drawAR1(F,iz,option) result(izp)
! 
! Given iz, will generate izp according to a discrete markov chain.
! If option==0, the Markov chain is taken to be F.
! If option==1, the Markov chain is given implicitly and F is treated as the cumulatively
!               summed transpose of it (with the convention that the markov matrix has 
!               pi(i,j) with i being today and j being tomorrow. 
! 
function drawAR1(F,iz,option) result(izp_out)
    use mod_matlab, only: cumsum,bsearch
    implicit none
    real(8), dimension(1:,1:), intent(in) :: F ! Either the transition matrix or the cumulatively summed transpose 
    integer, intent(in) :: iz, option
    integer :: izp_out
    ! local
!     integer :: izp
    real(8) :: randval
!     real(8), allocatable :: Ftmp(:)
    
    call random_number(randval)

    if (option==0) then
        izp_out = drawAR1core(cumsum(transpose(F)),iz,randval) 
    elseif (option==1) then
        izp_out = drawAR1core(F,iz,randval)
    end if
    
end function 

! Construct a Markov chain of specified length. Normally, F
! is assumed to be F_ij = F(z'=j|z=i). However, if opt is 
! present and equal to 1, then F is assumed to be the cumsum 
! of the transpose of the Markov matrix.
subroutine drawMarkov(iseq,F,i1,opt)
    use mod_matlab, only: cumsum,find
    implicit none
    integer, intent(inout) :: iseq(1:)
    real(8), dimension(1:,1:), intent(in) :: F ! The transition matrix
    integer, intent(in) :: i1
    integer, intent(in), optional :: opt
    ! local
    integer :: t,nseq
    real(8), allocatable :: Fcumtrans(:,:)
    logical :: isspecial

    nseq = size(iseq)

    if (present(opt)) then
        isspecial = opt==1
    else
        isspecial = .false.
    end if

    if (.not. isspecial) then
        allocate(Fcumtrans(size(F,1),size(F,2)))

        Fcumtrans = transpose(F)
        Fcumtrans = cumsum(Fcumtrans)

        iseq(1) = i1
        do t = 2,nseq
            iseq(t) = drawAR1(Fcumtrans,iseq(t-1),1)
        end do

        deallocate(Fcumtrans)
    else

        iseq(1) = i1
        do t = 2,nseq
            iseq(t) = drawAR1(F,iseq(t-1),1)
        end do

    end if

end subroutine drawMarkov

! Adda cooper algorithm. 
! Based on Martin Floden's 2005 code
! %[Z,PI] = addacooper(n,mu,rho,sigma)
! % Approximate n-state AR(1) process following Tauchen (1986) and Tauchen & Hussey (1991). 
! % See Adda & Cooper (2003) pp 57-.
! %
! % Z(t+1) = mu*(1-rho) + rho*Z(t) + eps(t+1)
! %
! % where  std(eps) = sigma
! %
! % Martin Floden, 2005
! 
subroutine sub_addacooper_grid(Z,mu,rho,sigma)
    use mod_matlab
    ! use mod_quad
    implicit none
    real(8), intent(out) :: Z(:)
    real(8), intent(in) :: mu,rho,sigma
    ! local
    integer :: i,n
    real(8) :: sigmaUNC
    real(8), allocatable :: E(:)

    n = size(Z)

    if (n==1) then
        Z = mu
        return
    elseif (n<=0) then
        STOP 'ERROR: sub_addacooper: n<=0'
    end if

    sigmaUNC = sigma/sqrt(1d0-rho**2)

    allocate(E(n+1))
    E = 0d0
    Z = 0d0

    E(1)   = mu - 8d0*sigmaUNC ! 10 unconditional standard deviations below mean
    E(n+1) = mu + 8d0*sigmaUNC
    do i = 2,n
        E(i) = sigmaUNC*normcdfinv(real(i-1,8)/real(n,8),0d0,1d0) + mu
    end do 

    do i = 1,n
        Z(i) = n*sigmaUNC*(stdnormpdf((E(i)-mu)/sigmaUNC) - stdnormpdf((E(i+1)-mu)/sigmaUNC)) + mu
    end do

end subroutine
subroutine sub_addacooper(Z,Fzz,mu,rho,sigma)
    use mod_matlab
    use mod_quad
    implicit none
    real(8), intent(out) :: Z(:),Fzz(:,:)
    real(8), intent(in) :: mu,rho,sigma
    ! local

    integer :: i,j,n
    real(8) :: sigmaUNC,E1,E2
    real(8), allocatable :: E(:),MFPI(:,:),xquad(:),wquad(:)
    integer, parameter :: nquad = 1000

    allocate(xquad(nquad),wquad(nquad))


    n = size(Z)
    if (size(Fzz,1)/=n) STOP 'ERROR: sub_addacooper size mismatch'
    if (size(Fzz,2)/=n) STOP 'ERROR: sub_addacooper size mismatch'


    if (n==1) then
        Z = mu
        Fzz = 1d0
        return
    elseif (n<=0) then
        STOP 'ERROR: sub_addacooper: n<=0'
    end if

    sigmaUNC = sigma/sqrt(1d0-rho**2);

    allocate(E(n+1))
    allocate(MFPI(n,n))
    E = 0d0
    Z = 0d0
    Fzz = 0d0
    MFPI = 0d0

    E(1)   = mu - 8d0*sigmaUNC ! 10 unconditional standard deviations below mean
    E(n+1) = mu + 8d0*sigmaUNC
    do i = 2,n
        E(i) = sigmaUNC*normcdfinv(real(i-1,8)/real(n,8),0d0,1d0) + mu
    end do 

    do i = 1,n
        Z(i) = n*sigmaUNC*(stdnormpdf((E(i)-mu)/sigmaUNC) - stdnormpdf((E(i+1)-mu)/sigmaUNC)) + mu
    end do

    do i = 1,n
        do j = 1,n
            E1 = E(j)
            E2 = E(j+1)
            ! PI(i,j) = quadl(th_fcn,E(i),E(i+1),1e-10)
            call chebyquad(xquad,wquad,E(i),E(i+1))
            Fzz(i,j) = dot_product(th_fcn(xquad),wquad)
!             if (any(isnan(xquad))) STOP 'ERROR'
!             if (any(isnan(wquad))) STOP 'ERROR'
! 
!             if (isnan(Fzz(i,j))) STOP 'ERROR'
!             print*,'Fzz(i,j)',Fzz(i,j)
            MFPI(i,j) = stdnormcdf((E(j+1)-mu*(1d0-rho)-rho*Z(i))/sigma) - stdnormcdf((E(j)-mu*(1d0-rho)-rho*Z(i))/sigma)
        end do
    end do

    do i = 1,n
        Fzz(i,:) = Fzz(i,:) / sum(Fzz(i,:))
        MFPI(i,:) = MFPI(i,:) / sum(MFPI(i,:))
    end do

contains

    function th_fcn(u) result(f)
        implicit none
        real(8), intent(in) :: u(:)
        real(8) :: f(size(u))

        f = real(n,8)/sqrt(2d0*pi*sigmaUNC**2) * (exp(-(u-mu)**2 / (2d0*sigmaUNC**2)) * &
            (stdnormcdf((E2-mu*(1d0-rho)-rho*u)/sigma) - stdnormcdf((E1-mu*(1d0-rho)-rho*u)/sigma)))

    end function

end subroutine
        


!Rouwenhorst method as presented in Kopecky and Suen RED 2010
! Based on Karen Kopecky's Code
subroutine sub_rouwenhorst(z,P,rho,sig,n)
    use mod_matlab
    implicit none
    real(8), intent(in) :: rho,sig
    integer, intent(in) :: n
    real(8), intent(out) :: z(n),P(n,n)
    !local
    real(8) :: mu_eps, q, nu
    real(8), allocatable :: P0(:,:),P1(:,:)
    integer :: i

    if (n<=0) then
        STOP 'sub_rouwenhorst: n must be > 0'
    elseif (n==1) then
        z = 0d0
        P = 1d0
        return
    end if
    
    mu_eps = 0.d0
    q = (rho+1.d0)/2.d0
    nu = ((real(n,8)-1.d0)/(1.d0-rho**2))**(.5d0) * sig
    
    allocate(P0(2,2))
    P0(1,:) = (/q, 1.d0-q/)
    P0(2,:) = (/1.d0-q, q/)
    
    do i = 2,n-1
        allocate(P1(i+1,i+1))
        P1 = 0.d0
        !        
        P1(1:i,1:i) = q*P0 + P1(1:i,1:i)
        !
        P1(1:i,2:i+1) = (1.d0-q)*P0 + P1(1:i,2:i+1)
        !
        P1(2:i+1,1:i) = (1.d0-q)*P0 + P1(2:i+1,1:i)
        !
        P1(2:i+1,2:i+1) = q*P0 + P1(2:i+1,2:i+1)
        !
        P1(2:i,:) = P1(2:i,:)/2.d0
        deallocate(P0)
        allocate(P0(i+1,i+1))
        P0 = P1
        deallocate(P1)
    end do
    
    P = P0
    deallocate(P0)
    
    z = linspace(mu_eps/(1.d0-rho) - nu, mu_eps/(1.d0-rho) + nu,n)
    

end subroutine sub_rouwenhorst

! subtauchen
! Generates Markov chain approximating an AR(1) function of the form
! y=ar1*y_{-1}+e,  e~N(0,sd_eps**2)
! ar1  is AR(1) coeff
! sd_eps is SD of normally distributed shock
! cover gives cover*SD coverage, e.g. 3 = 99% coverage, 2 = 95%
!
! Parts of this code may be motivated by Floden
subroutine sub_tauchen(y,transition,ar1,sd_eps,n,cover)
    use mod_matlab, only: linspace
    implicit none
    !Input/Output
    real(8), intent(in)                      :: ar1        !ar(1) coeff
    real(8), intent(in)                      :: sd_eps     !SD of normally distributed shocks
    real(8), intent(in)                      :: cover      !number of SDs to cover (3=>99%)  
    integer,  intent(in)                      :: n          !no of tauchen points
    real(8), dimension(1:n), intent(out)     :: y          !shock values
    real(8), dimension(1:n,1:n), intent(out) :: transition !transition matrix
    ! local
    real(8) :: sd_y            !Unconditional stdev of y

    if (n==1) then
        y = 0d0
        transition = 1d0
        return
    end if

    !Define discrete states of markov chain
    sd_y = sd_eps/sqrt(1-ar1**2)
    y = linspace(-cover*sd_y,cover*sd_y,n)

    ! Call tauchen with predetermined grid
    call sub_tauchen_fixedgrid(transition,y,ar1,sd_eps)

end subroutine sub_tauchen

! Iid shock version of tauchen, just call sub_tauchen and takes a row of the trans
subroutine sub_tauchen_iid(y,prob,sd_eps,n,cover)
    implicit none
    !Input/Output
    real(8), intent(in)                  :: sd_eps     !SD of normally distributed shocks
    real(8), intent(in)                  :: cover      !number of SDs to cover (3=>99%)  
    integer,  intent(in)                 :: n          !no of tauchen points
    real(8), dimension(1:n), intent(out) :: y          !shock values
    real(8), dimension(1:n), intent(out) :: prob !transition matrix
    ! local
    real(8), dimension(:,:), allocatable :: transition

    allocate(transition(n,n))
    call sub_tauchen(y,transition,0d0,sd_eps,n,cover)
    prob = transition(1,:)
    deallocate(transition)
    
end subroutine

subroutine sub_tauchen_iid_fixedgrid(prob,y,sd_eps)
    implicit none
    real(8), intent(in) :: sd_eps
    real(8), dimension(:), intent(in) :: y
    real(8), dimension(size(y)), intent(out) :: prob !transition matrix
    ! local 
    real(8), dimension(:,:), allocatable :: transition
    integer :: n

    n = size(y)

    allocate(transition(n,n))
    call sub_tauchen_fixedgrid(transition,y,0d0,sd_eps) ! no autocorrelation
    prob = transition(1,:)
    deallocate(transition)

end subroutine 

subroutine sub_tauchen_iid_fixedgrid_with_mean(prob,y,sd_eps,mean)
    implicit none
    real(8), intent(in) :: sd_eps,mean
    real(8), dimension(:), intent(in) :: y
    real(8), dimension(size(y)), intent(out) :: prob !transition matrix
    ! local 
    real(8), dimension(:,:), allocatable :: transition
    integer :: n

    n = size(y)

    allocate(transition(n,n))
    call sub_tauchen_fixedgrid_with_mean(transition,y,0d0,sd_eps,mean) ! no autocorrelation
    prob = transition(1,:)
    deallocate(transition)

end subroutine 

subroutine sub_tauchen_fixedgrid(transition,y,ar1,sd_eps)
    implicit none
    !Input/Output
    real(8), intent(in)                      :: ar1        !ar(1) coeff
    real(8), intent(in)                      :: sd_eps     !SD of normally distributed shocks
    real(8), dimension(1:), intent(in)      :: y          !shock values
    real(8), dimension(1:size(y),1:size(y)), intent(out) :: transition !transition matrix

    call sub_tauchen_fixedgrid_with_mean(transition,y,ar1,sd_eps,0d0)

end subroutine sub_tauchen_fixedgrid

! Approximates 
! y = mean + ar1*y_{-1} + sd_eps*eps, eps~N(0,1)
! using tauchen on the fixed grid y.  
! NOTE: mean is not the unconditional mean of y, but the conditional mean of y|(y_{-1}=0)
! 
subroutine sub_tauchen_fixedgrid_with_mean(transition,y,ar1,sd_eps,mean)
    implicit none
    !Input/Output
    real(8), intent(in) :: ar1, sd_eps, mean
    real(8), dimension(1:), intent(in) :: y !shock values
    real(8), dimension(1:size(y),1:size(y)), intent(out) :: transition !transition matrix
    !Variable Declaration
    integer :: n ! size of y
    integer :: y_ind            !shock
    integer :: yp_ind           !shock next period
    real(8) :: mean_cond_on_y  !mean of y' cdnal on y
    real(8) :: normcdfhigh     !cdf at ub of interval
    real(8) :: normcdflow      !cdf at lb of interval
    real(8), dimension(1:size(y)) :: betweenpts_high !interval ub
    real(8), dimension(1:size(y)) :: betweenpts_low  !interval lb

    n = size(y)

    if (n==1) then
        transition = 1d0
        return
    end if

    !Define intervals for tauchen points
    do y_ind = 1, n-1
        betweenpts_high(y_ind)=y(y_ind)+(y(y_ind+1) - y(y_ind))/2d0        
    enddo

    do y_ind = 2, n
        betweenpts_low(y_ind)=y(y_ind)-(y(y_ind) - y(y_ind-1))/2d0        
    enddo

    !Calculate transition probabilities
    do y_ind = 1,n
        mean_cond_on_y = mean + ar1*y(y_ind) 
        
        normcdfhigh=.5d0*(1.0d0+erf((betweenpts_high(1)-mean_cond_on_y)/(sd_eps*sqrt(2d0))))
        normcdflow=0.d0
        transition(y_ind, 1) = normcdfhigh-normcdflow
        
        normcdfhigh=1.d0
        normcdflow=.5d0*(1.0d0+erf((betweenpts_low(n)-mean_cond_on_y)/(sd_eps*sqrt(2d0))))
        transition(y_ind, n) = normcdfhigh-normcdflow
        
        do yp_ind = 2,n-1
            normcdfhigh=.5d0*(1.0d0+erf((betweenpts_high(yp_ind)-mean_cond_on_y)/(sd_eps*sqrt(2d0))))
            normcdflow=.5d0*(1.0d0+erf((betweenpts_low(yp_ind)-mean_cond_on_y)/(sd_eps*sqrt(2d0))))
            transition(y_ind,yp_ind) = (normcdfhigh - normcdflow)            
        enddo
    enddo

end subroutine sub_tauchen_fixedgrid_with_mean

end module mod_discreteAR1
