module mod_min1dim
    implicit none

    interface sub_gridsearch1dim
        module procedure sub_gridsearch1dim_nogrid, sub_gridsearch1dim_grid
    end interface
contains

! Computes the first and second derivative.
! Take as input the function f, the point where derivative is desired x, and the function 
! value at x. 
! Returns the first and second derivatives at x, fpx and fppx.
! The second derivative is computed at almost zero cost, but could be made optional
subroutine numderiv1dim(fx,f,x,fpx,fppx)
    implicit none
    real(8), intent(in) :: x,fx
    interface
        function f(x)
            implicit none
            real(8), intent(in) :: x
            real(8) :: f
        end function f
    end interface
    real(8), intent(out) :: fpx,fppx
    !local
    real(8) :: h,fxh,fxmh

    ! Compute the step size
    h = max(abs(x),1d0)*(epsilon(1.d0))**(1d0/3d0);

    ! Compute the function values at +/-h
    fxh = f(x+h)
    fxmh = f(x-h)

    ! Compute the derivatives
    fpx = (fxh-fxmh)/(2d0*h)
    fppx = ((fxmh + fxh) - 2d0*fx)/(h**2)

end subroutine numderiv1dim

! Grid search in 1 dimension
!
! flag  meaning
!    0  successfully converged
!   -1  max iterations reached
!
subroutine sub_gridsearch1dim_nogrid(x,f,a,b,n,printmod)
    use mod_matlab, only: linspace 
    implicit none
    real(8), intent(inout) :: x ! Final solution
    interface
        function f(x)
            implicit none
            real(8), intent(in) :: x
            real(8) :: f
        end function f
    end interface
    real(8), intent(in) :: a,b
    integer, intent(in) :: n,printmod
    ! local
    real(8), allocatable :: grid(:)

    allocate(grid(n))

    grid = linspace(a,b,n)

    call sub_gridsearch1dim_grid(x,f,grid,printmod)

    deallocate(grid)

end subroutine 
subroutine sub_gridsearch1dim_grid(x,f,grid,printmod)
    use mod_sort
    implicit none
    real(8), intent(inout) :: x ! Final solution
    interface
        function f(x)
            implicit none
            real(8), intent(in) :: x
            real(8) :: f
        end function f
    end interface
    real(8), intent(in) :: grid(1:)
    integer, intent(in) :: printmod
    ! local
    integer :: i,n
    real(8) :: xbest,fbest,tmp

    n = size(grid)

    if (printmod>0) write(*,'(A)') 'GS1:              x          f(x)         xbest      f(xbest)'
    fbest = huge(0d0)
    do i = 1,n
        tmp = f(grid(i))
        if (tmp<fbest) then
            xbest = grid(i)
            fbest = tmp
        end if
        if (printmod>0 .and. mod(i,printmod)==0) write(*,'(A,4e14.6)') 'GS1: ',grid(i),tmp,xbest,fbest
    end do
    x = xbest
    if (printmod>0) write(*,'(A)') 'GS1: finished'

end subroutine 


! Simplex routine in 1 dimension
!
! flag  meaning
!    0  successfully converged
!   -1  max iterations reached
!
! As of Mar 5, 2014, this routine is stable in that it only moves away
! from xstar if a new point is strictly better. 
subroutine sub_simplex1dim(xstar,f,e,xtoler,maxiter,printmod,flag)
    use mod_sort
    implicit none
    real(8), intent(inout) :: xstar ! Initial guess and final solution
    interface
        function f(x)
            implicit none
            real(8), intent(in) :: x
            real(8) :: f
        end function f
    end interface
    real(8), intent(in) :: e
    real(8), intent(in) :: xtoler
    integer, intent(in) :: maxiter,printmod
    integer, intent(out) :: flag
    ! local
    real(8) :: s(2), fs(2), xr, fxr, xe, fxe
    integer :: rank(2), iter

    if (printmod>0) write(*,'(A)') 'S1:              a             b            fa            fb'

    ! Get the initial simplex
    s = (/xstar,xstar+e/)
    fs = (/f(xstar),f(xstar+e)/)


    do iter = 1,maxiter

        ! Sort the points
!         call isort(fs,rank)
!         s = s(rank)

        ! Sort
        if (fs(2)<fs(1)) then
            xr = s(1) ! Uses xr for a temp
            fxr = fs(1) ! Uses fxr for a temp
            s(1) = s(2) ! Swap
            fs(1) = fs(2)
            s(2) = xr
            fs(2) = fxr
        end if
        
        ! Check for convergence
        if (abs(s(1)-s(2))<xtoler) then
            xstar = s(1)
            flag = 0
            if (printmod>0) write(*,'(A)') 'S1: converged successfully, simplex size<xtoler'
            return
        end if

        ! If desired, compute progress
        if (printmod>0 .and. mod(iter,printmod)==0) write(*,'(A,4e14.6)') 'S1: ',s(1),s(2),fs(1),fs(2)

        ! Now get rid of the worst point s(2)

        ! First, note the center of gravity is s(1)

        ! Now, compute the reflection point
        xr = 2d0*s(1) - s(2)
        fxr = f(xr)
        
        if (fxr<fs(1)) then 

            ! If reflection is better than current best point, be greedy---try to expand.
            xe = 3d0*s(1) - 2d0*s(2)
            fxe = f(xe)

            if (fxe<fs(1)) then
                ! Expand
                s(2) = xe
                fs(2) = fxe
            else
                ! Reflect
                s(2) = xr
                fs(2) = fxr
            end if
        else
            ! Otherwise, contract (which here is the same as contract all)
            s(2) = 1.5d0*s(1) - .5d0*s(2)
            fs(2) = f(s(2))
        end if

    end do

    ! Return best if haven't converged yet
    xstar = merge(s(1),s(2),fs(1)<fs(2)) 
    flag = -1
    if (printmod>0) write(*,'(A)') 'S1: failed to converge, max iterations reached'


end subroutine sub_simplex1dim

! Newton's method in 1 dim
!
! Output flag returns info about the routine
! flag     meaning
!    2     converged successfully to a maximum, fpp<0
!    1     converged succesffuly to a minimum, fpp>0
!  >=0     converged successfully as measured by abs(s)<xtoler*(1d0+abs(xstar))
!   -1     failed to converge, maxiter reached 
!   -2     failed to converge, bad function value encountered (NaN)
!   -3     failed to converge, implausibly large derivative value encountered
!   -4     failed to converge, second derivative of 0 encountered
subroutine sub_newton1dim(xstar,f,xtoler,maxiter,printmod,flag)
    implicit none
    real(8), intent(inout) :: xstar ! Initial guess and final solution
    interface
        function f(x)
            implicit none
            real(8), intent(in) :: x
            real(8) :: f
        end function f
    end interface
    real(8), intent(in) :: xtoler
    integer, intent(in) :: maxiter,printmod
    integer, intent(out) :: flag
    ! local
    real(8) :: s,fx,fpx,fppx
    integer :: iter

    if (printmod>0) write(*,'(A)') 'N1:              x            fx           fpx          fppx '

    ! Use Newton's method 
    do iter = 1,maxiter
        
        ! Compute the function value at x
        fx = f(xstar)

        ! Compute the derivatives at x
        call numderiv1dim(fx,f,xstar,fpx,fppx)

        ! Check for errors
!         if (isnan(fx) .or. isnan(fpx) .or. isnan(fppx)) then 
!             if (printmod>0) write(*,'(A)') 'N1: failed to converge, NaN encountered'
!             flag = -2
!             return
        if (abs(fpx)>1d50 .or. abs(fppx)>1d50) then
            if (printmod>0) write(*,'(A)') 'N1: failed to converge, huge derivative encountered' 
            flag = -3
            return
        elseif (abs(fppx)<1d-10) then
            if (printmod>0) write(*,'(A)') 'N1: failed to converge, second derivative is approx 0'
            flag = -4
            return
        end if

        ! Compute the Newton step
        s = fpx/fppx

        ! If desired, compute progress
        if (printmod>0 .and. mod(iter,printmod)==0) write(*,'(A,4e14.6)') 'N1: ',xstar,fx,fpx,fppx

        ! Check for convergence
        if (abs(s)<(1d0+abs(xstar))*xtoler) then
            if (fppx<0d0) then
                if (printmod>0) write(*,'(A)') 'N1: converged successfully to a max, step<toler'
                flag = 2
            else
                if (printmod>0) write(*,'(A)') 'N1: converged successfully to a min, step<toler'
                flag = 1
            end if
            return
        end if

        ! Otherwise, update the step
        xstar = xstar - s

    end do

    flag = -1
    if (printmod>0) write(*,'(A)') 'N1: failed to converge, maxiter reached'

end subroutine sub_newton1dim

! sub_brent_min
! This is the local minimization procedure described in Brent "Algorithms for 
! Minimization without Derivatives." The comments from the ALGOL procedure are
! the following (found on p.79):
!   If the function f is defined on the interval (a,b), then localmin finds an 
! approximation x to the point at which f attains its minimum (or the 
! appropriate limit point), and returns the value of f at x. t and eps define 
! a tolerance tol = eps|x| + t, and f is never evaluated at two points closer 
! together than tol. If f is delta-unimodal for some delta<tol, then x 
! approximates the global minimum of f with an error less than 3*tol. If f is 
! not delta-unimodal on (a,b), then x may approximate a local, but not global, 
! minimum. eps should be no smaller than 2*macheps, and preferably not much 
! less than sqrt (macheps), where macheps is the relative machine precision. t 
! should be [strictly] positive. For further details, see SEction 2.
!   The method used is a combination of golden section search and successive 
! parabolic interpolation. Convergence is never much slower than for a 
! Fibonacci search. If f has a continuous second derivative which is positive 
! at the minimum (not at a or b) then, ignoring rounding errors, convergence 
! is superlinear, and usually the order is at least 1.3247.
!
! [Grey:] Modifications: the original routine never evaluates f at the endpoints
! a and b. Consequently, a corner solution of a or b wil never be found exactly.
! I just add one check at the end if it looks like there may be a corner solution
! and return the corner solution if there is one.
! 
! I also consulted NR, but only after writing the code when it was producing golden
! steps only. I saw the code differed from the Algol procedure in Brent but not the
! Fortran procedure. Brent's Algol procedure had a bug! But once corrected, the 
! code works as expected. 
subroutine sub_brent_min(x,f,a_in,b_in,xtoler_in,printmod_in)
    implicit none
    real(8), intent(out) :: x ! Final argmin
    interface
        function f(x)
            implicit none
            real(8), intent(in) :: x
            real(8) :: f
        end function f
    end interface
    real(8), intent(in) :: a_in,b_in
    ! optional arguments
    real(8), intent(in), optional :: xtoler_in ! Relative stopping criterion for bracket size
    integer, intent(in), optional :: printmod_in ! Printing control. If <0, then no information is displayed
    ! local
    real(8) :: a,b,t,u,v,w,fu,fv,fw,fx,t2,r,tol,q,p,m,d,e
    integer :: iter,printmod,step
    character(len=10) :: msg
    real(8), parameter :: phi=(1d0+sqrt(5.d0))/2.d0, c = (3d0-sqrt(5d0))/2d0, eps = epsilon(0d0)

    ! Set defaults and replace if optional argument is present
    t = sqrt(epsilon(0d0)); if (present(xtoler_in)) t = xtoler_in ! Relative stopping criterion
    printmod = 1; if (present(printmod_in)) printmod = printmod_in ! Printing control

    if (t<=0d0) then
        STOP 'ERROR: sub_brent_min: xtoler_in must be > 0'
    end if

    ! Initialization
    a = a_in
    b = b_in

    x = a + c*(b-a)
    w = x
    v = x

    fx = f(x)
    fw = fx
    fv = fx

    e = 0d0

    ! Main loop
    do iter = 1,10000

        if (iter==1) then
            msg = ' initial'
        else
            msg = merge(' parabolic',' golden   ',step==1)
        end if

        if (printmod>0 .and. mod(iter,printmod)==0) &
            write(*,'(A,i5,4f15.9,A)') 'B1: i,a,b,x,fx',iter,a,b,x,fx,msg

        m = .5d0*(a+b)
        tol = eps*abs(x) + t
        t2 = 2d0*tol

        ! Check stopping criterion
        if (.not. (abs(x-m)>t2 - .5d0*(b-a))) exit

        if (abs(e)>tol) then

            ! Fit parabola
            r = (x-w)*(fx-fv)
            q = (x-v)*(fx-fw)
            p = (x-v)*q - (x-w)*r 
            q = 2d0*(q-r)

            if (q>0d0) then
                p = -p
            else
                q = -q
            end if

            r = e
            e = d
        
        else 
            
            ! Don't fit parabola
            r = 0d0
            q = 0d0
            p = 0d0

        end if

        ! if (abs(p)<abs(.5d0*q*r) .and. p<q*(a-x) .and. p<q*(b-x)) then
        ! NOTE: THE ALGOL PROCEDURE IN BRENT HAS A BUG AND WOULD BE TRANSLATED
        ! INTO THE ABOVE LINE OF CODE. THE BUG IS NOT PRESENT IN THE FORTRAN 
        ! CODE WHICH NR HAS CORRECT.
        if (abs(p)<abs(.5d0*q*r) .and. p>q*(a-x) .and. p<q*(b-x)) then

            ! Parabolic interpolation step
            d = p/q
            u = x + d

            ! f must not be evaluated too close to a or b
            if (u-a<t2 .or. b-u<t2) then
                d = merge(tol,-tol,x<m)
            end if

            step = 1

        else

            ! Golden section step
            e = merge(b,a,x<m) - x
            d = c*e

            step = 0

        end if

        ! f must not be evaluated too close to x

        u = x + merge(d,merge(tol,-tol,d>0d0),abs(d)>=tol)
        fu = f(u)

        ! Update a,b,v,w, and x
        if (fu<=fx) then

            ! Replace upper or lower bound
            if (u<x) then
                b = x
            else
                a = x
            end if

            ! Save previous values
            v = w
            fv = fw
            w = x
            fw = fx
            x = u
            fx = fu

        else
            
            ! Replace upper or lower bound
            if (u<x) then
                a = u
            else
                b = u
            end if

            ! Save previous values
            if (fu <= fw .or. w == x) then
                v = w 
                fv = fw
                w = u
                fw = fu
            else if (fu <= fv .or. v == x .or. v == w) then
                v = u
                fv = fu
            end if
        end if

    end do

    ! MODIFICATION: if close to boundary, compute the value at the boundary and take max
    tol = eps*abs(x) + t
    t2 = tol*2d0
    if (abs(x-a_in)<=t2) then
        fu = f(a_in) 
        if (fu<fx) then
            x = a_in
        end if

        msg = ' corner'
        iter = iter+1
        if (printmod>0 .and. mod(iter,printmod)==0) &
            write(*,'(A,i5,4f15.9,A)') 'B1: i,a,b,x,fx',iter,a,b,x,fx,msg

    elseif (abs(x-b_in)<=t2) then
        fu = f(b_in) 
        if (fu<fx) then
            x = b_in
        end if

        msg = ' corner'
        iter = iter+1
        if (printmod>0 .and. mod(iter,printmod)==0) &
            write(*,'(A,i5,4f15.9,A)') 'B1: i,a,b,x,fx',iter,a,b,x,fx,msg
    end if


end subroutine sub_brent_min

! Golden section search to minimize a function
! The algorithm is from Wikipedia 
! xtoler_in :: the algorithm stops when |c-a|<xtoler*(1+abs(b)) where [a,c] brackets a min
! 
subroutine sub_golden(xstar,f,xtoler_in,step_in,printmod_in)
    implicit none
    real(8), intent(inout) :: xstar ! Initial guess and final solution
    interface
        function f(x)
            implicit none
            real(8), intent(in) :: x
            real(8) :: f
        end function f
    end interface
    ! optional arguments
    real(8), intent(in), optional :: xtoler_in ! Relative stopping criterion for bracket size
    real(8), intent(in), optional :: step_in ! First step size when trying to bracket 
    integer, intent(in), optional :: printmod_in ! Printing control. If <0, then no information is displayed
    ! local
    real(8) :: a,b,c,fa,fb,fc,step,xtoler
    integer :: iter,printmod
    real(8), parameter :: phi=(1.d0+sqrt(5.d0))/2.d0

    ! Set defaults and replace if optional argument is present
    xtoler = sqrt(epsilon(0.d0)); if (present(xtoler_in)) xtoler = xtoler_in ! Relative stopping criterion
    step = 1d-6*(1.d0+abs(xstar)); if (present(step_in)) step = step_in ! Initial step size to search
    printmod = -1; if (present(printmod_in)) printmod = printmod_in ! Printing control

    ! Try to find a bracket centered around xstar consistent with golden section search
    b = xstar
    fb = f(b)

    ! Bracket a minimum with section lengths having the golden ratio
    if (printmod>0) print*,'G1: Finding a bracket'
    do iter = 1,100

        ! Choose right interval so it is of size step
        c = b + phi*step

        ! Choose left interval so that it is of size phi*step where phi is the golden ratio
        a = b - step

        ! Evaluate f at a and c
        fa = f(a)
        fc = f(c)

        if (printmod>0 .and. mod(iter,printmod)==0) then
            write(*,'(A,6f9.3)') 'G1: a,b,c,fa,fb,fc',a,b,c,fa,fb,fc
        end if

        ! If f(a)>=f(b) and f(b)<=f(c), have found a bracket
        if (fa>=fb .and. fb<=fc) then
            if (printmod>0) write(*,'(A)') 'G1: Bracket found. Beginning min search.'
            exit
        else
            step = step*10.d0
        end if

        if (iter==100) then
            write(*,'(A)') 'sub_golden: WARNING: couldn''t bracket a min, returning init guess'
            xstar = b
            return
        end if
    end do

    ! Call the core routine now that the first triple has been established
    call sub_golden_core(a,b,c,fa,fb,fc,f,xtoler,printmod)
    
    ! Return the b: this is better than (a+c)/2 because if c or a may not be feasible but b may be.
    xstar = b 

end subroutine sub_golden

! a<b<c bracket a minimum with f(a)>=f(b), f(b)<=f(c)
subroutine sub_golden_core(a,b,c,fa,fb,fc,f,xtoler,printmod)
    implicit none
    real(8), intent(inout) :: a,b,c,fa,fb,fc
    interface
        function f(x)
            implicit none
            real(8), intent(in) :: x
            real(8) :: f
        end function f
    end interface
    real(8), intent(in) :: xtoler ! Relative stopping criterion
    integer, intent(in) :: printmod ! Print control
    real(8), parameter :: phi=(1.d0+sqrt(5.d0))/2.d0
    ! local
    real(8) :: d,fd
    integer :: di
    integer :: iter

    do iter = 1,10000

        ! Check stopping criterion
        if (c-a<xtoler*(1.d0+abs(b))) then
            if (printmod>0) write(*,'(A)') 'G1: finished'
            return
        end if

        ! Compute the new point
        if (c-b>b-a) then
            d = b + (2.d0-phi)*(c-b)
        else
            d = b - (2.d0-phi)*(b-a)
        end if

        ! Compute the function value
        fd = f(d)

        ! Display info if desired
        if (printmod>0 .and. mod(iter,printmod)==0) then
            write(*,'(A,6f9.3)') 'G1: a,b,c,fa,fb,fc',a,b,c,fa,fb,fc
        end if

        ! Compute the new bracket depending on the value of fd 
        if (fd<fb) then
            if (c-b>b-a) then
                a = b;  fa = fb
                b = d;  fb = fd
            else
                c = b;  fc = fb
                b = d;  fb = fd
            end if
        else
            if (c-b>b-a) then
                c = d;  fc = fd
            else
                a = d;  fa = fd
            end if
        end if

    end do

end subroutine sub_golden_core 

end module mod_min1dim


module mod_min1dim_test
    use mod_min1dim
    implicit none 
contains

! Test functions
! Defined on R
! minimum at x = 5d0
function uniqMinQuad(x) result(y)
    real(8), intent(in) :: x
    real(8) :: y
    y = (x-5d0)**2 + 2d0
end function 
! Defined on R++, maximum at x = 1
function uniqMaxConc(x) result(y)
    real(8), intent(in) :: x
    real(8) :: y
    y = log(x) - x 
end function 
! Defined on R++, minimum at x = 1
function uniqMinConc(x) result(y)
    real(8), intent(in) :: x
    real(8) :: y
    y = -(log(x) - x )
end function 

subroutine test_brent_min
    implicit none
    real(8) :: x

    call sub_brent_min(x,uniqMinConc,0d0,100d0)
    if (abs(x-1d0)>1d-7) STOP 'ERROR: test_brent_min: failed test 1' 
    call sub_brent_min(x,uniqMinQuad,-100d0,100d0)
    if (abs(x-5d0)>1d-7) STOP 'ERROR: test_brent_min: failed test 2' 

    print*,'test_brent_min: passed'
    

end subroutine

subroutine test_simplex1dim
    implicit none
    real(8) :: x
    integer :: flag

    x = -10d0
    call sub_simplex1dim(x,uniqMinQuad,100d0,1d-6,10000,-1,flag)
    
    if (flag<0) STOP 'ERROR: test_simplex1dim: failed to converge' 

end subroutine
subroutine test_newton1dim
    implicit none
    real(8) :: x
    integer :: flag

    x = -10d0
    call sub_newton1dim(x,uniqMinQuad,1d-6,10000,-1,flag)
    if (flag<0) STOP 'ERROR: test_newton1dim: failed to converge' 
    x = .3d0
    call sub_newton1dim(x,uniqMaxConc,1d-6,10000,-1,flag)
    if (flag<0) STOP 'ERROR: test_newton1dim: failed to converge' 

end subroutine

end module 
