! Grey Gordon, 2012
! 
! For God so loved the world, that He gave His only begotten Son, 
! that whoever believes in Him shall not perish, but have eternal life.
!
module mod_minNdim_w_cache

    type minN_f_cache 
        real(8), allocatable :: x(:,:)
        real(8), allocatable :: y(:)
        integer :: n,maxn
!     contains
!         procedure, pass :: init => init_cache
!         procedure, pass :: put => put_cache
!         procedure, pass :: check => check_cache
!         procedure, pass :: free => empty_cache
    end type

    abstract interface
        function minN_f_template(x) result(y)
            implicit none
            real(8), intent(in) :: x(1:)
            real(8) :: y
        end function 
    end interface

contains

    ! The function g will first search the cache and if not found, will evaluate f
    function f_without_cache(x,f,cache) result(g)
        implicit none
        procedure(minN_f_template) :: f
        type(minN_f_cache), intent(inout) :: cache ! not used
        real(8), intent(in) :: x(1:)
        real(8) :: g

        g = f(x)

    end function

    function f_with_cache(x,f,cache) result(g)
        implicit none
        procedure(minN_f_template) :: f
        type(minN_f_cache), intent(inout) :: cache
        real(8), intent(in) :: x(1:)
        real(8) :: g
        ! local
        real(8) :: h
        integer :: flag

        call check_cache(cache,x,h,flag)

        if (flag == 1) then
            ! Cached value found
            g = h
            return
        else
            ! Cached value not found
            h = f(x)
            call put_cache(cache,x,h)
            g = h
            return
        end if

    end function

    subroutine init_cache(cache,nx,maxn)
        implicit none
        type(minN_f_cache), intent(out) :: cache
        integer, intent(in) :: nx, maxn

        cache%maxn = maxn
        cache%n = 0
        allocate(cache%x(nx,maxn))
        allocate(cache%y(maxn))

    end subroutine
     
    subroutine free_cache(cache)
        implicit none
        type(minN_f_cache), intent(out) :: cache
        if (allocated(cache%x)) deallocate(cache%x)
        if (allocated(cache%y)) deallocate(cache%y)
    end subroutine

    subroutine put_cache(cache,x,fx)
        implicit none
        real(8), intent(in) :: x(1:),fx
        type(minN_f_cache), intent(inout) :: cache

        if (cache%n>=cache%maxn) then
            print*,'WARNING: cache size exceeded, not caching more values' 
        else
            cache%n= cache%n + 1
            cache%x(:,cache%n) = x
            cache%y(cache%n) = fx
        end if

    end subroutine 

    subroutine check_cache(cache,x,fx,flag)
        implicit none
        type(minN_f_cache), intent(in) :: cache
        real(8), intent(in) :: x(1:)
        real(8), intent(inout) :: fx ! Overwritten if found in cache
        integer, intent(out) :: flag
        ! local
        integer :: i

        flag = 0
        ! Search the cache
        do i = 1,cache%n
            if (all(x==cache%x(:,i))) then
                flag = 1
                fx = cache%y(i)
                return
            end if
        end do
        
    end subroutine 
    
end module 

module mod_minNdim
    use mod_minNdim_w_cache
    use mod_mynlopt
    implicit none

    private :: minN_f_template

    abstract interface
        function minN_f_template_w_cache(x,f,cache) result(y)
           use mod_minNdim_w_cache ! only way I could figure out how to declare minN_f_template and minN_f_cache here
           implicit none
           real(8), intent(in) :: x(1:) 
           procedure(minN_f_template) :: f
           type(minN_f_cache), intent(inout) :: cache
           real(8) :: y
       end function
    end interface

    ! Options for minNdim 
    type minNdim_opts
        integer :: printmod = 1, maxiter = 10000
        real(8) :: xtoler = 1d-6
    end type 

contains

    function optimset(s1,x1,s2,x2,s3,x3) result(opts)
        implicit none
        character(len=*), intent(in), optional :: s1,s2,s3
        real(8), intent(in), optional :: x1,x2,x3
        type(minNdim_opts) :: opts

        if (present(s1) .and. present(x1)) call setoption(s1,x1)
        if (present(s2) .and. present(x2)) call setoption(s2,x2)
        if (present(s3) .and. present(x3)) call setoption(s3,x3)

    contains

        subroutine setoption(s,x)
            use mod_matlab
            implicit none
            character(len=*), intent(in) :: s
            real(8), intent(in) :: x

            if (strcmp(s,"xtoler")) then
                opts%xtoler = x
            elseif (strcmp(s,"printmod")) then
                opts%printmod = int(x)
            elseif (strcmp(s,"maxiter")) then
                opts%maxiter = int (x)
            end if

        end subroutine

    end function 

    ! Noisy Levenberg-Marquardt with MPI
    !
    ! solve min_{x \in R^{nx}} \sum^{nf}_{j=1} f^j(x)
    !
    ! This is a routine I made up (pdf is in noisyLM folder).
    ! Compute a linear projection of the function at hand via OLS and random points to obtain an 
    ! approximate Jacobian. Then this Jacobian is used in the usual Levenberg-Marquardt procedure.
    ! 
    subroutine sub_parnoisyLM(x,f,nf,stdev,ndraws,opts)
        use mod_mympi
        use mod_matlab
        implicit none
        integer, intent(in) :: nf
        real(8), dimension(:), intent(inout) :: x
        interface 
            ! f is vector valued taking x (a length nx vector) to a length nf vector.
            function f(x,nf) 
                implicit none
                integer, intent(in) :: nf
                real(8), dimension(:), intent(in) :: x
                real(8), dimension(nf) :: f
            end function
        end interface
        real(8), dimension(1:size(x)), intent(in), optional :: stdev
        integer, intent(in), optional :: ndraws
        type(minNdim_opts), optional :: opts
        !local
        type(minNdim_opts) :: o
        real(8), allocatable :: stdev_(:), xtrial(:), e_(:,:), Ymattranspose(:,:), Ymat(:,:), Xmat(:,:), J(:,:), fx(:), gL(:), s(:), eyex(:,:), ftrial(:), xtrials(:,:), ftrials(:,:)
        integer, allocatable :: request(:)
        logical, allocatable :: isFree(:)
        integer :: iter, nx, itmp(1), nL, iL, idraw, ndraws_, iproc
        logical :: haveRootDoIter1
        real(8) :: gradSSRSnorm,stepnorm,lambda,SSRSx,tmpSSRS,lambdabest

        if (mympi_nproc()==1) then
            write(*,*) 'WARNING: If only using one node, call sub_noisyLM instead'
            write(*,*) 'WARNING: Calling it for you with default parameters'
            call sub_noisyLM(x,f,nf)
            return
        end if

        nx = size(x)
        allocate(fx(nf))

    if (mympi_iproc()==0) then

        if (present(opts)) o = opts

        ndraws_ = 100 ! 10*nx
        if (present(ndraws)) ndraws_ = ndraws

        if (ndraws_<mympi_nproc()-1) then
            ! Increase ndraws so don't waste computation
            write(*,*) 'WARNING: Increasing ndraws to not waste nodes'
            ndraws_ = mympi_nproc()-1 
        end if
            
        ! Set nL so that node nodes are wasted and that at least 12 points are evaluated
        nL = max(12,mympi_nproc()-1)

        allocate(s(nx))
        allocate(gL(nL))
        allocate(stdev_(nx))
        allocate(xtrial(nx))
        allocate(ftrial(nf))
        allocate(xtrials(nx,nL))
        allocate(ftrials(nf,nL))
        allocate(e_(nx,ndraws_))
        allocate(Ymat(ndraws_,nf))
        allocate(Ymattranspose(nf,ndraws_))
        allocate(Xmat(ndraws_,nx))
        allocate(eyex(nx,nx))
        allocate(isFree(1:mympi_nproc()-1))
        allocate(request(1:mympi_nproc()-1))

        isFree = .true.
        eyex = eye(nx)

        if (nL==12) then
            gL(1:12) = [0d0,1d-6,1d-5,1d-4,1d-3,1d-2,1d-1,1d0,1d1,1d2,1d3,1d4]
        else
            gL(1) = 0d0
            gL(2:nL) = logspace(-6d0,4d0,nL-1)
        end if

        ! Give each process a random seed based on the clock
        call init_random_seed(mympi_iproc())

        !Set default values
        stdev_ = 1d-3

        !Overwrite if applicable
        if (present(stdev)) stdev_ = stdev

        ! Determine whether the root node should compute f on the first iteration
        ! This is done purely to avoid wasting resources because the first iteration requires 
        ! ndraws_+1 function evaluations whereas every subsequent iteration requires ndraws_.
        haveRootDoIter1 = ndraws_ == mympi_nproc()-1

        ! Main loop
        do iter = 1,o%maxiter

            ! Compute draws
            do idraw = 1,ndraws_
                e_(:,idraw) = stdev_*randn(nx)
            end do

            ! Compute function values
            do idraw = 1,merge(ndraws_+1,ndraws_,iter==1)

                if (idraw==ndraws_+1) then
                    ! Special case: first iteration, need to compute f of initial guess
                    xtrial = x
                else
                    ! Get xtrial
                    xtrial = x + e_(:,idraw)
                end if

                ! Compute the trial point
                if (idraw==ndraws_+1 .and. haveRootDoIter1) then

                    fx = f(xtrial,nf)

                else

                    ! If there are no free nodes, wait for one to be free
    ! write(*,'(A)',advance='no') '|'
                    do while (.not. any(isFree))

                        ! Test each node to see if free
                        do iproc = 1,mympi_nproc()-1
                            
                            ! Check whether finished
                            isFree(iproc) = mympi_test(request(iproc))

                        end do
                        
                    end do

                    ! Use the free node
                    if (any(isFree)) then

                        ! If there is a free node, get it
                        iproc = find1(isFree,'first')
                    
                    else
                        
                        call mympi_stop('ERROR: could not find free node')
                        
                    end if

                    ! Send free node xtrial
                    call mympi_send(xtrial,iproc)

                    ! Request the value of f(xtrial,nf) in Ymat(idraw,:)
                    if (idraw==ndraws_+1) then
                        ! print*,mympi_iproc(),': non-blocking receive of fx from ',iproc
                        call mympi_irecv(request(iproc),fx,iproc)
                    else
                        ! print*,mympi_iproc(),': non-blocking receive of Ymattran from ',iproc
                        call mympi_irecv(request(iproc),Ymattranspose(:,idraw),iproc)
                    end if

                    ! Mark as not free 
                    isFree(iproc) = .false.

                end if

            end do

            ! Force all requests to complete before continuing
            ! print*,mympi_iproc(),': waiting for all processes to complete'
            call mympi_waitall(pack(request,.not. isFree))
            isFree = .true.

            ! If first iteration, save SSRSx
            if (iter==1) then
                SSRSx = dot_product(fx,fx)
            end if

            ! Evaluate f at the trial points
            Ymat = transpose(Ymattranspose)
            Xmat = transpose(e_)
            J = transpose(mldivide(matmul(transpose(Xmat),Xmat),matmul(transpose(Xmat),Ymat)))

            gradSSRSnorm = sqrt(dot_product(matmul(transpose(J),fx),matmul(transpose(J),fx)))

            ! For lambda on the line search, compute x+s and f(x+s)
            do iL = 1,nL

                lambda = gL(iL)

                ! s = - mldivide(matmul(transpose(J),J) + lambda*eyex, matmul(transpose(J),fx))
                s = - mldivide(matmul(transpose(J),J) + lambda*diag(diag(matmul(transpose(J),J))), matmul(transpose(J),fx))

                xtrials(:,iL) = x + s

                ! If there are no free nodes, wait for one to be free
                do while (.not. any(isFree))

                    ! Test each node to see if free
                    do iproc = 1,mympi_nproc()-1
                        
                        ! Check whether finished
                        isFree(iproc) = mympi_test(request(iproc))

                    end do
                    
                end do

                ! If there is a free node, get it
                iproc = find1(isFree,'first')

                ! Send the free node xtrial
                call mympi_send(xtrials(:,iL),iproc)

                ! Request the value of f(xtrial,nf) in ftrials(:,iL)
                call mympi_irecv(request(iproc),ftrials(:,iL),iproc)

                ! Mark as not free 
                isFree(iproc) = .false.

                if (iL==nL) then

                    ! Wait for everyone to finish
                    call mympi_waitall(pack(request,.not. isFree))
                    isFree = .true.

                end if

            end do 

            itmp = minloc(sum(ftrials**2,dim=1))
            tmpSSRS = sum(ftrials(:,itmp(1))**2)

            if (tmpSSRS<SSRSx) then

                ! Best line search is better than current best
                lambdabest = gL(itmp(1))
                s = xtrials(:,itmp(1)) - x
                stepnorm = sqrt(dot_product(s,s))
                x = xtrials(:,itmp(1))
                fx = ftrials(:,itmp(1))
                SSRSx = tmpSSRS

            else

                ! Keep old value
                lambdabest = -1d0
                stepnorm = 0d0

            end if

            if (o%printmod>0 .and. mod(iter,o%printmod)==0) write(*,'(A,i8,4(A,d12.5))') ' NLM: iter',iter, ' step norm ',stepnorm,' SSRS ',SSRSx,' SSRS grad norm', gradSSRSnorm,' lambda ', lambdabest

            ! Check convergence criteria
            if (stepnorm<o%xtoler .and.  gradSSRSnorm<o%xtoler) exit
        
        end do

        ! Tell every node to stop by sending them -huge(0d0)
        do iproc = 1,mympi_nproc()-1
            xtrial = -huge(0d0)
            call mympi_send(xtrial,iproc)
        end do

    else
        
        ! If not the root node, enter an infinite loop where you wait to receive an x value. Compute f(x) when you do.
        do 

            ! Receive from root
            ! print*,mympi_iproc(),': receiving x from root'
            call mympi_recv(x,0) 

            ! Test whether x is less than -1d100, signaling it is time to stop the loop
            if (all(x<-1d100)) exit

            ! Evaluate f
            fx = f(x,nf)

            ! Send the value back to root
            ! print*,mympi_iproc(),': sending fx to root'
            call mympi_send(fx,0)

        end do


    end if

        ! Send x from root to every node
        call mympi_bcast(x,0)

    end subroutine 

    
    ! Noisy Levenberg-Marquardt
    !
    ! solve min_{x \in R^{nx}} \sum^{nf}_{j=1} f^j(x)
    !
    ! This is a routine I made up (pdf is in noisyLM folder).
    ! Compute a linear projection of the function at hand via OLS and random points to obtain an 
    ! approximate Jacobian. Then this Jacobian is used in the usual Levenberg-Marquardt procedure.
    ! 
    subroutine sub_noisyLM(x,f,nf,stdev,ndraws,opts)
        use mod_mympi
        use mod_matlab
        implicit none
        integer, intent(in) :: nf
        real(8), dimension(:), intent(inout) :: x
        interface 
            ! f is vector valued taking x (a length nx vector) to a length nf vector.
            function f(x,nf) 
                implicit none
                integer, intent(in) :: nf
                real(8), dimension(:), intent(in) :: x
                real(8), dimension(nf) :: f
            end function
        end interface
        real(8), dimension(1:size(x)), intent(in), optional :: stdev
        integer, intent(in), optional :: ndraws
        type(minNdim_opts), optional :: opts
        !local
        type(minNdim_opts) :: o
        real(8), allocatable :: stdev_(:), xtrial(:), e_(:,:), Ymat(:,:), Xmat(:,:), J(:,:), fx(:), gL(:), s(:), eyex(:,:), ftrial(:), fxbest(:), xbest(:)
        integer :: iter, nx, nL, iL, idraw, ndraws_
        real(8) :: gradSSRSnorm,stepnorm,lambda,SSRSx,tmpSSRS,lambdabest,SSRSbest

        if (present(opts)) o = opts

        nx = size(x)

        ndraws_ = 100 
        if (present(ndraws)) ndraws_ = ndraws

        nL = 12

        allocate(s(nx))
        allocate(xbest(nx))
        allocate(fx(nf))
        allocate(fxbest(nf))
        allocate(gL(nL))
        allocate(stdev_(nx))
        allocate(xtrial(nx))
        allocate(ftrial(nf))
        allocate(e_(nx,ndraws_))
        allocate(Ymat(ndraws_,nf))
        allocate(Xmat(ndraws_,nx))
        allocate(eyex(nx,nx))

        eyex = eye(nx)

        if (nL==12) then
            gL(1:12) = [0d0,1d-6,1d-5,1d-4,1d-3,1d-2,1d-1,1d0,1d1,1d2,1d3,1d4]
        else
            gL(1) = 0d0
            gL(2:nL) = logspace(-6d0,4d0,nL-1)
        end if

        ! Give each process a random seed based on the clock
        call init_random_seed(mympi_iproc())

        !Set default values
        stdev_ = 1d-3

        !Overwrite if applicable
        if (present(stdev)) stdev_ = stdev


        ! Initialize
        fx = f(x,nf) 
        SSRSx = dot_product(fx,fx)

        ! Main loop
        do iter = 1,o%maxiter

            do idraw = 1,ndraws_

                ! Compute draws
                e_(:,idraw) = stdev_*randn(nx)

                xtrial = x + e_(:,idraw)

                ! Evaluate f at the trial points
                Ymat(idraw,:) = f(xtrial,nf)

            end do

            Xmat = transpose(e_)
            J = transpose(mldivide(matmul(transpose(Xmat),Xmat),matmul(transpose(Xmat),Ymat)))

            gradSSRSnorm = sqrt(dot_product(matmul(transpose(J),fx),matmul(transpose(J),fx)))

            ! Do a line search
            ! Initialize values to current
            xbest = x
            fxbest = fx
            SSRSbest = SSRSx
            lambdabest = -1d0 ! keeping current value 
            stepnorm = 0d0
            do iL = 1,nL
                lambda = gL(iL)

                ! s = - mldivide(matmul(transpose(J),J) + lambda*eyex, matmul(transpose(J),fx))
                s = - mldivide(matmul(transpose(J),J) + lambda*diag(diag(matmul(transpose(J),J))), matmul(transpose(J),fx))

                xtrial = x + s
                ftrial = f(xtrial,nf)
                tmpSSRS = dot_product(ftrial,ftrial)

                if (tmpSSRS<SSRSbest) then
                    xbest = xtrial
                    fxbest = ftrial
                    SSRSbest = tmpSSRS
                    lambdabest = lambda
                    stepnorm = sqrt(dot_product(s,s))
                end if

            end do

            x = xbest
            fx = fxbest
            SSRSx = SSRSbest

            if (o%printmod>0 .and. mod(iter,o%printmod)==0) write(*,'(A,i8,4(A,d12.5))') ' NLM: iter',iter, ' step norm ',stepnorm,' SSRS ',SSRSx,' SSRS grad norm', gradSSRSnorm,' lambda ', lambdabest

            ! Check convergence criteria
            if (stepnorm<o%xtoler .and.  gradSSRSnorm<o%xtoler) exit
        
        end do

    end subroutine 


    ! MPI implementation of random search
    ! Beginning with an initial point x, draws N proposals x+eps where eps is N(0,factor^2*e^2)
    ! where N is the number of MPI ranks, e is given by the user, and factor is internally set.
    ! The factor begins at 1d0. If during an iteration an improvement is found, then factor = 
    ! factor*expand. O/w, factor = factor*shrink. The routine ends when factor*e<xtoler*abs(x) uniformly.
    ! 
    ! NOTE: this sets the random seed to distinct values for each node based on the clock. 
    subroutine sub_parrandom(x,f,e_in,maxiter_in,printmod_in,xtoler_in,shrink_in,expand_in)
        use mod_mympi
        use mod_matlab
        implicit none
        real(8), dimension(1:), intent(inout) :: x
        interface 
            function f(x) 
                implicit none
                real(8), dimension(1:), intent(in) :: x
                real(8) :: f
            end function f
        end interface
        real(8), dimension(1:size(x)), intent(in), optional :: e_in
        integer, intent(in), optional :: maxiter_in,printmod_in
        real(8), intent(in), optional :: xtoler_in, shrink_in, expand_in
        !local
        real(8), allocatable :: e(:), ftrials(:), xtrial(:) !, xtmp(:)
        real(8) :: xtoler, fx, expand, shrink, factor
        integer :: maxiter, printmod, iter, nproc, nx, itmp(1) !, iproc

        nx = size(x)

        ! Give each process a random seed based on the clock
        call init_random_seed(mympi_iproc())

        !Set default values
        allocate(e(nx))
        e = max(.1d0*abs(x),10d0*epsilon(0d0)) !the size of neighborhood about each point which is a 
                                   !hypersphere stretched to edge length e in each dimension
        maxiter = 100000
        printmod = 1
        xtoler = 1d-6
        shrink =  .9d0 
        expand = 1.1d0

        !Overwrite if applicable
        if (present(maxiter_in)) maxiter = maxiter_in
        if (present(printmod_in)) printmod = printmod_in
        if (present(e_in)) e = e_in
        if (present(xtoler_in)) xtoler = xtoler_in
        if (present(shrink_in)) shrink = shrink_in
        if (present(expand_in)) expand = expand_in

        ! Get MPI info
        nproc = mympi_nproc()
        allocate(ftrials(nproc))
        allocate(xtrial(nx))
        ! allocate(xtmp(nx))

        ! Initialize
        fx = huge(0d0) ! set large so replace for sure
        factor = 1d0

        ! Main loop
        do iter = 1,maxiter

            ! Compute a trial value
            xtrial = x + randn(nx)*e*factor
!             do iproc = 1,nproc
!                 if (iproc - 1 == mympi_iproc()) then
!                     xtrial = x + randn(nx)*e*factor
!                 else
!                     ! So that xtrial is definitely different across all nodes, call many randn calls
!                     xtmp = randn(nx)
!                 end if
!             end do

            ! Special case: first iteration, master node, compute value for x
            if (iter==1 .and. mympi_iproc()==0) then
                xtrial = x
            end if

            ! Compute the function value, but only for your rank
            ftrials(mympi_iproc()+1) = f(xtrial)

            ! Gather the function values only
            call mympi_allgatherv(ftrials)

            ! Select the best, check whether better than the current by a significant margin.
            if (minval(ftrials)<fx - (10d0*epsilon(0d0))*abs(fx)) then
                itmp = minloc(ftrials)
                fx = ftrials(itmp(1))
                
                ! If you are the best rank, send xtrial to everyone else. O/w, recieve.
                call mympi_bcast(xtrial,itmp(1)-1)
                
                ! Now all ranks have the copy of the best x
                x = xtrial
                factor = factor*expand
            else
                factor = factor*shrink
            end if

            if (mympi_iproc()==0 .and. printmod>0 .and. mod(iter,printmod)==0) write(*,'(A,i10,A,d12.5,A,d12.5)') ' PR: iter',iter, ' f',fx,' factor', factor

            ! Check stopping criteria
            if (all(factor*e<xtoler*abs(x))) exit
        
        end do

        if (mympi_iproc()==0 .and. printmod>0 .and. printmod==1) then
            write(*,*) 'PR: final fx', fx, 'final x',x
        end if

    end subroutine 

    ! Using mpi, assigns each node an initial guess xg = x + e*N(0,1). Subsequently, each node
    ! uses a user-specified minimization routine to minimize give that initial guess.
    ! This is a ridiculously parallel minization.
    ! 
    ! The minization routine is based on the argument opt
    ! opt | routine
    !   0   sub_simplex
    ! 
    subroutine sub_parmin(x,f,opt,e,a_in,b_in,xtoler_in,maxiter_in,maxfeval_in,printmod_in,uid_in)
        use mod_matlab
        use mod_mympi
        implicit none
        real(8), intent(inout) :: x(:)
        integer, intent(in) :: opt
        real(8), intent(in) :: e(:)
        real(8), intent(in), optional :: xtoler_in, a_in(:), b_in(:)
        integer, intent(in), optional :: maxiter_in,maxfeval_in,printmod_in,uid_in
        interface 
            function f(x) 
                implicit none
                real(8), dimension(:), intent(in) :: x
                real(8) :: f
            end function f
        end interface
        ! local
        real(8) :: xtoler, xtmp(size(x)), x0(size(x)), a(size(x)), b(size(x))
        integer :: i, maxfeval, printmod, maxiter, uid, iter

        ! Set default parameters and overwrite if optional argument is present
        xtoler = 1d-6;    if (present(xtoler_in)) xtoler = xtoler_in
        maxfeval = 10000; if (present(maxfeval_in)) maxfeval = maxfeval_in
        maxiter = 10000;  if (present(maxiter_in)) maxiter = maxiter_in
        printmod = 1;     if (present(printmod_in)) printmod = printmod_in
        uid = 6;          if (present(uid_in)) uid = uid_in
        a = -huge(0d0);   if (present(a_in)) a = a_in
        b = huge(0d0);    if (present(b_in)) b = b_in

        if (any(b<a)) call mympi_stop('ERROR: sub_parmin: b is less than a')
        if (any(x<a .or. x>b)) call mympi_stop('ERROR: sub_parmin: x is not in [a,b]')

        ! Generate initial points (here, burn through points to ensure each node gets a different x)
        x0 = x
        do i = 0,mympi_nproc()-1

            ! Generate a point in [a,b]
            do iter = 1,100000
                xtmp = x0 + e*randn(size(x)) 
                if (all(xtmp>=a .and. xtmp<=b)) exit
            end do
            if (iter>100000) then
                write(uid,*) 'WARNING: sub_parmin: could not find a point in [a,b]'
            end if

            ! Use as starting point if i is for your node
            if (i==mympi_iproc()) x = xtmp

        end do 

        ! Call the minimization routine in parallel
        select case (opt)
            case (0)
                call sub_simplex(x,f,e_in=e,xtoler_in=xtoler,maxiter_in=maxiter,printmod_in=printmod,uid_in=uid)
            case default 
                call mympi_stop('ERROR: sub_parmin: case not found')
        end select

        ! Call two more times to compare against initial x. This way, the routine never does worse than the input guess.
        if (f(x0)<f(x)) then
            x = x0
        end if

    end subroutine 

!     subroutine sub_subplex(x,f,e_in,xtoler_in,maxfeval_in,printmod_in)
!         implicit none
!         real(8), intent(inout) :: x(:)
!         real(8), intent(in), optional :: e_in(:),xtoler_in
!         integer, intent(in), optional :: maxfeval_in,printmod_in
!         interface 
!             function f(x) 
!                 implicit none
!                 real(8), dimension(:), intent(in) :: x
!                 real(8) :: f
!             end function f
!         end interface
!         ! local
!         integer :: n,maxfeval,mode,mdsing,mduser,mdcont,nfe,iflag,iter,printmod
!         real(8) :: xtoler,e(size(x)),fx
!         real(8), allocatable :: work(:)
!         integer, allocatable :: iwork(:)
!         character(len=100) :: nchar
! 
!         n = size(x)
!         write(nchar,'(i100)') n
! 
!         allocate(work(2*n + min(5,n)*(min(5,n)+4)+1))
!         allocate(iwork(n+int(n/min(2,n))))
! 
!         ! Set default parameters and overwrite if optional argument is present
!         xtoler = 1d-6;    if (present(xtoler_in)) xtoler = xtoler_in
!         maxfeval = 10000; if (present(maxfeval_in)) maxfeval = maxfeval_in
!         e = max(x,1d-3);  if (present(e_in)) e = e_in ! initial step sizes
!         printmod = 1;     if (present(printmod_in)) printmod = printmod_in
! 
!         ! Minimize
!         mdsing = 0 ! single step mode is off (0) or on (1)
!         mduser = 0 ! use default options
!         mdcont = 0 ! first call to subplx 
! 
!         do iter = 1,maxfeval
! 
!             mode = 4*mdsing + 2*mduser + mdcont
! 
!             call subplx(g,n,xtoler,iter,mode,e,x,fx,nfe,work,iwork,iflag) ! stop after each evaluation of f
! 
!             mdcont = 1 ! not the first call to subplx anymore
! 
!             if ((printmod>0 .and. iflag>=0) .or. mod(iter,printmod)==0) write(*,'(A,i8,A,f20.12,A,'//trim(nchar)//'f20.12)') 'SUB: iter',iter,' fx',fx,' x',x
!             if (iflag>=0 .or. iflag==-2) exit
!         
!         end do
! 
!         if (printmod>=0) then
!             select case (iflag)
!                 case (-2)
!                     write(*,'(A)') 'SUB: failed to converge: bad input'       
!                 case (-1)
!                     write(*,'(A)') 'SUB: failed to converge: maxfevals reached'       
!                 case (0)
!                     write(*,'(A)') 'SUB: successfully converged: xtoler exceeded'
!                 case (1)
!                     write(*,'(A)') 'SUB: successfully converged: machine precision reached'
!                 case (2)
!                     write(*,'(A)') 'SUB: successfully converged: fstop reached' 
!             end select
!         end if
! 
!     contains
!         function g(n,x)
!             implicit none
!             real(8) :: g
!             integer :: n
!             real(8) :: x(n)
!             g = f(x(1:n))
!         end function
! 
!     end subroutine

    ! Maximizes using a genetic algorithm
    ! Note that this routine is different from the rest because
    ! (1) the search space is integers
    ! (2) it maximizes (as this is more natural in thinking about fitness)
    !
    ! By default, searches in binary.
    ! 
    ! The stopping criterion is measured in terms of the standard deviations of the fitness
    ! in the population. When this falls below toler, execution is terminated and the 
    ! best point is returned.
    ! 
    ! The crossover method can be selected through the cross option
    ! cross             meaning
    ! 0                 single crossover point
    ! 1 (default)       double crossover point
    ! ... others not yet implemented
    subroutine sub_ga(x,f,npop_in,nelite_in,pmute_in,cross_in,toler_in,maxiter_in,printmod_in,&
                        maxfevals_in,openmp_in,mpi_in)
        use mod_matlab
        implicit none
        integer, dimension(:), intent(inout) :: x
        interface 
            function f(x) 
                implicit none
                integer, dimension(:), intent(in) :: x
                real(8) :: f
            end function f
        end interface
        real(8), intent(in), optional :: toler_in,pmute_in
        integer, intent(in), optional :: maxiter_in,printmod_in,maxfevals_in
        integer, intent(in), optional :: npop_in,nelite_in,cross_in
        logical, intent(in), optional :: openmp_in,mpi_in
        !local
        !controls
        integer :: maxiter,printmod,maxfevals,cross
        integer :: npop ! number of members in the population
        integer :: nelite ! number of elite members to carry over unchanged 
        logical :: openmp,mpi

        real(8) :: pmute ! pmute/nx is the probability of a particular gene mutating
        real(8), parameter :: pcross = .65d0 ! probability of two parents crossing
        
        !tools
        integer, dimension(:,:), allocatable :: pop ! population members
        integer, dimension(:,:), allocatable :: parents ! new population members
        integer, dimension(:,:), allocatable :: elite ! elite member
        real(8), dimension(:), allocatable :: fit ! fitness of population members
        real(8), dimension(:), allocatable :: probdraw ! used for drawing survivors
        real(8), dimension(:), allocatable :: fit_elite
        real(8) :: tmp,toler
        integer :: ipop,ix,nx,iter,itmp,itmp2,itmp3,fevals,ivec1(1)

        nx = size(x)        

        !Set defaults
        maxiter = 1000
        npop = 32
        nelite = 1
        pmute = 1d0
        cross = 1
        printmod = 1
        toler = 1d-12
        openmp = .false.
        mpi = .false.
        
        ! Overwrite if present
        if (present(maxiter_in)) maxiter = maxiter_in
        if (present(maxfevals_in)) maxfevals = maxfevals_in
        if (present(printmod_in)) printmod = printmod_in
        if (present(toler_in)) toler = toler_in 
        if (present(openmp_in)) openmp = openmp_in 
        if (present(mpi_in)) mpi = mpi_in  
        if (present(npop_in)) npop = npop_in
        if (present(cross_in)) cross = cross_in
        if (present(nelite_in)) nelite = nelite_in
        if (present(pmute_in)) pmute = pmute_in

        ! Set other defaults
        if (.not. present(maxfevals_in)) maxfevals = 1000*npop

        ! Check inputs
        if (mod(npop,2)/=0) STOP 'npop must be even'
        if (nelite>1) STOP 'nelite must less than or equal 1 or need to reprogram'

        ! Allocate space
        allocate(pop(nx,npop),parents(nx,npop),fit(npop),&
                 probdraw(npop))
        if (nelite>0) allocate(elite(nx,nelite),fit_elite(nelite))

        if (printmod>=0) write(*,'(A)') 'GA:________________________________________________________________________'            
        if (printmod>=0) write(*,'(A)') 'GA: Beginning Genetic Algorithm'            
        if (printmod>=0) write(*,'(A)') 'GA:  iter          best         worst           avg          stdev   fevals'

        !Create initial population by random draws except for input chromosone x
        pop(:,1) = x ! 
        do ipop = 2,npop
            do ix = 1,nx
                call random_number(tmp)
                pop(ix,ipop) = merge(1,0,tmp>.5d0) ! binary
            end do
        end do

        fevals = 0
        if (nelite>0) fit_elite = -huge(0d0)
        
        if (printmod==0) write(*,*) 'GA: Evaluating initial population: '

        ! Iterate
        do iter = 1,maxiter

            ! Evaluate the fitness of the population
            if (.not. openmp .and. .not. mpi) then
                do ipop = 1,npop
                    fit(ipop) = f(pop(:,ipop))
                end do
            elseif (openmp) then
                !$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(ipop)
                do ipop = 1,npop
                    fit(ipop) = f(pop(:,ipop))
                end do
                !$OMP END PARALLEL DO
            elseif (mpi) then
                STOP 'have not programmed mpi yet'
            end if
            fevals = fevals + npop


            ! Check that fitness is all positive
            if (any(fit<0d0)) STOP 'ERROR: GA: fitness is negative but should be positive'

            ! Drop the worst performing completely and add back in elite
            if (nelite==1) then
                if (maxval(fit)>fit_elite(1)) then
                    ivec1 = maxloc(fit)
                    elite(:,1) = pop(:,ivec1(1))
                    fit_elite(1) = fit(ivec1(1))
                else
                    ivec1 = minloc(fit)
                    pop(:,ivec1(1)) = elite(:,1)
                    fit(ivec1(1)) = fit_elite(1)
                end if
            end if

            ! Display some info
            if (printmod==0) call disp('GA: pop fitness: ',fit)
            if (printmod==0 .or. (printmod>0 .and. mod(iter,printmod)==0)) &
                write(*,'(A,i6,A,e12.5,A,e12.5,A,e12.5,A,e12.5,A,i6)') &
                'GA:',iter,'  ', maxval(fit),'  ', minval(fit),'  ', mean(fit),'   ',std(fit),'   ',fevals

            ! Check stopping criteria
            if (iter==maxiter .or. fevals>=maxfevals .or. std(fit)<=toler) then
                print*,'iter',iter,'maxiter',maxiter
                print*,'fevals',fevals,'maxfevals',maxfevals
                print*,'std(fit)',std(fit),'toler',toler
                print*,'fit',fit
                exit
            end if

            ! Generate a new population 
            probdraw = cumsum(fit)/sum(fit)

            ! Roulette selection
            !!! Draw parents according to cumsum(fit)/sum(fit)
            do ipop = 1,npop
                call random_number(tmp)
                itmp = find1(tmp<=probdraw,'first')
                parents(:,ipop) = pop(:,ipop)
            end do

            ! Crossover the parents
            select case (cross)
                case (0)
                    !!! One-point crossover
                    do ipop = 1,npop-1,2

                        call random_number(tmp)
                        if (tmp>pcross) then 

                            ! Don't cross
                            pop(:,ipop) = parents(:,ipop)
                            pop(:,ipop+1) = parents(:,ipop+1)

                        else

                            !!! Draw one crossover 
                            call random_number(tmp)
                            itmp = nint(1d0 + tmp*(nx-1)) ! scale to 1,nx and round to nearest int

                            ! Now have a crossover point, so do it
                            pop(1:itmp,ipop) = parents(1:itmp,ipop)
                            pop(itmp+1:nx,ipop) = parents(itmp+1:nx,ipop+1)

                            pop(1:itmp,ipop+1) = parents(1:itmp,ipop+1)
                            pop(itmp+1:nx,ipop+1) = parents(itmp+1:nx,ipop)

                        end if

                    end do
                case (1)
                    !!! Two-point crossover
                    do ipop = 1,npop-1,2

                        call random_number(tmp)
                        if (tmp>pcross) then 

                            ! Don't cross
                            pop(:,ipop) = parents(:,ipop)
                            pop(:,ipop+1) = parents(:,ipop+1)

                        else

                            !!! Two-point crossover. Draw two crossover points
                            call random_number(tmp)
                            itmp = nint(1d0 + tmp*(nx-1)) ! scale to 1,nx and round to nearest int
                            call random_number(tmp)
                            itmp2 = nint(1d0 + tmp*(nx-1)) ! ditto

                            itmp3 = itmp
                            itmp = min(itmp3,itmp2)
                            itmp2 = max(itmp3,itmp2)

                            ! Now have two points for cross over. They may be equal.
                            pop(1:itmp,ipop) = parents(1:itmp,ipop)
                            pop(itmp+1:itmp2,ipop) = parents(itmp+1:itmp2,ipop+1)
                            pop(itmp2+1:nx,ipop) = parents(itmp2+1:nx,ipop)

                            pop(1:itmp,ipop+1) = parents(1:itmp,ipop+1)
                            pop(itmp+1:itmp2,ipop+1) = parents(itmp+1:itmp2,ipop)
                            pop(itmp2+1:nx,ipop+1) = parents(itmp2+1:nx,ipop+1)

                        end if

                    end do

!                 case (2)
!                     !!! WARNING: this is not actually uniform crossover
!                     !!! Uniform crossover
!                     do ipop = 1,npop-1,2
!                         do ix = 1,nx
!                             !!! W.p. 1/2, swap the gene
!                             call random_number(tmp)
! 
!                             if (tmp>.5d0) then
!                                 pop(ix,ipop) = parents(ix,ipop)
!                                 pop(ix,ipop+1) = parents(ix,ipop+1)
!                             else
!                                 pop(ix,ipop) = parents(ix,ipop+1)
!                                 pop(ix,ipop+1) = parents(ix,ipop)
!                             end if
! 
!                         end do
!                     end do
                case default
                    STOP 'ERROR: crossover case not found'
            end select

            ! Mutate the population
            do ipop = 1,npop
                do ix = 1,nx
                    call random_number(tmp)
                    if (tmp<pmute) pop(ix,ipop) = merge(1,0,pop(ix,ipop)==0) ! mutate
                end do
            end do

        end do

        ! Return the best point and exit
        ivec1 = maxloc(fit)
        x = pop(:,ivec1(1))

!         if (printmod>0 .AND. printmod==1) then
!             write(*,*) 'GA:Final minimum x: '
!             call disp(x)            
!             write(*,*) 'GA:Final minimum f: '
!             write(*,*) 'GA:',minval(f_simplex)
!         end if
        if (printmod>=0) write(*,'(A)') 'GA:________________________________________________________________________'           

    end subroutine sub_ga


!     ! sub_sequentialmin
!     ! Cycles through each dimension and optimizes it before moving onto the next dimension
!     ! Good for problems where one-dimension may be more important than others or the problem
!     ! is mostly independent. This is definitely *not* guaranteed to find a global min or even
!     ! a local min because doing so may require changing x in more than one direction. 
!     ! However, it can still be useful!
!     subroutine sub_sequentialmin(x,f,e_in,shrink_in,toler_in,xtoler_in,maxiter_in,printmod_in)
!         use mod_min1dim
!         use mod_matlab
!         implicit none
!         real(8), dimension(1:), intent(inout) :: x
!         interface 
!             function f(x) 
!                 implicit none
!                 real(8), dimension(1:), intent(in) :: x
!                 real(8) :: f
!             end function f
!         end interface
!         real(8), dimension(1:size(x)), intent(in), optional :: e_in
!         real(8), intent(in), optional :: toler_in, xtoler_in
!         integer, intent(in), optional :: maxiter_in,printmod_in
!         real(8), intent(in), optional :: shrink_in
!         !local
!         real(8) :: toler, xtoler, shrink, lastf, enorm
!         integer :: maxiter,printmod
!         real(8), dimension(1:size(x)) :: e
!         integer :: n,i,iter,flag
!         
!         n = size(x)        
! 
!         !Set defaults
!         maxiter = 1000
!         e = max(.1d0*abs(x),.01d0) !inital step size
!         shrink = .5d0
!         xtoler = 1d-12
!         toler = -huge(0.d0) !stop iterations if f(x)<toler for some x (note that f may be negative)
!         printmod = 1
! 
!         !Overwrite if present
!         if (present(e_in)) e = e_in
!         if (present(shrink_in)) shrink = shrink_in
!         if (present(xtoler_in)) xtoler = xtoler_in
!         if (present(toler_in)) toler = toler_in
!         if (present(maxiter_in)) maxiter = maxiter_in
!         if (present(printmod_in)) printmod = printmod_in
! 
!         if (printmod>0) then
!             write(*,'(A)') 'SM:__________________________________________________________________'            
!             write(*,'(A)') 'SM:Beginning Sequential Minimization Method  '        
!             write(*,'(A)') 'SM:    iter   dim   f(x)   enorm                                  '
!         end if 
! 
!         ! Do the search
!         do iter = 1,maxiter
! 
!             enorm = sqrt(sum(e**2))
!            
!             ! Exit if xtoler exceeded 
!             if (enorm < xtoler) exit 
! 
!             ! Search in each direction
!             do i = 1,n
!                 
!                 ! 
!                 call sub_simplex1dim(x(i),fonedim,e(i),1d-6,10,-1,flag)
! 
!                 ! Update status 
!                 if (printmod>0 .and. (iter==1 .or. mod(iter,printmod)==0)) &
!                     write(*,'(A,2i5,2f13.7)') 'SM:   ',iter,i,lastf,enorm
! 
!             end do
!             
!             ! So it will eventually converge, shrink e
!             e = shrink*e
! 
!         end do
! 
!         if (printmod>0) then
!             write(*,'(A)',advance='no') 'SM: Final x:'
!             write(*,'(99d16.6)',advance='no') x
!             write(*,*)
!             write(*,'(A)',advance='no') 'SM: Final e:'
!             write(*,'(99d16.6)',advance='no') e
!             write(*,*)
!             write(*,'(A)') 'SM:__________________________________________________________________'            
!         end if
! 
!     contains
!         
!         function fonedim(xi)
!             implicit none
!             real(8), intent(in) :: xi
!             real(8) :: fonedim
! 
!             x(i) = xi
!             lastf = f(x) ! Used for stats
!             fonedim = lastf
! 
!         end function 
! 
!     end subroutine sub_sequentialmin


    ! Stochastic Gradient Descent
    ! 
    ! This minimizes a function that can only be measured noisily.
    ! That is, min_x f(x) where f(x) = E(g(x,Z)) and z is a measurement error.
    ! Is essential a noisy version of gradient descent.
    !
    subroutine sub_stochGradDescent(x,f,e_in,a_in,b_in,lambda_in,scaling_in,xtoler_in,maxiter_in,printmod_in)
        use mod_deriv
        use mod_matlab
        implicit none
        real(8), dimension(1:), intent(inout) :: x
        interface 
            function f(x) 
                implicit none
                real(8), dimension(1:), intent(in) :: x
                real(8) :: f
            end function f
        end interface
        real(8), dimension(1:size(x)), intent(in), optional :: e_in,a_in,b_in
        real(8), intent(in), optional :: xtoler_in, scaling_in, lambda_in
        integer, intent(in), optional :: maxiter_in,printmod_in
        !local
        real(8) :: xtoler, f0
        integer :: maxiter,printmod
        real(8), dimension(1:size(x)) :: e, a, b, grad, scaling_init, scaling
        real(8) :: snorm, lambda, lambda_init, tmp
        integer :: n,iter,fevals
        
        n = size(x)        

        ! Set defaults and overwrite if present
        e = max(.1d0*abs(x),.01d0); if (present(e_in)) e = e_in
        a = -huge(0d0);             if (present(a_in)) a = a_in
        b = huge(0d0);              if (present(b_in)) b = b_in
        maxiter = 1000;             if (present(maxiter_in)) maxiter = maxiter_in
        xtoler = 1d-6;              if (present(xtoler_in)) xtoler = xtoler_in
        printmod = 1;               if (present(printmod_in)) printmod = printmod_in
        lambda_init = 1d0;          if (present(lambda_in)) lambda_init = lambda_in
        scaling_init = 1d0;         if (present(scaling_in)) scaling_init = scaling_in

        ! Init
        fevals = 0

        if (printmod>0) then
            write(*,'(A)') 'SGD:__________________________________________________________________'            
            write(*,'(A)') 'SGD:Beginning Stochastic Gradient Descent Method       '            
            write(*,'(A)') 'SGD:    iter  fevals         f(x)        snorm'
        end if 

    
        do iter = 1,maxiter
            
            ! Compute the gradient estimate
            scaling = scaling_init*real(iter,8)**(-1d0/3d0)
            call sub_numGradHess(x,f,scaling,grad,f0=f0)
            fevals = fevals + n + 1
            ! Compute the new point: use line search with a maximum step of lambda
            tmp = lambda_init/real(iter,8)
            call sub_lineSearch(x,lambda,-grad,f,f0,grad,lambda_init_in=tmp,maxiter_in=10)
            ! x = x - lambda*grad

            ! Compute the norm of the step size
            snorm = lambda*sqrt(sum(grad**2))

            ! Update status 
            if (printmod>0 .and. (iter==1 .or. mod(iter,printmod)==0)) &
                write(*,'(A,i5,i8,2f13.7)') 'SGD:   ',iter, fevals, f0, snorm

            ! Exit if converged
            if (snorm<xtoler) exit

        end do 

        if (printmod>0) then
            write(*,'(A)',advance='no') 'SGD: Final x:'
            write(*,'(99d16.6)',advance='no') x
            write(*,*)
            write(*,'(A)') 'SGD:__________________________________________________________________'            
        end if

        
    end subroutine 


    ! Smolyak search
    ! 
    ! Smolyak search computes a smolyak collocation projection of the function f on a hypercube
    ! defined by x +- e. Subsequently, the projection (which is an infinitely differentiable
    ! function) is optimized using derivative based methods. After the projection is optimized,
    ! the hypercube is recentered around the new x. The projection is applied again on the 
    ! hypercube x +- e. If little improvement is made, the hypercube is shrunk: e = shrink*e and
    ! the process begins again (terminated when the L2 norm of e falls below xtoler).
    ! 
    ! The philosophy behind this is that if the underlying function is "roughly smooth" --- 
    ! i.e. almost differentiable but maybe not because of noise --- that we can get a close
    ! approximation by the smolyak function and then optimize that. The smolyak projection (verse
    ! a tensor product approx) is used because it is suitable for high dimension problems.
    ! 
    ! This is totally heuristic --- although I believe it could be formalized. I believe it is
    ! my creation, but it is a fairly obvious.
    !
    ! The user may choose e, shrink, and also "mu", which is the degree of the smolyak 
    ! approximation used. mu should generally be at least 2, because mu==1 has no crossterms
    ! in the approximation. Also, for higher order approximations, there may be a lot of local
    ! minima. Generally, the default value mu==2 is recommended. a,b, if present, are bounds
    ! on the problem that are enforced in both in the smolyak approximation construction and 
    ! the inner maximization (which is the simplex method currently).
    ! 
    subroutine sub_smolyaksearch(x,f,e_in,a_in,b_in,shrink_in,mu_in,xtoler_in,maxiter_in,printmod_in)
        use mod_matlab
        use mod_smolyak
        implicit none
        real(8), dimension(1:), intent(inout) :: x
        interface 
            function f(x) 
                implicit none
                real(8), dimension(1:), intent(in) :: x
                real(8) :: f
            end function f
        end interface
        real(8), dimension(1:size(x)), intent(in), optional :: e_in,a_in,b_in
        real(8), intent(in), optional :: xtoler_in
        integer, intent(in), optional :: maxiter_in,printmod_in
        real(8), intent(in), optional :: shrink_in
        integer, intent(in), optional :: mu_in
        !local
        type(smolyakFcn) :: sApprox ! Smolyak approximation
        real(8) :: xtoler
        integer :: maxiter,printmod
        real(8), dimension(1:size(x)) :: e, xoldbest, xbest, a, b, hcubelb, hcubeub
        real(8) :: shrink, enorm, snorm, fbest, foldbest
        integer :: n,i,iter,fevals,mu,inner, tmpi1(1)
        real(8), allocatable :: coll(:,:),fcoll(:)
        
        n = size(x)        

        ! Set defaults and overwrite if present
        mu = 2;                     if (present(mu_in)) mu = mu_in
        e = max(.1d0*abs(x),.01d0); if (present(e_in)) e = e_in
        a = -huge(0d0);             if (present(a_in)) a = a_in
        b = huge(0d0);              if (present(b_in)) b = b_in
        shrink = .5d0;              if (present(shrink_in)) shrink = shrink_in
        maxiter = 1000;             if (present(maxiter_in)) maxiter = maxiter_in
        xtoler = 1d-6;              if (present(xtoler_in)) xtoler = xtoler_in
        printmod = 1;               if (present(printmod_in)) printmod = printmod_in

        ! Check inputs and try to recover when appropriate
        if (any(e<0d0)) then
            print*,'WARNING: e has some negative, replacing'
            e = abs(e)
        end if
        if (any(e==0d0)) STOP 'SS: ERROR: e is equal to 0'

        ! Init
        fevals = 0

        if (printmod>0) then
            write(*,'(A)') 'SS:__________________________________________________________________'            
            write(*,'(A)') 'SS:Beginning Smolyak Projection Search Method       '            
            write(*,'(A)') 'SS:    iter  fevals    avg.fcoll        fbest        enorm        snorm'
        end if 

        xoldbest = x
        foldbest = f(x) ! inefficient
        fevals = fevals + 1

        ! Do the smolyak search
        OUTER: do iter = 1,maxiter

            ! Exit if xtoler exceeded 
            enorm = sqrt(sum(e**2))
            if (enorm < xtoler) exit 

            ! Inner loop: updating the hypercube center successively 
            do inner = 1,20

                ! Get the smolyak grid
                hcubelb = max(xoldbest-e,a)
                hcubeub = min(xoldbest+e,b)
                call smolyakapprox_step1(coll,sApprox,n,mu,hcubelb,hcubeub)

                ! Evaluate the function at each collocation point
                if (allocated(fcoll)) deallocate(fcoll)
                allocate(fcoll(size(coll,1)))
                do i = 1,size(fcoll)
                    fcoll(i) = f(coll(i,:))
                    fevals = fevals + 1
                end do

                ! Construct the smolyak approximation
                call smolyakapprox_step2(fcoll,sApprox)

                ! Minimize the approximation within the hypercube 
                !!! Start from every collocation point and find the best of the approx
                do i = 1,size(fcoll)
                    x = coll(i,:)                 
                    call sub_simplex(x,fapprox,xtoler_in=1d-11,printmod_in=-1)
                    if (i==1 .or. fapprox(x)<fbest) then
                        xbest = x
                        fbest = fapprox(x)
                    end if
                end do

                ! This is inefficient to do, but for info sake and ease, do it here 
                ! (could save this value since will be a collocation point next time
                fbest = f(xbest)
                fevals = fevals + 1

                ! To ensure we always descend according to the *real* function values rather
                ! than the *approximate* function values, we can do this: 
                ! If this new "xbest" is worse than one of the collocation points we checked,
                ! replace it. 
                ! Note that with this, smolyak is at worst a grid search in several important
                ! dimensions.
                tmpi1 = minloc(fcoll)
                if (fcoll(tmpi1(1))<fbest) then
                    xbest = coll(tmpi1(1),:) 
                    fbest = fcoll(tmpi1(1))
                end if

                snorm = sqrt(sum((xbest-xoldbest)**2))

                xoldbest = xbest
                foldbest = fbest

                ! Update status 
                if (printmod>0 .and. (iter==1 .or. mod(iter,printmod)==0)) &
                    write(*,'(A,i5,i8,4f13.7)') 'SS:   ',iter, fevals, mean(fcoll), fbest, enorm, snorm

                ! Exit inner loop if best x has not changed much
                ! if (xnorm<maxval(abs(xbest))*(1d0 + 1d-3)) exit
                if (snorm<1d2*xtoler) exit

            end do 
                
            ! Shrink the hypercube dimensions
            e = shrink*e

        end do OUTER

        if (printmod>0) then
            write(*,'(A)',advance='no') 'SS: Final x:'
            write(*,'(99d16.6)',advance='no') x
            write(*,*)
            write(*,'(A)',advance='no') 'SS: Final e:'
            write(*,'(99d16.6)',advance='no') e
            write(*,*)
            write(*,'(A)') 'SS:__________________________________________________________________'            
        end if

    contains 
        
        function fapprox(x)
            implicit none
            real(8), intent(in) :: x(1:)
            real(8) :: fapprox
            ! local
            real(8) :: fvec(1),xmat(1,size(x))

            if (any(x<hcubelb) .or. any(x>hcubeub)) then
                fapprox = 1d100
                return
            end if

            xmat(1,:) = x

            ! Evaluate the smolyak approximation at x
            call smolyakapprox_step3(fvec,sApprox,xmat)
            fapprox = fvec(1)

        end function
    
    end subroutine sub_smolyaksearch


    ! Compass search
    ! 
    ! This is a comparison method that searches in each coordinate direction successfully,
    ! shrinking the step size only when no improvement is found. It can be awfully slow,
    ! but it is also reliable --- more so than the simplex algorithm. In my testing on the 
    ! rosenbrock function, SA got stuck far from the global optimium while CS was in the 
    ! right area (but without a precise estimate).
    ! 
    ! The step sizes in each direction are initially given by "e". At each step, the new e
    ! is e*shrink. The routine ceases when the L2 norm of e falls below xtoler or maxiter
    ! is reached.
    ! 
    subroutine sub_compass(x,f,e_in,shrink_in,toler_in,xtoler_in,maxiter_in,printmod_in,usecache_in)
        use mod_matlab
        implicit none
        real(8), dimension(1:), intent(inout) :: x
        procedure(minN_f_template) :: f
        procedure(minN_f_template_w_cache), pointer :: g
        type(minN_f_cache) :: cache
        real(8), dimension(1:size(x)), intent(in), optional :: e_in
        real(8), intent(in), optional :: toler_in, xtoler_in
        integer, intent(in), optional :: maxiter_in,printmod_in
        real(8), intent(in), optional :: shrink_in
        logical, intent(in), optional :: usecache_in
        !local
        real(8) :: toler, xtoler
        integer :: maxiter,printmod
        real(8), dimension(1:size(x)) :: e,xbest,xtrial
        real(8) :: shrink,fbest,ftrial,avgftrial,enorm, fpos, fneg
        integer :: n,i,iter,nimprove,ntrial,fevals
        logical :: usecache
        
        n = size(x)        

        !Set defaults
        maxiter = 1000
        e = max(.1d0*abs(x),.01d0) !inital step size
        shrink = .5d0
        xtoler = 1d-12
        toler = -huge(0.d0) !stop iterations if f(x)<toler for some x (note that f may be negative)
        printmod = 1
        usecache = .true.

        !Overwrite if present
        if (present(e_in)) e = e_in
        if (present(shrink_in)) shrink = shrink_in
        if (present(xtoler_in)) xtoler = xtoler_in
        if (present(toler_in)) toler = toler_in
        if (present(maxiter_in)) maxiter = maxiter_in
        if (present(printmod_in)) printmod = printmod_in
        if (present(usecache_in)) usecache = usecache_in

        ! Point g to either use f with a cache or without one
        if (usecache) then
            ! Create a cache
            call init_cache(cache,size(x),1 + 2*n*maxiter)
            g => f_with_cache
        else
            ! Proceed with a cache
            g => f_without_cache 
        end if

        ! Init
        fevals = 0

        if (printmod>0) then
            write(*,'(A)') 'CS:__________________________________________________________________'            
            write(*,'(A)') 'CS:Beginning Compass Search Method       '            
            write(*,'(A)') 'CS:    iter  fevals #better   avg.ftrial        fbest        xsize'
        end if 

        ! Compute the function value at the initial point
        xbest = x
        fbest = g(x,f,cache)

        ! Do the compass search
        do iter = 1,maxiter

            ! Search in each direction
            nimprove = 0
            avgftrial = 0d0
            ntrial = 0
            do i = 1,n

                ! Try positive direction
                xtrial = xbest
                xtrial(i) = xtrial(i) + e(i)
                fpos = g(xtrial,f,cache)
                avgftrial = avgftrial + fpos
                ntrial = ntrial + 1

                ! Try negative direction
                xtrial = xbest
                xtrial(i) = xtrial(i) - e(i)
                fneg = g(xtrial,f,cache)
                avgftrial = avgftrial + fneg
                ntrial = ntrial + 1
                
                ! Get the best of the two directions
                if (fpos<fneg) then
                    xtrial = xbest
                    xtrial(i) = xtrial(i) + e(i)
                    ftrial = fpos
                else
                    xtrial = xbest
                    xtrial(i) = xtrial(i) - e(i)
                    ftrial = fneg
                end if

                ! Store if better 
                if (ftrial<fbest) then
                    xbest = xtrial
                    fbest = ftrial
                    nimprove = nimprove+1
!                 else
!                     ! Shrink that direction
!                     e(i) = shrink*e(i)
                end if    

            end do

            ! If didn't find a better, shrink
            if (nimprove==0) then
                e = shrink*e
            end if

            ! Compute L2 norm of compass size e
            enorm = sqrt(sum(e**2))

            ! Update status 
            if (printmod>0 .and. (iter==1 .or. mod(iter,printmod)==0 .or. enorm<xtoler)) &
              write(*,'(A,i5,2i8,3f13.7)') 'CS:   ',iter, fevals, nimprove, avgftrial/real(ntrial,8), fbest, enorm
           
            ! Exit if xtoler exceeded 
            if (enorm < xtoler) exit 

        end do

        x = xbest

        if (printmod>0) then
            write(*,'(A)',advance='no') 'CS: Final x:'
            write(*,'(99d16.6)',advance='no') x
            write(*,*)
            write(*,'(A)',advance='no') 'CS: Final e:'
            write(*,'(99d16.6)',advance='no') e
            write(*,*)
            write(*,'(A)') 'CS:__________________________________________________________________'            
        end if

        if (usecache) then
            ! Free cache
            call free_cache(cache)
        end if

    end subroutine sub_compass


    ! sub_lineSearch
    ! 
    ! This is a utility routine that, given a function, initial point (x), and direction (s),
    ! minimizes the function along the direction.
    !
    subroutine sub_lineSearch(x,lambda,s,f,fx,gradx,lambda_init_in,toler_in,xtoler_in,maxiter_in,printmod_in)
        use mod_min1dim
        use mod_matlab
        implicit none
        real(8), dimension(1:), intent(inout) :: x
        real(8), dimension(1:), intent(in) :: s, gradx ! direction and gradient at initial point x
        real(8), intent(in) :: fx ! f at initial point x
        real(8), intent(out) :: lambda ! step size chosen
        interface 
            function f(x) 
                implicit none
                real(8), dimension(1:), intent(in) :: x
                real(8) :: f
            end function f
        end interface
        real(8), intent(in), optional :: toler_in, xtoler_in, lambda_init_in
        integer, intent(in), optional :: maxiter_in,printmod_in
        !local
        integer :: maxiter,printmod,iter,n
        real(8) :: toler, xtoler, gp0, g0, flambda, glambda, bestf, bestlambda, lambda_init
        real(8), parameter :: sig = 2d0, eps = 1d-4

        n = size(x)        

        ! Set defaults and overwrite if present
        maxiter = 40; if (present(maxiter_in)) maxiter = maxiter_in
        xtoler = 1d-6;  if (present(xtoler_in)) xtoler = xtoler_in
        toler = 1d-6;   if (present(toler_in)) toler = toler_in
        printmod = -1;   if (present(printmod_in)) printmod = printmod_in
        lambda_init = 1d0;  if (present(lambda_init_in)) lambda_init = lambda_init_in 

        ! We can use an inexact line search, the Armijo rule, to try to get a good value quickly
        ! Get the total derivative of f at x. This is the derivative of g(lambda) at lambda = 0  
        g0 = fx
        gp0 = dot_product(gradx,s)

        ! Begin with lambda = 1d0, and creep it down until find an acceptable
        lambda = lambda_init 
        bestf = huge(0d0)
        bestlambda = 1d0
        do iter = 1,maxiter
            flambda = f(x+lambda*s)
            glambda = g0 + lambda*eps*gp0
            if (flambda<=glambda) then
                x = x+lambda*s
                return
            else
                lambda = lambda/sig
            end if
            if (flambda<bestf) then
                bestlambda = lambda
            end if
        end do
        
        lambda = bestlambda
        x = x + lambda*s

    end subroutine sub_lineSearch

    ! NOTE: THIS HAS NOT BEEN TESTED MUCH
    ! Broyden-Fletcher-Goldfarb-Shanno (BFGS) method
    ! 
    ! This uses line search with a Hessian estimate.
    ! 
    subroutine sub_BFGS(x,f,scaling_in,toler_in,xtoler_in,maxiter_in,printmod_in,hess_out)
        use mod_deriv
        use mod_matlab
        implicit none
        real(8), dimension(1:), intent(inout) :: x
        interface 
            function f(x) 
                implicit none
                real(8), dimension(1:), intent(in) :: x
                real(8) :: f
            end function f
        end interface
        real(8), intent(in), optional :: scaling_in(1:size(x)),toler_in, xtoler_in
        integer, intent(in), optional :: maxiter_in,printmod_in
        real(8), intent(out), optional :: hess_out(size(x),size(x))
        !local
        integer :: maxiter,printmod,iter,n
        real(8) :: toler, xtoler, f0, snorm, gnorm, lambda
        real(8), allocatable :: x0(:),s(:), hess(:,:), scaling(:), grad(:), grad0(:), y(:)
        
        n = size(x)        
        allocate(x0(n),s(n),grad(n),hess(n,n),grad0(n),y(n),scaling(n))

        !Set defaults and overwrite if present
        maxiter = 1000; if (present(maxiter_in)) maxiter = maxiter_in
        xtoler = 1d-6;  if (present(xtoler_in)) xtoler = xtoler_in
        toler = 1d-6;   if (present(toler_in)) toler = toler_in
        printmod = 1;   if (present(printmod_in)) printmod = printmod_in
        scaling = 1d0;  if (present(scaling_in)) scaling = scaling_in

        if (printmod>0) then
            write(*,'(A)') 'BFGS:__________________________________________________________'            
            write(*,'(A)') 'BFGS: Beginning Broyden-Fletcher-Goldfarb-Shanno Method        '          
            write(*,'(A)') 'BFGS: Computing initial function value and gradient            '         
        end if

        ! Set the initial hessian to do steepest descent
        hess = eye(N)

        ! Compute the initial gradient
        x0 = x
        call sub_numGradHess(x0,f,scaling,grad0,f0=f0)
        gnorm = sum(grad0)**2
        
        if (printmod>0) then
            write(*,'(A)') 'BFGS:  iter          f0       snorm       gnorm      lambda    '
        end if

        ! Do the gradient descent method
        do iter = 1,maxiter
         
            ! Compute the direction using the hessian estimate and the gradient
            s = -mldivide(hess,grad0)

            ! Compute the step via line search and the new x
            x = x0
            call sub_lineSearch(x,lambda,s,f,f0,grad0) ! x overwritten
            s = x-x0 ! Now s is the step (not the direction)
            snorm = sum(s**2)

            ! Display output
            if (printmod>0 .and. mod(iter,printmod)==0) &
                write(*,'(A,i6,5e12.5)') 'BFGS:',iter,f0,snorm,gnorm,lambda

            ! Check for convergence
            if (snorm<xtoler*(1d0+sum(x0**2))) then
                if (printmod>0) then
                    if (gnorm<toler*(1d0+abs(f0))) then
                        write(*,'(A)') 'BFGS: Converged to critical point: gradient is nearly 0'
                    else
                        write(*,'(A)') 'BFGS: Converged to non optimal point: gradient is not nearly 0, but step is small'
                    end if
                end if
                exit
            end if

            ! Update the gradient and hessian
            !!! Compute the new gradient
            call sub_numGradHess(x,f,scaling,grad,f0=f0)
            gnorm = sum(grad**2)
            !!! Update the hessian
            y = grad-grad0
            if (abs(dot_product(y,s))>1d-10 .and. sum(y**2)/dot_product(y,s)<1d10) then
                hess = hess - matmul(vec2mat(matmul(hess,s)),transpose(vec2mat(matmul(s,hess))))/&
                              dot_product(s,matmul(hess,s)) + matmul(vec2mat(y),transpose(vec2mat(y)))/dot_product(y,s)
            else
                ! hess = hess
            end if

            !!! Update the gradient
            grad0 = grad

            ! Update
            x0 = x

            if (iter==maxiter) then
                write(*,'(A)') 'BFGS: Did not converge before max iter reached'
            end if

        end do

        if (present(hess_out)) hess_out = hess

        if (printmod>0) write(*,'(A)') 'BFGS:__________________________________________________________'

    end subroutine sub_BFGS


    ! NOTE: THIS HAS NOT BEEN TESTED MUCH
    ! Gradient descent (GD)
    ! 
    ! Gradient descent method employing a line search
    ! 
    subroutine sub_gradDescent(x,f,scaling_in,toler_in,xtoler_in,maxiter_in,printmod_in)
        use mod_deriv
        use mod_matlab
        implicit none
        real(8), dimension(1:), intent(inout) :: x
        interface 
            function f(x) 
                implicit none
                real(8), dimension(1:), intent(in) :: x
                real(8) :: f
            end function f
        end interface
        real(8), intent(in), optional :: scaling_in(1:size(x)),toler_in, xtoler_in
        integer, intent(in), optional :: maxiter_in,printmod_in
        !local
        integer :: maxiter,printmod,iter,n
        real(8) :: toler, xtoler, f0, snorm, gnorm, lambda
        real(8), allocatable :: x0(:),s(:), hess(:,:), grad(:), scaling(:)
        
        n = size(x)        
        allocate(x0(n),s(n),grad(n),hess(n,n),scaling(n))

        !Set defaults and overwrite if present
        maxiter = 1000; if (present(maxiter_in)) maxiter = maxiter_in
        xtoler = 1d-6;  if (present(xtoler_in)) xtoler = xtoler_in
        toler = 1d-6;   if (present(toler_in)) toler = toler_in
        printmod = 1;   if (present(printmod_in)) printmod = printmod_in
        scaling = 1d0;  if (present(scaling_in)) scaling = scaling_in

        if (printmod>0) then
            write(*,'(A)') 'GS:________________________________________________________'            
            write(*,'(A)') 'GS:Beginning Gradient Descent Method       '          
            write(*,'(A)') 'GS:iter      f0       snorm      gnorm     '
        end if

        ! Do the gradient descent method
        do iter = 1,maxiter
         
            ! Compute the derivative information 
            call sub_numGradHess(x,f,scaling,grad,f0=f0)
            gnorm = sum(grad**2)

! print*,'x is ',x

            ! Compute the step via line search
            x0 = x
            grad = -grad
            call sub_lineSearch(x0,lambda,grad,f,f0,grad) ! x0 overwritten
            grad = -grad
            s = x-x0
            snorm = sum(s**2)

            ! Display output
            if (printmod>0 .and. mod(iter,printmod)==0) &
                write(*,'(A,i6,4e12.5)') 'GS:',iter,f0,snorm,gnorm

            ! Check for convergence
            if (snorm<xtoler*(1d0+sum(x**2))) then
                if (gnorm<toler*(1d0+abs(f0))) then
                    print*,'GS: Converged to critical point: gradient is nearly 0'
                else
                    print*,'GS: Converged to non optimal point: gradient is not nearly 0, but step is small'
                end if
                exit
            end if

            ! Update
            x = x0

            if (iter==maxiter) then
                print*,'GS: Did not converge before max iter reached'
            end if

        end do

        if (printmod>0) write(*,'(A)') 'GS:________________________________________________________'

    end subroutine sub_gradDescent


    ! Newton's method in N dimensions (NN)
    ! 
    ! Pure newton's method, calculating the hessian, without linesearch.
    ! 
    subroutine sub_newtonNdim(x,f,scaling_in,toler_in,xtoler_in,maxiter_in,printmod_in)
        use mod_matlab
        use mod_deriv
        implicit none
        real(8), dimension(1:), intent(inout) :: x
        interface 
            function f(x) 
                implicit none
                real(8), dimension(1:), intent(in) :: x
                real(8) :: f
            end function f
        end interface
        real(8), intent(in), optional :: scaling_in(1:size(x)),toler_in, xtoler_in
        integer, intent(in), optional :: maxiter_in,printmod_in
        !local
        integer :: maxiter,printmod,iter,n
        real(8) :: toler, xtoler, f0, snorm, gnorm
        real(8), allocatable :: s(:), hess(:,:), grad(:), scaling(:)
        
        n = size(x)        
        allocate(s(n),grad(n),hess(n,n),scaling(n))

        !Set defaults and overwrite if present
        maxiter = 1000; if (present(maxiter_in)) maxiter = maxiter_in
        xtoler = 1d-6;  if (present(xtoler_in)) xtoler = xtoler_in
        toler = 1d-6;   if (present(toler_in)) toler = toler_in
        printmod = 1;   if (present(printmod_in)) printmod = printmod_in
        scaling = 1d0;  if (present(scaling_in)) scaling = scaling_in

        if (printmod>0) then
            write(*,'(A)') 'NN:________________________________________________________'            
            write(*,'(A)') 'NN:Beginning Newton''s Method                     '          
            write(*,'(A)') 'NN:  iter          f0       snorm       gnorm     '
        end if

        ! Do Newton's method
        do iter = 1,maxiter
         
            ! Compute the derivative information 
            call sub_numGradHess(x,f,scaling,grad,hess,f0)
            gnorm = sum(grad**2)

            ! Compute the step
            s = -mldivide(hess,grad)
            snorm = sum(s**2)

            ! Display output
            if (printmod>0 .and. mod(iter,printmod)==0) &
                write(*,'(A,i6,4e12.5)') 'NN:',iter,f0,snorm,gnorm

            ! Check for convergence
            if (snorm<xtoler*(1d0+sum(x**2))) then
                if (gnorm<toler*(1d0+abs(f0))) then
                    if (sum(sum(hess,1))<0d0) then
                        print*,'NN: Converged to min'
                    else
                        print*,'NN: Converged to max'
                    end if
                else
                    print*,'NN: Converged to non optimal point: gradient is not nearly 0, but step is small'
                end if
                exit
            end if

            ! Update
            x = x + s

            if (iter==maxiter) then
                print*,'NN: Did not converge before max iter reached'
            end if

        end do

        if (printmod>0) write(*,'(A)') 'NN:________________________________________________________'

    end subroutine sub_newtonNdim


    !Nelder-Mead algorithm
    ! 
    ! I should change that so it just treats anywhere outside of [a,b] as infinite.
    !
    ! This code essentially follows the algorithm in wikipedia
    ! However, I try a small modification where I accept the reflection point if it is better
    ! than the worst point, not the second worse point. This does not affect the volume of the 
    ! simplex, but gets around one pathology I encountered with a non-continuous function.
    ! 
    ! There are two routines here: 
    !   sub_simplex: full featured but slow. Ideal for flexibility, debugging, and calibration. 
    !   sub_simplex_core: not as many options and info, but faster (and so suitable for core routines)
    ! I recently added a new option: "randomize". If this is selected, the initial simplex is 
    ! drawn randomly with x_j = x + randn*e so that is e is the standard deviation of the shock
    !
    subroutine sub_simplex_core(x,f,e,xtoler,maxiter,printmod)
        use mod_sort
        implicit none
        real(8), dimension(1:), intent(inout) :: x
        interface 
            function f(x) 
                implicit none
                real(8), dimension(1:), intent(in) :: x
                real(8) :: f
            end function f
        end interface
        real(8), dimension(1:size(x)), intent(in) :: e
        real(8), intent(in) :: xtoler
        integer, intent(in) :: maxiter,printmod
        !local
        !controls
        real(8), parameter :: alpha = 1.d0, beta = 1.d0, gamma = .5d0, eta = .5d0
        !tools
        real(8), dimension(1:size(x)+1,1:size(x)) :: x_simplex
        real(8), dimension(1:size(x)+1) :: f_simplex
        real(8), dimension(1:size(x)) :: x_gravity,x_reflect,x_expand,x_contract
        real(8) :: f_reflect,f_expand,f_contract
        integer, dimension(1:size(x)+1) :: ranking
        real(8) :: dist
        integer :: n, i, iter
        logical :: doNotTryReflect

        n = size(x)        

        !Create initial simplex and evaluate at simplex corners
        x_simplex(1,:) = x 
        do i = 2,n+1
            x_simplex(i,:) = x
            x_simplex(i,i-1) = x(i-1) + e(i-1)            
        end do

        !Evaluate simplex at corners
        do i = 1,n+1
            f_simplex(i) = f(x_simplex(i,:))
        end do

        if (printmod>0) write(*,'(A)') 'NM:iter     min(f)        max(f)        simplex size'

        !Main routine
        doNotTryReflect = .false.
        do iter = 1,maxiter

            ! Reorder the simplex from smallest to largest (ascending sort)
            !!! Use insertion sort because almost surely this will be a very small list
            call isort(f_simplex,ranking) 
            x_simplex = x_simplex(ranking,:)
 
            ! Output
            if (printmod>0 .AND. mod(iter,printmod)==0) write(*,'(A,i6,A,e12.5,A,e12.5,A,e12.5)') &
                'NM:',iter,'  ', f_simplex(1),'  ', f_simplex(n+1),'  ', dist

            !Check whether length of simplex sides from best point are less than xtoler
            dist = 0.d0
            do i = 2,n+1
                dist = max(dist,sqrt(sum((x_simplex(i,:)-x_simplex(1,:))**2)))                
            end do

            ! Check for convergence or end of loop
            if (dist<xtoler .or. iter==maxiter) then
                if (printmod>0) write(*,'(A,i6,A,e12.5,A,e12.5,A,e12.5)') &
                    'NM:',iter,'  ', f_simplex(1),'  ', f_simplex(n+1),'  ', dist
                ! Finished
                x = x_simplex(1,:)
                return
            end if

            !Try to replace the worst point, n+1, by another better point

            !Calc center of gravity of n other pts
            x_gravity = sum(x_simplex(1:n,:),1)/real(n,8) 

            !Reflect xr = xg + alpha*(xg-x(n+1)) and calculate value
            x_reflect = x_gravity + alpha*(x_gravity-x_simplex(n+1,:))
            if (.not. doNotTryReflect) then
                f_reflect = f(x_reflect)
            else
                f_reflect = huge(0d0)
                doNotTryReflect = .false.
            end if
            
            !Determine what to do with reflected point
            !if (f_simplex(1)<f_reflect .and. f_reflect<f_simplex(n)) then ! The old algorithm had this
            if (f_simplex(1)<f_reflect .and. f_reflect<f_simplex(n+1)) then 
                ! NOTE: next time, reflect will not satisfy this criterion---could optimize by avoiding it.
                ! If this is true, then next time reflect is tried, it will be the old point.
                ! We already know this is worse -- so don't waste an feval!
                if (f_reflect>=f_simplex(n)) then
                    doNotTryReflect = .true.
                end if 

                !Reflect: If f(x1)>f(xr)>f(xn) then replace and continue
                x_simplex(n+1,:) = x_reflect
                f_simplex(n+1) = f_reflect

            elseif (f_reflect>=f_simplex(n)) then

                ! If f_reflect is worst point, generate a new point by contraction
                x_contract = x_gravity + gamma*(x_simplex(n+1,:)-x_gravity)
                f_contract = f(x_contract)

                if (f_contract<f_simplex(n+1)) then

                    ! Contract: found a better point, so use it
                    x_simplex(n+1,:) = x_contract
                    f_simplex(n+1) = f_contract
                
                else
                
                    ! Contract all: no better point found, contract the entire simplex toward the best point
                    do i = 2,n+1                        
                        x_simplex(i,:) = x_simplex(1,:) + eta*(x_simplex(i,:)-x_simplex(1,:))
                        f_simplex(i) = f(x_simplex(i,:))
                    end do

                end if 

            elseif (f_reflect<=f_simplex(1)) then
                
                ! If f_reflect is best point, then try expanding in this direction for further improve
                x_expand = x_reflect + beta*(x_reflect-x_gravity)
                f_expand = f(x_expand)

                if (f_expand<f_reflect) then
                    ! Expand
                    x_simplex(n+1,:) = x_expand
                    f_simplex(n+1) = f_expand
                else
                    ! Reflect
                    x_simplex(n+1,:) = x_reflect
                    f_simplex(n+1) = f_reflect
                end if
            else 
                !Case not found
                STOP 'simplex: case not found'
            end if

        end do

    end subroutine sub_simplex_core
    subroutine sub_simplex(x,f,e_in,a_in,b_in,toler_in,xtoler_in,maxiter_in,printmod_in,randomize_in,contract_in,uid_in)
        use mod_matlab
        implicit none
        real(8), dimension(1:), intent(inout) :: x
        interface 
            function f(x) 
                implicit none
                real(8), dimension(1:), intent(in) :: x
                real(8) :: f
            end function f
        end interface
        real(8), dimension(1:size(x)), intent(in), optional :: e_in,a_in,b_in
        real(8), intent(in), optional :: toler_in, xtoler_in, contract_in
        integer, intent(in), optional :: maxiter_in,printmod_in,uid_in
        logical, intent(in), optional :: randomize_in
        !local
        !controls
        real(8) :: toler, xtoler
        integer :: maxiter,printmod
        real(8), dimension(1:size(x)) :: e
        real(8), parameter :: alpha = 1.d0, beta = 1.d0
        real(8) :: gamma, eta
        
        !tools
        real(8), dimension(1:size(x)+1,1:size(x)) :: x_simplex
        real(8), dimension(1:size(x)+1) :: f_simplex, tmp
        real(8), dimension(1:size(x)) :: x_gravity,x_reflect,x_expand,x_contract
        real(8) :: f_reflect,f_expand,f_contract
        integer, dimension(1:size(x)+1) :: ranking
        real(8) :: dist
        integer :: n, i, iter, fevals, uid
        character(len=20) :: action
        logical :: is_constrained, doNotTryReflect, randomize

        n = size(x)        

        !Set defaults
        maxiter = 1000
        e = max(.1d0*abs(x),.0001d0) !inital step size
        xtoler = 1d-12
        toler = -huge(0.d0) !stop iterations if f(x)<toler for some x (note that f may be negative)
        printmod = 1
        randomize = .false.
        gamma = .5d0 ! contraction param
        eta = .5d0 ! contraction param      
        uid = 6 ! stdout

        !Overwrite if present
        if (present(e_in)) e = e_in
        if (present(xtoler_in)) xtoler = xtoler_in
        if (present(toler_in)) toler = toler_in
        if (present(maxiter_in)) maxiter = maxiter_in
        if (present(printmod_in)) printmod = printmod_in
        if (present(randomize_in)) randomize = randomize_in
        if (present(uid_in)) uid = uid_in
        if (present(contract_in)) then
            gamma = (1d0-contract_in)
            eta = (1d0-contract_in)
            if (contract_in<=0d0 .or. contract_in>=1d0) then
                STOP 'ERROR: sub_simplex: contract_in invalid'
            end if
        end if

        if (printmod>0) write(uid,'(A)') 'NM:________________________________________________________'            
        if (printmod>0) write(uid,'(A)') 'NM:Beginning Simplex Method (Nelder-Mead)'            

        is_constrained = present(a_in) .and. present(b_in)

        !Create initial simplex and evaluate at simplex corners
        x_simplex(1,:) = x 
        if (.not. randomize) then
            do i = 2,n+1
                x_simplex(i,:) = x
                x_simplex(i,i-1) = x(i-1) + e(i-1)            
            end do
        else
            do i = 2,n+1
                x_simplex(i,:) = x + randn(n)*e
            end do
        end if

        !Evaluate simplex at corners
        do i = 1,n+1
            if (.not. is_constrained) then
                f_simplex(i) = f(x_simplex(i,:))
            else
                if (any(x_simplex(i,:)<a_in .or. x_simplex(i,:)>b_in)) then
                    f_simplex(i) = huge(0d0)
                else
                    f_simplex(i) = f(x_simplex(i,:))
                end if
            end if
        end do
        fevals = n+1
    
        if (printmod>0 .AND. printmod==1) then
            write(uid,'(A)') 'NM:Initial simplex: '
            do i = 1,n+1
                write(uid,'(A,i4,A,99(e12.5))') 'NM: i ',i,' x ',x_simplex(i,:)
            end do
            write(uid,'(A)') 'NM: Initial simplex values: '
            write(uid,'(A,99(e12.5))') 'NM: ',f_simplex
        end if

        dist = -1d0

        !Main routine
        if (printmod>0) write(uid,'(A)') 'NM:iter     min(f)        max(f)        simplex size  fevals   action    '            
        action = 'Init'
        doNotTryReflect = .false.
        do iter = 1,maxiter

            !Reorder the simplex from smallest to largest (ascending sort)
            call sort2(tmp,ranking,f_simplex)
            f_simplex = f_simplex(ranking)
            x_simplex = x_simplex(ranking,:)
            
! print*,'f_simplex',f_simplex
! print*,'x_simplex'
! call disp(x_simplex)

            if (printmod>0 .AND. mod(iter,printmod)==0) write(uid,'(A,i6,A,e12.5,A,e12.5,A,e12.5,A,i6,2A)') &
                'NM:',iter,'  ', f_simplex(1),'  ', f_simplex(n+1),'  ', dist,'   ',fevals,'   ', action

            !Check whether length of simplex sides from best point are less than xtoler
            dist = 0.d0
            do i = 2,n+1
                dist = max(dist,sqrt(sum((x_simplex(i,:)-x_simplex(1,:))**2)))                
            end do
            if (dist<xtoler) then
                if (printmod>0) &
                    write(uid,'(A,i6,A,e12.5,A,e12.5,A,e12.5,A,i6,2A)') &
                    'NM:',iter,'  ', minval(f_simplex),'  ', maxval(f_simplex),'  ', dist,'   ',fevals,'   ', action
                if (printmod>0) &
                    write(uid,'(A)') 'NM:Simplex finished: simplex size less than xtoler'
                exit
            end if
            if (f_simplex(1)<toler) then
                if (printmod>0) &
                    write(uid,'(A,i6,A,e12.5,A,e12.5,A,e12.5,A,i6,2A)') &
                    'NM:',iter,'  ', minval(f_simplex),'  ', maxval(f_simplex),'  ', dist,'   ',fevals,'   ', action
                if (printmod>0) &
                    write(uid,'(A)') 'NM:Simplex finished: f is less than toler'
                exit
            end if            

            !Try to replace the worst pt, N+1, by another better point

            !Calc center of gravity of n other pts
            x_gravity = sum(x_simplex(1:n,:),1)/real(n,8) 

            !Reflect xr = xg + alpha*(xg-x(n+1)) and calculate value
            x_reflect = x_gravity + alpha*(x_gravity-x_simplex(n+1,:))
! print*,'x_reflect', x_reflect            
            if (.not. doNotTryReflect) then
                if (.not. is_constrained) then
                    f_reflect = f(x_reflect)
                else
                    if (any(x_reflect<a_in .or. x_reflect>b_in)) then
                        f_reflect = huge(0d0)
                    else
                        f_reflect = f(x_reflect)
                    end if
                end if
! print*,'f_reflect', f_reflect            
                fevals = fevals+1
            else
                f_reflect = huge(0d0)
                doNotTryReflect = .false.
            end if
            
            !Determine what to do with reflected point
            !if (f_simplex(1)<f_reflect .and. f_reflect<f_simplex(n)) then ! The old algorithm had this
            if (f_simplex(1)<f_reflect .and. f_reflect<f_simplex(n+1)) then 
                ! NOTE: next time, reflect will not satisfy this criterion---could optimize by avoiding it.
                ! If this is true, then next time reflect is tried, it will be the old point.
                ! We already know this is worse -- so don't waste an feval!
                if (f_reflect>=f_simplex(n)) then
                    doNotTryReflect = .true.
                end if 
                action = 'Reflected'
                !If f(x1)>f(xr)>f(xn) then replace and continue
                x_simplex(n+1,:) = x_reflect
                f_simplex(n+1) = f_reflect

            elseif (f_reflect>=f_simplex(n)) then
                !If f_reflect is worst point, generate a new point by contraction
                x_contract = x_gravity + gamma*(x_simplex(n+1,:)-x_gravity)
                if (.NOT. is_constrained) then
                    f_contract = f(x_contract)
                else
                    if (any(x_contract<a_in .or. x_contract>b_in)) then
                        f_contract = huge(0d0)
                    else
                        f_contract = f(x_contract)
                    end if
                end if
                fevals = fevals+1
                if (f_contract<f_simplex(n+1)) then
                    action = 'Contracted'
                    !If found a better point, then use it
                    x_simplex(n+1,:) = x_contract
                    f_simplex(n+1) = f_contract
                else
                    !If no better point found, contract the entire simplex toward the best point
                    action = 'Contracted All'
                    do i = 2,n+1                        
                        x_simplex(i,:) = x_simplex(1,:) + eta*(x_simplex(i,:)-x_simplex(1,:))
                        !f_simplex(i) = f(x_simplex(i,:))
                        if (.not. is_constrained) then
                            f_simplex(i) = f(x_simplex(i,:))
                        else
                            if (any(x_simplex(i,:)<a_in .or. x_simplex(i,:)>b_in)) then
                                f_simplex(i) = huge(0d0)
                            else
                                f_simplex(i) = f(x_simplex(i,:))
                            end if
                        end if
                        fevals = fevals+1
                    end do
                end if 
            elseif (f_reflect<=f_simplex(1)) then
                
                !If f_reflect is best point, then try expanding in this direction for further improve
                x_expand = x_reflect + beta*(x_reflect-x_gravity)
                !f_expand = f(x_expand)
                if (.not. is_constrained) then
                    f_expand = f(x_expand)
                else
                    if (any(x_expand<a_in .or. x_expand>b_in)) then
                        f_expand = huge(0d0)
                    else
                        f_expand = f(x_expand)
                    end if
                end if
                fevals = fevals+1
                if (f_expand<f_reflect) then
                    action = 'Expanded'
                    x_simplex(n+1,:) = x_expand
                    f_simplex(n+1) = f_expand
                else
                    action = 'Reflected'
                    x_simplex(n+1,:) = x_reflect
                    f_simplex(n+1) = f_reflect
                end if
            else 
                !Case not found
                print*,'f_reflect:',f_reflect
                print*,'f_simplex:',f_simplex
                STOP 'simplex: case not found'
            end if

        end do
 
        !Set out
        x = x_simplex(minloc(f_simplex,1),:)

        if (printmod>0 .AND. printmod==1) then
            write(uid,'(A)') 'NM: Final simplex: '
            do i = 1,n+1
                write(uid,'(A,i4,A,99(e12.5))') 'NM: i ',i,' x ',x_simplex(i,:)
            end do
            write(uid,'(A)') 'NM: Final simplex values: '
            write(uid,'(A,99(e12.5))') 'NM: ',f_simplex
            write(uid,'(A)') 'NM:Final minimum x: '
            write(uid,'(A,99(e12.5))') 'NM: ',x
            write(uid,'(A)') 'NM:Final minimum f: '
            write(uid,'(A,e12.5)') 'NM:',minval(f_simplex)
        end if
        if (printmod>0) write(uid,'(A)') 'NM:________________________________________________________'

    end subroutine sub_simplex

    subroutine sub_parallelsimplex(x,f,e_in,toler_in,xtoler_in,maxiter_in,printmod_in,randomize_in,contract_in)
        use mod_mympi
        use mod_matlab
        implicit none
        real(8), dimension(1:), intent(inout) :: x
        interface 
            function f(x) 
                implicit none
                real(8), dimension(1:), intent(in) :: x
                real(8) :: f
            end function f
        end interface
        real(8), dimension(1:size(x)), intent(in), optional :: e_in
        real(8), intent(in), optional :: toler_in, xtoler_in, contract_in
        integer, intent(in), optional :: maxiter_in,printmod_in
        logical, intent(in), optional :: randomize_in
        !local
        !controls
        real(8) :: toler, xtoler
        integer :: maxiter,printmod
        real(8), dimension(1:size(x)) :: e
        real(8), parameter :: alpha = 1.d0, beta = 1.d0
        real(8) :: gamma, eta
        
        !tools
        real(8), dimension(1:size(x)+1,1:size(x)) :: x_simplex
        real(8), dimension(1:size(x)+1) :: f_simplex, tmp
        real(8), dimension(1:size(x)) :: x_gravity,x_reflect,x_expand,x_contract
        real(8) :: f_reflect,f_expand,f_contract
        integer, dimension(1:size(x)+1) :: ranking
        real(8) :: dist
        integer :: n, i, iter, fevals, ip, np, ilo, ihi
        logical :: randomize
        logical, dimension(1:size(x)+1) :: didImprove
        real(8), dimension(1:size(x),1:size(x)+1) :: x_simplex_transpose

        n = size(x)        

        ! Get number of nodes
        np = min(n,mympi_nproc())

        !Set defaults
        maxiter = 1000
        e = max(.1d0*abs(x),.0001d0) !inital step size
        xtoler = 1d-12
        toler = -huge(0.d0) !stop iterations if f(x)<toler for some x (note that f may be negative)
        printmod = 1
        randomize = .false.
        gamma = .5d0 ! contraction param
        eta = .5d0 ! contraction param      

        !Overwrite if present
        if (present(e_in)) e = e_in
        if (present(xtoler_in)) xtoler = xtoler_in
        if (present(toler_in)) toler = toler_in
        if (present(maxiter_in)) maxiter = maxiter_in
        if (present(printmod_in)) printmod = printmod_in
        if (present(randomize_in)) randomize = randomize_in
        if (present(contract_in)) then
            gamma = (1d0-contract_in)
            eta = (1d0-contract_in)
            if (contract_in<=0d0 .or. contract_in>=1d0) then
                STOP 'ERROR: sub_simplex: contract_in invalid'
            end if
        end if

        if (randomize) STOP 'ERROR: sub_parallelsimplex: not sure randomization is okay across multiple processes'

        printmod = merge(printmod,-1,mympi_iproc()==0) 

        if (printmod>0) write(*,'(A)') 'PNM:________________________________________________________'            
        if (printmod>0) write(*,'(A)') 'PNM:Beginning Parallel Simplex Method (Nelder-Mead)'            

        !Create initial simplex and evaluate at simplex corners
        x_simplex(1,:) = x 
        if (.not. randomize) then
            do i = 2,n+1
                x_simplex(i,:) = x
                x_simplex(i,i-1) = x(i-1) + e(i-1)            
            end do
        else
            do i = 2,n+1
                x_simplex(i,:) = x + randn(n)*e
            end do
        end if

        !Evaluate simplex at corners
        do i = mympi_a(n+1),mympi_b(n+1)
            f_simplex(i) = f(x_simplex(i,:))
        end do
        call mympi_allgatherv(f_simplex)

        fevals = n+1
    
        if (printmod>0 .AND. printmod==1) then
            write(*,'(A)') 'PNM:Initial simplex: '
            do i = 1,n+1
                write(*,'(A,i4,A,99(e12.5))') 'PNM: i ',i,' x ',x_simplex(i,:)
            end do
            write(*,'(A)') 'PNM: Initial simplex values: '
            write(*,'(A,99(e12.5))') 'PNM: ',f_simplex
        end if

        dist = -1d0

        !Main routine
        if (printmod>0) write(*,'(A)') 'PNM:iter     min(f)        max(f)        simplex size  '            
        do iter = 1,maxiter

            !Reorder the simplex from smallest to largest (ascending sort)
            call sort2(tmp,ranking,f_simplex)
            f_simplex = f_simplex(ranking); x_simplex = x_simplex(ranking,:)

            if (printmod>0 .AND. mod(iter,printmod)==0) write(*,'(A,i6,A,e12.5,A,e12.5,A,e12.5)') 'PNM:',iter,'  ', f_simplex(1),'  ', f_simplex(n+1),'  ', dist

            !Check whether length of simplex sides from best point are less than xtoler
            dist = 0.d0
            do i = 2,n+1
                dist = max(dist,sqrt(sum((x_simplex(i,:)-x_simplex(1,:))**2)))                
            end do
            if (dist<xtoler) then
                if (printmod>0) &
                    write(*,'(A,i6,A,e12.5,A,e12.5,A,e12.5)') 'PNM:',iter,'  ', minval(f_simplex),'  ', maxval(f_simplex),'  ', dist
                if (printmod>0) &
                    write(*,'(A)') 'PNM:Simplex finished: simplex size less than xtoler'
                exit
            end if
            if (f_simplex(1)<toler) then
                if (printmod>0) &
                    write(*,'(A,i6,A,e12.5,A,e12.5,A,e12.5)') 'PNM:',iter,'  ', minval(f_simplex),'  ', maxval(f_simplex),'  ', dist
                if (printmod>0) &
                    write(*,'(A)') 'PNM:Simplex finished: f is less than toler'
                exit
            end if            

            !Calc center of gravity of n other pts
            x_gravity = sum(x_simplex(1:n+1-np,:),1)/real(n+1-np,8) 

            !!!!! PARALLEL ALGORITHM !!!!!!
            ! Try to replace the worst P points, N-P,N-P+1,...,N+1, with other points.
            do ip = n+1-np+ mympi_a(np),n+1-np+ mympi_b(np)

                !Reflect xr = xg + alpha*(xg-x(n+1)) and calculate value
                x_reflect = x_gravity + alpha*(x_gravity-x_simplex(ip,:))
                f_reflect = f(x_reflect)
                
                !Determine what to do with reflected point
                if (f_reflect<=f_simplex(1)) then
                    
                    !If f_reflect is best point, then try expanding in this direction for further improve
                    x_expand = x_reflect + beta*(x_reflect-x_gravity); f_expand = f(x_expand)

                    if (f_expand<f_reflect) then
                        x_simplex(ip,:) = x_expand;  f_simplex(ip) = f_expand
                    else
                        x_simplex(ip,:) = x_reflect; f_simplex(ip) = f_reflect
                    end if

                    didImprove(ip) = .true.

                elseif (f_simplex(1)<f_reflect .and. f_reflect<f_simplex(n+1-np)) then 

                    x_simplex(ip,:) = x_reflect; f_simplex(ip) = f_reflect
                    didImprove(ip) = .true.

                elseif (f_reflect>=f_simplex(n+1-np)) then

                    !If f_reflect is worst point, generate a new point by contraction
                    if (f_reflect>=f_simplex(ip)) then
                        x_reflect = x_simplex(ip,:); f_reflect = f_simplex(ip)
                    end if

                    x_contract = x_gravity + gamma*(x_reflect-x_gravity)
                    f_contract = f(x_contract)

                    if (f_contract<f_reflect) then
                        !If found a better point, then use it
                        x_simplex(ip,:) = x_contract; f_simplex(ip) = f_contract
                        didImprove(ip) = .true.
                    else
                        ! These are not actually the reflection points, but the original point
                        x_simplex(ip,:) = x_reflect; f_simplex(ip) = f_reflect  
                        didImprove(ip) = .false.
                    end if 
                else 
                    !Case not found
                    print*,'f_reflect:',f_reflect
                    print*,'f_simplex:',f_simplex
                    STOP 'simplex: case not found'
                end if

            end do

            ! Now every process has updated a simplex point. Use all gather to collect them
            ilo = n+1-np+1
            ihi = n+1
            call mympi_allgatherv(f_simplex(ilo:ihi))
            x_simplex_transpose = transpose(x_simplex)
            call mympi_allgatherv(x_simplex_transpose(:,ilo:ihi))
            x_simplex = transpose(x_simplex_transpose)
            call mympi_allgatherv(didImprove(ilo:ihi))

            if (.not. any(didImprove(ilo:ihi))) then

                ! Contract all
                do i = 2,n+1
                    x_simplex(i,:) = x_simplex(1,:) + eta*(x_simplex(i,:)-x_simplex(1,:))
                end do

                ! Evaluate at the new simplex 
                do i = 1+mympi_a(n),1+mympi_b(n)
                    f_simplex(i) = f(x_simplex(i,:))
                end do
                call mympi_allgatherv(f_simplex(2:n+1))

            end if

        end do
 
        !Set out
        x = x_simplex(minloc(f_simplex,1),:)

        if (printmod>0 .AND. printmod==1) then
            write(*,'(A)') 'PNM: Final simplex: '
            do i = 1,n+1
                write(*,'(A,i4,A,99(e12.5))') 'PNM: i ',i,' x ',x_simplex(i,:)
            end do
            write(*,'(A)') 'PNM: Final simplex values: '
            write(*,'(A,99(e12.5))') 'PNM: ',f_simplex
            write(*,'(A)') 'PNM:Final minimum x: '
            write(*,'(A,99(e12.5))') 'PNM: ',x
            write(*,'(A)') 'PNM:Final minimum f: '
            write(*,'(A,e12.5)') 'PNM:',minval(f_simplex)
        end if
        if (printmod>0) write(*,'(A)') 'PNM:________________________________________________________'

    end subroutine sub_parallelsimplex

    !Shrinking hypercube
    ! Starting with an initial guess and a hypercube centered on that guess, 
    ! this draws a random point from within the hypercube.
    ! If the value is better (less) than the current value, the hypercube edges are doubled
    ! If the value is worse (greater), the hypercube edges are decreased by a small amount
    ! 
    ! I originally got this idea from the web. It is discussed in http://linneus20.ethz.ch:8080/4_4_2.html
    ! which looks like it was copied from somewhere. If you know where the original article is from, please
    ! let me know and I'll credit appropriately.
    subroutine sub_shrinkingcube(x,f,e_in,maxiter_in,printmod_in,xtoler_in,shrinkFactor_in)
        implicit none
        real(8), dimension(1:), intent(inout) :: x
        interface 
            function f(x) 
                implicit none
                real(8), dimension(1:), intent(in) :: x
                real(8) :: f
            end function f
        end interface
        real(8), dimension(1:size(x)), intent(in), optional :: e_in
        integer, intent(in), optional :: maxiter_in,printmod_in
        real(8), intent(in), optional :: xtoler_in, shrinkFactor_in
        !local
        integer :: maxiter, printmod, iter
        real(8), dimension(1:size(x)) :: e, bestx, draw
        real(8) :: xtoler, fx, bestf, shrinkFactor
        integer :: seedSize,seed(100)

        ! Get the current state of the seed and save it to not interfere with a main program
        call random_seed(size=seedSize)
        if (seedSize>size(seed)) STOP 'sub_shrinkingcube: must increase seed size'
        call random_seed(get=seed(1:seedSize))        

        !Set default values
        e = .02d0*max(abs(x),1d-5) !the size of neighborhood about each point which is a 
                                   !hypersphere stretched to edge length e in each dimension
        maxiter = 1000
        printmod = 1
        xtoler = 1d-8
        shrinkFactor = .99d0 !Default suggested on website

        !Overwrite if applicable
        if (present(maxiter_in)) maxiter = maxiter_in
        if (present(printmod_in)) printmod = printmod_in
        if (present(e_in)) e = e_in
        if (present(xtoler_in)) xtoler = xtoler_in
        if (present(shrinkFactor_in)) shrinkFactor = shrinkFactor_in

        if (printmod>0) write(*,'(A)') 'SH: ________________________________________________________'            
        if (printmod>0) write(*,'(A)') 'SH: Beginning Shrinking Hypercube'            

        ! Initialize
        fx = f(x)
        bestf = fx
        bestx = x        

        ! 
        do iter = 1,maxiter

            if (printmod>0) write(*,'(A,i10,A,d12.5,A,d12.5,A,d12.5)') &
                'SH: iter',iter, ' f',fx,' bestf',bestf, ' cubeVol', product(e)

            ! If a cube edge is less than xtoler stop
            if (any(e<xtoler)) exit
        
            ! Take random draw from within cube
            call random_number(draw)
            draw = draw*2.d0-1.d0 !now uniform in [-.5,.5]
            x = bestx + e*draw

            ! Evalute current point
            fx = f(x)

            ! If best save it and expand, otherwise shrink it
            if (fx<bestf) then
                bestf = fx
                bestx = x
                e = e*2.d0
            else
                e = e*shrinkFactor
            end if

        end do

        x = bestx

        if (printmod>0 .AND. printmod==1) write(*,'(A)') 'SH: Final minimum x: '
        if (printmod>0 .AND. printmod==1) write(*,*) 'SH:', x !call disp(x)            
        if (printmod>0 .AND. printmod==1) write(*,'(A)') 'SH: Final minimum f: '
        if (printmod>0 .AND. printmod==1) write(*,'(A,d12.5)') 'SH:',bestf
        if (printmod>0) write(*,'(A)') 'SH:________________________________________________________'

    end subroutine sub_shrinkingcube

    ! Simulated Annealing (SA)
    ! 
    ! Algorithm description: 
    ! Begins at x and draws random normal innovations to generate x'. x' is accepted whenever f(x')<f(x)
    ! Even when f(x')>f(x), it still accepted according to the probability that U < exp((fx-f(x'))/T).
    ! The temperature T "cools" slowly trending to 0. The innovations are drawn with a decreasing variance.
    ! 
    ! The temperature begins at T0 and cools with T(k) = T0/(1+log(k)).
    ! The variance begins at Var(0)=e**2 and decreases with Var(k) = Var(0)/(1+log(k)).
    !
    ! Global convergence is guaranteed for T0 "large enough". Unfortunately, this is not really helpful.
    ! Part of the program output is an acceptance rate. One may try to tune T0 and e so that this is not much
    ! more than 30% but also not 0%. Also, the "currf" and "bestf" should not be too far apart
    ! 
    ! Optional input:
    ! e: initial standard deviation of the shocks to x.
    ! T0: initial temperature.
    ! a: lower bound on parameter space
    ! b: upper bound on parameter space
    ! maxiter: self-explanatory
    ! printmod: self-explanatory
    ! 
    subroutine sub_simanneal(x,f,e_in,a_in,b_in,T0_in,maxiter_in,printmod_in,aggressive_in)
        use mod_matlab
        implicit none
        real(8), dimension(1:), intent(inout) :: x
        interface 
            function f(x) 
                implicit none
                real(8), dimension(1:), intent(in) :: x
                real(8) :: f
            end function f
        end interface
        real(8), dimension(1:size(x)), intent(in), optional :: e_in,a_in,b_in
        real(8), intent(in), optional :: T0_in
        integer, intent(in), optional :: maxiter_in,printmod_in
        logical, intent(in), optional :: aggressive_in
        !local
        real(8), parameter :: fracEps = .02d0
        integer :: maxiter, printmod
        real(8) :: T,T0
        real(8), dimension(1:size(x)) :: a, b, e, y, xbest
        real(8) :: fx,fy,fxbest
        integer :: seedSize,seed(100)
        integer :: n, iter, naccept, nreject
        real(8) :: randnDraw(size(x)), randuDraw
        logical :: aggressive

        n = size(x)

        ! Check seedSize is large enough
        call random_seed(size=seedSize)
        if (seedSize>size(seed)) STOP 'sub_simanneal: must increase seed size'

        !Set defaults
        maxiter = 10000
        printmod = 1
        a = -huge(0.d0) !the lowerbound on the state space
        b = huge(0.d0)  !the upperbound on the state space
        e = .01d0*max(abs(x),1d-4) !the size of neighborhood about each point which is a hypersphere stretched to diameter e in each dimension
        T0 = 1d0 !temperature scaling  (setting T0 sufficiently ensures a.s. convergence)
        aggressive = .false.

        !Overwrite if present
        if (present(maxiter_in)) maxiter = maxiter_in
        if (present(printmod_in)) printmod = printmod_in
        if (present(a_in)) a = a_in
        if (present(b_in)) b = b_in
        if (present(e_in)) e = e_in
        if (present(T0_in)) T0 = T0_in
        if (present(aggressive_in)) aggressive = aggressive_in

        ! Check inputs
        if (any(x<a) .or. any(x>b)) STOP 'sub_simanneal: x was not initialized to be between a and b!'

        ! Evaluate f at x
        fx = f(x)

        ! Initialize best values
        fxbest = fx
        xbest = x

        naccept = 0
        nreject = 0

        if (printmod>0) then
            write(*,'(A)') 'SA: ________________________________________________________'            
            write(*,'(A)') 'SA: Beginning Simulated Annealing Method '            
        end if
        do iter = 1,maxiter

!             ! Save the seed state, draw a random normal (size n) and random uniform 
!             !!! Get the current state of the seed and save it to not interfere with a main program
!             call random_seed(get=seed(1:seedSize))
            !!! Draw the variables
            randnDraw = randn(n)
            call random_number(randuDraw)
!             !!! Return the state of the seed to what it was before sim annealing was called
!             call random_seed(put=seed(1:seedSize))

            ! Set the current temperature
            T = T0/(1d0+log(real(iter,8))) ! T0 to 0
! T = T0/(1d0+real(iter-1,8)) ! T0 to 0

            ! Draw a newpoint within the neighborhood of current x
            !!! I saw here http://stats.stackexchange.com/questions/27923/should-the-proposal-distribution-in-simulated-annealing-depend-on-the-temperatur
            !!! that can draw from this variance decreasing in T and will still get convergence (for large enough T0)
            if (.not. aggressive) then
                y = x + e*randnDraw*sqrt(T/T0) ! Initial innovations are drawn from N(0,e**2). Subsquent, N(0,e**2/(1+log(k))**2)
            else
                y = x + e*randnDraw*T/T0 ! Initial innovations are drawn from N(0,e**2). Subsquent, N(0,e**2/(1+log(k))**2)
            end if


            ! Bound y so it remains in the parameter space
            y = min(max(y,a),b)

            ! Evaluate f at y 
            fy = f(y)

            ! Determine accept reject
            if (fy<fx) then 
                ! In this variant we always accept better moves
                x = y
                fx = fy
                naccept = naccept + 1
            else
                if (randuDraw<exp((fx-fy)/T)) then
                    ! Move from x to y
                    x = y
                    fx = fy
                    naccept = naccept + 1
                else
                    ! Keep current x
                    nreject = nreject + 1
                end if
            end if

            ! Try
            if (printmod>0 .and. mod(iter,printmod)==0) then
                write(*,'(A,i6,A,d12.5,A,d12.5,A,d12.5,A,d12.5,A,d12.5)') &
                    'SA: iter ',iter,' ftrial',fy,' fcurr',fx,' fbest',fxbest,' Temp', T,&
                    ' Accept%', 100.d0*real(naccept,8)/real(naccept+nreject,8)
            end if

            ! Save best f
            if (fx<fxbest) then 
                fxbest = fx
                xbest = x
            end if

        end do

        ! Return the best function value
        if (printmod>0) then
            if (fxbest/=fx) then
                write(*,'(A,d12.5,A)') 'SA: best value ever encountered, ',fxbest,', being returned, it is /= fx'
            end if
        end if
        x = xbest
        fx = fxbest

    end subroutine sub_simanneal

    ! Parallel Simulated Annealing (PSA)
    ! 
    ! All the argumetns are the same as the regular simulated annealing. 
    ! The algorithm is the same except that at each iteration, P proposals are drawn at the same time.
    ! The best of the proposals is taken. 
    ! 
    subroutine sub_parallelsimanneal(x,f,e_in,a_in,b_in,T0_in,maxiter_in,printmod_in,aggressive_in)
        use mod_mympi
        use mod_matlab
        implicit none
        real(8), dimension(1:), intent(inout) :: x
        interface 
            function f(x) 
                implicit none
                real(8), dimension(1:), intent(in) :: x
                real(8) :: f
            end function f
        end interface
        real(8), dimension(1:size(x)), intent(in), optional :: e_in,a_in,b_in
        real(8), intent(in), optional :: T0_in
        integer, intent(in), optional :: maxiter_in,printmod_in
        logical, intent(in), optional :: aggressive_in
        !local
        real(8), parameter :: fracEps = .02d0
        integer :: maxiter, printmod, n, iter, naccept, nreject, np, ip, itmp(1), seedSize, seed(100)
        real(8) :: fx,fy,fxbest,T,T0,randuDraw
        real(8), dimension(1:size(x)) :: a, b, e, y, xbest
        real(8), allocatable :: z(:,:),fz(:)
        logical :: aggressive

        n = size(x)
        np = mympi_nproc()

        allocate(z(n,np),fz(np))

        ! Check seedSize is large enough
        call random_seed(size=seedSize)
        if (seedSize>size(seed)) STOP 'sub_simanneal: must increase seed size'

        !Set defaults
        maxiter = 10000
        printmod = 1
        a = -huge(0.d0) !the lowerbound on the state space
        b = huge(0.d0)  !the upperbound on the state space
        e = .01d0*max(abs(x),1d-4) !the size of neighborhood about each point which is a hypersphere stretched to diameter e in each dimension
        T0 = 1d0 !temperature scaling  (setting T0 sufficiently ensures a.s. convergence)
        aggressive = .false.

        !Overwrite if present
        if (present(maxiter_in)) maxiter = maxiter_in
        if (present(printmod_in)) printmod = printmod_in
        if (present(a_in)) a = a_in
        if (present(b_in)) b = b_in
        if (present(e_in)) e = e_in
        if (present(T0_in)) T0 = T0_in
        if (present(aggressive_in)) aggressive = aggressive_in

        printmod = merge(printmod,-1,mympi_iproc()==0)

        ! Check inputs
        if (any(x<a) .or. any(x>b)) STOP 'sub_simanneal: x was not initialized to be between a and b!'

        ! Initialize best values
        fxbest = huge(0d0)
        xbest = 0d0

        naccept = 0
        nreject = 0

        if (printmod>0) then
            write(*,'(A)') 'PSA: ________________________________________________________'            
            write(*,'(A)') 'PSA: Beginning Parallel Simulated Annealing Method '            
        end if
        do iter = 1,maxiter

            ! Set the current temperature
            T = T0/(1d0+log(real(iter,8))) ! T0 to 0

            !!! Draw the variables
            ! Draw a newpoint within the neighborhood of current x
            !!! I saw here http://stats.stackexchange.com/questions/27923/should-the-proposal-distribution-in-simulated-annealing-depend-on-the-temperatur
            !!! that can draw from this variance decreasing in T and will still get convergence (for large enough T0)
            do ip = 1,np

                if (iter==1 .and. ip==1) then
                    z(:,ip) = x
                else
                    do while (.true.)

                        if (.not. aggressive) then
                            z(:,ip) = x + e*randn(n)*sqrt(T/T0) ! Initial innovations are drawn from N(0,e**2). Subsquent, N(0,e**2/(1+log(k))**2)
                        else
                            z(:,ip) = x + e*randn(n)*T/T0 ! Initial innovations are drawn from N(0,e**2). Subsquent, N(0,e**2/(1+log(k))**2)
                        end if

                        ! Use rejection sampling to get the new point
                        if (all(z(:,ip)>=a) .and. all(z(:,ip)<=b)) exit

                    end do
                end if

            end do

            ! Evaluate f in parallel and share
            do ip = mympi_a(np),mympi_b(np)
                fz(ip) = f(z(:,ip))
            end do

            call mympi_allgatherv(fz)

            ! Take the best
            fy = minval(fz)
            itmp = minloc(fz)
            y = z(:,itmp(1))

            ! Determine accept reject
            if (iter==1) then
                x = y
                fx = fy
            elseif (fy<fx) then 
                ! In this variant we always accept better moves
                x = y
                fx = fy
                naccept = naccept + 1
            else
                call random_number(randuDraw)
                if (randuDraw<exp((fx-fy)/T)) then
                    ! Move from x to y
                    x = y
                    fx = fy
                    naccept = naccept + 1
                else
                    ! Keep current x
                    nreject = nreject + 1
                end if
            end if

            ! Try
            if (printmod>0 .and. mod(iter,printmod)==0) then
                write(*,'(A,i6,A,d12.5,A,d12.5,A,d12.5,A,d12.5,A,d12.5)') &
                    'PSA: iter ',iter,' ftrial',fy,' fcurr',fx,' fbest',fxbest,' Temp', T,&
                    ' Accept%', 100.d0*real(naccept,8)/real(naccept+nreject,8)
            end if

            ! Save best f
            if (fx<fxbest) then 
                fxbest = fx
                xbest = x
            end if

        end do

        ! Return the best function value
        if (printmod>0) then
            if (fxbest/=fx) then
                write(*,'(A,d12.5,A)') 'PSA: best value ever encountered, ',fxbest,', being returned, it is /= fx'
            end if
        end if
        x = xbest
        fx = fxbest

    end subroutine sub_parallelsimanneal

    ! Grid search (GS)
    ! 
    ! Performs a brute force grid search. The tensor product grid is formed from the variables "grids" and "n".
    ! The grid in the ith dimension is grids(1:n(i),i). 
    subroutine sub_gridsearch(x,f,n,grids,printmod_in)
        use mod_matlab
        implicit none
        real(8), dimension(1:), intent(inout) :: x
        interface 
            function f(x) 
                implicit none
                real(8), dimension(1:), intent(in) :: x
                real(8) :: f
            end function f
        end interface
        real(8), dimension(:,:), intent(in) :: grids
        integer, dimension(:), intent(in) :: n
        integer, intent(in), optional :: printmod_in
        !local
        real(8), dimension(size(x)) :: xbest
        real(8) :: fxbest,fx
        integer :: printmod, d, j, i
        integer, dimension(size(x)) :: ivec

        ! Get dimension of the problem
        d = size(x)

        ! Check 
        if (d/=size(n)) STOP 'ERROR: sub_gridsearch: n has wrong size'
        if (d/=size(grids,2)) STOP 'ERROR: sub_gridsearch: grids has wrong size 2'
        if (maxval(n)>size(grids,1)) STOP 'ERROR: sub_gridsearch: grids has wrong size 1'

        ! Set defaults
        printmod = 1

        ! Overwrite if present
        if (present(printmod_in)) printmod = printmod_in

        ! Check inputs
!         if (any(b<a)) STOP 'sub_gridsearch: ERROR: b<a' 
        if (any(n<=0)) STOP 'sub_gridsearch: ERROR: n<=0'

        if (printmod>0) then
            write(*,'(A)') 'GS: ________________________________________________________'            
            write(*,'(A)') 'GS: Beginning Grid Search'            
        end if

        ! Do the grid search
        fxbest = huge(0d0)
        ivec(:) = 1
        do i = 1,product(n)

            ! Get x
            do j = 1,d
                x(j) = grids(ivec(j),j)
            end do

            ! Evaluate 
            fx = f(x)

            ! Store the best
            if (fx<fxbest) then
                xbest = x
                fxbest = fx
            end if

            ! Try
            if (printmod>0 .and. mod(i,printmod)==0) then
                write(*,'(A,i6,A,d12.5,A,d12.5)') 'GS: iter ',i,' f',fx,' fbest',fxbest
            end if

            ! Increment ivec
            do j = 1,d
                if (mod(ivec(j),n(j))==0) then
                    ivec(j) = 1 
                else
                    ivec(j) = ivec(j)+1
                    exit
                end if
            end do
        end do
    
        ! Return the best
        x = xbest

        ! Return the best function value
        if (printmod>0) then
            write(*,'(A)') 'GS: finished.'
        end if

    end subroutine sub_gridsearch


!     ! Grid search (GS)
!     ! 
!     ! Performs a basic grid search over the domain [a,b] with n giving the points in each dimension
!     !
!     subroutine sub_gridsearch(x,f,a,b,n,printmod_in)
!         use mod_matlab
!         implicit none
!         real(8), dimension(1:), intent(inout) :: x
!         interface 
!             function f(x) 
!                 implicit none
!                 real(8), dimension(1:), intent(in) :: x
!                 real(8) :: f
!             end function f
!         end interface
!         real(8), dimension(1:size(x)), intent(in) :: a,b
!         integer, dimension(1:size(x)), intent(in) :: n
!         integer, intent(in), optional :: printmod_in
!         !local
!         real(8), dimension(size(x)) :: xbest
!         real(8) :: fxbest,fx
!         integer :: printmod, d, j, i
!         integer, dimension(size(x)) :: ivec
! 
!         ! Get dimension of the problem
!         d = size(x)
! 
!         ! Set defaults
!         printmod = 1
! 
!         ! Overwrite if present
!         if (present(printmod_in)) printmod = printmod_in
! 
!         ! Check inputs
!         if (any(b<a)) STOP 'sub_gridsearch: ERROR: b<a' 
!         if (any(n<=0)) STOP 'sub_gridsearch: ERROR: n<=0'
! 
!         if (printmod>0) then
!             write(*,'(A)') 'GS: ________________________________________________________'            
!             write(*,'(A)') 'GS: Beginning Grid Search'            
!         end if
! 
!         ! Do the grid search
!         fxbest = huge(0d0)
!         ivec(:) = 1
!         do i = 1,product(n)
! 
! !             print*,'ivec',ivec
! 
!             ! Get x
!             do j = 1,d
!                 x(j) = a(j) + real(ivec(j)-1,8)/real(n(j)-1,8)*(b(j)-a(j))
!             end do
! 
!             ! Evaluate 
!             fx = f(x)
! 
!             ! Store the best
!             if (fx<fxbest) then
!                 xbest = x
!                 fxbest = fx
!             end if
! 
!             ! Try
!             if (printmod>0 .and. mod(i,printmod)==0) then
!                 write(*,'(A,i6,A,d12.5,A,d12.5)') 'GS: iter ',i,' f',fx,' fbest',fxbest
!             end if
! 
!             ! Increment ivec
!             do j = 1,d
!                 if (mod(ivec(j),n(j))==0) then
!                     ivec(j) = 1 
!                 else
!                     ivec(j) = ivec(j)+1
!                     exit
!                 end if
!             end do
!         end do
!     
!         ! Return the best
!         x = xbest
! 
!         ! Return the best function value
!         if (printmod>0) then
!             write(*,'(A)') 'GS: finished.'
!         end if
! 
!     end subroutine sub_gridsearch

end module mod_minNdim

module mod_minNdim_test
    use mod_minNdim
    implicit none
contains
    function testfn_noisyLM(x,nf) result(y)
        use mod_matlab
        implicit none
        real(8), intent(in) :: x(:)
        integer, intent(in) :: nf
        real(8) :: y(nf)

        ! y = sin(x) - real(colon(1,size(x)),8)*(1d0-cos(x)) + sum(cos(x)) - size(x) ! + floor(mod(exp(x)-1+.2,1)) + exp(.3*x)-1

!         y = log(abs(x(1)*x(2) + sin(x(1)) - exp(cos(x(2)))))

        ! x = exp(x)
        y = x - sin(x)**2



    end function 
    subroutine test_parnoisyLM
        use mod_mympi
        implicit none
        real(8) :: x(2)

        x = [-5d0,-5d0]
        call mympi_barrier()
        call sub_parnoisyLM(x,testfn_noisyLM,size(x),1d-3 + 0d0*x,ndraws=1000,opts=optimset('maxiter',300d0))
        write(*,*) 'x: ',x
        write(*,*) 'fx: ',testfn_noisyLM(x,size(x))

    end subroutine 
    subroutine test_noisyLM
        use mod_mympi
        implicit none
        real(8) :: x(2)
        if (mympi_iproc()==0) then
            x = [-5d0,-5d0]
            call sub_noisyLM(x,testfn_noisyLM,size(x),1d-3 + 0d0*x,opts=optimset('maxiter',300d0))
            write(*,*) 'x: ',x
            write(*,*) 'fx: ',testfn_noisyLM(x,size(x))
        end if
    end subroutine 
    subroutine test_parmin
        implicit none
        real(8) :: x(3)
        x = (/.3d0, .1d0, .5d0 /)
        call sub_parmin(x,rosenbrock,opt=0,e=abs(x)*.5d0)
        write(*,*) 'x: ',x
        write(*,*) 'fx: ',rosenbrock(x)
    end subroutine 
!     subroutine test_subplex
!         implicit none
!         real(8) :: x(3)
!         x = (/.3d0, .1d0, .5d0 /)
!         call sub_subplex(x,rosenbrock)
!         write(*,*) 'x: ',x
!         write(*,*) 'fx: ',rosenbrock(x)
!     end subroutine 
    subroutine test_simplex
        implicit none
        real(8) :: x(3)
        x = (/.3d0, .1d0, .5d0 /)
        ! call sub_simplex(x,rosenbrock,toler_in=1d-6,maxiter_in=10000,printmod_in=50,a_in=a,b_in=b,randomize_in=.true.)
        call sub_simplex(x,rosenbrock,toler_in=1d-6,maxiter_in=10000,printmod_in=1,randomize_in=.false.)
        write(*,*) 'x: ',x
        write(*,*) 'fx: ',rosenbrock(x)
    end subroutine 
    subroutine test_parallelsimplex
        use mod_mympi
        implicit none
        real(8) :: x(3)
        x = (/.3d0, .1d0, .5d0 /)
        call sub_parallelsimplex(x,rosenbrock,toler_in=1d-6,maxiter_in=10000,printmod_in=1)
        write(*,*) 'x: ',x
        write(*,*) 'fx: ',rosenbrock(x)
    end subroutine 


    subroutine test_parallelsimanneal
        implicit none
        real(8) :: x(3),a(3),b(3)
        x = (/.3d0,.1d0,.5d0/)
        a = 0.d0
        b = (/1.2d0,3d0,99d0/)
        call random_seed
        call sub_parallelsimanneal(x,rosenbrock,a_in=a,b_in=b,e_in=1d0*(/.01d0,.01d0,.01d0/),printmod_in=1000,maxiter_in=10000,T0_in=.001d0)
        write(*,*) 'x: ',x
        write(*,*) 'fx: ',rosenbrock(x)
    end subroutine 

    subroutine test_simanneal
        implicit none
        real(8) :: x(3),a(3),b(3)
        x = (/.3d0,.1d0,.5d0/)
        a = 0.d0
        b = (/1.2d0,3d0,99d0/)
        call random_seed
        call sub_simanneal(x,rosenbrock,a_in=a,b_in=b,e_in=1d0*(/.01d0,.01d0,.01d0/),printmod_in=1000,maxiter_in=10000,T0_in=.001d0)
        write(*,*) 'x: ',x
        write(*,*) 'fx: ',rosenbrock(x)
    end subroutine 

    subroutine test_gridsearch
        use mod_matlab, only: linspace
        implicit none
        real(8) :: x(3),grids(40,3)
        integer :: n(3)

        n = (/19,29,05/)
        grids(1:n(1),1) = linspace(0d0,2d0,n(1))
        grids(1:n(2),2) = linspace(0d0,2d0,n(2))
        grids(1:n(3),3) = linspace(0d0,2d0,n(3))
        call sub_gridsearch(x,rosenbrock,n,grids,printmod_in=1000)

        write(*,*) 'x: ',x
        write(*,*) 'fx: ',rosenbrock(x)

    end subroutine

    subroutine test_shrinkingcube
        implicit none
        real(8) :: x(3)
        x = (/.3d0, .1d0, .5d0 /)
        call sub_shrinkingcube(x,rosenbrock)
        write(*,*) 'x: ',x
        write(*,*) 'fx: ',rosenbrock(x)        
    end subroutine 

    subroutine test_parrandom
        implicit none
        real(8) :: x(3)
        x = (/.3d0, .1d0, .5d0 /)
        call sub_parrandom(x,rosenbrock,printmod_in=100)
        write(*,*) 'x: ',x
        write(*,*) 'fx: ',rosenbrock(x)        
    end subroutine 

    subroutine test_newtonNdim
        implicit none
        real(8) :: x(3)
        x = (/.3d0, .1d0, .5d0 /)
        call sub_newtonNdim(x,rosenbrock)
        write(*,*) 'x: ',x
        write(*,*) 'fx: ',rosenbrock(x)        
    end subroutine

    subroutine test_compass
        implicit none
        real(8) :: x(3)
        x = (/.3d0, .1d0, .5d0 /)
        call sub_compass(x,rosenbrock,shrink_in=.5d0,printmod_in=50)
        write(*,*) 'x: ',x
        write(*,*) 'fx: ',rosenbrock(x)        
    end subroutine

!     subroutine test_sequentialmin
!         implicit none
!         real(8) :: x(3)
!         x = (/.3d0, .1d0, .5d0 /)
!         call sub_sequentialmin(x,rosenbrock,shrink_in=.5d0,printmod_in=1)
!         write(*,*) 'x: ',x
!         write(*,*) 'fx: ',rosenbrock(x)        
!     end subroutine 

    subroutine test_gradDescent
        implicit none
        real(8) :: x(3)
        x = (/.3d0, .1d0, .5d0 /)
        call sub_gradDescent(x,rosenbrock,maxiter_in=10000,xtoler_in=1d-12,toler_in=1d-12,printmod_in=1000)
        write(*,*) 'x: ',x
        write(*,*) 'fx: ',rosenbrock(x)        
    end subroutine

    subroutine test_stochGradDescent
        implicit none
        real(8) :: x(3)
        x = (/.3d0, .1d0, .5d0 /)
        call sub_stochGradDescent(x,rosenbrock,maxiter_in=10000,xtoler_in=1d-12,printmod_in=1000,lambda_in=.1d0)
        write(*,*) 'x: ',x
        write(*,*) 'fx: ',rosenbrock(x)        
    end subroutine

    subroutine test_BFGS
        implicit none
        real(8) :: x(3)
        x = (/.3d0, .1d0, .5d0 /)
        call sub_BFGS(x,rosenbrock)
        write(*,*) 'x: ',x
        write(*,*) 'fx: ',rosenbrock(x)        
    end subroutine

    subroutine test_smolyaksearch
        implicit none
        real(8) :: x(3)
        x = (/.3d0, .1d0, .5d0 /)
        call sub_smolyaksearch(x,rosenbrock,e_in=.1d0*(/10d0,10d0,10d0/),mu_in=2,maxiter_in=1000,xtoler_in=1d-6)
        write(*,*) 'x: ',x
        write(*,*) 'fx: ',rosenbrock(x)        
    end subroutine

    subroutine test_numGradHess
        use mod_deriv
        use mod_matlab
        implicit none
        real(8) :: x(3), grad(3), hess(3,3), gradT(3), hessT(3,3), err

        hessT = reshape((/2d0,3d0,0d0,3d0,2d0,3d0,0d0,3d0,2d0/),(/3,3/)) ! Not a function of x

        x = 0d0
        gradT = 0d0
        call sub_numGradHess(x,quadratic,(/1d0,1d0,1d0/),grad,hess)
        err = maxval(abs(gradT-grad))
        err = max(err,maxval(abs(hessT-hess)))
!         print*,'Gradient',grad
!         print*,'Hess'
!         call disp(hess)

        x = 1d0
        gradT = (/5d0,8d0,5d0/)
        call sub_numGradHess(x,quadratic,(/1d0,1d0,1d0/),grad,hess)
        err = max(err,maxval(abs(gradT-grad)))
        err = max(err,maxval(abs(hessT-hess)))
!         print*,'Gradient',grad
!         print*,'Hess'
!         call disp(hess)

        if (err>1d-07) then
            print*,'Error was: ',err
            STOP 'test_numGradHess: failed'
        else
            print*,'test_numGradHess: passed'
        end if

    end subroutine

    function quadratic(x) result(f)
        implicit none
        real(8), dimension(1:), intent(in) :: x
        real(8) :: f
        !local
        real(8) :: tmp
        integer :: i

        tmp = sum(x**2)
        do i = 1,size(x)-1
            tmp = tmp + 3d0*x(i)*x(i+1)
        end do
        f = tmp

    end function

    function rosenbrock(x) result(f)
        implicit none
        real(8), dimension(1:), intent(in) :: x
        real(8) :: f
        !local
        integer :: i,n
        real(8) :: f_tmp
        n = size(x)
        f_tmp = 0.d0
        do i = 1,n-1
             f_tmp = f_tmp + 100.d0*(x(i+1)-x(i)**2)**2 + (1.d0-x(i))**2
        end do
        f = f_tmp
    end function
end module mod_minNdim_test

