! Routines for projection
module mod_project
    implicit none

contains

! Constructs a chebyshev approximation to the function "fcn" and evaluates
! it at points x. The approximation is controlled by the domain [a,b] and 
! the highest order Chebyshev polynomial used in the basis. This is n>=0.
function chebApprox(fcn,x,a,b,n) result(fvals)
    use mod_matlab
    implicit none
    interface 
        function fcn(x)
            real(8), intent(in) :: x
            real(8) :: fcn
        end function
    end interface
    real(8), intent(in) :: x(1:),a,b
    integer, intent(in) :: n
    real(8) :: fvals(size(x))
    ! local
    real(8), allocatable :: colloc(:),fcolloc(:),theta(:)
    integer :: i

    allocate(colloc(n+1),fcolloc(n+1),theta(n+1))

    colloc = chebZeros(n+1,a,b)

    do i = 1,n+1
        fcolloc(i) = fcn(colloc(i))
    end do
    
    theta = chebCoef(fcolloc)
    fvals = chebSumEval(x,theta,a,b)

! call disp('colloc',colloc)
! call disp('fcolloc',fcolloc)
! call disp('theta',theta)
! call disp('fvals',fvals)

end function 

! Given f, i.e. the n+1 function values at the x from chebZeros, produces
! the n+1 coefficients of the projection.
function chebCoef(f) result(theta)
    use mod_matlab
    implicit none
    real(8), intent(in) :: f(1:)
    real(8) :: theta(1:size(f))
    ! local
    integer :: i,n
    real(8), allocatable :: T(:,:),z(:)

    n = size(f)-1

    allocate(z(n+1),T(n+1,n+1))

    ! Get Chebyshev zeros
    z = chebZeros(n+1,-1d0,1d0)

    ! For each order = 0 to n, get the Cheby poly values at z
    T(:,1) = 1d0
    if (n>=1) then
        T(:,2) = z
    end if
    if (n>=2) then
        do i = 3,n+1
            T(:,i) = 2d0*z*T(:,i-1) - T(:,i-2)
        end do
    end if 

    ! Compute the coefficients
    ! theta = 2d0/real(n+1,8)*matmul(transpose(T),f)
    call mydgemm(theta,T,f,n+1,n+1,n+1,1,tA='T',alpha=2d0/real(n+1,8))
    theta(1) = theta(1)/2d0

end function

! Gives the m zeros of the Chebyshev polynomial of degree m scaled to [a,b]
function chebZeros(m,a,b) result(x)
    use mod_matlab, only: pi
    implicit none
    integer, intent(in) :: m
    real(8), intent(in) :: a,b
    real(8) :: x(m) ! m points
    ! local
    integer :: i

    do i = 1,m
        x(i) =(   -cos(pi*real(2*i-1,8)/real(2*m,8))  +1d0 )*(b-a)/2d0+a
    end do
    ! x = (x+1d0)*(b-a)/2d0 + a

end function

! Gives the m+1 extrema of the Chebyshev polynomial of degree m scaled to [a,b]
function chebExtrema(m,a,b) result(x)
    use mod_matlab, only: pi
    implicit none
    integer, intent(in) :: m
    real(8), intent(in) :: a,b
    real(8) :: x(m+1) ! m points
    ! local
    integer :: i

    do i = 1,m-1
        x(i+1) =( -cos(real(i,8)/real(m,8)*pi) + 1d0 )*(b-a)/2d0 + a
    end do
    x(1) = a
    x(m+1) = b ! Exact

    ! x = (x+1d0)*(b-a)/2d0 + a

end function

! Evaluates the Chebyshev approximation determined by coefficients theta 
! and domain [a,b] at points x. The approximation is
! f(x) = \sum^n_{i=0} theta(i)*phi_i(z(x;a,b)) where {phi_i} 
! are the Chebyshev polynomials. f should have the same size as x
function chebSumEval(x,theta,a,b) result(f)
    use mod_matlab
    implicit none
    real(8), intent(in) :: x(1:),theta(1:),a,b
    real(8) :: f(1:size(x))
    ! local
    real(8), allocatable :: z(:),T(:),Tm1(:),Tm2(:)
    integer :: n,i

    allocate(z(size(x)),T(size(x)),Tm1(size(x)),Tm2(size(x)))

    z = (x-a)*2d0/(b-a) - 1d0
    
    n = size(theta) - 1

    if (n==0) then
        f = theta(1)
    elseif (n==1) then
        f = theta(1) + theta(2)*z
    else

        Tm2 = 1d0
        Tm1 = z

        f = theta(1) + theta(2)*z

        do i = 3,n+1

            T = 2d0*z*Tm1 - Tm2
            f = f + theta(i)*T

            Tm2 = Tm1
            Tm1 = T

        end do
    end if

end function


end module mod_project

module mod_project_test
    use mod_project
    implicit none
contains

    subroutine test_cheb
        use mod_matlab
        implicit none
        real(8) :: tmp(1000),tmp2(size(tmp))
        integer :: i

        intrinsic :: dlog
    
        do i = 0,50
            tmp = linspace(1d0,9d0,size(tmp))
            tmp2 = chebApprox(dlog,tmp,0d0,10d0,i)
            ! call plot(tmp,tmp2)
        end do
        if (maxval(abs(log(tmp)-tmp2))>.004d0) &
            STOP 'ERROR: test_cheb: chebyshev approximation too poor'

        print*,'test_cheb: passed'

    end subroutine 

end module 
