module mod_quad
    implicit none
    private :: en_r2_11_1, en_r2_11_1_size
contains

! Integrates 
! \int^b_a f(x) dx 
! using Chebyshev quadrature adapted from [-1,1] with weighting function 1/sqrt(1-x**2).
subroutine chebyquad(x,w,a,b)
    use mod_matlab
    implicit none
    real(8), intent(out) :: x(:),w(:)
    real(8), intent(in) :: a,b
    ! local
    real(8) :: tmp
    integer :: n,i

    n = size(x)

    if (size(w)/=n) STOP 'ERROR: chebyquad: x,w don''t conform'

    do i = 1,n
        tmp = (2d0*real(i,8)-1d0)/2d0/real(n,8)*pi
        tmp = cos(tmp)
        x(i) = a + (b-a)*(tmp+1d0)/2d0
        w(i) = (b-a)/2d0*pi/real(n,8)*sqrt(1d0-tmp**2)
    end do

    if (abs(sum(w)-(b-a))>1d-6*abs(b-a)) then
        print*,'w',w
        print*,'x',x
        print*,'b',b
        print*,'a',a
        print*,'b-a',b-a
        print*,'sum(w)',sum(w)
        STOP 'ERROR: chebyquad: weights do not sum to b-a'
    end if


end subroutine

! gauss-legendre here has
!  \int^1_{-1} f(x) dx 
! but I use a change of variables to handle
!  \int^b_a f(x) dx
! n-point rules exactly integrate a 2n-1 degree polynomial wrt a weighting function
! n is inferred from the size of x and w
subroutine legendre(x,w,a,b)
    implicit none
    real(8), intent(out) :: x(:),w(:)
    real(8), intent(in) :: a,b
    ! local

    integer :: n

    n = size(x)

    if (size(w)/=n) STOP 'ERROR: x,w don''t conform'

    ! Get the nodes and weights for [-1,1] 
    select case (n)
        case (1)
            x = 0d0
            w = 2d0
        case (2)
            x = [-sqrt(1d0/3d0),sqrt(1d0/3d0)]
            w = 1d0
        case (3)
            x(1) = -sqrt(3d0/5d0)
            x(2) = 0d0
            x(3) = sqrt(3d0/5d0)
            w(1) = 5d0/9d0
            w(2) = 8d0/9d0
            w(3) = 5d0/9d0
!         case (5) 
!             x(1) = = 
!             x(3) = 0d0
! 
!             w(1) =
!             w(3) = 128d0/225d0
!             x(4) = -x(2)
!             w(4) = w(2)
!             x(5) = -x(1)
!             w(5) = w(1)

        case default
            STOP 'ERROR: legendre: formula for requested n not programmed yet'
    end select

    ! Adapt to [a,b]
    w = (b-a)/2d0*w
    x = (b-a)/2d0*x + (a+b)/2d0

end subroutine


! Trapezoidal rule for a predetermined x grid
! This specifies the quadrature weights for the trapezoidal rule
! at the nodes x. Note: if y = f(x) is piecewise linear, then
! this integrates the function exactly.
function trapzquad_fixedx(x) result(w)
    implicit none
    real(8), intent(in) :: x(1:)
    real(8) :: w(size(x))
    ! local
    integer :: n,i

    n = size(x)

    w(1) = .5d0*h(1)
    w(n) = .5d0*h(n-1)
    do i = 2,n-1
        w(i) = .5d0*(h(i-1)+h(i))
    end do

contains
    real(8) function h(i)
        integer, intent(in) :: i
        h = x(i+1)-x(i)
    end function
end function 

! Simpson's rule 
! NOTE: n must be odd
subroutine sub_simpsonquad(x,w,a,b,n)
    use mod_matlab
    implicit none
    integer, intent(in) :: n
    real(8), intent(in) :: a,b
    real(8), intent(out) :: x(n),w(n)
    ! local
    real(8) :: h
    integer :: i

    if (mod(n,2)==0) STOP 'ERROR: simpson: n must be odd'
    if (n<3) STOP 'ERROR: simpson: n must be >= 3'

    x = linspace(a,b,n)

    ! Step size from x(i) to x(i+1)
    h = (b-a)/real(n-1,8) 

    ! End point weights
    w(1) = 1d0
    w(n) = 1d0

    ! Even interior values
    do i = 2,n-1,2
        w(i) = 4d0
    end do

    ! Odd interior values
    do i = 3,n-2,2
        w(i) = 2d0
    end do

    ! Scale by h/3
    w = w*h/3d0

end subroutine sub_simpsonquad

! Integrate the piecewise linear function f(x) over domain [a,b] wrt the Lebesgue measure
! That is, \int^b_a f(x) dx. It is assumed the knots x are strictly increasing. 
! If a or b is outside the domain of x, linear extrapolation is used
function intPiecewiseLinear(xin,yin,a,b) result(qout)
    use mod_matlab
    implicit none
    real(8), intent(in) :: a,b,xin(1:),yin(1:)
    real(8) :: qout
    ! local
    real(8), allocatable :: x(:),y(:)
    real(8) :: q,m,xlo,xhi
    integer :: nx,ialo,ibhi,i

    nx = size(xin) 
    if (nx/=size(yin)) STOP 'x and y must have the same size'
    if (nx<2) STOP 'x must have at least length 2'
    if (b<a) STOP 'b must be >= a'

    allocate(x(nx),y(nx))

    x = xin
    y = yin

    ! Since linearly extrapolating, can just extend the range of x to a and b if a<x(1) or b>x(nx)
    if (a<x(1)) then
        ! Replace (x(1),y(1)) with (a,f(a))
        m = (y(2)-y(1))/(x(2)-x(1))       
        y(1) = y(1) + m*(a-x(1))
        x(1) = a
    end if
    if (b>x(nx)) then
        ! Replace (x(nx),y(nx)) with (b,f(b))
        m = (y(nx)-y(nx-1))/(x(nx)-x(nx-1))       
        y(nx) = y(nx-1) + m*(b-x(nx-1))
        x(nx) = b
    end if
    
    ! Now, everything is interior with x(1)<=a<=b<=x(n)
    call bsearch(ialo,a,x)
    call bsearch(ibhi,b,x)
    if (b>x(ibhi)) ibhi = ibhi + 1 ! Note: bsearch always returns lower bracket. We want higher.

    !!! Check that conditions hold
    if (a<x(1)) STOP 'ERROR: intPiecewiseLinear: should have x(1)<=a'
    if (b<a) STOP 'ERROR: intPiecewiseLinear: should have a<=b'
    if (x(nx)<b) STOP 'ERROR: intPiecewiseLinear: should have b<=x(nx)'
    if (.not. (a>=x(ialo) .and. a<=x(ialo+1))) STOP 'ERROR: intPiecewiseLinear: error in getting ialo'
    if (.not. (b>=x(ibhi-1) .and. b<=x(ibhi))) STOP 'ERROR: intPiecewiseLinear: error in getting ibhi'

    ! Integrate looping through the intervals
    q = 0d0
    do i = ialo,ibhi-1
        ! In range [max(a,x(i)),min(b,x(i+1))], integrate
        xlo = max(a,x(i))
        xhi = min(b,x(i+1))

        m = (y(i+1)-y(i))/(x(i+1)-x(i))

        q = q + y(i)*(xhi-xlo) + m*(xhi-x(i))**2/2d0 - m*(xlo-x(i))**2/2d0

    end do

    qout = q

end function

! Provides quadrature nodes (x) and weights (w) that can be used to 
! integrate a function f of dim d with respect to a multivariate normal
! distribution N(mean,varcov) using monomial rules taken from Stroud.
! There are several methods available
! 
! Method name      Description
! 'deg3B'          A degree 3 rule using 2*d spread out equally weighted points 
! ! 'deg5'           A degree 5 rule using 2d^2+1 pts at various distances
subroutine myquad(x,w,d,mean,varcov,method)
    use mod_matlab, only: chol,det,pi,strcmp
    implicit none
    real(8), allocatable, intent(out) :: x(:,:),w(:)
    integer, intent(in) :: d
    real(8), intent(in) :: mean(d),varcov(d,d)
    character(len=*), intent(in) :: method
    !local
    real(8) :: Omega(d,d), absdetOmega

    ! Construct the cholesky decomposition of varcov
    Omega = transpose(chol(varcov))
    absdetOmega = abs(det(Omega))

    ! Get the weights for standard form
    if (strcmp(method,'deg3B')) then
        call int_deg3ruleB(x,w,d)
!     elseif (strcmp(method,'deg5')) then
!         call int_deg5rule(x,w,d)
    else
        STOP 'myquad: Method not found.  Remember it''s case sensitive.'
    end if

    ! Convert to non-standard form
    x = sqrt(2.d0)*x*transpose(Omega) + transpose(spread(mean,2,size(x,1)))
    w = w*(2.d0**(real(d,8)/2.d0))*absdetOmega*1.d0/sqrt(abs(det(varcov)))*(2.d0*pi)**(-real(d,8)/2.d0)

end subroutine myquad

subroutine int_deg3ruleB(x,w,d)
    use mod_matlab, only: pi, diag
    implicit none
    real(8), intent(out), allocatable :: x(:,:),w(:)
    integer, intent(in) :: d
    !local
    real(8) :: r,V

    allocate(x(2*d,d),w(2*d))

    r = sqrt(real(d,8)/2.d0);    
    x(1:d,:) = diag(spread(r,1,d))
    x(d+1:2*d,:) = diag(spread(-r,1,d))

    V = pi**(real(d,8)/2.d0)
!     E = V/(2*d)*sum(f);
    w = V/(2.d0*real(d,8))
end subroutine int_deg3ruleB


subroutine test_quad_deg5rule
    use mod_matlab, only: randn, mean
    implicit none
    integer, parameter :: dint = 3
    real(8) :: x(2*dint**2+1,dint) !
    real(8) :: w(2*dint**2+1) !
    
    call sub_quad_deg5rule(x,w,dint)

    ! integrate sum(2*(.9*x-3)**5+1,2) using formula
    write(*,*) 'E according to quad rule', dot_product(sum(2.d0*(.9d0*x-3.d0)**5+1.d0,dim=2),w)

    write(*,*) 'E according to MC       ', mean(sum(2.d0*(.9d0*randn(100000,dint)-3.d0)**5+1.d0,dim=2))

end subroutine test_quad_deg5rule

! Integrate \int_{R^d} f(x) (2\pi)^{-d/2} e^{-(1/2) \sum^d_{i=1} x^2_i dx}
! Which is the integral of f wrt a multivariate unit normal distribution. 
!
! Based on integrating \int_{R^d} f(x) e^{-\sum^d_{i=1} x^2_i dx} but uses
! a change of variables to get it into a more useful form.
! Uses a degree 5 rule as in Judd p.275.
! 
! x are the quadrature points  2*d**2+1 by d
! w are the quadrature weights  2*d**2+1
! d is the dimension
!
subroutine sub_quad_deg5rule(x,w,dint)
    use mod_matlab, only: pi, diag
    implicit none
    integer, intent(in) :: dint
    real(8), intent(inout) :: x(1:,1:) !(2*dint**2+1,dint)
    real(8), intent(inout) :: w(1:) !(2*dint**2+1)
    !local
    integer :: N, ind, i, j
    real(8) :: d, V, A, B, C !, E
    real(8) :: r,s, tmpPP(dint),tmpPM(dint),tmpMP(dint),tmpMM(dint)

    ! Check for correct sizes
    if (size(x,1)/=2*dint**2+1) STOP 'sub_quad_deg5rule: x has wrong size1'
    if (size(x,2)/=dint) STOP 'sub_quad_deg5rule: x has wrong size2'
    if (size(w)/=2*dint**2+1) STOP 'sub_quad_deg5rule: w has wrong size'

    ! Get quadrature points
    N = 2*dint**2+1 !Number of points used in formula
    d = real(dint,8)
    r = sqrt(1.d0+.5d0*d)
    s = sqrt(.5d0+.25d0*d)
    
    !x = 0            
    x(1,:) = 0.d0
    ! x= r*ei
    x(2:2+dint-1,:) = diag(spread(r,1,dint))
    ! x= -r*ei
    x(2+dint:2+dint+dint-1,:) = diag(spread(-r,1,dint))
    ! x = +-s*ei +-s*ej
    ind = 2+dint+dint
    do i = 1,dint-1
        do j = i+1,dint
            tmpPP = 0.d0
            tmpPP(i) = s
            tmpPP(j) = s
            tmpPM = 0.d0
            tmpPM(i) = s
            tmpPM(j) = -s
            tmpMP = 0.d0
            tmpMP(i) = -s
            tmpMP(j) = s
            tmpMM = 0.d0
            tmpMM(i) = -s
            tmpMM(j) = -s

            x(ind,:) = tmpPP
            x(ind+1,:) = tmpPM
            x(ind+2,:) = tmpMP
            x(ind+3,:) = tmpMM
            ind = ind+4
        end do
    end do

    ! Change of variables to integrate f wrt unit normal
    x = sqrt(2.d0)*x

    ! Get quadrature weights
    N = 2*dint**2+1 !Number of points used in formula
    d = real(dint,8)

    V = pi**(d/2.d0)
    A = 2.d0/(d + 2.d0)*V
    B = (4.d0-d)/(2.d0*(d+2.d0)**2)*V
    C = V/(d+2.d0)**2

    w(1) = A
    w(2:2+dint-1) = B
    w(2+dint:2+dint+dint-1) = B
    w(2+dint+dint:2*dint**2+1) = C

    ! Use change of variables formula
    w = 1.d0/(sqrt(pi)**dint)*w  

end subroutine sub_quad_deg5rule

!Integrate \int_{R^d} f(x) e^{-\sum^d_{i=1} x^2_i dx}
!using a degree 5 rule as in Judd p.275.
!Proceed in two steps.  In the first step, you passed values of x for which
!you must evaluate f.  Then passing the values back, the integration is calculated.
!
function int_deg5rule_step1(dint) result(x)
    use mod_matlab
    implicit none
    integer, intent(in) :: dint
    real(8) :: x(2*dint**2+1,dint)
    !local
    integer :: N, ind, i, j
    real(8) :: d,r,s, tmpPP(dint),tmpPM(dint),tmpMP(dint),tmpMM(dint)
    !Values of x that need to be evaluated    
    !Variable definitions using monomial approach due to Stroud (1971) and found in Judd
    !integer, parameter :: dint = 4 !underscore J is for Judd

    N = 2*dint**2+1 !Number of points used in formula
    d = real(dint,8)
    r = sqrt(1.d0+.5d0*d)
    s = sqrt(.5d0+.25d0*d)
    
    !x = 0            
    x(1,:) = 0.d0
    ! x= r*ei
    x(2:2+dint-1,:) = diag(spread(r,1,dint))
    ! x= -r*ei
    x(2+dint:2+dint+dint-1,:) = diag(spread(-r,1,dint))

    ! x = +-s*ei +-s*ej
    ind = 2+dint+dint
    do i = 1,dint-1
        do j = i+1,dint
            tmpPP = 0.d0
            tmpPP(i) = s
            tmpPP(j) = s
            tmpPM = 0.d0
            tmpPM(i) = s
            tmpPM(j) = -s
            tmpMP = 0.d0
            tmpMP(i) = -s
            tmpMP(j) = s
            tmpMM = 0.d0
            tmpMM(i) = -s
            tmpMM(j) = -s

            x(ind,:) = tmpPP
            x(ind+1,:) = tmpPM
            x(ind+2,:) = tmpMP
            x(ind+3,:) = tmpMM
            ind = ind+4
        end do
    end do

end function

! See help for int_deg5rule_step1
function int_deg5rule_step2(dint,fx) result(E)
    use mod_matlab
    implicit none
    integer, intent(in) :: dint
    real(8), dimension(1:), intent(in) :: fx
    real(8) :: E
    !local
    integer :: N
    real(8) :: d, V, A, B, C

    N = 2*dint**2+1 !Number of points used in formula
    d = real(dint,8)

    V = pi**(d/2d0)
    A = 2d0/(d + 2d0)*V
    B = (4d0-d)/(2d0*(d+2d0)**2)*V
    C = V/(d+2d0)**2

    E = (A*fx(1)+ B*sum(fx(2:2+dint-1)+fx(2+dint:2+dint+dint-1)) + C*sum(fx(2+dint+dint:size(fx,1))))

end function


! Degree 11 rule due to Stroud
function deg11rule_pts(d) result(x)
    implicit none
    integer, intent(in) :: d
    real(8), allocatable :: x(:,:)
    !local
    real(8), allocatable :: w(:)
    real(8), allocatable :: xlocal(:,:)
    integer :: o
    if (d<3 .or. d>5) STOP 'deg11rule can only handle a dimension of 3,4, or 5'

!     N    O
!    __  ___
!     3  151
!     4  417
!     5  983
    if (d==3) o = 151
    if (d==4) o = 417
    if (d==5) o = 983

    allocate(xlocal(o,d),x(o,d),w(o))

    call en_r2_11_1(d,1,o,xlocal,w)
    
    ! Modify the points so they can be thought of as innovations to a normal distribution
    x = sqrt(2.d0)*xlocal

end function deg11rule_pts


! THE TWO ROUTINES ARE ADAPTED TO FORTRAN FROM JON BURKARDT'S WEBSITE:
! http://people.sc.fsu.edu/~jburkardt/m_src/stroud/en_r2_11_1.m
subroutine en_r2_11_1 ( n, option, o, x, w )
!*****************************************************************************80
!
!! EN_R2_11_1 implements the Stroud rule 11.1 for region EN_R2.
!
!  Discussion:
!
!    The rule has order 
!
!      O = ( 4 * N^5 - 20 * N^4 + 140 * N^3 - 130 * N^2 + 96 * N + 15 ) / 15.
!
!    The rule has precision P = 11.
!
!    EN_R2 is the entire N-dimensional space with weight function
!
!      w(x) = product ( 1 <= i <= n ) ( exp ( - x(i)^2 ) 
!
!    There are two versions of each rule, chosen by setting the
!    OPTION variable to 1 or 2.
!
!    The rule as tabulated by Stenger is available for N = 2 through 20.
!    This function accepts N = 3 through 5.
!
!     N    O
!    __  ___
!     3  151
!     4  417
!     5  983
!
!  Licensing:
!
!    This code is distributed under the GNU LGPL license.
!
!  Modified:
!
!    23 January 2010
!
!  Author:
!
!    John Burkardt
!
!  Reference:
!
!    Arthur Stroud,
!    Approximate Calculation of Multiple Integrals,
!    Prentice Hall, 1971,
!    ISBN: 0130438936,
!    LC: QA311.S85.
!
!  Parameters:
!
!    Input, integer ( kind = 4 ) N, the spatial dimension.
!    3 <= N <= 5.
!
!    Input, integer ( kind = 4 ) OPTION, chooses rule option 1 or 2.
!
!    Input, integer ( kind = 4 ) O, the order.
!
!    Output, real ( kind = 8 ) X(N,O), the abscissas.
!
!    Output, real ( kind = 8 ) W(O), the weights.
!
  implicit none

  integer ( kind = 4 ) n
  integer ( kind = 4 ) o

  real    ( kind = 8 ) b0
  real    ( kind = 8 ) b1
  real    ( kind = 8 ) b2
  real    ( kind = 8 ) b3
  real    ( kind = 8 ) b4
  real    ( kind = 8 ) b5
  real    ( kind = 8 ) b6
  real    ( kind = 8 ) b7
  real    ( kind = 8 ) b8
  real    ( kind = 8 ) b9
  real    ( kind = 8 ) b10
  real    ( kind = 8 ) b11
  real    ( kind = 8 ) b12
  real    ( kind = 8 ) b13
  real    ( kind = 8 ) b14
  real    ( kind = 8 ) b15
  integer ( kind = 4 ) i
  integer ( kind = 4 ) i1
  integer ( kind = 4 ) i2
  integer ( kind = 4 ) i3
  integer ( kind = 4 ) i4
  integer ( kind = 4 ) i5
  integer ( kind = 4 ) j
  integer ( kind = 4 ) k
  integer ( kind = 4 ) l
  integer ( kind = 4 ) m
!   logical              more
  integer ( kind = 4 ) option
  real    ( kind = 8 ), parameter :: pi = 3.141592653589793D+00
  real    ( kind = 8 ) u
  real    ( kind = 8 ) v
  real    ( kind = 8 ) volume
  real    ( kind = 8 ) w2
  real    ( kind = 8 ) w(o)
  real    ( kind = 8 ) x(n,o)

  if ( n < 3 .or. 5 < n ) then
    write ( *, '(a)' ) ' '
    write ( *, '(a)' ) 'EN_R2_11_1 - Fatal error!'
    write ( *, '(a)' ) '  3 <= N <= 5 required.'
    stop
  end if

  if ( option < 1 .or. 2 < option ) then
    write ( *, '(a)' ) ' '
    write ( *, '(a)' ) 'EN_R2_11_1 - Fatal error!'
    write ( *, '(a)' ) '  1 <= OPTION <= 2 required.'
    stop
  end if

  volume = sqrt ( pi ** n )

  if ( n == 3 .and. option == 1 ) then
    u =     0.235060497367449D+01
    v =     0.436077411927617D+00
    w2 =    0.133584907401370D+01
    b0 =  - 0.881591029957858D+01
    b1 =  - 0.751996143360650D-01
    b2 =    0.621743189471515D+01
    b3 =    0.241426451456494D+00
    b4 =  - 0.120709739276065D-02
    b5 =  - 0.427751221210138D+01
    b6 =    0.550169924840163D-01
    b7 =    0.237084999634707D-01
    b8 =  - 0.169791992887741D-02
    b9 =  - 0.252266276123350D-04
    b10 =   0.326777873717691D+01
    b11 =   0.968469949206802D-02
    b12 =   0.789754514877422D-03
    b13 =   0.000000000000000D+00
    b14 =   0.000000000000000D+00
    b15 =   0.000000000000000D+00
  else if ( n == 3 .and. option == 2 ) then
    u =     0.235060497367449D+01
    v =     0.133584907401370D+01
    w2 =    0.436077411927617D+00
    b0 =  - 0.141214037032900D+02
    b1 =  - 0.803730274707282D-01
    b2 =    0.235546545595906D+00
    b3 =    0.888123191556611D+01
    b4 =    0.142467131155533D-03
    b5 =    0.582993124006494D-01
    b6 =  - 0.561099173155661D+01
    b7 =  - 0.204028691521686D-02
    b8 =    0.252880089932256D-01
    b9 =  - 0.814378678627283D-04
    b10 =   0.804353953375146D-02
    b11 =   0.393451849690453D+01
    b12 =   0.171183493169724D-03
    b13 =   0.000000000000000D+00
    b14 =   0.000000000000000D+00
    b15 =   0.000000000000000D+00
  else if ( n == 4 .and. option == 1 ) then
    u =     0.235060497367449D+01
    v =     0.436077411927617D+00
    w2 =    0.133584907401370D+01
    b0 =    0.241502736147339D+03
    b1 =  - 0.196095938531478D+00
    b2 =  - 0.128675737999280D+03
    b3 =    0.307568784278696D+00
    b4 =  - 0.480908422319460D-02
    b5 =    0.698087019367085D+02
    b6 =    0.631837143743771D-01
    b7 =    0.392226151971179D-01
    b8 =  - 0.300948471646799D-02
    b9 =  - 0.650235306755170D-04
    b10 = - 0.386951974646715D+02
    b11 =   0.171656829095787D-01
    b12 =   0.139980343116450D-02
    b13 =   0.101552487093372D-04
    b14 =   0.222435922356439D+02
    b15 =   0.000000000000000D+00
  else if ( n == 4 .and. option == 2 ) then
    u =     0.235060497367449D+01
    v =     0.133584907401370D+01
    w2 =    0.436077411927617D+00
    b0 =  - 0.151944464736584D+03
    b1 =  - 0.223498438689039D+00
    b2 =    0.243574919068010D+00
    b3 =    0.634373877008693D+02
    b4 =  - 0.782065187814018D-04
    b5 =    0.911833754536616D-01
    b6 =  - 0.238927288245914D+02
    b7 =  - 0.422314408318853D-02
    b8 =    0.448218289217760D-01
    b9 =  - 0.138053374667391D-03
    b10 =   0.607473265800655D-02
    b11 =   0.697375246129742D+01
    b12 =   0.303414841680135D-03
    b13 = - 0.314574391771792D-05
    b14 =   0.409103498175100D-02
    b15 =   0.000000000000000D+00
  else if ( n == 5 .and. option == 1 ) then
    u =     0.235060497367449D+01
    v =     0.436077411927617D+00
    w2 =    0.133584907401370D+01
    b0 =    0.255885269311763E+04
    b1 =  - 0.439598677491526D+00
    b2 =  - 0.106541406144610E+04
    b3 =    0.453540909054264D+00
    b4 =  - 0.132100905623778D-01
    b5 =    0.418606568954203D+03
    b6 =    0.511394563043680D-01
    b7 =    0.645581013845604D-01
    b8 =  - 0.533417277494500D-02
    b9 =  - 0.137981626254496D-03
    b10 = - 0.147436933189884D+03
    b11 =   0.304253807765057D-01
    b12 =   0.248108698207828D-02
    b13 =   0.113652094546015D-04
    b14 =   0.394257407160391D+02
    b15 =   0.331725011358320D-05
  else if ( n == 5 .and. option == 2 ) then
    u =     0.235060497367449D+01
    v =     0.133584907401370D+01
    w2 =    0.436077411927617D+00
    b0 =  - 0.761305347548192D+03
    b1 =  - 0.536360805019297D+00
    b2 =    0.110669832078736D+00
    b3 =    0.246421088923968D+03
    b4 =  - 0.773649327968607D-03
    b5 =    0.169088641205970D+00
    b6 =  - 0.670700680243651D+02
    b7 =  - 0.856090560229205D-02
    b8 =    0.794446232770302D-01
    b9 =  - 0.220272863263544D-03
    b10 = - 0.373515812228225D-02
    b11 =   0.123606544052884D+02
    b12 =   0.537788804557843D-03
    b13 = - 0.122101861480881D-04
    b14 =   0.725117070759373D-02
    b15 =   0.331725011358320D-05
  end if

  x(1:n,1:o) = 0.0D+00

  k = 0
!
!  1 point.
!
  k = k + 1
! x(1:n,k) = 0.0D+00
  w(k) = b0
!
!  2 * N points.
!
  do i = 1, n
    k = k + 1
    x(i,k) = - u
    w(k) = b1
    k = k + 1
    x(i,k) = + u
    w(k) = b1
  end do
!
!  2 * N points.
!
  do i = 1, n
    k = k + 1
    x(i,k) = - v
    w(k) = b2
    k = k + 1
    x(i,k) = + v
    w(k) = b2
  end do
!
!  2 * N points.
!
  do i = 1, n
    k = k + 1
    x(i,k) = - w2
    w(k) = b3
    k = k + 1
    x(i,k) = + w2
    w(k) = b3
  end do
!
!  4 * ( N * ( N - 1 ) / 2 ) points.
!
  do i = 1, n - 1
    do j = i + 1, n
      k = k + 1
      x(i,k) = - u
      x(j,k) = - u
      w(k) = b4
      k = k + 1
      x(i,k) = - u
      x(j,k) = + u
      w(k) = b4
      k = k + 1
      x(i,k) = + u
      x(j,k) = - u
      w(k) = b4
      k = k + 1
      x(i,k) = + u
      x(j,k) = + u
      w(k) = b4
    end do
  end do
!
!  4 * ( N * ( N - 1 ) / 2 ) points.
!
  do i = 1, n - 1
    do j = i + 1, n
      k = k + 1
      x(i,k) = - v
      x(j,k) = - v
      w(k) = b5
      k = k + 1
      x(i,k) = - v
      x(j,k) = + v
      w(k) = b5
      k = k + 1
      x(i,k) = + v
      x(j,k) = - v
      w(k) = b5
      k = k + 1
      x(i,k) = + v
      x(j,k) = + v
      w(k) = b5
    end do
  end do
!
!  4 * ( N * ( N - 1 ) / 2 ) points.
!
  do i = 1, n - 1
    do j = i + 1, n
      k = k + 1
      x(i,k) = - w2
      x(j,k) = - w2
      w(k) = b6
      k = k + 1
      x(i,k) = - w2
      x(j,k) = + w2
      w(k) = b6
      k = k + 1
      x(i,k) = + w2
      x(j,k) = - w2
      w(k) = b6
      k = k + 1
      x(i,k) = + w2
      x(j,k) = + w2
      w(k) = b6
    end do
  end do
!
!  4 * ( N * ( N - 1 ) ) points.
!
  do i = 1, n - 1
    do j = i + 1, n
      k = k + 1
      x(i,k) = - u
      x(j,k) = - v
      w(k) = b7
      k = k + 1
      x(i,k) = - u
      x(j,k) = + v
      w(k) = b7
      k = k + 1
      x(i,k) = + u
      x(j,k) = - v
      w(k) = b7
      k = k + 1
      x(i,k) = + u
      x(j,k) = + v
      w(k) = b7
      k = k + 1
      x(i,k) = - v
      x(j,k) = - u
      w(k) = b7
      k = k + 1
      x(i,k) = - v
      x(j,k) = + u
      w(k) = b7
      k = k + 1
      x(i,k) = + v
      x(j,k) = - u
      w(k) = b7
      k = k + 1
      x(i,k) = + v
      x(j,k) = + u
      w(k) = b7
    end do
  end do
!
!  4 * ( N * ( N - 1 ) ) points.
!
  do i = 1, n - 1
    do j = i + 1, n
      k = k + 1
      x(i,k) = - u
      x(j,k) = - w2
      w(k) = b8
      k = k + 1
      x(i,k) = - u
      x(j,k) = + w2
      w(k) = b8
      k = k + 1
      x(i,k) = + u
      x(j,k) = - w2
      w(k) = b8
      k = k + 1
      x(i,k) = + u
      x(j,k) = + w2
      w(k) = b8
      k = k + 1
      x(i,k) = - w2
      x(j,k) = - u
      w(k) = b8
      k = k + 1
      x(i,k) = - w2
      x(j,k) = + u
      w(k) = b8
      k = k + 1
      x(i,k) = + w2
      x(j,k) = - u
      w(k) = b8
      k = k + 1
      x(i,k) = + w2
      x(j,k) = + u
      w(k) = b8
    end do
  end do
!
!  8 * ( N * ( N - 1 ) * ( N - 2 ) / 6 ) points.
!
  do i = 1, n - 2
    do j = i + 1, n - 1
      do l = j + 1, n
        k = k + 1
        x(i,k) = - u
        x(j,k) = - u
        x(l,k) = - u
        w(k) = b9
        k = k + 1
        x(i,k) = - u
        x(j,k) = - u
        x(l,k) = + u
        w(k) = b9
        k = k + 1
        x(i,k) = - u
        x(j,k) = + u
        x(l,k) = - u
        w(k) = b9
        k = k + 1
        x(i,k) = - u
        x(j,k) = + u
        x(l,k) = + u
        w(k) = b9
        k = k + 1
        x(i,k) = + u
        x(j,k) = - u
        x(l,k) = - u
        w(k) = b9
        k = k + 1
        x(i,k) = + u
        x(j,k) = - u
        x(l,k) = + u
        w(k) = b9
        k = k + 1
        x(i,k) = + u
        x(j,k) = + u
        x(l,k) = - u
        w(k) = b9
        k = k + 1
        x(i,k) = + u
        x(j,k) = + u
        x(l,k) = + u
        w(k) = b9
      end do
    end do
  end do
!
!  8 * ( N * ( N - 1 ) * ( N - 2 ) / 6 ) points.
!
  do i = 1, n - 2
    do j = i + 1, n - 1
      do l = j + 1, n
        k = k + 1
        x(i,k) = - v
        x(j,k) = - v
        x(l,k) = - v
        w(k) = b10
        k = k + 1
        x(i,k) = - v
        x(j,k) = - v
        x(l,k) = + v
        w(k) = b10
        k = k + 1
        x(i,k) = - v
        x(j,k) = + v
        x(l,k) = - v
        w(k) = b10
        k = k + 1
        x(i,k) = - v
        x(j,k) = + v
        x(l,k) = + v
        w(k) = b10
        k = k + 1
        x(i,k) = + v
        x(j,k) = - v
        x(l,k) = - v
        w(k) = b10
        k = k + 1
        x(i,k) = + v
        x(j,k) = - v
        x(l,k) = + v
        w(k) = b10
        k = k + 1
        x(i,k) = + v
        x(j,k) = + v
        x(l,k) = - v
        w(k) = b10
        k = k + 1
        x(i,k) = + v
        x(j,k) = + v
        x(l,k) = + v
        w(k) = b10
      end do
    end do
  end do
!
!  8 * ( N * ( N - 1 ) * ( N - 2 ) / 6 ) points.
!
  do i = 1, n - 2
    do j = i + 1, n - 1
      do l = j + 1, n
        k = k + 1
        x(i,k) = - w2
        x(j,k) = - w2
        x(l,k) = - w2
        w(k) = b11
        k = k + 1
        x(i,k) = - w2
        x(j,k) = - w2
        x(l,k) = + w2
        w(k) = b11
        k = k + 1
        x(i,k) = - w2
        x(j,k) = + w2
        x(l,k) = - w2
        w(k) = b11
        k = k + 1
        x(i,k) = - w2
        x(j,k) = + w2
        x(l,k) = + w2
        w(k) = b11
        k = k + 1
        x(i,k) = + w2
        x(j,k) = - w2
        x(l,k) = - w2
        w(k) = b11
        k = k + 1
        x(i,k) = + w2
        x(j,k) = - w2
        x(l,k) = + w2
        w(k) = b11
        k = k + 1
        x(i,k) = + w2
        x(j,k) = + w2
        x(l,k) = - w2
        w(k) = b11
        k = k + 1
        x(i,k) = + w2
        x(j,k) = + w2
        x(l,k) = + w2
        w(k) = b11
      end do
    end do
  end do
!
!  8 * ( N * ( N - 1 ) * ( N - 2 ) / 2 ) points.
!
  do i = 1, n - 2
    do j = i + 1, n - 1
      do l = j + 1, n
        k = k + 1
        x(i,k) = - u
        x(j,k) = - u
        x(l,k) = - v
        w(k) = b12
        k = k + 1
        x(i,k) = - u
        x(j,k) = - u
        x(l,k) = + v
        w(k) = b12
        k = k + 1
        x(i,k) = - u
        x(j,k) = + u
        x(l,k) = - v
        w(k) = b12
        k = k + 1
        x(i,k) = - u
        x(j,k) = + u
        x(l,k) = + v
        w(k) = b12
        k = k + 1
        x(i,k) = + u
        x(j,k) = - u
        x(l,k) = - v
        w(k) = b12
        k = k + 1
        x(i,k) = + u
        x(j,k) = - u
        x(l,k) = + v
        w(k) = b12
        k = k + 1
        x(i,k) = + u
        x(j,k) = + u
        x(l,k) = - v
        w(k) = b12
        k = k + 1
        x(i,k) = + u
        x(j,k) = + u
        x(l,k) = + v
        w(k) = b12
        k = k + 1
        x(i,k) = - u
        x(j,k) = - v
        x(l,k) = - u
        w(k) = b12
        k = k + 1
        x(i,k) = - u
        x(j,k) = - v
        x(l,k) = + u
        w(k) = b12
        k = k + 1
        x(i,k) = - u
        x(j,k) = + v
        x(l,k) = - u
        w(k) = b12
        k = k + 1
        x(i,k) = - u
        x(j,k) = + v
        x(l,k) = + u
        w(k) = b12
        k = k + 1
        x(i,k) = + u
        x(j,k) = - v
        x(l,k) = - u
        w(k) = b12
        k = k + 1
        x(i,k) = + u
        x(j,k) = - v
        x(l,k) = + u
        w(k) = b12
        k = k + 1
        x(i,k) = + u
        x(j,k) = + v
        x(l,k) = - u
        w(k) = b12
        k = k + 1
        x(i,k) = + u
        x(j,k) = + v
        x(l,k) = + u
        w(k) = b12
        k = k + 1
        x(i,k) = - v
        x(j,k) = - u
        x(l,k) = - u
        w(k) = b12
        k = k + 1
        x(i,k) = - v
        x(j,k) = - u
        x(l,k) = + u
        w(k) = b12
        k = k + 1
        x(i,k) = - v
        x(j,k) = + u
        x(l,k) = - u
        w(k) = b12
        k = k + 1
        x(i,k) = - v
        x(j,k) = + u
        x(l,k) = + u
        w(k) = b12
        k = k + 1
        x(i,k) = + v
        x(j,k) = - u
        x(l,k) = - u
        w(k) = b12
        k = k + 1
        x(i,k) = + v
        x(j,k) = - u
        x(l,k) = + u
        w(k) = b12
        k = k + 1
        x(i,k) = + v
        x(j,k) = + u
        x(l,k) = - u
        w(k) = b12
        k = k + 1
        x(i,k) = + v
        x(j,k) = + u
        x(l,k) = + u
        w(k) = b12
      end do
    end do
  end do
!
!  16 * ( N * ( N - 1 ) * ( N - 2 ) * ( N - 3 ) / 24 ) points.
!
  do i = 1, n - 3
    do j = i + 1, n - 2
      do l = j + 1, n - 1
        do m = l + 1, n
          k = k + 1
          x(i,k) = - u
          x(j,k) = - u
          x(l,k) = - u
          x(m,k) = - u
          w(k) = b13
          k = k + 1
          x(i,k) = - u
          x(j,k) = - u
          x(l,k) = - u
          x(m,k) = + u
          w(k) = b13
          k = k + 1
          x(i,k) = - u
          x(j,k) = - u
          x(l,k) = + u
          x(m,k) = - u
          w(k) = b13
          k = k + 1
          x(i,k) = - u
          x(j,k) = - u
          x(l,k) = + u
          x(m,k) = + u
          w(k) = b13
          k = k + 1
          x(i,k) = - u
          x(j,k) = + u
          x(l,k) = - u
          x(m,k) = - u
          w(k) = b13
          k = k + 1
          x(i,k) = - u
          x(j,k) = + u
          x(l,k) = - u
          x(m,k) = + u
          w(k) = b13
          k = k + 1
          x(i,k) = - u
          x(j,k) = + u
          x(l,k) = + u
          x(m,k) = - u
          w(k) = b13
          k = k + 1
          x(i,k) = - u
          x(j,k) = + u
          x(l,k) = + u
          x(m,k) = + u
          w(k) = b13
          k = k + 1
          x(i,k) = + u
          x(j,k) = - u
          x(l,k) = - u
          x(m,k) = - u
          w(k) = b13
          k = k + 1
          x(i,k) = + u
          x(j,k) = - u
          x(l,k) = - u
          x(m,k) = + u
          w(k) = b13
          k = k + 1
          x(i,k) = + u
          x(j,k) = - u
          x(l,k) = + u
          x(m,k) = - u
          w(k) = b13
          k = k + 1
          x(i,k) = + u
          x(j,k) = - u
          x(l,k) = + u
          x(m,k) = + u
          w(k) = b13
          k = k + 1
          x(i,k) = + u
          x(j,k) = + u
          x(l,k) = - u
          x(m,k) = - u
          w(k) = b13
          k = k + 1
          x(i,k) = + u
          x(j,k) = + u
          x(l,k) = - u
          x(m,k) = + u
          w(k) = b13
          k = k + 1
          x(i,k) = + u
          x(j,k) = + u
          x(l,k) = + u
          x(m,k) = - u
          w(k) = b13
          k = k + 1
          x(i,k) = + u
          x(j,k) = + u
          x(l,k) = + u
          x(m,k) = + u
          w(k) = b13
        end do
      end do
    end do
  end do
!
!  16 * ( N * ( N - 1 ) * ( N - 2 ) * ( N - 3 ) / 24 ) points.
!
  do i = 1, n - 3
    do j = i + 1, n - 2
      do l = j + 1, n - 1
        do m = l + 1, n
          k = k + 1
          x(i,k) = - v
          x(j,k) = - v
          x(l,k) = - v
          x(m,k) = - v
          w(k) = b14
          k = k + 1
          x(i,k) = - v
          x(j,k) = - v
          x(l,k) = - v
          x(m,k) = + v
          w(k) = b14
          k = k + 1
          x(i,k) = - v
          x(j,k) = - v
          x(l,k) = + v
          x(m,k) = - v
          w(k) = b14
          k = k + 1
          x(i,k) = - v
          x(j,k) = - v
          x(l,k) = + v
          x(m,k) = + v
          w(k) = b14
          k = k + 1
          x(i,k) = - v
          x(j,k) = + v
          x(l,k) = - v
          x(m,k) = - v
          w(k) = b14
          k = k + 1
          x(i,k) = - v
          x(j,k) = + v
          x(l,k) = - v
          x(m,k) = + v
          w(k) = b14
          k = k + 1
          x(i,k) = - v
          x(j,k) = + v
          x(l,k) = + v
          x(m,k) = - v
          w(k) = b14
          k = k + 1
          x(i,k) = - v
          x(j,k) = + v
          x(l,k) = + v
          x(m,k) = + v
          w(k) = b14
          k = k + 1
          x(i,k) = + v
          x(j,k) = - v
          x(l,k) = - v
          x(m,k) = - v
          w(k) = b14
          k = k + 1
          x(i,k) = + v
          x(j,k) = - v
          x(l,k) = - v
          x(m,k) = + v
          w(k) = b14
          k = k + 1
          x(i,k) = + v
          x(j,k) = - v
          x(l,k) = + v
          x(m,k) = - v
          w(k) = b14
          k = k + 1
          x(i,k) = + v
          x(j,k) = - v
          x(l,k) = + v
          x(m,k) = + v
          w(k) = b14
          k = k + 1
          x(i,k) = + v
          x(j,k) = + v
          x(l,k) = - v
          x(m,k) = - v
          w(k) = b14
          k = k + 1
          x(i,k) = + v
          x(j,k) = + v
          x(l,k) = - v
          x(m,k) = + v
          w(k) = b14
          k = k + 1
          x(i,k) = + v
          x(j,k) = + v
          x(l,k) = + v
          x(m,k) = - v
          w(k) = b14
          k = k + 1
          x(i,k) = + v
          x(j,k) = + v
          x(l,k) = + v
          x(m,k) = + v
          w(k) = b14
        end do
      end do
    end do
  end do
!
!  All quintuples UUUUU with 32 sign combinations.
!
  do i1 = 1, n - 4
    do i2 = i1 + 1, n - 3
      do i3 = i2 + 1, n - 2
        do i4 = i3 + 1, n - 1
          do i5 = i4 + 1, n
            k = k + 1
            x(i1,k) = - u
            x(i2,k) = - u
            x(i3,k) = - u
            x(i4,k) = - u
            x(i5,k) = - u
            w(k) = b15
            k = k + 1
            x(i1,k) = - u
            x(i2,k) = - u
            x(i3,k) = - u
            x(i4,k) = - u
            x(i5,k) = + u
            w(k) = b15
            k = k + 1
            x(i1,k) = - u
            x(i2,k) = - u
            x(i3,k) = - u
            x(i4,k) = + u
            x(i5,k) = - u
            w(k) = b15
            k = k + 1
            x(i1,k) = - u
            x(i2,k) = - u
            x(i3,k) = - u
            x(i4,k) = + u
            x(i5,k) = + u
            w(k) = b15
            k = k + 1
            x(i1,k) = - u
            x(i2,k) = - u
            x(i3,k) = + u
            x(i4,k) = - u
            x(i5,k) = - u
            w(k) = b15
            k = k + 1
            x(i1,k) = - u
            x(i2,k) = - u
            x(i3,k) = + u
            x(i4,k) = - u
            x(i5,k) = + u
            w(k) = b15
            k = k + 1
            x(i1,k) = - u
            x(i2,k) = - u
            x(i3,k) = + u
            x(i4,k) = + u
            x(i5,k) = - u
            w(k) = b15
            k = k + 1
            x(i1,k) = - u
            x(i2,k) = - u
            x(i3,k) = + u
            x(i4,k) = + u
            x(i5,k) = + u
            w(k) = b15
            k = k + 1
            x(i1,k) = - u
            x(i2,k) = + u
            x(i3,k) = - u
            x(i4,k) = - u
            x(i5,k) = - u
            w(k) = b15
            k = k + 1
            x(i1,k) = - u
            x(i2,k) = + u
            x(i3,k) = - u
            x(i4,k) = - u
            x(i5,k) = + u
            w(k) = b15
            k = k + 1
            x(i1,k) = - u
            x(i2,k) = + u
            x(i3,k) = - u
            x(i4,k) = + u
            x(i5,k) = - u
            w(k) = b15
            k = k + 1
            x(i1,k) = - u
            x(i2,k) = + u
            x(i3,k) = - u
            x(i4,k) = + u
            x(i5,k) = + u
            w(k) = b15
            k = k + 1
            x(i1,k) = - u
            x(i2,k) = + u
            x(i3,k) = + u
            x(i4,k) = - u
            x(i5,k) = - u
            w(k) = b15
            k = k + 1
            x(i1,k) = - u
            x(i2,k) = + u
            x(i3,k) = + u
            x(i4,k) = - u
            x(i5,k) = + u
            w(k) = b15
            k = k + 1
            x(i1,k) = - u
            x(i2,k) = + u
            x(i3,k) = + u
            x(i4,k) = + u
            x(i5,k) = - u
            w(k) = b15
            k = k + 1
            x(i1,k) = - u
            x(i2,k) = + u
            x(i3,k) = + u
            x(i4,k) = + u
            x(i5,k) = + u
            w(k) = b15
            k = k + 1
            x(i1,k) = + u
            x(i2,k) = - u
            x(i3,k) = - u
            x(i4,k) = - u
            x(i5,k) = - u
            w(k) = b15
            k = k + 1
            x(i1,k) = + u
            x(i2,k) = - u
            x(i3,k) = - u
            x(i4,k) = - u
            x(i5,k) = + u
            w(k) = b15
            k = k + 1
            x(i1,k) = + u
            x(i2,k) = - u
            x(i3,k) = - u
            x(i4,k) = + u
            x(i5,k) = - u
            w(k) = b15
            k = k + 1
            x(i1,k) = + u
            x(i2,k) = - u
            x(i3,k) = - u
            x(i4,k) = + u
            x(i5,k) = + u
            w(k) = b15
            k = k + 1
            x(i1,k) = + u
            x(i2,k) = - u
            x(i3,k) = + u
            x(i4,k) = - u
            x(i5,k) = - u
            w(k) = b15
            k = k + 1
            x(i1,k) = + u
            x(i2,k) = - u
            x(i3,k) = + u
            x(i4,k) = - u
            x(i5,k) = + u
            w(k) = b15
            k = k + 1
            x(i1,k) = + u
            x(i2,k) = - u
            x(i3,k) = + u
            x(i4,k) = + u
            x(i5,k) = - u
            w(k) = b15
            k = k + 1
            x(i1,k) = + u
            x(i2,k) = - u
            x(i3,k) = + u
            x(i4,k) = + u
            x(i5,k) = + u
            w(k) = b15
            k = k + 1
            x(i1,k) = + u
            x(i2,k) = + u
            x(i3,k) = - u
            x(i4,k) = - u
            x(i5,k) = - u
            w(k) = b15
            k = k + 1
            x(i1,k) = + u
            x(i2,k) = + u
            x(i3,k) = - u
            x(i4,k) = - u
            x(i5,k) = + u
            w(k) = b15
            k = k + 1
            x(i1,k) = + u
            x(i2,k) = + u
            x(i3,k) = - u
            x(i4,k) = + u
            x(i5,k) = - u
            w(k) = b15
            k = k + 1
            x(i1,k) = + u
            x(i2,k) = + u
            x(i3,k) = - u
            x(i4,k) = + u
            x(i5,k) = + u
            w(k) = b15
            k = k + 1
            x(i1,k) = + u
            x(i2,k) = + u
            x(i3,k) = + u
            x(i4,k) = - u
            x(i5,k) = - u
            w(k) = b15
            k = k + 1
            x(i1,k) = + u
            x(i2,k) = + u
            x(i3,k) = + u
            x(i4,k) = - u
            x(i5,k) = + u
            w(k) = b15
            k = k + 1
            x(i1,k) = + u
            x(i2,k) = + u
            x(i3,k) = + u
            x(i4,k) = + u
            x(i5,k) = - u
            w(k) = b15
            k = k + 1
            x(i1,k) = + u
            x(i2,k) = + u
            x(i3,k) = + u
            x(i4,k) = + u
            x(i5,k) = + u
            w(k) = b15
          end do
        end do
      end do
    end do
  end do

  return
end subroutine

subroutine en_r2_11_1_size ( n, option, o )

!*****************************************************************************80
!
!! EN_R2_11_1_SIZE sizes the Stroud rule 11.1 for region EN_R2.
!
!  Discussion:
!
!    The rule has order 
!
!      O = ( 4 * N^5 - 20 * N^4 + 140 * N^3 - 130 * N^2 + 96 * N + 15 ) / 15.
!
!    The rule has precision P = 11.
!
!    EN_R2 is the entire N-dimensional space with weight function
!
!      w(x) = product ( 1 <= i <= n ) ( exp ( - x(i)^2 ) 
!
!    There are two versions of each rule, chosen by setting the
!    OPTION variable to 1 or 2.
!
!    The rule as tabulated by Stenger is available for N = 2 through 20.
!    This function accepts N = 3 through 5.
!
!     N    O
!    __  ___
!     3  151
!     4  417
!     5  983
!
!  Licensing:
!
!    This code is distributed under the GNU LGPL license.
!
!  Modified:
!
!    27 January 2010
!
!  Author:
!
!    John Burkardt
!
!  Reference:
!
!    Arthur Stroud,
!    Approximate Calculation of Multiple Integrals,
!    Prentice Hall, 1971,
!    ISBN: 0130438936,
!    LC: QA311.S85.
!
!  Parameters:
!
!    Input, integer ( kind = 4 ) N, the spatial dimension.
!    3 <= N <= 5.
!
!    Input, integer ( kind = 4 ) OPTION, chooses rule option 1 or 2.
!
!    Output, integer ( kind = 4 ) O, the order.
!
  implicit none

  integer ( kind = 4 ) n
  integer ( kind = 4 ) o
  integer ( kind = 4 ) option

  if ( n < 3 .or. 5 < n ) then
    write ( *, '(a)' ) ' '
    write ( *, '(a)' ) 'EN_R2_11_1_SIZE - Fatal error!'
    write ( *, '(a)' ) '  3 <= N <= 5 required.'
    stop
  end if

  if ( option < 1 .or. 2 < option ) then
    write ( *, '(a)' ) ' '
    write ( *, '(a)' ) 'EN_R2_11_1_SIZE - Fatal error!'
    write ( *, '(a)' ) '  1 <= OPTION <= 2 required.'
    stop
  end if

  o = ( 4 * n ** 5 - 20 * n ** 4 + 140 * n ** 3 - 130 * n ** 2 &
    + 96 * n + 15 ) / 15


  return
end subroutine

end module mod_quad


module mod_quad_test
    use mod_quad
contains

    subroutine test_sub_simpsonquad()
        implicit none
        real(8) :: x(11),w(11),y(11)
        real(8) :: lo,hi, Eapprox,E
        
        lo = 2d0
        hi = 5d0
        call sub_simpsonquad(x,w,lo,hi,11)

        ! Integrate a quadratic function (should be exact)
        ! y = \int^h_l a + bx + cx**2 
        !   = ax + bx**2/2 + c x**3/3 |^h_l
        y = -.5d0 - 5d0*x + 3d0*x**2
        Eapprox = dot_product(w,y)
        E = -.5d0*hi - 5d0*hi**2/2d0 + 3d0*hi**3/3d0
        E = E - (-.5d0*lo - 5d0*lo**2/2d0 + 3d0*lo**3/3d0)

        if (abs(E-Eapprox)>1d-10) STOP 'ERROR: test_sub_simpsonquad: not exact'

        print*,'test_sub_simpsonquad: passed'

    end subroutine

    subroutine test_intPiecewiseLinear()
        implicit none
        real(8) :: x(4),y(4),a,b,q,qtrue,q2
        integer :: i

        x = [-3d0,0d0,10d0,20d0]
        y = -sqrt(2d0)*x + 1d0  ! Linear function so should integrate exactly

        a = x(1)
        b = x(size(x))
        qtrue = (b-a) - sqrt(2d0)/2d0*(b**2-a**2)
        q2 = dot_product(y,trapzquad_fixedx(x))
        if (abs(qtrue-q2)>1d-12) then
            print*,'qtrue',qtrue
            print*,'q2',q2
            print*,'sum(trapzquad_fixedx(x))',sum(trapzquad_fixedx(x))
            STOP 'qtrue and q2 are far apart'
        end if

        do i = 1,10000

            call random_number(a)
            call random_number(b)

            ! Get a in range [-25d0,25d0]
            a = a*50d0 - 25d0
            ! Get b-a in range [0d0,30d0]
            b = a + b*30d0

            ! Integrate
            q = intPiecewiseLinear(x,y,a,b)
            qtrue = (b-a) - sqrt(2d0)/2d0*(b**2-a**2)

            if (abs(qtrue-q)>1d-12) then
                print*,'i',i,'a',a,'b',b
                print*,'qtrue',qtrue
                print*,'q',q
                STOP 'qtrue and q are far apart'
            end if


        end do
        print*,'test_intPiecewiseLinear: passed'

    end subroutine
end module mod_quad_test 
