module mod_root1dim
    implicit none
    real(8), parameter, private :: xtoler_def = 1d-6, toler_def = 1d-6
    integer, parameter, private :: printmod_def = -1, maxiter_def = 100
contains

! Tests whether the two functions values bracket a root -- i.e., have different signs
function bracketsRoot(fa,fb) result(tf)
    logical :: tf
    real(8) :: fa,fb

    ! tf = sign(fa,fb)/=fa
    tf = sign(fa,fb)/=fa .or. fa==0d0 .or. fb==0d0 

end function bracketsRoot

! Attempts to find n roots within a specific interval, [a,b], by checking n subintervals
! [a_i,a_{i+1}], i = 1,n where a_1:=a and a_{n+1}:=b. One must specify which intervals
! are to be checked by giving the n+1 points {a_i} points. Within each sub interval, 
! sub_brent is called with the toler, xtoler, maxiter, and printmod values desired.
subroutine sub_findallroots(roots,nroots,f,knots,toler_in,maxiter_in,xtoler_in,printmod_in)
    use mod_matlab
    implicit none    
    real(8), intent(out) :: roots(:)
    integer, intent(out) :: nroots
    interface
        function f(x)
        implicit none
        real(8), intent(in) :: x
        real(8) :: f
        end function f
    end interface
    real(8), intent(in) :: knots(:)
    real(8), intent(in), optional :: toler_in,xtoler_in
    integer, intent(in), optional :: maxiter_in,printmod_in
    ! local
    integer :: maxiter, printmod
    real(8) :: toler, xtoler
    real(8) :: fi, fip1
    integer :: i,n,iroots

    toler = 0d0; if (present(toler_in)) toler = toler_in ! Same custom toler used in brent
    xtoler = xtoler_def; if (present(xtoler_in)) xtoler = xtoler_in
    maxiter = maxiter_def; if (present(maxiter_in)) maxiter = maxiter_in    
    printmod = printmod_def; if (present(printmod_in)) printmod = printmod_in

    if (size(roots)<size(knots)-1) STOP 'ERROR: sub_findallroots: roots vector not big enough'
    if (size(knots)<2) STOP 'ERROR: sub_findallroots: knots vector must have length of at least two'

    n = size(knots)-1 ! number of intervals


    fi = f(knots(1))
    iroots = 0

    do i = 1,n

        fip1 = f(knots(i+1))

        if (printmod>=1) then
            write(*,*) 'FA: Checking Interval ',i,' of ',n,'. Root found: ',bracketsRoot(fi,fip1)
        end if

        if (bracketsRoot(fi,fip1)) then
            iroots = iroots + 1
            call sub_brent(roots(iroots),f,knots(i),knots(i+1),toler_in=toler,xtoler_in=xtoler,maxiter_in=maxiter,printmod_in=printmod)
        ! else, no root!
        end if

        fi = fip1

    end do

    nroots = iroots
    
    


end subroutine sub_findallroots

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! sub_bisect
!
! Uses bisection to find a root f(x) = 0 in a bracket [a,b]
! f should be defined as in the below interface block as just f(x)
!
! Optional arguments are the evaluation of f at a AND or b.  Alternatively,
! just the correct sign of fa AND fb can be passed.
!
subroutine sub_bisect(x_out,f,a_in,b_in,toler_in,maxiter_in,fa_in,fb_in,xtoler_in,printmod_in)
    implicit none
    real(8), intent(out) :: x_out
    interface
        function f(x)
        implicit none
        real(8), intent(in) :: x
        real(8) :: f
        end function f
    end interface
    real(8), intent(in) :: a_in,b_in
    real(8), intent(in), optional :: toler_in, fa_in,fb_in, xtoler_in
    integer, intent(in), optional :: maxiter_in, printmod_in
    !local
    real(8) :: a,b,x,fa,fb,fx,res,toler,xtoler
    integer :: maxiter,printmod, iter
    
    ! Get initial bracket
    a = a_in
    b = b_in
    
    ! Get or set parameter values
    printmod = printmod_def; if (present(printmod_in)) printmod = printmod_in
    xtoler = xtoler_def; if (present(xtoler_in)) xtoler = xtoler_in
    toler = toler_def; if (present(toler_in)) toler = toler_in
    maxiter = maxiter_def; if (present(maxiter_in)) maxiter = maxiter_in
    
    ! Get or set bracket function values
    if (present(fa_in)) then 
        fa = fa_in
    else
        fa = f(a)
    end if
    if (present(fb_in)) then 
        fb = fb_in
    else
        fb = f(b)
    end if

    ! Check whether a zero is at one bracket 
    if (abs(fa)<toler) then
        if (printmod>0) write(*,'(A)') 'bisect: Converged successfully: |f(a)|<toler'
        x_out = a
        return
    elseif (abs(fb)<toler) then
        if (printmod>0) write(*,'(A)') 'bisect: Converged successfully: |f(b)|<toler'
        x_out = b
        return
    end if
    
    ! Check for bracketing
    if (.not. bracketsRoot(fa,fb)) then
        write(*,'(A)') 'bisect: WARNING: zero not bracketed, returning best endpoint'
        if (abs(fa)<abs(fb)) then
            x_out = a
        else
            x_out = b
        end if
        return
    end if
    
    ! Main loop
    do iter = 1,maxiter
        
        ! Output progress
        if (printmod>0 .and. mod(iter,printmod)==0) &
            write(*,"(A,i4,A,e13.5,A,e13.5,A,e13.5,A,e13.5)") &
            'bisect: iter ',iter,' a',a,' b',b,' fa',fa,' fb',fb

        ! Check convergence criterion
        if (abs(b-a)<=xtoler) then 
            if (printmod>0) then
                write(*,*) 'bisect: Failed to converge: |b-a|<xtoler. Returning midpoint.'
            end if

            x_out = (a+b)/2d0
            exit
        end if
    
        ! Midpoint evaluation
        x = (a+b)/2d0
        fx = f(x)
        
        ! Check other convergence criterion
        if (abs(fx)<toler) then
            if (printmod>0) then 
                write(*,'(A,e13.5,A,e13.5)') 'bisect: Converged successfully: |f(x)|<toler. x ', x,' fx',fx
            end if
            x_out = x
            return
        end if

        ! Replace appropriate endpoint 
        if (bracketsRoot(fa,fx)) then
            b = x
            fb = fx
        else
            a = x
            fa = fx
        end if

    end do

end subroutine sub_bisect

! Ridder's method
! There are not very good notes out there on this method. 
subroutine sub_ridders(x_out,f,a_in,b_in,toler_in,maxiter_in,fa_in,fb_in,xtoler_in,printmod_in)
    implicit none
    real(8), intent(out) :: x_out
    interface
        function f(x)
        implicit none
        real(8), intent(in) :: x
        real(8) :: f
        end function f
    end interface
    real(8), intent(in) :: a_in,b_in
    real(8), intent(in), optional :: toler_in,fa_in,fb_in,xtoler_in
    integer, intent(in), optional :: maxiter_in, printmod_in
    !local
    real(8) :: toler, xtoler
    integer :: maxiter, printmod, iter, minind(1)
    real(8) :: a, b, fa, fb, m, fm, r, fr
    character(len=6) :: step
    logical :: brack(3)
    real(8) :: dist(3)
    

    ! Set and replace default values
    toler = toler_def; if (present(toler_in)) toler = toler_in
    xtoler = xtoler_def; if (present(xtoler_in)) xtoler = xtoler_in
    maxiter = maxiter_def; if (present(maxiter_in)) maxiter = maxiter_in
    printmod = printmod_def; if (present(printmod_in)) printmod = printmod_in

    ! Copy initial bracket
    a = a_in
    b = b_in
    
    ! Evaluate at brackets
    if (present(fa_in)) then
        fa = fa_in
    else
        fa = f(a)
    end if
    if (present(fb_in)) then
        fb = fb_in
    else
        fb = f(b)
    end if

    ! Check that a<b. If not, switch them
    if (a>b) then
        ! Use m for temp storage of b
        m = b 
        fm = fb
        ! Replace b with a
        b = a
        fb = fa
        ! Replace a with b
        a = m
        fa = fm
    end if

    ! Check for covergence
    if (abs(fa)<toler) then
       x_out = a
       return
    elseif (abs(fb)<toler) then
        x_out = b
        return
    end if

    ! If not bracketed, return error
    if (.not. bracketsRoot(fa,fb)) then
        x_out = merge(a,b,abs(fa)<abs(fb))
        print*,'sub_ridders: WARNING: root is not bracketed, returning best'
        return
    end if

    ! Main loop
    step = 'Init'
    do iter = 1,maxiter

        ! Report progress
        if (printmod>0 .and. mod(iter,printmod)==0) &
            write(*,'(A,i3,A,e13.5,A,e13.5,A,e13.5,A,e13.5,A,A)') &
            'ridders: iter ',iter,' a',a, ' b',b,' fa',fa, ' fb',fb, ' step ', trim(step) 
        
        ! Check stopping criterion
        if (abs(b-a)<xtoler) then
            x_out = (b+a)/2d0
            if (printmod>0) write(*,'(A)') 'ridders: Failed to converge: |b-a|<xtoler'
            return
        end if

        ! Midpoint
        m = (b+a)/2d0
        fm = f(m)
        if (abs(fm)<toler) then
            x_out = m
            if (printmod>0) write(*,'(A,e13.5,A,e13.5)') 'ridders: Converged: |f(m)|<toler. m',m,' fm',fm 
            return
        end if

        ! Exponential point (ridders pt)
        if (fa>0d0) then
            r = m + (m-a)*fm/sqrt(fm**2-fa*fb)
        else
            r = m - (m-a)*fm/sqrt(fm**2-fa*fb)
        end if
        fr = f(r)
        if (abs(fr)<toler) then
            x_out = r
            if (printmod>0) write(*,'(A,e13.5,A,e13.5)') 'ridders: Converged: |f(r)|<toler. r ', r, ' fr',fr 
            return
        end if

        ! Now among possible bracketing intervals, find the one with least length. 
        ! There are 5 possible brackets. But I believe after extensive testing that 
        ! [a,m] and [m,b] are always dominated. So, I am just not even considering these.
        ! However, if I'm wrong this could cause problems with slow convergence (it will still
        ! always convergence, but may do so slowly). 
        brack(1) = bracketsRoot(fr,fm)
        brack(2) = bracketsRoot(fa,fr)
        brack(3) = bracketsRoot(fr,fb)
        
        dist(1) = abs(r-m)
        dist(2) = abs(a-r)
        dist(3) = abs(r-b)
        
        ! Get location of smallest bracket 
        minind = minloc(dist,mask=brack)
        
        ! Replace bracket
        select case (minind(1))
            case (1)
                ! [r,m] best
                if (r<m) then
                    a = r
                    fa = fr
                    b = m
                    fb = fm
                else
                    a = m
                    fa = fm
                    b = r
                    fb = fr
                end if
                step = 'Best'
            case (2)
                ! [a,r] best
                b = r
                fb = fr
                step = 'Good'
            case (3)
                ! [r,b] best
                a = r
                fa = fr
                step = 'Good'
        end select

    end do

end subroutine sub_ridders

! Written using the FORTRAN program zero in "Algorithms for Minimization 
! without Derivatives" by Brent. 
! 
! Description: 
!       Returns a zero x of the function f in the given interval [a,b], to within a tolerance 6macheps|x| + 2t, 
! where macheps is the relative machine precision and t is a positive tolerance. The procedure assumes that 
! f(a) and f(b) have different signs.
subroutine sub_brent(x,f,a_in,b_in,toler_in,maxiter_in,fa_in,fb_in,xtoler_in,printmod_in)
    use mod_matlab
    implicit none    
    real(8), intent(out) :: x
    interface
        function f(x)
        implicit none
        real(8), intent(in) :: x
        real(8) :: f
        end function f
    end interface
    real(8), intent(in) :: a_in,b_in
    real(8), intent(in), optional :: toler_in,fa_in,fb_in,xtoler_in
    integer, intent(in), optional :: maxiter_in, printmod_in
    ! local
    real(8), parameter :: machep = epsilon(0d0)
    real(8) :: a,b,c,fa,fb,fc,toler,xtoler,e,d,m,p,q,tol,t,r,s
    integer :: maxiter,printmod,iter
    character(len=6) :: step

    ! Set of get parameters
    toler = 0d0; if (present(toler_in)) toler = toler_in ! Better to use custom toler here
    xtoler = xtoler_def; if (present(xtoler_in)) xtoler = xtoler_in
    maxiter = maxiter_def; if (present(maxiter_in)) maxiter = maxiter_in    
    printmod = printmod_def; if (present(printmod_in)) printmod = printmod_in

    ! Set the user chosen tolerance t to xtoler
    if (xtoler<0d0) then
        if (printmod>=-1) print*,'WARNING: xtoler must be positive. Resetting xtoler.'
        xtoler = 0d0
    end if
    t = xtoler
    
    ! Get initial bracket
    a=a_in
    b=b_in
    if (present(fa_in)) then
        fa = fa_in
    else
        fa = f(a)
    end if

    if (printmod==1) write(*,"(2(A,e13.5))") 'brent: init       a', a, ' fa', fa

    if (present(fb_in)) then
        fb = fb_in
    else
        fb = f(b)
    end if

    if (printmod==1) write(*,"(2(A,e13.5))") 'brent: init       b', b, ' fb', fb

    ! Test whether root is bracketed
    if (.not. bracketsRoot(fa,fb)) then
        if (abs(fa)<abs(fb)) then
            if (printmod>=-1) write(*,*) 'brent: WARNING: root is not bracketed, returning best endpoint a'
            x = a
        else
            if (printmod>=-1) write(*,*) 'brent: WARNING: root is not bracketed, returning best endpoint b'
            x = b
        end if
        return
    end if

    step = 'init'

    ! At any point in time, b is the best guess of the root, a is the previous value of b, 
    ! and the root is bracketed by b and c.
    do iter = 1,maxiter

        if (iter==1 .or. (fb>0d0 .and. fc>0d0) .or. (fb<=0d0 .and. fc<=0d0)) then
            c = a
            fc = fa
            e = b - a
            d = e
        end if

        ! If c is strictly better than b, swap b and c so b is the best guess. 
        if (abs(fc)<abs(fb)) then
            a = b
            b = c
            c  = a
            fa = fb
            fb = fc
            fc = fa
        end if

        ! Set the tolerance. Note: brent is very careful with these things, so don't deviate from this.
        tol = 2d0*machep*abs(b) + t

        ! Determine what half the length of the bracket [b,c] is
        m = .5d0*(c-b)

        ! If taking a bisection step would move the guess of the root less than tol, then return b the best guess.
        if ((abs(m)<=tol) .or. (fb==0d0)) then
            x = b
            return
        end if

        ! Display info
        if (printmod>0 .and. mod(iter,printmod)==0) &
            write(*,"(A,i5,A,e13.5,A,e13.5,A,A)") &
            'brent: iter ',iter,' b',b,' fb',fb,' ',step

        ! If still here, then check whether need to do bisection or can do interpolation
        if ((abs(e)>=tol) .and. (abs(fa)>abs(fb))) then
            s = fb/fa
            if (a/=c) then
                ! Inverse quadratic interpolation
                q = fa/fc
                r = fb/fc
                p = s*(2d0*m*q*(q-r) - (b-a)*(r-1d0))
                q = (q-1d0)*(r-1d0)*(s-1d0)
                
                step = 'quad'
            else
                ! Linear interpolation
                p = 2d0*m*s
                q = 1d0-s

                step = 'linear'
            end if

            ! Ensure p is positive
            if (p<=0d0) then
                p = -p
            else
                q = -q
            end if

            s = e
            e = d
            if ((2d0*p>=3d0*m*q-abs(tol*q)) .or. &
                (p>=abs(.5d0*s*q))) then
                ! Interpolation step failed to produce good step, bisect instead
                e = m
                d = m ! m is half the distance between b and c
                step = 'bisect'
            else
                !  Do interpolation step (either quadratic or linear)
                d = p/q
            end if
        else

            ! Do bisection step
            e = m 
            d = m
            
        end if

        ! Get new points. 
        !! Replace a (the old b) with b.
        a = b
        fa = fb

        !!! Increment b by d if that is greater than the tolerance. O/w, increment by tol.
        if (abs(d)<=tol) then
            ! m is .5*(c-b) with the bracket either [b,c] or [c,b]. 
            if (m > 0d0) then
                ! If m>0d0, then bracket is [b,c] so move towards c by tol
                b = b + tol
            else
                ! If m<=0d0, then bracket is [c,b] so move towards c by tol
                b = b - tol
            end if
        else
            b = b + d
        end if

        !!! Evaluate at the new point
        fb = f(b)

        ! Check my custom tolerance 
        if (abs(fb)<toler) then
            x = b
            return
        end if
            
    end do

end subroutine sub_brent

!Adapted from Numerical Recipes
! 
! Because of the license, I am able to use this ON ONE COMPUTER!!! And that is only if I bought the book.
! So I did order the book, and I used this for one project (the garnish one). I can not use it on any
! other projects (since I'll be working on different computers). Use bisection or program my own routine
! for the next one --- this one is far too close to the original.
! 
subroutine sub_brent_old(x_out,f,a_in,b_in,toler_in,maxiter_in,fa_in,fb_in,xtoler_in,printmod_in,c_in,fc_in)
    use mod_matlab
    implicit none    
    real(8), intent(out) :: x_out
    interface
        function f(x)
        implicit none
        real(8), intent(in) :: x
        real(8) :: f
        end function f
    end interface
    real(8), intent(in) :: a_in,b_in
    real(8), intent(in), optional :: c_in,toler_in,fa_in,fb_in,fc_in,xtoler_in
    integer, intent(in), optional :: maxiter_in, printmod_in
    !local
    real(8), parameter :: eps = epsilon(0.d0) 
    integer :: maxiter
    real(8) :: a,b,c,d,e,fa,fb,fc,p,q,r,s,tol1,xm,zbrent
    real(8) :: toler,xtoler 
    integer :: iter,printmod,hist_iter

    ! Set of get parameters
    toler = toler_def; if (present(toler_in)) toler = toler_in
    xtoler = xtoler_def; if (present(xtoler_in)) xtoler = xtoler_in
    maxiter = maxiter_def; if (present(maxiter_in)) maxiter = maxiter_in    
    printmod = printmod_def; if (present(printmod_in)) printmod = printmod_in
    
    ! Get initial bracket
    a=a_in
    b=b_in

    if (.not. present(fa_in)) then
        fa = f(a)
    else
        fa = fa_in
    end if
    if (.not. present(fb_in)) then
        fb = f(b)
    else
        fb = fb_in
    end if

    if (.not. bracketsRoot(fa,fb)) then
        if (abs(fa)<abs(fb)) then
            write(*,*) 'brent: WARNING: root is not bracketed, returning best endpoint lb'
        else
            write(*,*) 'brent: WARNING: root is not bracketed, returning best endpoint ub'
        end if
        x_out = merge(a,b,abs(fa)<abs(fb))         
        return
    end if

    c = b
    fc = fb
    
    !main solver
    do iter=1,maxiter

        if((fb>0.d0 .AND. fc>0.d0).or.(fb<0.d0 .AND. fc<0.d0)) then
            c=a !Rename a, b, c AND adjust bounding interval d.
            fc=fa
            d=b-a
            e=d
        endif
        if(abs(fc)<abs(fb)) then
            a=b
            b=c
            c=a
            fa=fb
            fb=fc
            fc=fa
        endif
        
        if (printmod>0 .and. mod(iter,printmod)==0) write(*,"(A,i5,A,e13.5,A,e13.5)") 'brent: iter ',iter,' b',b,' fb',fb
        
        tol1=2.d0*EPS*abs(b)+0.5d0*toler !Convergence check.
        xm=.5d0*(c-b)

        !ORIGINAL CODE HAD THIS TERMINATION CRITERION
!         if(abs(xm)<=tol1 .or. fb==0.d0) then
!             zbrent=b
!             exit !return
!         endif

        !MY CODE HAS THE TERM CRITERION
        if (abs(fb)<toler) then
            if (printmod>0) write(*,'(A)') 'brent: Converged successfully: |f(x)|<toler'
            x_out = b
            return
        elseif (abs(xm)<=tol1) then
            if (printmod>0) write(*,'(A)') 'brent: Failed to converge: change(x)<xtoler'
            x_out = b
            return
        end if

!         !MY CODE HAS THIS TERMINATION CRITERION
!         if (abs(fb)<toler .OR. abs(c-b)<xtoler) then
!             zbrent = b
!             exit
!         end if

        if(abs(e)>=tol1 .AND. abs(fa)>abs(fb)) then
            s=fb/fa !Attempt inverse quadratic interpolation.
            if(a==c) then
                p=2.d0*xm*s
                q=1.d0-s
            else
                q=fa/fc
                r=fb/fc
                p=s*(2.d0*xm*q*(q-r)-(b-a)*(r-1.d0))
                q=(q-1.d0)*(r-1.d0)*(s-1.d0)
            endif
            if(p>0.d0) q=-q !Check whether in bounds.
            p=abs(p)
            if(2.d0*p < min(3.d0*xm*q-abs(tol1*q),abs(e*q))) then
                e=d !Accept interpolation.
                d=p/q
            else
                d=xm !Interpolation failed, use bisection.
                e=d
            endif
        else !Bounds decreasing too slowly, use bisection.
            d=xm
            e=d
        endif
        a=b !Move last best guess to a.
        fa=fb
        if(abs(d) > tol1) then !Evaluate new trial root.
            b=b+d
        else
            b=b+sign(tol1,xm)
        endif
        fb=f(b)
    enddo

    if (printmod>0) write(*,*) 'brent: Failed to converge: max iters reached'
    x_out = b
    
end subroutine sub_brent_old

!OLD BISECTION ROUTINE
! subroutine sub_bisect(x_out,f,a_in,b_in,toler_in,maxiter_in,fa_in,fb_in,xtoler_in,printmod_in)
!     implicit none
!     real(8), intent(out) :: x_out
!     interface
!         function f(x)
!         implicit none
!         real(8), intent(in) :: x
!         real(8) :: f
!         end function f
!     end interface
!     real(8), intent(in) :: a_in,b_in
!     real(8), intent(in), optional :: toler_in, fa_in,fb_in, xtoler_in
!     integer, intent(in), optional :: maxiter_in, printmod_in
!     !local
!     real(8) :: a,b,x,fa,fb,fx,res,toler,xtoler, bestx,bestf
!     logical :: fapos, faneg, case1,case2,case3,case4,case5
!     integer :: maxiter,printmod, iter
!     
!     a = a_in
!     b = b_in
!     
!     !default values
!     printmod = 1
!     xtoler = 1.d-9
!     toler = 1.d-6
!     maxiter = 1000
!     if (present(printmod_in)) printmod = printmod_in
!     if (present(xtoler_in)) xtoler = xtoler_in
!     if (present(toler_in)) toler = toler_in
!     if (present(maxiter_in)) maxiter = maxiter_in
!     !check input values
!     if (xtoler<epsilon(0.d0)*10.d0) then
!         write(*,*) 'sub_bisect: warning: xtoler is set beyond machine precision.  increasing. '
!         xtoler = epsilon(0.d0)*10.d0
!     end if
!     
!     if (present(fa_in)) then 
!         fa = fa_in
!     else
!         fa = f(a)
!     end if
!     if (present(fb_in)) then 
!         fb = fb_in
!     else
!         fb = f(b)
!     end if
!     
!     if ((fa>0.d0 .AND. fb>0.d0) .OR. (fa<0.d0 .AND. fb<0.d0)) then
!         write(*,*) 'sub_bisect: zero not bracketed:', fa, fb
!         x_out = merge(a,b,abs(fa)<abs(fb))
!         return
!     end if
!     
!     !Main routine
!     bestx = merge(a,b,abs(fa)<abs(fb))
!     bestf = merge(fa,fb,abs(fa)<abs(fb))
!     fapos = fa>=0.d0
!     faneg = fa<=0.d0
!     res = 1d10
!     if (printmod/=-1) write(*,"('|~~~~~~~~~~~~~~Beginning Bisection Algorithm~~~~~~~~~~~~~~~~~~|')")
!     if (printmod/=-1) write(*,"('| iter,            a,            b,            x,           fx|')")
!     iter = 1
!     do while(res>toler)
!     
!         x = .5d0*a + .5d0*b    
!         if (abs(x-a)<=xtoler) then 
!             write(*,*) 'sub_bisect: Warning: xtoler exceeded'
!             exit
!         end if
!         fx = f(x)
!     
!         bestx = merge(bestx,x,abs(bestf)<abs(fx))
!         bestf = merge(bestf,fx,abs(bestf)<abs(fx))
!         
!         if (printmod/=-1) write(*,"('|',i5,4e14.5,'|')") iter,a,b,x,fx
!     
!         case1 = fx<0.d0 .AND. faneg
!         case2 = fx>0.d0 .AND. faneg
!         case3 = fx<0.d0 .AND. fapos
!         case4 = fx>0.d0 .AND. fapos
!         case5 = fx==0.d0
!     
!         if (case1) then
!             fa = fx
!             a = x
!         elseif (case2) then
!             fb = fx
!             b = x
!         elseif (case3) then
!             fb = fx
!             b = x
!         elseif (case4) then
!             fa = fx
!             a = x
!         elseif (case5) then
!             exit
!         end if
!          
!         res = abs(fx)
!         if (res<=toler) then 
!             if (printmod/=-1) write(*,"('|',i5,4e14.5,'|')") 0, 0.d0, 0.d0, x, fx
!             if (printmod/=-1) write(*,"('|~~~~~~~~~~~~~~Converged Successfully~~~~~~~~~~~~~~~~~~~~~~~~~|')")
!         end if
!     
!         iter = iter+1
!     
!         if (iter>maxiter) then
!             write(*,*) 'Max iterations exceeded'
!             exit
!         end if
!     
!     end do
!     
!     !Return the best found.  
!     x_out = bestx
!     !If fx /= bestf, want to call the routine with the best x in case the function has side effects
!     if (fx/=bestf) fx = f(bestx)
!     
!     if (printmod/=-1) write(*,"('|~~~~~~~~~~~~~~End Bisection Algorithm~~~~~~~~~~~~~~~~~~~~~~~~|')")
! 
! end subroutine sub_bisect

! Uses secant method with the bisection method as a backup
! subroutine sub_secant(x,f,a_in,b_in,toler_in,maxiter_in,fa_in,fb_in,xtoler_in,printmod_in)
!     use mod_matlab
!     implicit none   
!     interface    
!         function f(x)
!             real(8), intent(in) :: x
!             real(8) :: f
!         end function f
!     end interface
!     real(8), intent(inout) :: x
!     real(8), intent(in), optional :: a_in,b_in
!     real(8), intent(in), optional :: toler_in,fa_in,fb_in,xtoler_in
!     integer, intent(in), optional :: maxiter_in, printmod_in
!     !local
!     real(8) :: e, fx, fxpe, dfx, change_x, change_x_old, s, x_old, fx_old
!     real(8), dimension(1:1000) :: xhist,fhist
!     real(8) :: toler, xtoler
!     integer :: maxiter, iter, flag, feval, hist_iter, loc(1), printmod
!     
! 
!     !default values
!     printmod = 1
!     xtoler = 1.d-12
!     toler = 1.d-6
!     maxiter = 200
!     if (present(printmod_in)) printmod = printmod_in
!     if (present(xtoler_in)) xtoler = xtoler_in
!     if (present(toler_in)) toler = toler_in
!     if (present(maxiter_in)) maxiter = maxiter_in
!     
!     flag = 1
!     feval = 1
! 
!     change_x_old = huge(0.d0)
!     change_x = huge(0.d0)
!     hist_iter = 0
!     if (printmod>=1) write(*,*) 'sub_secant______________________________________________'
!     do iter = 1,maxiter
! 
!         !Compute fx at guess
!         fx = f(x)
! 
!         !Tracking progress
!         feval = feval + 1
!         hist_iter = hist_iter+1
!         xhist(hist_iter) = x
!         fhist(hist_iter) = fx        
!         if (hist_iter>size(xhist)) STOP 'increase secant history length'
! 
!         if (printmod>=1) write(*,'(A,i4,A,e12.5,A,e12.5)') ' iter ',iter,'  x: ',x,'  fx: ',fx
! 
!         !Check for goodness of fit
!         if (abs(fx)<toler) then
!             flag = 0
!             exit
!         end if
! 
!         !Check if have cycling
!         if (abs(change_x)>abs(change_x_old)) then
!             if ((change_x<0.d0 .AND. change_x_old<0.d0) .or. (change_x>0.d0 .AND. change_x_old>0.d0)) then
!                 if (printmod>=1) write(*,*) 'It appears you are accelerating away from a root.  Switching to sub_bisect.'
!             else
!                 if (printmod>=1) write(*,*) 'It appears you are cycling around a root.  Switching to sub_bisect.'
!             end if
!             call bisect()
!             hist_iter = hist_iter+1
!             xhist(hist_iter) = x
!             fhist(hist_iter) = 0.d0 !This esnures x is selected below
!             exit
!         end if
!         change_x_old = change_x
! 
!         !Approximate derivative if first step
!         if (iter==1) then
!             e = .0001d0 !Step size        
!             fxpe = f(x+e)
!             dfx = (fxpe-fx)/e        
!             s = -fx/dfx
!             !Tracking progress
!             feval = feval + 1
!             hist_iter = hist_iter+1
!             xhist(hist_iter) = x+e
!             fhist(hist_iter) = fxpe        
!             if (hist_iter==size(xhist)) STOP 'increase secant history length'
!         else
!             s = -fx*(x-x_old)/(fx-fx_old)
!         end if
! 
!         !Set x' = x + s
!         x_old = x
!         fx_old = fx
!         change_x = s
!         x = x_old + s
! 
!     end do
! 
!     loc = minloc(abs(fhist(1:hist_iter)))
!     x = xhist(loc(1))
!     fx = fhist(loc(1))
! 
!     select case (flag)
!         case (0)
!             if (printmod>=1) write(*,*) 'Converged successfully, fx: ', fx
!         case default
!             if (printmod>=1) write(*,*) 'Failed to converge, fx: ', fx
!      end select
!     if (printmod>=1) write(*,*) '________________________________________________________'
! 
!     if (printmod>=1) call plot(xhist(1:hist_iter),fhist(1:hist_iter))
! 
! contains
!     !Locate bounds a,b if they exist and call sub_bisect
!     subroutine bisect()
!         implicit none
!         real(8) :: a,b,fa,fb
!         logical :: call_bisect
! 
!         !Get bounds from xhist if possible
!         if (any(fhist(1:hist_iter)<0.d0) .and. any(fhist(1:hist_iter)>0.d0)) then
!             call_bisect = .TRUE.
! 
!             loc = maxloc(fhist(1:hist_iter),fhist(1:hist_iter)<0.d0)
!             a = xhist(loc(1))
!             fa = fhist(loc(1))
!             loc = minloc(fhist(1:hist_iter),fhist(1:hist_iter)>0.d0)
!             b = xhist(loc(1))
!             fb = fhist(loc(1))
! 
!         !If that failed, use bounds if present
!         elseif (present(a_in) .AND. present(b_in)) then
!             call_bisect = .TRUE.
! 
!             a = a_in
!             b = b_in
!             if (present(fa_in) .AND. present(fb_in)) then
!                 fa = fa_in
!                 fb = fb_in
!             else 
!                 fa = f(a)
!                 fb = f(b)
!             end if
!         
!         !If neither present alert the programmer and do nothing
!         else
!         
!             write(*,*) 'sub_secant_1dim: WARNING: bracket not provided and not found endogenously.  Trying large bounds.'
!             call_bisect = .TRUE.
!             a = -1d10 
!             b = 1d10
! 
!         end if
! 
!         if (call_bisect) then
!             call sub_bisect(x,f,a,b,toler,maxiter,fa,fb,xtoler,printmod)
!         end if
! 
!     end subroutine bisect
! 
! end subroutine sub_secant


end module mod_root1dim

! Contains test routines for mod_root1dim
module mod_root1dim_test
    use mod_root1dim
    implicit none
contains
    
    ! One dimensional test functions
    function f0(x)
        implicit none
        real(8) :: f0
        real(8), intent(in) :: x
        f0 = sin(x) + cos(x)
    end function 
    function f1(x)
        real(8) :: f1
        real(8), intent(in) :: x
        f1 = sign(sqrt(abs(x-5d0)),x-5d0)
    end function 
    function f2(x) 
        real(8) :: f2
        real(8), intent(in) :: x
        f2 = -10d0 + 2d0*x + x**2 
    end function 
    function f3(x)
        real(8) :: f3
        real(8), intent(in) :: x
        f3 = exp(x) - 2d0 
    end function 
    ! f4 has multiple roots
    function f4(x)
        real(8) :: f4
        real(8), intent(in) :: x
        f4 = sin(x) 
    end function 


    ! Main routine
    subroutine test_root1dim() 
        implicit none
        real(8) :: a,b
        logical :: tf(3)
        
        a = -10; b = 20; if (.not. all(passed(f0))) STOP 'test_root1dim: f0 test failed'
        a = -100; b = 1000; if (.not. all(passed(f1))) STOP 'test_root1dim: f1 test failed'
        a = -6; b = 2; if (.not. all(passed(f2))) STOP 'test_root1dim: f2 test failed' 
        a = -100; b = 100; if (.not. all(passed(f3))) STOP 'test_root1dim: f3 test failed' 
        a = -100; b = 100; if (.not. all(passed(f4))) STOP 'test_root1dim: f4 test failed' 

        print*, 'test_root1dim: passed'

    contains 
        function passed(f) result(tfout)
            implicit none
            logical :: tf(3), tfout(3)
            real(8) :: xstar(3)
            interface
                function f(x)
                    implicit none
                    real(8), intent(in) :: x
                    real(8) :: f
                end function f
            end interface
            
            tf = .true.
            
            call sub_bisect( xstar(1),f,a,b,xtoler_in=2d0*epsilon(0d0),printmod_in=1) 
            call sub_brent(  xstar(2),f,a,b,xtoler_in=2d0*epsilon(0d0),printmod_in=1) 
            call sub_ridders(xstar(3),f,a,b,xtoler_in=2d0*epsilon(0d0),printmod_in=1)
            
            if (abs(f(xstar(1)))>1d-4) tf(1) = .false.
            if (abs(f(xstar(2)))>1d-3) tf(2) = .false.
            if (abs(f(xstar(3)))>1d-4) tf(3) = .false.

!             print*,'xstar',xstar
!             print*,'f(xstar)',f(xstar(1)),f(xstar(2)),f(xstar(3))
!             print*,'tf',tf

            tfout = tf

        end function passed
        
    end subroutine test_root1dim
    
    subroutine test_findallroots()
        use mod_matlab, only: linspace,pi
        implicit none
        real(8) :: roots(1000)
        integer :: nroots,iroots

        call sub_findallroots(roots,nroots,f4,linspace(-5d0,5d0,1000))

        ! since f4 is just the sin function, the roots are any multiples of pi.
        ! with a range of -5d0,5d0, there are three roots (-pi,0,pi)
    
        if (nroots/=3d0) STOP 'test_findallroots: test failed'
        if (any(abs(roots(1:3)-[-pi,0d0,pi])>1d-7)) STOP 'test_findallroots: test 2 failed'

        print*, 'test_findallroots: passed'

!         do iroots = 1,nroots
!             print*,'roots(iroots)',roots(iroots)
!             print*,'f4(roots(iroots))',f4(roots(iroots))
!         end do

    end subroutine

end module mod_root1dim_test
