!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! mod_rootNdim
! Contains subroutines for solving the nonlinear system of equations
! f(x) = 0 where f:Rn->Rn.
!
! Powell uses the MINPACK routine hybrd and is based off hybrd1.  
!
module mod_rootNdim
    implicit none

    abstract interface
        ! Function
        function rootN_f_template(x) result(y)
            implicit none
            real(8), intent(in) :: x(1:)
            real(8) :: y(size(x))
        end function 

        ! Jacobian
        ! NOTE: the jacobian is organized such that the partial derivative of f wrt 
        !       x_j should be in column J
        function rootN_fjac_template(x) result(J)
            implicit none
            real(8), intent(in) :: x(1:)
            real(8) :: J(size(x),size(x))
        end function
    end interface



contains

! Hompack's fixpdf method
subroutine sub_fixpdf(x,f,fjac_in) !,toler_in,xtoler_in,maxiter_in,printmod_in,eps_in,maxfev_in,factor_in,fjac_in)
    use mod_hompack_helper
    implicit none   
    procedure(rootN_f_template) :: f
    procedure(rootN_fjac_template), optional :: fjac_in
    real(8), dimension(1:), intent(inout) :: x    

    ! local
    real(8) :: arctol,eps
    real(8), dimension(size(x)+1) :: yp,ypold,tz,pivot,y,wt,p
    real(8), dimension(size(x),size(x)+1) :: qr
    real(8), dimension(size(x)+1,16) :: phi
    real(8), dimension(size(x)) :: alpha
    integer :: n,iflag,trace,ndima
    real(8) :: a(size(x)),par(1),arclen
    integer :: ipar(1),nfe

    real(8), parameter :: step = 1d-6

    n = size(x)
    iflag = -1
    arctol = 1d-5 !1d-10 !1d-8
    eps = 1d-6 !0d0 !1d-10
    trace = 6
    ndima = size(a)

    ! point sub_hompack_function to the internal routine subf 
    ! then calls to fixpdf will use subf and implicitly f.
    sub_hompack_function => subf

    if (present(fjac_in)) then
        sub_hompack_jac_function => subfjac
    else
        sub_hompack_jac_function => subfjacdiff
    end if
    
    Y(2:n+1) = x
    call fixpdf(N,Y,IFLAG,ARCTOL,EPS,TRACE,A,NDIMA,NFE,&
         ARCLEN,YP,YPOLD,QR,ALPHA,TZ,PIVOT,WT,PHI,P,PAR,IPAR)
    x = Y(2:n+1)

    print*,'lambda',y(1)
    print*,'iflag',iflag
    print*,'nfe',nfe
    print*,'x',x

    select case (iflag)
        case (-2)
             print*,'CAUSES FIXPDF TO INITIALIZE EVERYTHING FOR THE PROBLEM '//&
             'RHO(A,LAMBDA,X) = 0 (USE ON FIRST CALL).'
        case (-1)
             print*,'CAUSES FIXPDF TO INITIALIZE EVERYTHING FOR THE PROBLEM '//&
             'F(X) = 0 (USE ON FIRST CALL).'
        case ( 0)
             print*,'CAUSES  FIXPDF  TO INITIALIZE EVERYTHING FOR THE PROBLEM '//&
             'X = F(X) (USE ON FIRST CALL).'
        case ( 1)
             print*,'NORMAL RETURN.'
        case ( 2)
             print*,'SPECIFIED ERROR TOLERANCE CANNOT BE MET. EPS HAS BEEN'//&
             'INCREASED TO A SUITABLE VALUE. TO CONTINUE, JUST CALL ' //&
             'FIXPDF  AGAIN WITHOUT CHANGING ANY PARAMETERS.'
        case ( 3)
             print*,'STEPS  HAS BEEN CALLED 1000 TIMES.  TO CONTINUE, CALL '//&
             'FIXPDF  AGAIN WITHOUT CHANGING ANY PARAMETERS.'
        case ( 4)
             print*,'JACOBIAN MATRIX DOES NOT HAVE FULL RANK.  THE ALGORITHM '//&
             'HAS FAILED (THE ZERO CURVE OF THE HOMOTOPY MAP CANNOT BE '//&
             'FOLLOWED ANY FURTHER).'
        case ( 5)
             print*,'EPS  (OR  ARCTOL ) IS TOO LARGE.  THE PROBLEM SHOULD BE '//&
             'RESTARTED BY CALLING  FIXPDF  WITH A SMALLER  EPS  (OR '//&
             'ARCTOL ) AND  IFLAG = 0 (-1, -2).'
        case ( 6)
             print*,'I - DF(X)  IS NEARLY SINGULAR AT THE FIXED POINT (DF(X) IS'//&
             'NEARLY SINGULAR AT THE ZERO, OR  D RHO(A,LAMBDA,X)/DX  IS'//&
             'NEARLY SINGULAR AT  LAMBDA = 1 ).  ANSWER MAY NOT BE'//&
             'ACCURATE.'
        case ( 7)
             print*,'ILLEGAL INPUT PARAMETERS, A FATAL ERROR.'
        case default
    end select

contains

    subroutine subf(x,v)
        implicit none
        real(8) :: x(1),v(1)

        v(1:n) = f(x(1:n))

!         iter = iter+1

        write(*,'(A)',advance='no') 'subf: x,v'
        write(*,'(99f10.6)') x(1:n),v(1:n)

    end subroutine

    subroutine subfjac(x,v,k)
        implicit none
        real(8) :: x(1),v(1)
        integer :: k
        ! local
        real(8) :: jac(n,n)

        jac = fjac_in(x(1:n))
        v(1:n) = jac(:,k)

!         write(*,'(A)',advance='no') 'subfjac: x,v'
!         write(*,'(99f10.6)') x(1:n),v(1:n)

!         iter = iter+1

    end subroutine

    subroutine subfjacdiff(x,v,k)
        implicit none
        real(8) :: x(1),v(1)
        integer :: k
        ! local
        real(8) :: xscaltmp,h

        v(1:n) = f(x(1:n))

        xscaltmp = x(k) 

        h = max(1d0,abs(xscaltmp))*step
        x(k) = x(k) + h

        v(1:n) = (f(x(1:n)) - v(1:n))/h

        x(k) = xscaltmp

!         write(*,'(A)',advance='no') 'subfjacdiff: x,v'
!         write(*,'(99f10.6)') x(1:n),v(1:n)

!         iter = iter+1

    end subroutine

!     !optional parameters
!     real(8), intent(in), optional :: toler_in, xtoler_in, eps_in, factor_in
!     integer, intent(in), optional :: maxiter_in, printmod_in, maxfev_in
!     !local 
!     real(8) :: toler, xtoler, eps, factor
!     integer :: maxiter, printmod, maxfev, mode
!     ! more local
!     procedure(rootN_f_template), pointer :: pow_f_ptr => NULL()
!     procedure(rootN_fjac_template), pointer :: pow_fjac_ptr => NULL()
!     integer :: pow_f_eval_count, pow_f_printmod
!     real(8) :: pow_f_toler
!     real(8) :: pow_bestx(1000) ! max size
! 
!     ! Error check
!     if (omp_get_num_threads()/=1) STOP 'ERROR: sub_powell might not be thread safe, '//&
!                                     'depends on handling of threadprivate procedure pointers'
!     if (size(pow_bestx)<size(x)) STOP 'ERROR: sub_powell: size of pow_bestx must be increased'
!     
!     ! Point the module variable pow_f_ptr to the function f
!     pow_f_ptr => f 
!     if (present(fjac_in)) pow_fjac_ptr => fjac_in
! 
!     !set defaults for optional parameters
!     toler = 1d-12   !function termination criterion |f|<toler
!     eps = 1d-12    !precision of f function.  if <=0, then use machine precision ??or is this the tolerence??
!     xtoler = 1d-12   !x termination criterion. it is relative the values of x
!     maxiter = 200  !max # of iterations in powell
!     maxfev = 200*size(x) !max # of function evaluations
!     printmod = 1   !print modulus.  if -1, don't print anything.    
!     factor = 100d0 !Used in determining initial step bound.  Should be in (.1d0,100d0).  100d0 recommended.
!     mode = 1       !mode 1 <=> var scaled internally, mode 2 <=> scale specified by input diag
! 
!     !override default values if argument present
!     if (present(toler_in)) toler = toler_in
!     if (present(xtoler_in)) xtoler = xtoler_in
!     if (present(eps_in)) eps = eps_in
!     if (present(maxfev_in)) maxfev = maxfev_in
!     if (present(maxiter_in)) maxiter = maxiter_in
!     if (present(printmod_in)) printmod = printmod_in
!     if (present(factor_in)) factor = factor_in
!        
!     !check for inappropriate parameters
!     if (toler<=0d0 .OR. xtoler<=0d0) STOP 'ERROR: POW: inappropriate parameters'
! 
!     pow_f_eval_count = 0
!     pow_f_printmod = printmod
!     pow_f_toler = toler
!     if (.not. present(fjac_in)) then
!         ! Finite difference jacobian, call hybrd
!         call hybrd_driver(x,xtoler,maxfev,eps,mode,factor,0) ! the 0 is a print control
!     else
!         ! Analytic jacobian, call hybrj
!         call hybrj_driver(x,xtoler,maxfev,mode,factor,1) ! the 1 is a print control
!     end if
! 
! contains 
! 
!     ! To call this routine, the user should have already pointed pow_f_ptr
!     ! to the correct function.
!     subroutine hybrd_driver(x,xtol,maxfev,epsfcn,mode,factor,nprint)
!         implicit none
!         real(8), intent(inout) :: x(1:)
!         integer, intent(in) :: maxfev,mode,nprint
!         real(8), intent(in) :: xtol,factor,epsfcn
!         ! local
!         integer :: n,ldfjac,nfev,lr,info,ml,mu
!         real(8), dimension(:), allocatable :: fvec,diag,r,qtf,wa1,wa2,wa3,wa4
!         real(8), dimension(:,:), allocatable :: fjac
! 
!         n = size(x)
!         ldfjac = n
!         lr = (n*(n+1))/2
! 
!         ! For non-banded jacobian, set ml and mu to at least n-1
!         ml = n-1
!         mu = n-1
!         
!         allocate(fvec(n),fjac(ldfjac,n),diag(n),r(lr),&
!                  qtf(n),wa1(n),wa2(n),wa3(n),wa4(n))
! 
!         if (pow_f_printmod>0) then 
!             write(*,*) 'POW:________________________________________________________________'
!             write(*,*) 'POW: Beginning Powell''s Hybrid Method                              '
!             write(*,*) 'POW:   iter       normx       normf'
!         end if
! 
!         call hybrd(hybrd_private,n,x,fvec,xtol,maxfev,ml,mu,epsfcn,diag,mode, &
!                    factor,nprint,info,nfev,fjac,ldfjac,r,lr, &
!                    qtf,wa1,wa2,wa3,wa4)
! 
!         !output results    
!         if (pow_f_printmod>0) then
!             select case (info)
!                 case (0)
!                     write(*,*) 'POW: Improper input parameters.'
!                 case (1)                
!                     write(*,*) 'POW: Successfully converged: relative error between two iterates is < xtol'
!                 case (2)
!                     write(*,*) 'POW: Failed to converge: max fevals reached '
!                 case (3)
!                     write(*,*) 'POW: Failed to converge: xtol is too small, no further improvement possible '
!                 case (4)
!                     write(*,*) 'POW: Failed to converge: iteration is not making good progress'
!                     write(*,*) 'POW: as measured by the improvement from the last five jacobian evals.'
!                 case (5)
!                     write(*,*) 'POW: Failed to converge: iteration is not making good progress'
!                     write(*,*) 'POW: as measured by the improvement from the last ten iterations.'
!                 case (-1)
!                     ! This is a user-defined (me) case
!                     x = pow_bestx(1:n)
!                     write(*,*) 'POW: Successfully converged: max|f|<toler. Note: last output may not '
!                     write(*,*) 'POW: correspond to final x.'
!             end select
! 
!         elseif (pow_f_printmod==0 .and. abs(info) /= 1) then        
! 
!             write(*,*) 'POW: Error in powell.  Set printmod>0 for more info.  Exit flag:', info
! 
!         endif
! 
!     end subroutine 

end subroutine sub_fixpdf

!     ! Smolyak root finding 
!     ! 
!     ! Smolyak root finding computes a smolyak collocation projection of the function f on a hypercube
!     ! defined by x +- e. Subsequently, the projection (which is an infinitely differentiable
!     ! function) is optimized using derivative based methods. After the projection is optimized,
!     ! the hypercube is recentered around the new x. The projection is applied again on the 
!     ! hypercube x +- e. If little improvement is made, the hypercube is shrunk: e = shrink*e and
!     ! the process begins again (terminated when the L2 norm of e falls below xtoler).
!     ! 
!     ! The philosophy behind this is that if the underlying function is "roughly smooth" --- 
!     ! i.e. almost differentiable but maybe not because of noise --- that we can get a close
!     ! approximation by the smolyak function and then optimize that. The smolyak projection (verse
!     ! a tensor product approx) is used because it is suitable for high dimension problems.
!     ! 
!     ! This is totally heuristic --- although I believe it could be formalized. I believe it is
!     ! my creation, but it is a fairly obvious.
!     !
!     ! The user may choose e, shrink, and also "mu", which is the degree of the smolyak 
!     ! approximation used. mu should generally be at least 2, because mu==1 has no crossterms
!     ! in the approximation. Also, for higher order approximations, there may be a lot of local
!     ! minima. Generally, the default value mu==2 is recommended. a,b, if present, are bounds
!     ! on the problem that are enforced in both in the smolyak approximation construction and 
!     ! the inner maximization (which is the simplex method currently).
!     ! 
!     subroutine sub_smolyakroot(x,f,e_in,a_in,b_in,shrink_in,mu_in,xtoler_in,maxiter_in,printmod_in)
!         use mod_matlab
!         use mod_smolyak
!         use mod_minNdim
!         implicit none
!         real(8), dimension(1:), intent(inout) :: x
!         procedure(rootN_f_template) :: f
!         real(8), dimension(1:size(x)), intent(in), optional :: e_in,a_in,b_in
!         real(8), intent(in), optional :: xtoler_in
!         integer, intent(in), optional :: maxiter_in,printmod_in
!         real(8), intent(in), optional :: shrink_in
!         integer, intent(in), optional :: mu_in
!         !local
!         type(smolyakFcn) :: sApprox ! Smolyak approximation
!         real(8) :: xtoler
!         integer :: maxiter,printmod
!         real(8), dimension(1:size(x)) :: e, xbest, a, b, hcubelb, hcubeub, xtmp
!         real(8) :: shrink, enorm, snorm, fnorm, fbestnorm
!         integer :: n,i,iter,fevals,mu,inner,ncoll
!         real(8), allocatable :: coll(:,:),fcoll(:,:)
!         
!         n = size(x)        
! 
!         ! Set defaults and overwrite if present
!         mu = 2;                     if (present(mu_in)) mu = mu_in
!         e = max(.1d0*abs(x),.01d0); if (present(e_in)) e = e_in
!         a = -1d5;                   if (present(a_in)) a = a_in
!         b = 1d5;                    if (present(b_in)) b = b_in
!         shrink = .5d0;              if (present(shrink_in)) shrink = shrink_in
!         maxiter = 1000;             if (present(maxiter_in)) maxiter = maxiter_in
!         xtoler = 1d-6;              if (present(xtoler_in)) xtoler = xtoler_in
!         printmod = 1;               if (present(printmod_in)) printmod = printmod_in
! 
!         ! Check inputs and try to recover when appropriate
!         if (any(e<0d0)) then
!             print*,'WARNING: e has some negative, replacing'
!             e = abs(e)
!         end if
!         if (any(e==0d0)) STOP 'SS: ERROR: e is equal to 0'
! 
!         ! Init
!         fevals = 0
! 
!         if (printmod>0) then
!             write(*,'(A)') 'SS:__________________________________________________________________'            
!             write(*,'(A)') 'SS:Beginning Smolyak Projection Root Method                          '              
!             write(*,'(A)') 'SS:    iter inner  fevals    fbestnorm        snorm                '          
!         end if 
! 
!         ! Do the smolyak search
!         OUTER: do iter = 1,maxiter
! 
!             ! Exit if xtoler exceeded 
!             enorm = sqrt(sum(e**2))
!             if (enorm < xtoler) exit 
! 
!             ! Inner loop: updating the hypercube center successively 
!             do inner = 1,20
! 
!                 ! Get the smolyak grid
!                 hcubelb = max(x-e,a)
!                 hcubeub = min(x+e,b)
!                 call smolyakapprox_step1(coll,sApprox,n,mu,hcubelb,hcubeub)
!                 ncoll = size(coll,1)
! 
!                 ! Evaluate the function at each collocation point
!                 if (allocated(fcoll)) deallocate(fcoll)
!                 allocate(fcoll(n,ncoll))
!                 do i = 1,ncoll
!                     fcoll(:,i) = f(coll(i,:))
!                     fevals = fevals + 1
!                 end do
! 
!                 ! Construct the smolyak approximation
!                 call smolyakapprox_step2(fcoll,sApprox,tF='T')
! 
! ! call disp(hcubelb)
! ! call disp(hcubeub)
! ! call disp(fcoll)
! ! call disp(transpose(fcoll))
! ! do i = 1,ncoll
! !     print*, (fapprox(coll(i,:)))
! ! end do
! ! STOP 'fcoll'
! 
!                 ! Minimize the approximation within the hypercube 
!                 !!! Start from every collocation point and find the best of the approx
!                 do i = 1,ncoll
!                     xtmp = coll(i,:)
!                     call sub_lsqnonlin(xtmp,fapprox,xtoler_in=1d-11,printmod_in=-1,a_in=hcubelb,b_in=hcubeub,maxiter_in=1000)
!                     coll(i,:) = xtmp
!                 end do
! 
!                 !!! Now have ncoll proposal roots. Evaluate each proposal using the actual f function.
!                 fbestnorm = sqrt(sum(f(coll(1,:))**2)) 
!                 xbest = coll(1,:)
!                 do i = 2,ncoll
!                     ! If there is *significant* disagreement between the current best min and the next, 
!                     ! evaluate it using the true function.
!                     if (sqrt(sum((coll(i,:) - xbest)**2))>xtoler) then
!                         fnorm = sqrt(sum(f(coll(i,:))**2))
!                         fevals = fevals + 1
!                         if (fnorm<fbestnorm) then
!                             xbest = coll(i,:)
!                             fbestnorm = fnorm
!                         end if
!                     end if
!                 end do
! 
!                 ! Also try to do simulated annealing to ensure we find the global
!                 xtmp = xbest
!                 call sub_simanneal(xtmp,fSSRSapprox,e_in=.01d0*(hcubeub-hcubelb),maxiter_in=10000,printmod_in=-1)
!                 fnorm = sqrt(sum(f(xtmp)**2))
!                 fevals = fevals + 1
!                 if (fnorm<fbestnorm) then
!                     xbest = xtmp
!                     fbestnorm = fnorm
!                 end if
!                 
! 
! ! print*,'xbest'
! ! call disp(xbest)
! ! print*,'fbestnorm',fbestnorm
! 
!                 ! This is inefficient to do, but for info sake and ease, do it here (could save this value since will be a 
!                 ! collocation point next time
! !                 fbest = f(xbest)
! !                 fevals = fevals + 1
! 
!                 ! To ensure we always descend according to the *real* function values rather
!                 ! than the *approximate* function values, we can do this: 
!                 ! If this new "xbest" is worse than one of the collocation points we checked,
!                 ! replace it. 
!                 ! Note that with this, smolyak is at worst a grid search in several important
!                 ! dimensions.
! !                 tmpi1 = minloc(fcoll)
! !                 if (fcoll(tmpi1(1))<fbest) then
! !                     xbest = coll(tmpi1(1),:) 
! !                     fbest = fcoll(tmpi1(1))
! !                 end if
! ! 
! !                 snorm = sqrt(sum((xbest-xoldbest)**2))
! ! 
! !                 xoldbest = xbest
! !                 foldbest = fbest
! ! STOP 'checking'
! 
!                 snorm = sqrt(sum(xbest - x)**2)
! 
!                 x = xbest
! 
!                 ! Update status 
!                 if (printmod>0 .and. (iter==1 .or. mod(iter,printmod)==0)) &
!                     write(*,'(A,i5,i6,i8,2f13.7)') 'SS:   ',iter, inner, fevals, fbestnorm, snorm
! 
!                 ! Exit inner loop if best x has not changed much
!                 ! if (xnorm<maxval(abs(xbest))*(1d0 + 1d-3)) exit
!                 if (snorm<100d0*xtoler) exit
! 
!             end do 
!                 
!             ! Shrink the hypercube dimensions
!             e = shrink*e
! 
!         end do OUTER
! 
!         if (printmod>0) then
!             write(*,'(A)',advance='no') 'SS: Final x:'
!             write(*,'(99d16.6)',advance='no') x
!             write(*,*)
!             write(*,'(A)',advance='no') 'SS: Final e:'
!             write(*,'(99d16.6)',advance='no') e
!             write(*,*)
!             write(*,'(A)') 'SS:__________________________________________________________________'            
!         end if
! 
!     contains 
!         
!         function fSSRSapprox(x)
!             implicit none
!             real(8), intent(in) :: x(1:)
!             real(8) :: fSSRSapprox
!             
!             if (any(x<hcubelb) .or. any(x>hcubeub)) then
!                 fSSRSapprox = 1d10
!                 return
!             end if
!             
!             fSSRSapprox = sum(fapprox(x)**2)
! 
!         end function 
!         function fapprox(x)
!             implicit none
!             real(8), intent(in) :: x(1:)
!             real(8) :: fapprox(size(x))
!             ! local
!             ! real(8) :: fvec(n,1) ! ,xmat(1,size(x))
! 
!             if (any(x<hcubelb) .or. any(x>hcubeub)) then
!                 fapprox = 1d10
!                 return
!             end if
! 
!             ! xmat(1,:) = x
! 
!             ! Evaluate the smolyak approximation at x
!             call smolyakapprox_step3(fapprox,sApprox,x)
!             ! fapprox = fvec(:,1)
! 
!         end function
!     
!     end subroutine sub_smolyakroot

    ! Least Squares Nonlinear Minimization 
    ! 
    ! Uses the MKL Trust-Region Algorithm
    ! 
    ! Minimize the sum of squared residuals sum f_i^2 of a function f:R^n->R^m with or without
    ! boundary constraints
    ! 
    ! If printmod_in = 0, then all information is printed
    ! eps_in: controls the step size when taking the jacobian. should be the same order of magnitude
    !         as error in the function.
!     subroutine sub_lsqnonlin(x,f,a_in,b_in,xtoler_in,maxiter_in,printmod_in,eps_in,robust_in)
!         use mod_deriv
!         use mod_matlab
!         implicit none
!         include "mkl_rci.fi" 
!         real(8), dimension(1:), intent(inout) :: x
!         procedure(rootN_f_template) :: f
!         real(8), dimension(1:size(x)), intent(in), optional :: a_in,b_in
!         real(8), intent(in), optional :: xtoler_in,eps_in
!         integer, intent(in), optional :: maxiter_in,printmod_in
!         logical, intent(in), optional :: robust_in
!         !local
!         integer*8 :: handle 
!         integer :: res, maxIterTrial, n, maxiter, printmod, m, RCI_Request, info(6), iter, i, fevals
!         real(8), dimension(1:size(x)) :: a, b, lastx
!         real(8) :: eps(6), initTrust, xtoler, relErrF
!         real(8), allocatable :: fvec(:),fjac(:,:),wa(:,:)
!         logical :: robust
! 
!         n = size(x)        
!         m = n ! I am only enforcing this for convenience. Can have m>=n with this algorithm
! 
!         allocate(fvec(m),fjac(m,n),wa(n,2))
!         fvec = 0d0
!         fjac = 0d0
! 
!         ! Set defaults and overwrite if present
!         a = -huge(0d0);             if (present(a_in)) a = a_in
!         b = huge(0d0);              if (present(b_in)) b = b_in
!         maxiter = 1000;             if (present(maxiter_in)) maxiter = maxiter_in
!         xtoler = 1d-8;              if (present(xtoler_in)) xtoler = xtoler_in
!         printmod = 1;               if (present(printmod_in)) printmod = printmod_in
!         relErrF = 1d-12;            if (present(eps_in)) relErrF = eps_in  ! This is used in the Jacobian calculation
!                                                                            ! routine to obtain a good step size
!         robust = .false.;           if (present(robust_in)) robust = robust_in 
!         initTrust = 100d0;  ! Should be between .1 and 100
!         maxIterTrial = 20 
! 
!         ! Stopping criteria 
!         eps(1) = 1d-10 ! trustRegionArea < eps(1)
!         eps(2) = 1d-10   ! L2 norm of f < eps(2)
!         eps(3) = 1d-10 ! L2 norm of Jacobian matrix(1:m,j) < eps(3) for j=1,n
!         eps(4) = xtoler ! L2 norm of trial step < eps(4)
!         eps(5) = 1d-10 ! Change in L2 norm of change in function value < eps(5)
!         eps(6) = 1d-4 ! The trial step precision  (eps(6)=0 => precision<=1d-10)
! 
!         ! Initialize the solver
!         res = dtrnlspbc_init(handle, n, m, x, a, b, eps, maxiter, maxIterTrial, initTrust)
!         select case (res)
!             case (TR_SUCCESS)
!             case (TR_INVALID_OPTION)
!                 STOP 'TR: ERROR: invalid option given'
!             case (TR_OUT_OF_MEMORY)
!                 STOP 'TR: ERROR: out of memory'
!             case default
!                 STOP 'TR: ERROR: did not find case when initializing'
!         end select
!         
!         if (printmod>=0) then
!             write(*,'(A)') 'LSQ:________________________________________________________________'
!             write(*,'(A)') 'LSQ: Beginning Least Squares Minimization Routine                   '
!         end if
!         
!         ! Solve
!         iter = 1
!         fevals = 0
!         lastx = -huge(0d0)
!         do 
! 
!             ! Check setup was okay
!             res = dtrnlspbc_check(handle, n, m, fjac, fvec, a, b, eps, info)
!             call checkInfo()
! 
!             ! Get the next step
!             res = dtrnlspbc_solve(handle, fvec, fjac, RCI_Request)
! 
!             ! Check for error in solve
!             if (res/=TR_SUCCESS) STOP 'LSQ: ERROR: LSQ method encountered an error in dtrnlsp_solve'
! 
!             ! Perform the requested action
!             select case (RCI_Request)
!                 case (2)
!                     if (printmod==0) write(*,'(A)') 'LSQ: Calculating the jacobian'
!                     ! Request for jacobian matrix computation
!                     if (any(x/=lastx)) then
!                         ! Need the fvec at x. If x/=lastx (which corresponds to most recent fvec value), then need to compute it
!                         fvec = f(x)
!                     end if
!                     if (.not. robust) then
!                         call sub_jacobian(x,f,fvec,fjac,eps_in=relErrF)
!                         fevals = fevals + n
!                     else
!                         call sub_jacobian_robust(x,f,fvec,fjac,eps_in=relErrF)
!                         fevals = fevals + 3*n
!                     end if
!                     if (printmod==0) then
!                         write(*,'(A)') 'LSQ: Jacobian:' 
!                         do i = 1,n
!                             write(*,'(A,99f16.7)') 'LSQ: ',fjac(i,:)
!                         end do
!                     end if
!                 case (1)
!                     ! Request for calculation of f at x
!                     fvec = f(x)
!                     lastx = x
!                     fevals = fevals + 1
!                     if (printmod==0) then
!                         write(*,'(A)') 'LSQ: x'
!                         write(*,'(A,99f16.7)') 'LSQ: ',x
!                         write(*,'(A)') 'LSQ: f at x'
!                         write(*,'(A,99f16.7)') 'LSQ: ',fvec
!                     end if
!                 case (0)
!                     ! One successful iteration step on the trust region (x may not have changed)
!                     if (printmod==0 .or. (printmod>0 .and. mod(iter,printmod)==0)) &
!                         write(*,'(A,i5,A,i5,A,f16.7)') 'LSQ: iter ',iter,' fevals ',fevals,' fnorm ',sum(fvec**2) 
!                     iter = iter + 1
!                 case (-1)
!                     ! Max iter reached
!                     if (printmod>=0) write(*,'(A)') 'LSQ: Max iterations reached.'
!                 case (-2)
!                     ! eps(1) exceeded
!                     if (printmod>=0) write(*,'(A)') 'LSQ: Trust region area too small'
!                 case (-3)
!                     ! eps(2) exceeded
!                     if (printmod>=0) write(*,'(A)') 'LSQ: L2 norm of f sufficiently small'
!                 case (-4)
!                     ! eps(3) exceeded
!                     if (printmod>=0) write(*,'(A)') 'LSQ: Jacobian matrix singular'
!                 case (-5)
!                     ! eps(4) exceeded
!                     if (printmod>=0) write(*,'(A)') 'LSQ: xtoler exceeded'
!                 case (-6)
!                     ! eps(5) exceeded
!                     if (printmod>=0) write(*,'(A)') 'LSQ: change in function value too small'
!                 case default
!                     STOP 'LSQ: ERROR: did not find case'
!             end select
! 
!             if (RCI_Request<0) exit
! 
!         end do
! 
!         ! Release info
!         res = dtrnlspbc_delete(handle)
!         if (res/=TR_SUCCESS) STOP 'LSQ: ERROR: could not free memory'
! 
! 
!     contains
!         
!         subroutine checkInfo()
!             implicit none
!             if (all(info==0)) return
! 
!             if (info(1)==1) STOP 'LSQ: ERROR: invalid handle'
!             if (info(2)==1) STOP 'LSQ: ERROR: fjac is not allocated'
!             if (info(2)==2) STOP 'LSQ: ERROR: fjac contains NaN'
!             if (info(2)==3) STOP 'LSQ: ERROR: fjac contains Inf'
!             if (info(3)==1) STOP 'LSQ: ERROR: fvec is not allocated'
!             if (info(3)==2) STOP 'LSQ: ERROR: fvec contains NaN'
!             if (info(3)==3) STOP 'LSQ: ERROR: fvec contains Inf'
!             if (info(4)==1) STOP 'LSQ: ERROR: a is not allocated'
!             if (info(4)==2) STOP 'LSQ: ERROR: a contains NaN'
!             if (info(4)==3) STOP 'LSQ: ERROR: a contains Inf'
!             if (info(4)==4) then
!                 print*,'a:'
!                 call disp(a)
!                 print*,'b:'
!                 call disp(b)
!                 STOP 'LSQ: ERROR: b<=a'
!             end if
!             if (info(5)==1) STOP 'LSQ: ERROR: b is not allocated'
!             if (info(5)==2) STOP 'LSQ: ERROR: b contains NaN'
!             if (info(5)==3) STOP 'LSQ: ERROR: b contains Inf'
!             if (info(5)==4) then
!                 print*,'a:'
!                 call disp(a)
!                 print*,'b:'
!                 call disp(b)
!                 STOP 'LSQ: ERROR: b<=a'
!             end if
!             if (info(6)==1) STOP 'LSQ: ERROR: eps array not allocated'
!             if (info(6)==2) STOP 'LSQ: ERROR: eps contains NaN'
!             if (info(6)==3) STOP 'LSQ: ERROR: eps contains Inf'
!             if (info(6)==4) STOP 'LSQ: ERROR: eps contains value <= 0'
!         
!         end subroutine
! 
!     end subroutine


! subroutine sub_homotopy(f,x)
!     use mod_matlab
!     implicit none
!     interface 
!         function f(x) result(feval)
!             implicit none 
!             real(8), intent(in) :: x(1:)
!             real(8) :: feval(1:size(x))
!         end function f
!     end interface
!     real(8), intent(inout) :: x(1:)
!     !local
!     integer :: n !Dim of problem
!     real(8) :: hs, hy(size(x)+1) !Step for parametric path (hs) and for derivative
!     real(8) :: t !Weighting parameter between easy and complicated function
!     real(8) :: y(size(x)+1), yp(size(x)+1)  ! (x,t) and x't'
!     real(8) :: J(size(x),size(x)+1) !Jacobian of the system
!     real(8) :: Jnoi(size(x),size(x)) !Jacobian of the system with the ith column removed
!     real(8) :: dyds(size(x)+1) !derivative of y wrt the parameter
!     real(8) :: gy(size(x)), gyp(size(x)) !Value at y and yp
!     real(8) :: xp(size(x)), tp, x0(size(x))
!     integer :: ind, iter, printmod
!     
!     printmod = 1
! 
!     n = size(x)
!     hs = 1d-4
!     hy = 1d-6
! 
!     !Starting conditions
!     t = 0
!     x0 = x !x0 is initial guess
!     y = (/x,t/)
! 
!     !!!!!!!!!!!!!!!!!! END INIT
! 
!     do iter = 1,1000000
! 
!         x = y(1:n)
!         t = y(n+1)
! 
!         ! Compute the residual at the point
!         gy = (1.d0-t)*(x-x0) + t*f(x)
!         
! !         ! If the point is too far away, call fsolve to try to get closer
! !         if (maxval(abs(y))>1d-8) then
! !             y(1:n) = fsolve(g,y(1:n))
! !             x = y(1:n)
! !             t = y(n+1)
! !             gy = (1.d0-t)*(x-x0) + t*f(x)            
! !         end if
! 
!         ! Compute the jacobian of the system
!         do ind = 1,n+1
!             yp = y
!             yp(ind) = yp(ind) + hy(ind)
!             xp = yp(1:n)
!             tp = yp(n+1)
!             gyp = (1.d0-tp)*(xp-x0) + tp*f(xp)
!             J(:,ind) =  (gyp-gy)/hy(ind) ! indth row of jacobian
!         end do
! 
!         ! Compute how much to change y based on the Jacobian
!         do ind = 1,n+1
!             Jnoi(:,1:ind-1) = J(:,1:ind-1)
!             Jnoi(:,ind:n) = J(:,ind+1:n+1)
!             dyds(ind) = (-1.d0)**ind * det(Jnoi)
!         end do
!             
!         ! Compute an update for the state
!         yp = y + hs*dyds
! 
!         ! Write update
!         if (printmod>0 .and. mod(iter,printmod)==0) write(*,'(A,i6,A,e14.6,A,e14.6)') &
!             '* iter ',iter,' max(abs(x-fx)) ',maxval(abs(gy)), ' t ', t
! 
!         ! Check if t>1 (in which case we've jumped past t=1 which is the solution of interest)
!         if (yp(n+1)>1.d0) then
!             ! Set x to the best guess we have for t=1
!             t = y(n+1)
!             tp = yp(n+1)
!             x = y(1:n)
!             xp = yp(1:n)
!             x = x*(1.d0-t)/(tp-t) + xp*(tp-1.d0)/(tp-t)
! 
!             exit
!         end if
! 
!         ! Update y
!         y = yp
! 
!     end do
! 
! contains
! 
! function g(x) result(gy)
!     implicit none
!     real(8), dimension(1:), intent(in) :: x
!     real(8), dimension(1:size(y)) :: gy
! 
!     gy = (1.d0-y(n+1))*(x(1:n)-x0) + y(n+1)*f(x(1:n))    
! 
! end function
! 
! end subroutine

!Solves using either the gauss-seidel algorithm or the gauss-jacobi algorithm
!Repeatedly solves for 1 root to fi(x~)=0 where x~(j) = x(j) for j/=i
!If the solution is known to be bracketed, this information should be provided and can then be used.
! 
! Right now, I am assuming any value greater absolute value than "thresh" is an error. If 
! such a value is encountered, the routine will try creeping this value toward a feasible
! region to not have an error
subroutine sub_gauss(x,f,a,b,toler_in,xtoler_in,printmod_in,maxiter_in,method_in,thresh_in)
    use mod_root1dim
    implicit none
    procedure(rootN_f_template) :: f
    real(8), dimension(1:), intent(inout) :: x
    real(8), dimension(1:size(x)), intent(in) :: a,b

    !optional parameters
    real(8), intent(in), optional :: toler_in, xtoler_in, thresh_in
    integer, intent(in), optional :: maxiter_in, printmod_in
    character(len=*), intent(in), optional :: method_in

    !default parameters
    real(8) :: toler, xtoler
    integer :: maxiter, printmod
    character(len=30) :: method


    !local
    real(8), dimension(1:size(x)) :: fx, xold
    real(8) :: thresh, fai, fbi, ai,bi
    integer :: i, nx, iter
    integer :: j
    
    nx = size(x)
    maxiter = 1000
    toler = 1d-6
    xtoler = 1d-6 
    method = 'seidel'
    thresh = huge(0d0)
    printmod = 1

    if (present(toler_in)) toler = toler_in
    if (present(xtoler_in)) xtoler = xtoler_in
    if (present(thresh_in)) thresh = thresh_in
    if (present(maxiter_in)) maxiter = maxiter_in
    if (present(printmod_in)) printmod = printmod_in
    if (present(method_in)) method = method_in

    if (method/='seidel' .and. method/='jacobi') then 
        STOP 'GM:sub_gauss: no appropriate method specified.'
    end if

    if (printmod>=1) then 
        write(*,*) 'GM:____________________________________________________________'
        write(*,*) 'GM:sub_gauss___________________________________________________'
        write(*,*) 'GM:Using the gauss-',trim(method),' method'
        write(*,*) 'GM:iter   euclid(x)     euclid(fx)    min(fx)       max(fx)'
    end if
    do iter = 1,maxiter

        xold = x
        
        !Iteratively solve each equation
        do i = 1,nx
            if (printmod>=1) write(*,'(A,i5)') 'GM: solving f_i(x(i))=0 for i=', i 
            if (method=='seidel') then
                ai = a(i)
                bi = b(i)
                do j = 1,10000
                    fai = fi(ai)
                    if (abs(fai)>=thresh) then 
                        ai = min(a(i) + .00001d0*(b(i)-a(i)),b(i))
                    else
                        exit
                    end if
                    if (j==10000) STOP 'GM: ERROR: could not find a non error value'
                end do 
                do j = 1,10000
                    fbi = fi(bi)
                    if (abs(fbi)>=thresh) then
                        bi = max(b(i) - .00001d0*(b(i)-a(i)),a(i))
                    else
                        exit
                    end if
                    if (j==10000) STOP 'GM: ERROR: could not find a non error value'
                end do 
                call sub_ridders(x(i),fi,a(i),b(i),fa_in=fai,fb_in=fbi,toler_in=toler,xtoler_in=xtoler,&
                                 printmod_in=-1,maxiter_in=20)
            elseif (method=='jacobi') then
                ai = a(i)
                bi = b(i)
                do j = 1,10000
                    fai = fi(ai)
                    if (abs(fai)>=thresh) then 
                        ai = min(a(i) + .00001d0*(b(i)-a(i)),b(i))
                    else
                        exit
                    end if
                    if (j==10000) STOP 'GM: ERROR: could not find a non error value'
                end do 
                do j = 1,10000
                    fbi = fi(bi)
                    if (abs(fbi)>=thresh) then
                        bi = max(b(i) - .00001d0*(b(i)-a(i)),a(i))
                    else
                        exit
                    end if
                    if (j==10000) STOP 'GM: ERROR: could not find a non error value'
                end do 
                call sub_ridders(x(i),gi,a(i),b(i),fa_in=fai,fb_in=fbi,toler_in=toler,xtoler_in=xtoler,&
                                 printmod_in=-1,maxiter_in=20)
            end if
        end do
        
        !Evaluate f at x
        fx = f(x)

        !Report
        if (printmod>=1) write(*,'(A,i5,4e14.5)') &
            'GM:',iter, sqrt(sum(x**2)), sqrt(sum(fx**2)), minval(fx), maxval(fx)

        !Check convergence
        if (all(abs(fx)<toler)) exit
        if (all(abs(x-xold)<xtoler)) exit

    end do
    if (printmod>=1) then
        if (all(abs(fx)<toler)) then
            write(*,*) 'GM:Converged successfully______________________________________'
        elseif (all(abs(x-xold)<xtoler)) then
            write(*,*) 'GM:Failed to converge: xtoler exceeded_________________________'
        elseif (iter>=maxiter) then
            write(*,*) 'GM:Failed to converge: max iterations reached__________________'
        else
            write(*,*) 'GM:Failed to converge: unknown termination_____________________'
        end if
        write(*,*) 'GM:____________________________________________________________'
    end if

contains

    !This is totally illegal fortran, we'll see if ifort allows it
    !It compiles...
    function fi(xi)
        implicit none
        real(8), intent(in) :: xi
        real(8) :: fi
        !local
        real(8), dimension(1:nx) :: x_local, res
    
        x_local(1:i-1) = x(1:i-1)
        x_local(i+1:nx) = x(i+1:nx)
        x_local(i) = xi
        
        res = f(x_local)
        fi = res(i)        
    end function

    function gi(xi)
        implicit none
        real(8), intent(in) :: xi
        real(8) :: gi
        !local
        real(8), dimension(1:nx) :: x_local, res
    
        x_local(1:i-1) = xold(1:i-1)
        x_local(i+1:nx) = xold(i+1:nx)
        x_local(i) = xi
        
        res = f(x_local)
        gi = res(i)        
    end function

end subroutine sub_gauss


! Powell's hybrid method
!
! Input parameters:
! eps_in: determines step length for forward differencing. Errors in the function should be 
!         on the order of eps_in, so increase eps_in for larger step sizes. 
!         
subroutine sub_powell(x,f,toler_in,xtoler_in,maxiter_in,printmod_in,eps_in,maxfev_in,factor_in,fjac_in,info_out,uid_in,derivstep_in)
include 'include_powell.f90'

!!! There are two sub_powells because it is not thread safe and I some time need
!!! to call it more than one instance at a time.
subroutine sub_powell2(x,f,toler_in,xtoler_in,maxiter_in,printmod_in,eps_in,maxfev_in,factor_in,fjac_in,info_out,uid_in,derivstep_in)
include 'include_powell.f90'
subroutine sub_powell3(x,f,toler_in,xtoler_in,maxiter_in,printmod_in,eps_in,maxfev_in,factor_in,fjac_in,info_out,uid_in,derivstep_in)
include 'include_powell.f90'

! Broyden computes the Jacobian only one time and subsequently updates it.
! Powell tends to work better though. Also, when the Jacobian is singular, 
! this has problems.
subroutine sub_broyden(x,f,xtoler_in,maxiter_in,printmod_in,eps_in,relax_in)
    use mod_deriv
    use mod_matlab
    implicit none   
    procedure(rootN_f_template) :: f
    real(8), dimension(1:), intent(inout) :: x
    real(8), intent(in), optional :: eps_in, xtoler_in, relax_in
    integer, intent(in), optional :: maxiter_in, printmod_in
    !local
    real(8), allocatable :: J(:,:), s(:), fx(:), fxp(:)
    real(8), dimension(1:size(x),1) ::  s_mat, fxp_mat, fx_mat
    real(8) :: xtoler, eps, relax
    integer :: n, maxiter, iter, fevals, printmod
    
    n = size(x)

    allocate(J(n,n),s(n),fx(n),fxp(n))

    ! Set defaults and overwrite if present
    maxiter = 1000;   if (present(maxiter_in)) maxiter = maxiter_in
    xtoler = 1d-8;    if (present(xtoler_in)) xtoler = xtoler_in
    printmod = 1;     if (present(printmod_in)) printmod = printmod_in
    eps = 1d-4;       if (present(eps_in)) eps = eps_in  ! This is used in the Jacobian calculation routine to obtain 
                                                         ! a good step size
    relax = 0d0;      if (present(relax_in)) relax = relax_in


    write(*,'(A)') 'BR:________________________________________________________________'
    write(*,'(A)') 'BR: Beginning Broyden''s Method     '
    write(*,'(A)') 'BR:   iter       normx       normf'
    do iter = 1,maxiter

        if (iter==1) then

            !Compute fx at initial point
            fx = f(x)
            fevals = 1

            !If first iteration, compute the true Jacobian
            call sub_jacobian(x,f,fx,J,eps)
            fevals = fevals + n

            !Solve for s in J*s = -f(x)
            s = mldivide(J,-fx)

            !Set x' = x + s
            x = x + s*(1d0-relax)

        else

            !Compute residual after Newton step
            fx = f(x)

            !Solve for s in J*s = -f(x)
            s = mldivide(J,-fx)

            !Set x' = x + s
            x = x + s*(1d0-relax)
        
            !Compute f(x')
            fxp = f(x)
            fevals = fevals + 1

            ! Use already computed Jacobian to update the Jacobian
            s_mat(:,1) = s*(1d0-relax)
            fxp_mat(:,1) = fxp
            fx_mat(:,1) = fx
            J = J + matmul((fxp_mat-fx_mat - matmul(J,s_mat)),transpose(s_mat))/dot_product(s,s)
            
            !Update fx
            fx = fxp

        end if

        if (printmod==0 .or. (printmod>0 .and. mod(iter,printmod)==0)) &
            write(*,'(A,i5,A,i5,A,f16.7)') 'BR: iter ',iter,' fevals ',fevals,' fnorm ',sqrt(sum(fx**2))
        if (sqrt(sum(s**2))<xtoler) exit

    end do

    write(*,'(A)') 'BR:________________________________________________________________'

end subroutine sub_broyden
! 
! subroutine sub_newton(f,x,res)
!     use mod_matlab
!     implicit none   
!     interface    
!         function f(x)
!             real(8), dimension(1:), intent(in) :: x
!             real(8), dimension(1:size(x)) :: f
!         end function f
!     end interface
!     real(8), dimension(1:), intent(inout) :: x
!     real(8), dimension(1:), intent(inout), optional :: res
!     !local
!     real(8), dimension(1:size(x)) :: e
!     real(8), dimension(1:size(x),1:size(x)) :: J
!     real(8), dimension(1:1+size(x),1:size(x)) :: fhist,xhist
!     real(8), dimension(1:size(x)) ::  s, fx, best_x, best_fx
!     real(8) :: ftol
!     integer :: n, maxiter, iter, flag, feval
!     
!     n = size(x)
!     maxiter = 200
!     ftol = 1d-6
! 
!     flag = 1
!     feval = 1
! 
!     best_x = x
!     best_fx = -huge(0.d0)
! 
!     write(*,*) 'sub_newton______________________________________________'
!     do iter = 1,maxiter
! 
!         !Compute fx at guess
!         fx = f(x)
! 
!         !Check for goodness of fit
!         if (sqrt(sum(fx**2))<ftol) then
!             flag = 0
!             exit
!         end if
! 
!         write(*,'(A,i4,A,i6,A,e12.5,A,e12.5)') ' iter ',iter,'  feval ',feval,'  max(res): ',maxval(fx), &
! '  min(res) ',minval(fx)
! 
!         !Get Jacobian
!         e = .01d0 !Step size            
!         call sub_jacobi(J,f,x,e,fx,fhist,xhist)
!         feval = feval + 1 + n
! 
!         !Solve for s in J*s = -f(x)
!         s = mldivide(J,-fx)
! 
!         !Set x' = x + s
!         x = x + s
! 
!     end do
! 
!     select case (flag)
!         case (0)
!             write(*,*) 'Converged successfully, SSRS: ', sqrt(sum(fx**2))
!         case default
!             write(*,*) 'Failed to converge, SSRS: ', sqrt(sum(fx**2))
!      end select
!     write(*,*) '________________________________________________________'
!     if (present(res)) res = fx
! 
! 
! end subroutine sub_newton
! 
! 
! ! Calculate the jacobian of a matrix using first differences with a step size (vector) e
! ! option
! subroutine sub_jacobi(J,f,x,e,fx,option) 
!     implicit none   
!     interface    
!         function f(x)
!             real(8), dimension(1:), intent(in) :: x
!             real(8), dimension(1:size(x)) :: f
!         end function f
!     end interface
!     real(8), dimension(1:), intent(in) :: x, e, fx
!     real(8), dimension(1:size(x),1:size(x)), intent(out) :: J
!     integer, intent(in), optional :: option
!     !local
!     real(8), dimension(1:size(x)) :: fvec
!     integer :: n, iflag, i, switch
! 
!     n = size(x)
! 
!     !Jacobian is 
!     ! [dy1/dx1 dy1/dx2 ...
!     !  dy2/dx1 dy2/dx2 ...
!     !  ...             
! 
!     do i = 1,n
! 
!         xhist(i,:) = x
!         xhist(i,i) = x(i) + e(i)
!         fhist(i,:) = f(xhist(i,:))        
! 
!         if (switch == 1 .AND. all(abs(fhist(i,:))< abs(fx))) then
!             J = 0.d0            
!             fhist = spread(fhist(i,:),1,n)
!             xhist = spread(xhist(i,:),1,n)  
!             return 
!         end if
! 
!         J(:,i)= (fhist(i,:) - fx)/e(i)
!     end do
! 
!     if (present(fhist_out)) fhist_out = fhist
!     if (present(xhist_out)) xhist_out = xhist
! 
! end subroutine sub_jacobi


end module mod_rootNdim

module mod_rootNdim_test
    use mod_rootNdim
    implicit none
contains

    ! The function is f_i = i - prod^i_{j=1}x_j
    ! has a unique recursive solution. x_1 = 1 and x_i = i/prod(x_j|j<i)
    function testFn1(x) result(y)
        implicit none
        real(8), intent(in) :: x(1:)
        real(8) :: y(1:size(x))
        ! local
        integer :: i
    
        if (any(x<0)) then
            y = 1d100
            return 
        end if

        do i = 1,size(x)
            y(i) = real(i,8) - product(x(1:i))
        end do

    end function testFn1

    ! Minimize f_j = sum_i (i*x_i)**8
    function testFn2(x) result(y)
        implicit none
        real(8), intent(in) :: x(1:)
        real(8) :: y(1:size(x))
        ! local
        integer :: i
    
        do i = 1,size(x)
            y(i) = sum(x**8) 
                             
        end do

    end function testFn2

    ! Analytic Jacobian
    function testJac2(x) result(J)
        implicit none
        real(8), intent(in) :: x(1:)
        real(8) :: J(size(x),size(x))
        ! local
        integer :: i
        
        do i = 1,size(x)
            J(i,:) = 8d0*x**7
        end do
            ! f_j(x) = \sum_i x_i^8 (all f_j are identical)
            ! df_j/dx_i = 8*x_i**7

    end function


!     subroutine test_smolyakroot
!         implicit none
!         real(8) :: x(3)
! 
!         x = 2d0
! 
!         call sub_smolyakroot(x,testFn1,printmod_in=100)
!         print*,'x is',x
!         print*,'fx is',testFn1(x)
!         call sub_smolyakroot(x,testFn2,printmod_in=100)
!         print*,'x is',x
!         print*,'fx is',testFn2(x)
! 
!     end subroutine
    subroutine test_powell
        implicit none
        real(8) :: x(3)


        x = 10d0
        call sub_powell(x,testFn1,printmod_in=100)
        print*,'x is',x
        print*,'fx is',testFn1(x)

        x = 10d0
        call sub_powell(x,testFn2,printmod_in=100)
        print*,'x is',x
        print*,'fx is',testFn2(x)

        x = 10d0
        call sub_powell(x,testFn2,printmod_in=100,fjac_in=testJac2)
        print*,'x is',x
        print*,'fx is',testFn2(x)

        x = 10d0
        call sub_powell(x,testFn2,printmod_in=1,derivstep_in=x*0d0 + 1d-2)
        print*,'x is',x
        print*,'fx is',testFn2(x)

    end subroutine
!     subroutine test_lsqnonlin
!         implicit none
!         real(8) :: x(3)
!         
!         x = 10d0
! 
!         call sub_lsqnonlin(x,testFn1,printmod_in=100,robust_in=.true.)
!         print*,'x is',x
!         print*,'fx is',testFn1(x)
!         call sub_lsqnonlin(x,testFn2,printmod_in=100)
!         print*,'x is',x
!         print*,'fx is',testFn2(x)
!     end subroutine
    subroutine test_broyden
        implicit none
        real(8) :: x(3)
        
        x = 10d0

        call sub_broyden(x,testFn1,printmod_in=100,relax_in=.9d0)
        print*,'x is',x
        print*,'fx is',testFn1(x)

        ! Singular jacobian case has Broyden fail
        ! print*,'NOTE: the jacobian here is singular (noninvertible)'
        ! call sub_broyden(x,testFn2,printmod_in=1,maxiter_in=10)
        ! print*,'x is',x
        ! print*,'fx is',testFn2(x)
    end subroutine

    subroutine test_fixpdf
        implicit none
        real(8) :: x(3)


        print*,'testing fixpdf'
        x = 10d0
        call sub_fixpdf(x,testFn1)
        print*,'x is',x
        print*,'fx is',testFn1(x)

        x = 10d0
        call sub_fixpdf(x,testFn2)
        print*,'x is',x
        print*,'fx is',testFn2(x)

        x = 10d0
        call sub_fixpdf(x,testFn2,testJac2)
        print*,'x is',x
        print*,'fx is',testFn2(x)

!         STOP 'checking test_fixpdf'


    end subroutine


end module
