module mod_sparse
    use mod_matlab, only: mtimes
    use mod_myprint
    implicit none

    ! CSR format (3-array formulation) (default)
    type sparse_mat
        integer :: nnz !number of nonzero elts
        integer :: nr !number of rows
        integer :: nc !number of columns
        real(8), dimension(:), allocatable :: val !nonzero values
        integer, dimension(:), allocatable :: j !same length as val
        integer, dimension(:), allocatable :: i !length nr+1
    end type sparse_mat

    ! Coordinate format
    type spmat_coo
        integer :: nnz !number of nonzero elts
        integer :: nr !number of rows
        integer :: nc !number of cols
        real(8), dimension(:), allocatable :: val !nonzero values
        integer, dimension(:), allocatable :: row,col !row and column locations for each value (so each has nnz length)
    end type spmat_coo

    interface spmtimes
        module procedure mtimes_spmat_dvec, mtimes_spmat_dmat !, mtimes_dvec_spmat
    end interface
    interface transpose
        module procedure sptranspose
    end interface transpose
    interface myprint
        module procedure myprint_spmat
    end interface myprint
    interface myread
        module procedure myread_spmat
    end interface myread

    interface operator (+)
        module procedure mplus_spmat_spmat
    end interface
    interface operator (*)
        module procedure mtimes_dscal_spmat, mtimes_spmat_dscal
    end interface

    interface operator (-)
        module procedure mminus_spmat_spmat
    end interface


contains

    ! Print the sparse double prec matrix A
    subroutine myprint_spmat(name,A)
        implicit none
        type(sparse_mat), intent(in) :: A
        character(len=*), intent(in) :: name
        !local
        type(spmat_coo) :: Acoo

        if (.not. allocated(A%val)) STOP 'myread_spmat: tried to read into a sparse matrix that is not allocated'

        ! Convert to coordinate
        Acoo = convert_csr_to_coo(A)

        ! Print the three components
        call myprint(name//'.sparseVAL',Acoo%val)
        call myprint(name//'.sparseROW',Acoo%row)
        call myprint(name//'.sparseCOL',Acoo%col)
        call myprint(name//'.sparseNNZ',Acoo%nnz)
        call myprint(name//'.sparseNNR',Acoo%nr)
        call myprint(name//'.sparseNNC',Acoo%nc)

    end subroutine myprint_spmat

    ! Read the sparse double prec matrix A
    subroutine myread_spmat(name,A)
        implicit none
        type(sparse_mat), intent(out) :: A
        character(len=*), intent(in) :: name
        !local
        type(spmat_coo) :: Acoo
        integer :: nnz, nr, nc

        ! Get dimensions
        call myread(name//'.sparseNNZ',nnz)
        call myread(name//'.sparseNNR',nr)
        call myread(name//'.sparseNNC',nc)

        ! Allocate sparse matrices
        call allocate_sparse_coo(Acoo,nnz,nr,nc)
        call allocate_sparse(A,nnz,nr,nc)

        ! Read the remaining components
        call myread(name//'.sparseVAL',Acoo%val)
        call myread(name//'.sparseROW',Acoo%row)
        call myread(name//'.sparseCOL',Acoo%col)

        ! Convert to CSR
        A = convert_coo_to_csr(Acoo)

    end subroutine myread_spmat

    subroutine allocate_sparse(spmat,nnz,nr,nc)
        implicit none
        type(sparse_mat), intent(out) :: spmat
        integer, intent(in) :: nnz, nr, nc
        
        spmat%nnz = nnz
        spmat%nr = nr
        spmat%nc = nc
        allocate(spmat%val(1:nnz),spmat%j(1:nnz),spmat%i(1:nr+1))
               
    end subroutine allocate_sparse
    subroutine deallocate_sparse(spmat)
        implicit none
        type(sparse_mat), intent(out) :: spmat

        deallocate(spmat%val,spmat%j,spmat%i)
        spmat%nnz = -1
        spmat%nr = -1
        spmat%nc = -1        

    end subroutine deallocate_sparse
    subroutine allocate_sparse_COO(spmat,nnz,nr,nc)
        implicit none
        type(spmat_coo), intent(out) :: spmat
        integer, intent(in) :: nnz, nr, nc
        
        spmat%nnz = nnz
        spmat%nr = nr
        spmat%nc = nc
        allocate(spmat%val(1:nnz),spmat%row(1:nnz),spmat%col(1:nnz))
               
    end subroutine allocate_sparse_COO


    function speye(n) result(A)
        implicit none
        integer, intent(in) :: n
        type(sparse_mat) :: A
        !local
        integer :: ind

        call allocate_sparse(A,n,n,n)
        A%val = 1.d0
        do ind = 1,n
            A%j(ind) = ind  !Column locations
        end do
        A%i(1:n) = A%j !Location in A%val of first non zero element in row i
        A%i(n+1) = A%nnz + 1

    end function 

    function sptranspose(A) result(B)
        implicit none
        type(sparse_mat), intent(in) :: A
        type(sparse_mat) :: B
        !local
        type(spmat_coo) :: Acoo, Bcoo
        
        Acoo = convert_csr_to_coo(A)

        call allocate_sparse_coo(Bcoo,Acoo%nnz,Acoo%nc,Acoo%nr)

        ! Perform the actual transpose
        Bcoo%val = Acoo%val
        Bcoo%row = Acoo%col
        Bcoo%col = Acoo%row

        B = convert_coo_to_csr(Bcoo)
        
    end function sptranspose

    ! Convert a full matrix to sparse representation
    function sparse(A) result(B)
        implicit none
        real(8), dimension(1:,1:), intent(in) :: A
        type(sparse_mat) :: B
        !local
        type(spmat_coo) :: Bcoo
        integer :: nnz, job(6), info, i, j, valind

!         nnz = count(abs(A)>1d-30)
        nnz = count(A/=0.d0)

!         call allocate_sparse_coo(Bcoo,nnz,size(A,1),size(A,2))
!         valind = 1
!         do j = 1,size(A,2)
!             do i = 1,size(A,1)
! !                 if (abs(A(i,j))>1d-30) then
!                 if (A(i,j)/=0.d0) then
!                     Bcoo%val(valind) = A(i,j)
!                     Bcoo%row(valind) = i
!                     Bcoo%col(valind) = j
!                     valind = valind+1
!                 end if
!             end do
!         end do
! 
!         B = convert_coo_to_csr(Bcoo)

        job(1) = 0 !Full to sparse
        job(2) = 1 !One based indexing (Fortran style)
        job(3) = 1 !One based indexing (Fortran style)
        job(4) = 2 !adns is a whole matrix
        job(5) = nnz !max number of nonzero elements in CSR format
        job(6) = 1 !generate all arrays (not just ia)

        call allocate_sparse(B,nnz,size(A,1),size(A,2))

        call mkl_ddnscsr(job, size(A,1), size(A,2), A, size(A,1), B%val, B%j, B%i, info)

        if (info/=0) STOP 'sparse: conversion failed'

    end function sparse

    function full(A) result(B)
        implicit none
        type(sparse_mat), intent(in) :: A
        real(8), dimension(A%nr,A%nc) :: B
        !local
        type(spmat_coo) :: Acoo
        integer :: ind
        write(*,*) 'Converting to full matrix will require ', (8*A%nr*A%nc)/(1028**2), ' MB'
!         allocate(B(A%nr,A%nc))

        Acoo = convert_csr_to_coo(A)

        B = 0
        do ind = 1,Acoo%nnz
            B(Acoo%row(ind),Acoo%col(ind)) = Acoo%val(ind)
        end do

    end function full
!     function full_sp(A) result(B)
!         implicit none
!         type(sparse_mat), intent(in) :: A
!         real(4), dimension(:,:), allocatable :: B
!         !local
!         integer :: ind
!         write(*,*) 'Converting to full matrix will require ', (4*A%nr*A%nc)/(1028**2), ' MB'
!         allocate(B(A%nr,A%nc))
!         B = 0
!         do ind = 1,A%nnz
!             B(A%row(ind),A%col(ind)) = A%val(ind)
!         end do
!     end function

    !Computes A*b (or b*A) for A sparse and b scalar
    function mtimes_spmat_dscal(A,b) result(C)
        implicit none
        type(sparse_mat), intent(in) :: A
        real(8), intent(in) :: b
        type(sparse_mat) :: C
        call allocate_sparse(C,A%nnz,A%nr,A%nc)
        C%val = b*A%val
        C%i = A%i
        C%j = A%j
    end function
    function mtimes_dscal_spmat(b,A) result(C)
        implicit none
        real(8), intent(in) :: b
        type(sparse_mat), intent(in) :: A
        type(sparse_mat) :: C
        C = mtimes_spmat_dscal(A,b)
    end function

    !Computes A*b for A sparse and b dense vector
    function mtimes_spmat_dvec(A,b) result(c)
        implicit none
        type(sparse_mat), intent(in) :: A
        real(8), dimension(1:), intent(in) :: b
        real(8), dimension(1:A%nr) :: c
        if (A%nc/=size(b)) then
            STOP 'mtimes: sparse matrix does not have same #rows as elts in dense vector.'
        end if
        call mkl_dcsrgemv('N', A%nr, A%val, A%i, A%j, b, c)        
    end function

    !Computes C = A*B for A sparse and B dense matrix
    function mtimes_spmat_dmat(A,B) result(C)
        implicit none
        type(sparse_mat), intent(in) :: A
        real(8), dimension(1:,1:), intent(in) :: B
        real(8), dimension(1:A%nr,1:size(B,2)) :: C
        !local
        character(len=6) :: matdescra

        if (A%nc/=size(B,1)) then
            STOP 'mtimes: sparse matrix does not have same #columns as rows in dense matrix'
        end if

        matdescra = 'G00F00' ! General matrix, Fortran (ones-based) indexing

        call mkl_dcsrmm('N', A%nr, A%nc, size(B,2), 1.d0, matdescra, A%val, A%j, A%i(1:A%nr), A%i(2:A%nr+1), B, size(B,1), 1.d0, C, A%nr)

    end function


!     !Computes b*A for A sparse and b dense vector
!     function mtimes_dvec_spmat(b,A) result(c)
!         implicit none
!         type(sparse_mat), intent(in) :: A
!         real(8), dimension(1:), intent(in) :: b
!         real(8), dimension(1:A%nc) :: c
!         if (A%nr/=size(b)) then
!             STOP 'mtimes: sparse matrix does not have same #rows as elts in dense vector.'
!         end if
! 
!         ! Can calculate b*A as transpose(A)*transpose(b)
!         call mkl_dcsrgemv('T', A%nr, A%val, A%i, A%j, b, c)
! 
!     end function


    function mminus_spmat_spmat(A,B) result(C)
        implicit none
        type(sparse_mat), intent(in) :: A,B
        type(sparse_mat) :: C
        !local
        type(sparse_mat) :: negB
        call allocate_sparse(negB,B%nnz,B%nr,B%nc)
        negB%val = -B%val
        negB%i = B%i
        negB%j = B%j
        C = mplus_spmat_spmat(A,negB)
    end function

    !Adds two sparse matrices in CSR format 
    function mplus_spmat_spmat(A,B) result(C)
        implicit none
        type(sparse_mat), intent(in) :: A,B
        type(sparse_mat) :: C
        !local
!         type(sparse_mat) :: A,B,C
        integer :: request,m,n,sort,nzmax,info
        
        !Check matrices agree

        !Convert matrices to CSR format
!         A = convert_coo_to_csr(Ain)
!         B = convert_coo_to_csr(Bin)

        m = A%nr
        n = A%nc
        sort = 3 !Sort, this will be (significantly) inefficient if A,B are already sorted
        nzmax = 0 !Used only if request == 0, which it isn't

        request = 1 !Write only ic which has known length

        allocate(C%i(m+1))
        call mkl_dcsradd('N', request, sort, m, n, A%val, A%j, A%i, 1.d0, B%val, B%j, B%i, C%val, C%j, C%i, nzmax, info)
        if (info/=0) then
            write(*,*) 'info: ',info
            write(*,*) C%i
            STOP 'mplus: error in step 1 of addition procedure '
        end if

        C%nnz = C%i(m+1)-1
        C%nr = A%nr
        C%nc = A%nc
        allocate(C%j(C%nnz),C%val(C%nnz))
        

        request = 2 !With space allocated, now fill in values

        call mkl_dcsradd('N', request, sort, m, n, A%val, A%j, A%i, 1.d0, B%val, B%j, B%i, C%val, C%j, C%i, nzmax, info)

        if (info/=0) then
            STOP 'mplus: error in step 2 of addition procedure '
        end if

        !Convert back
!         Cout = convert_csr_to_coo(C)
    end function

    !Convert a sparse matrix in coordinate format (default) to csr format
    function convert_coo_to_csr(coo) result(csr)
        implicit none
        type(spmat_coo), intent(in) :: coo
        type(sparse_mat) :: csr
        !local
        integer :: job(8), info

        job(1) = 1 !Convert to CSR
        job(2) = 1 !One-based indexing
        job(3) = 1 !One-based indexing
        job(5) = coo%nnz !number of nonzero elts of the matrix A
        job(6) = 0 !fill all arrays

        call allocate_sparse(csr,coo%nnz,coo%nr,coo%nc)

        call mkl_dcsrcoo(job, coo%nr, csr%val, csr%j, csr%i, csr%nnz, coo%val, coo%row, coo%col, info)

        if (info/=0) then
            STOP 'Converting to sparse matrix in CSR format failed'
        end if

    end function

    !Convert a sparse matrix in coordinate format (default) to csr format
    function convert_csr_to_coo(csr) result(coo)
        implicit none
        type(sparse_mat), intent(in) :: csr
        type(spmat_coo) :: coo
        !local
        integer :: job(8), info

        job(1) = 0 !Convert from CSR
        job(2) = 1 !One-based indexing
        job(3) = 1 !One-based indexing
        job(5) = csr%nnz !number of nonzero elts of the matrix A
        job(6) = 3 !fill all arrays
                
        call allocate_sparse_coo(coo,csr%nnz,csr%nr,csr%nc)

        call mkl_dcsrcoo(job, coo%nr, csr%val, csr%j, csr%i, csr%nnz, coo%val, coo%row, coo%col, info)
        if (info/=0) then
            STOP 'Converting to sparse matrix in COO format failed'
        end if

    end function

    subroutine disp_sparse(A)
        implicit none
        type(sparse_mat) :: A
        !local
        type(spmat_coo) :: Acoo
        integer :: i

        Acoo = convert_csr_to_coo(A)

        write(*,*) '  Value                           i           j'
        do i = 1,Acoo%nnz
            write(*,*) Acoo%val(i),Acoo%row(i),Acoo%col(i)
        end do
    end subroutine disp_sparse

!     !Finds invariant distribution of a sparse transition matrix P
!     !ie x in P'x=x (which is given by (I-P')x=0 but can't be solved by that method w/o svd)
!     subroutine spinvar(xinout,Ain,xtol,maxiter)
!         type(spmat_coo) :: Ain
!         real(8), dimension(Ain%nc), intent(inout) :: xinout !Guess on dsn and output
!         real(8), intent(in), optional :: xtol
!         integer, intent(in), optional :: maxiter
!         !local
!         type(spmat_coo) :: At
!         real(8), dimension(Ain%nc) :: xt,xtp1
!         integer :: tind
!         real(8) :: xtoleff
!         integer :: maxitereff
! 
!         xtoleff = 1d-10
!         maxitereff = 200
! 
!         if (present(xtol)) xtoleff = xtol
!         if (present(xtol)) maxitereff = maxiter
! 
!         !Transpose A
!         At = sptranspose(Ain)
! 
!         !Update
!         xt = xinout
!         do tind = 1,maxitereff
!             xtp1 = mtimes_spmat_dvec(At,xt)
!             if (maxval(abs(xtp1-xt))<xtoleff) then
!                 exit
!             end if
!             xt = xtp1
!         end do
!         xinout = xt
!         
!     end subroutine

! mkl_dcsradd !computes sum of two sparse matrices
! mkl_dcsrmultcsr !product of two sparse matrices
! mkl_dcsrmultd !prod two sparse and stored in dense
! 
! mkl_dcsrcoo !Convert a sparse too and from CSR and coordinate format

    subroutine test_sparse
        implicit none
        type(sparse_mat) :: A,D,E
        real(8) :: bdense(4), cdense(4)
        real(4), dimension(:,:), allocatable :: B
        real(8) :: Dfull(4,4)
        write(*,*) 'Beginning test'

        call allocate_sparse(A,4,4,4)
        call allocate_sparse(D,8,4,4)
        write(*,*) 'Allocated sparse'

        write(*,*) 'Set value of A to identity matrix'
        A = speye(4)

        Dfull = 0.d0
        Dfull(2,3) = 3.d0
        D = sparse(Dfull)

        print*,'A'
        call disp_sparse(A)
        print*,'D'
        call disp_sparse(D)

        E = A+D
        print*,'E'
        call disp_sparse(E)

        E = sptranspose(E)
        print*,'Etranspose'
        call disp_sparse(E)

        Dfull = full(E)
        print*,'Dfull'
        print*,Dfull

        bdense = (/1,2,3,4/)
        write(*,*) 'Compute matrix vector product'
        cdense = spmtimes(A,bdense)
        write(*,*) 'cdense: should be (/1,2,3,4/)'
        write(*,*) cdense

        print*,'E*Dfull'
        Dfull = mtimes_spmat_dmat(E,Dfull)
        print*,Dfull

        call myprint('Esparsetest',E)

!         Bcoo = convert_coo_to_csr(A)
!         write(*,*) 'Converted to csr'
! 
!         write(*,*) 'Bcoo%val', Bcoo%val
!         write(*,*) 'Bcoo%j', Bcoo%j
!         write(*,*) 'Bcoo%i', Bcoo%i
! 
!         A%val = 0
!         write(*,*) 'Set value of A to 0'
! 
!         A = convert_csr_to_coo(Bcoo)
!         write(*,*) 'Converted to coo'
! 
!         write(*,*) 'Should produce all ones'
!         write(*,*) A%val
!         
!         write(*,*) 'Should be 1,2,3,4'
!         write(*,*) A%row
!         write(*,*) 'Should be 1,2,3,4'
!         write(*,*) A%col
!         
!         bdense = (/1,2,3,4,5,6,7,8,9,10/)
! 
!         write(*,*) 'Compute matrix vector product'
!         cdense = mtimes_spmat_dvec(A,bdense)
!         write(*,*) 'cdense: should be (/1,2,3,4/) and then 16 zeros'
!         write(*,*) cdense
! 
!         write(*,*) 'Adding A and D together'
!         E = A+D
!         write(*,*) 'Should have 1 on first four diag, and 2 on first 8 subdiag'
!         call disp_sparse(E)
!         
!         write(*,*) 'Subtracting D from A+D'
!         E = E-D
!         write(*,*) 'Should have 1 on first four diag'
!         call disp_sparse(E)
! 
!         write(*,*) 'Adding D back and multiplying by 2'
!         E = 2.d0*(E + D)
!         write(*,*) 'Should have 2 on first four diag, and 4 on first 8 subdiag'
!         call disp_sparse(E)
! 
!         write(*,*) 'Constructing a 2x2 Identity matrix'
!         E = speye(2)
!         call disp_sparse(E)
! 
!         write(*,*) 'Forming a 3x3 transition matrix'
!         call allocate_sparse(E,7,3,3)
!         E%col = (/1,1,2,2,2,3,3/)
!         E%row = (/1,2,1,2,3,2,3/)        
!         E%val = (/.9d0,.05d0,.1d0,.9d0,.1d0,.05d0,.9d0/)
!         call disp_sparse(E)
!                 
!         write(*,*) 'Finding null space of above'
!         invar = 1.d0/3.d0
!         call spinvar(invar,E)
!         write(*,*) 'invar', invar
! 
!         write(*,*) 'Making a large sparse matrix'
!         A = speye(100*100)
! 
!         write(*,*) 'Converting to a large full matrix'
!         B = full_sp(A)

    end subroutine test_sparse

end module mod_sparse

!         !paradiso data
!         integer :: pt(64), maxfct, mnum, mtype, phase, neqns, nrhs, perm(Ain%nr), iparm(64), msglvl, error
! solution of (eye-trans')*pi=0
!         !call pardiso (pt, maxfct, mnum, type, phase, n, a, ia, ja, perm, nrhs, iparm, msglvl, b, x, error)
!         xt = xinout
! 
!         pt = 0
!         maxfct = 1
!         mnum = 1
!         mtype = 11 !real and unsymmetric matrix
!         neqns = A%nr !number of equations 
!         nrhs = 1
!         msglvl = 1
!         do ind = 1,A%nr
!             perm(ind) = ind
!         end do
! 
!         !If want defaults, just set iparm(1) = 0
!         iparm(1) = 1
!         iparm(2) = 2 !2 is default value, 3 is parallel
!         iparm(3) = 0 !not used
!         iparm(4) = 0
!         iparm(5) = 0
!         iparm(6) = 0 
!         iparm(7) = 0
!         iparm(8) = 0
!         iparm(9) = 0
!         iparm(10) = 8 !default is 8
!         iparm(11) = 1 !default is 1
!         iparm(12) = 0
!         iparm(13) = 1 !default is 1
!         iparm(14:17) = 0
!         iparm(18) = -1 !<0 reports nnz
!         iparm(19) = 0 
!         iparm(20) = 0
!         iparm(21) = 1
!         iparm(22:23) = 0
!         iparm(24) = 0
!         iparm(25:26) = 0
!         iparm(27) = 1 !Checks integer arrays
!         iparm(28) = 0 
!         iparm(29:30) = 0
!         iparm(31) = 0
!         iparm(32:34) = 0
!         iparm(35) = 0 !Fortran style indexing
!         iparm(36:59) = 0
!         iparm(60) = 0 !default
!         iparm(61:64) = 0
! 
!         write(*,*) 'Calling paradiso'
!         phase = 11
!         call pardiso (pt, maxfct, mnum, mtype, phase, neqns, A%val, A%i, A%j, &
!                        perm, nrhs, iparm, msglvl, zeros, x, error)  
! 
!         !Because Ax=0 has a trivial solution, I instead solve 
! 
!         phase = 33
!         x = 1.d0
! 
!         write(*,*) 'A'
!         write(*,*) A%val
!         write(*,*) A%i
!         write(*,*) A%j
! 
!         call pardiso (pt, maxfct, mnum, mtype, phase, neqns, A%val, A%i, A%j, &
!                       perm, nrhs, iparm, msglvl, zeros, x, error)  
! 
!         write(*,*) x
!         write(*,*) zeros
!         write(*,*) 'Finished with paradiso'
