module mod_main
    use mod_law
    use mod_param
    use mod_myprint
    use mod_SS, only: sub_SS
    use mod_matlab
    use mod_minNdim
    use mod_rootNdim
    use mod_shocks
    use mod_stochsimul
    use mod_nonstochsimul
    use mod_price_fcns
    use mod_value_fcns
    use mod_matlab
    use mod_data
    use mod_stat, only: R2
    implicit none
contains

! solveSSandBC 
! if doSS, compute SS. o/w, load data.
! if doBC, compute BC. o/w, do nothing.
subroutine solveSSandBC(doSS,doBC)
    implicit none
    real(8) :: K_SS,N_SS,w_SS
    logical, intent(in) :: doSS,doBC

    if (doSS) then

        ! Compute the median efficiency in case doing BAPCPA
        medianBAPCPA = fn_medianE_SS()

        ! If partial equilibrium set qbarPE to the steady state qbar value in the PEdir
        if (isPE) then 
            call usedir(trim(PEdir)) ! Choose from where to load the SS interest rate and law of motion
            call myread('qbar_SS',qbarPE)
            call revertdir() ! Revert to the dir
        end if

        ! Calculate the SS.
        write(*,*) '*Calling Steady State Program*' 
        call sub_SS(.true.) ! true means write results

    else
        
        ! load data
        call load_data_SS()

    end if

    ! Simulate panel
    call myread('w_SS',w_SS)
    call sub_panel(w_SS=w_SS)

    ! If only calculating the SS, return
    if (.not. doBC) return

    ! Setup grids on moments.  Do this by approximating around the steady state
    call myread('K_SS',K_SS)
    call myread('N_SS',N_SS)

    ! Setup eshift to exactly match volatility in labor supply
    call setup_N(eshift)

    ! GE or PE?
    if (.not. isPE) then

        ! If general equilibrium, solve the model
        grid_KN = K_SS/N_SS*linspace(.85d0,1.15d0,len_KN) ! K/N grid
        grid_WW = linspace(.50d0,.52d0,len_WW) ! equity premium grid

    else

       ! If doing partial equilibrium, set the prices
        call usedir(trim(PEdir)) ! Choose from where to load the SS interest rate and law of motion
        call init_law(1.d0,1.d0) !Read in the law from the directory
        call myread('grid_KN',grid_KN)  ! Also load the grids so there is no discrepancy in the weighting
        call myread('grid_WW',grid_WW)

        call revertdir() ! Revert to the dir
        call save_law() ! Save the law to your directory as an ``initial guess''

    end if

    ! Save the grids
    call myprint('grid_KN',grid_KN)
    call myprint('grid_WW',grid_WW)
    call myprint('eshift',eshift)

    !Initialize moment structure
    call init_m(grid_z,grid_KN,grid_WW) !sets up grid_m

    !Initialize law
    write(*,*) 'Initializing law'
    call init_law(K_SS,N_SS)

    ! Print parameter configuration to file
    call printConfig() 

    ! Solve the business cycle model
    write(*,*) '*Calling KS Program*'
    call tic()
    call sub_BC()
    call toc()

    ! Law is now saved inside sub_BC
!     !Save law
!     call save_law()   

end subroutine solveSSandBC

function fn_SSRS(param) result(SSRs)
    implicit none
    real(8), intent(in) :: param(1:)
    real(8) :: SSRs

    SSRS = sum(fn_resvec_core(param)**2)

end function 


function fn_resvec(param,nf) result(resvec)
    implicit none
    integer, intent(in) :: nf
    real(8), intent(in) :: param(:)
    real(8) :: resvec(nf)
    ! local
    real(8) :: tmp(100)

    if (size(param)/=7) STOP 'ERROR: cannot use fn_resvec'

    tmp = fn_resvec_core(param)
    resvec = tmp([1,2,3,4,5,6,10])

end function

function fn_resvec_core(param) result(SSRvec)
    use mod_stats
    implicit none
    real(8), intent(in) :: param(1:)
    real(8) :: SSRvec(100)
    !local
!     real(8) :: statlist(100)
    type(stat_type) :: s

    !!!!!!!!!!!!!!!!!!!!!!!!
    ! Targets
    ! 
    ! include 'calibtargets_lowdebt.inc'
    include 'calibtargets_hidebt.inc'
    !!! Targets not related to debt
    real(8), parameter :: t_KY = 3.08d0 ! K/Y
    real(8), parameter :: t_earnTop5 = .301d0, t_earnTop20 = .602d0, t_earn4020 = .229d0, & 
                          t_earn6040 = .130d0, t_earn8060 = .040d0, t_earnGini = .61d0
    real(8), parameter :: t_wealthTop5 = .578d0, t_wealthTop20 = .817d0, t_wealth4020 = .122d0, & 
                          t_wealth6040 = .050d0, t_wealth8060 = .013d0, t_wealthGini = .80d0
    real(8), parameter :: t_wealth_mean_to_median = 4.03d0
    real(8), parameter :: t_earn_mean_to_median = 1.57d0
                          
    !!!!!!!!!!!!!!!!!!!!!!!! 


    ! Set the parameter values
    call set_param(param)

    !Check for bad parameter values
    if (checkParam()/=0) then
        print*,'SSRS bad param'
        SSRvec = 1d10
        return
    end if

    ! Set shocks
    call sub_setshocks()
    
    ! Solve the model
    call sub_SS(.false.,s_out=s) !false means don't write results to file (speeds up computation)

    ! For calibration, output a residual vector
    SSRvec = 0d0
    SSRvec(1) = 1d0 - s%pop_in_debt/t_pop_in_debt ! pop in debt, -a+x>0
    SSRvec(2) = 1d0 - s%h0_pop_def/t_pct_hhs_filing ! pop with h=0,d=1
    SSRvec(3) = 1d0 - s%total_debt/s%Y/t_debt_to_output ! total debt (-a+x>0 any h)/Y ratio
    SSRvec(4) = 10d0*(1d0 - s%K/s%Y/t_KY) !K/Y
    SSRvec(5) = 1d0 - (s%meanWealth/s%medianWealth)/t_wealth_mean_to_median ! wealth mean to median
    SSRvec(6) = 1d0 - (s%meanEarn/s%medianEarn)/t_earn_mean_to_median ! earnings mean to median
!     SSRvec(7) = 1d0 - s%wealthTop5/t_wealthTop5
!     SSRvec(8) = 1d0 - s%earnTop5/t_earnTop5
!     SSRvec(9) = 1d0 - s%wealthTop20/t_wealthTop20
    SSRvec(10) = 1d0 - s%earnTop20/t_earnTop20 ! earnings of top 20
!     SSRvec(11) = 1d0 - s%wealth4020/t_wealth4020
!     SSRvec(12) = 1d0 - s%earn4020/t_earn4020
!     SSRvec(13) = 1d0 - s%wealth6040/t_wealth6040
!     SSRvec(14) = 1d0 - s%earn6040/t_earn6040
!     SSRvec(15) = 1d0 - s%wealth8060/t_wealth8060
!     SSRvec(16) = 1d0 - s%earn8060/t_earn8060
!     SSRvec(17) = 1d0 - s%frac_def_below_median/t_frac_def_below_median
!     SSRvec(18) = 1d0 - s%debt_discharge/s%Y/t_debt_discharge_to_output
!     SSRvec(19) = 1d0 - s%debt_to_earn_def/t_debt_to_earn_def
!     SSRvec(20) = 1d0 - (s%total_debt/s%Y/s%pop_in_debt)/(t_debt_to_output/t_pop_in_debt) ! debt/y / pop in debt
!     SSRvec(21) = 1d0 - s%earnGini/t_earnGini
!     SSRvec(22) = 1d0 - s%wealthGini/t_wealthGini

    if (any(isNaN(SSRvec))) then
        SSRvec = 1d10
    end if

!21 is earnings gini
!22 is wealth gini
! 7 is wealth of top 5
! 8 is earn of top 5
! 4,5,6,9,10

!     SSRvec = 0d0
!     SSRvec([1,2,3,4,5,6,9,10]) = resvec([1,2,3,4,5,6,9,10])


!     print*, 'Contribution to total sum of squared residuals'
!     call disp(pack(SSRvec**2/sum(SSRvec**2),SSRvec**2>0d0))
! 
    write(*,'(A,f30.12,A,7(f17.10,A))') 'SSRS ',sum(SSRvec**2),' param = [',param(1),'d0,',param(2),'d0,',param(3),'d0,',param(4),'d0,',param(5),'d0,',param(6),'d0,',param(7),'d0]'

end function 

subroutine sub_BC()
    use mod_mympi
    implicit none
    integer, dimension(nT) :: zseq_ind
    type(mom), dimension(nT) :: mseq
    real(8), dimension(nT) :: Kseq, Nseq !, Aseq
    type(mom), dimension(nT-(len_t0+1)+1-1) :: m_fcast1,m_real1
    type(mom), dimension(nT-(len_t0+1)+1-50) :: m_fcast50,m_real50
    real(8) :: R2_KN,R2_WW
    integer :: iter
    integer, parameter :: maxiter = 100
    real(8) :: R2History(maxiter,2)
    real(8) :: invar_dsn_m(nM), invar_dsn_m_cdnal_g(nM), invar_dsn_m_cdnal_b(nM)
    real(8) :: diffLaw(8)
    real(8) :: K_SS

    call myread('K_SS',K_SS)

    !Initialize {z}
    ! call init_zseq(zseq_ind,trans_z)

    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    !!! WRITE THE Z INDS ONLY ONCE FOR REPRODUCIBLE RESULTS
    !!! ACROSS DIFFERENT COMPILERS
    ! if (mympi_iproc()==0) then
    !   call init_zseq(zseq_ind,trans_z)
    !   open(12322,file=trim(rootdatadir)//'iZs.dat',status='replace')
    !   write(12322,'(I3)') zseq_ind
    !   close(12322)
    ! end if
    ! call mympi_barrier()
    ! STOP 'just generating iZs.dat'
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    
    open(12322+mympi_iproc(),file=trim(rootdatadir)//'iZs.dat',status='old',action='read')
    read(12322+mympi_iproc(),'(I3)') zseq_ind
    close(12322+mympi_iproc())

    !Calculate the exogenous sequence of labor supply
    call sub_nonstoch_simul_eonly(Nseq,zseq_ind,gE,trans_e,invar_dsn_e_SS,init_dsn_e)
    ! write(*,*) 'std(hp(log(N)))',std(hpfilter(log(Nseq(len_t0+5:nT-5)),100.d0)),'eshift',eshift

    !Get wage implied by aggregate moments
    w = grid_m%z*(1d0-alpha_K)*grid_m%KN**alpha_K

    R2_KN = -1d0
    ! Iterate to compute GE unless is PE
    do iter = 1,merge(maxiter,1,.not.isPE)

        !Get the transition associated with moments and z shock (see comments of m_law)
        call m_law(trans_m,trans_m_cdnal_zp,trans_z)

        !Get prices implied by the law
        call qbargqbarb_law(qbar)

        !Solve for the partial equilibrium price schedule and value functions 
        call subVFI(v,c,apg,apb,d,q,qbar,w,gA,gE,trans_e,trans_ez,trans_m_cdnal_zp,grid_m%z_ind)
        
        !Simulate the economy using policies and z sequence
        write(*,*) 'Beginning Simulation of the Economy'
        call sub_nonstoch_simul(mseq,Kseq,Nseq,zseq_ind,K_SS,doComputeStats=isPE)

        !Check forecast errors 
        call sub_fcastJ(m_fcast1,m_real1,1,mseq)
        call sub_fcastJ(m_fcast50,m_real50,50,mseq)
        R2_KN = R2(m_real1%KN,m_fcast1%KN)
        R2_WW = R2(m_real1%WW,m_fcast1%WW)
        write(*,*)           '*Forecast Errors (1-step ahead)-------------------------*'
        write(*,'(A,i4)')    'iter:',iter
        write(*,*)           'KN forecast errors'
        write(*,'(A,f11.4)') 'R2: KN', R2_KN
        write(*,'(A,f11.4)') 'Max % Error ', 100.d0*maxval(abs((m_fcast1%KN-m_real1%KN)/m_real1%KN))
        write(*,'(A,f11.4)') 'Mean % Error', 100.d0*mean(abs((m_fcast1%KN-m_real1%KN)/m_real1%KN))    
        write(*,*)           'WW forecast errors'
        write(*,'(A,f11.4)') 'R2: WW', R2_WW
        write(*,'(A,f11.4)') 'Max % Error ', 100.d0*maxval(abs((m_fcast1%WW-m_real1%WW)/m_real1%WW))
        write(*,'(A,f11.4)') 'Mean % Error', 100.d0*mean(abs((m_fcast1%WW-m_real1%WW)/m_real1%WW))    
        write(*,*)           '*-------------------------------------------------------*'
        write(*,*)
        write(*,*)           '*Forecast Errors (50-steps ahead)-------------------------*'
        write(*,*)           'AA forecast errors'
        write(*,'(A,f11.4)') 'R2 50: KN', R2(m_real50%KN,m_fcast50%KN)
        write(*,'(A,f11.4)') 'R2 50: WW', R2(m_real50%WW,m_fcast50%WW)
        write(*,*)           '*-------------------------------------------------------*'
        R2History(iter,1) = R2_KN
        R2History(iter,2) = R2_WW
        call myprint('R2History',R2History)

        ! Compute what the updated law would be (but don't update!) 
        call sub_compute_law_update(mseq,diffLaw)
        print*, 'Change in law is ', maxval(diffLaw)

        ! Update the law, check for convergence and exit if there is convergence
        if (.not. isPE) then
            if (maxval(diffLaw)>5d-5) then 

                ! Update the law: 0 means don't relax, 1 means relax completely (don't update)
                ! call sub_update_law(.75d0,.75d0) 
                if (iter>20) then
                    lawUpdateRelax = max(lawUpdateRelax,.95d0) ! if taking a long time to converge, increase relaxation
                end if
                call sub_update_law(lawUpdateRelax,lawUpdateRelax) 

                ! Writes the law to file (only if iter>1 so that can solve the model with good initial 
                ! guesses without changing them (because they converge in one iteration).
                if (iter>1) call save_law()

            else
                ! Don't update the law and exit
                

                ! Note: if this happens on the first try (b/c of good initial guesses), 
                ! then "sub_update_law" is never called and so the beta values are never changed
                ! What will happen is that the old law will be printed out along with the value
                ! functions and everything else associated with it.
                print*, 'Law converged successfully', maxval(diffLaw)

                ! Simulate a final time, but collect additional statistics
                call sub_nonstoch_simul(mseq,Kseq,Nseq,zseq_ind,K_SS,doComputeStats=.true.)
                exit

            end if
        else
            exit
        end if

    end do

    !Simulate a panel for some extra statistics using the mseq found nonstochastically
    print*,'Simulating a panel of households'
    call sub_panel(mseq)

    print*,'Printing results'        
    if (.not. onlySaveLaw) call print_all ! to save memory on the cluster, this need not be saved
    call sub_print_law() ! prints variables in mod_law, different from save_law

contains

    subroutine print_all
        use mod_myprint   
        implicit none

        !lengths
        call myprint('nA',nA)
        call myprint('nAmm',nAmm)
        call myprint('nE',nE)

        !grids
        call myprint('gA',gA)
        call myprint('gE',gE)
        call myprint('init_dsn_e',init_dsn_e)
        call myprint('grid_KN',grid_KN)

        !law of motion
        call myprint('trans_m',trans_m)
        call myprint('trans_m_cdnal_zp',trans_m_cdnal_zp)

        invar_dsn_m = nullspace(eye(nM)-transpose(trans_m))
        invar_dsn_m = invar_dsn_m/sum(invar_dsn_m)
        call myprint('invar_dsn_m',invar_dsn_m)

        invar_dsn_m_cdnal_b = merge(invar_dsn_m,0.d0,grid_m%z_ind==b_ind)/sum(invar_dsn_m,mask=grid_m%z_ind==b_ind)
        invar_dsn_m_cdnal_g = merge(invar_dsn_m,0.d0,grid_m%z_ind==g_ind)/sum(invar_dsn_m,mask=grid_m%z_ind==g_ind)
        call myprint('invar_dsn_m_cdnal_b',invar_dsn_m_cdnal_b)
        call myprint('invar_dsn_m_cdnal_g',invar_dsn_m_cdnal_g)
        
        !transition matrices
        call myprint('trans_z',trans_z)

        !prices
        call myprint('w',w)
        call myprint('qbar', qbar)

        !value and policy functions 
        call myprint('Ed',E_over_m_5d_logical(d,invar_dsn_m))
        call myprint('Ev',E_over_m_5d(v,invar_dsn_m))
        call myprint('Ec',E_over_m_5d(c,invar_dsn_m))
        call myprint('Eapg',E_over_m_5d(apg,invar_dsn_m))
        call myprint('Eapb',E_over_m_5d(apb,invar_dsn_m))
        call myprint('Ev_cdnal_g',E_over_m_5d(v,invar_dsn_m_cdnal_g)) ! Note: m dsn is different
        call myprint('Ev_cdnal_b',E_over_m_5d(v,invar_dsn_m_cdnal_b)) ! Note: m dsn is different
        call myprint('vNewborn',v(nAmm+1,:,1,1,:))

        !Expected prices
        call myprint('Ew',E_over_m_1d(w,invar_dsn_m))
        call myprint('Eqg',E_over_m_5d(q(:,:,:,:,:,g_ind),invar_dsn_m))
        call myprint('Eqb',E_over_m_5d(q(:,:,:,:,:,b_ind),invar_dsn_m))
        call myprint('Eqbarg',E_over_m_1d(qbar(:,g_ind),invar_dsn_m))
        call myprint('Eqbarb',E_over_m_1d(qbar(:,b_ind),invar_dsn_m))
        call myprint('EKN',E_over_m_1d(grid_m%KN,invar_dsn_m))
        call myprint('EWW',E_over_m_1d(grid_m%WW,invar_dsn_m))
        call myprint('Ez',E_over_m_1d(grid_m%z,invar_dsn_m))

        !sequences
        call myprint('zseq_ind',zseq_ind)
        call myprint('WWseq',mseq%WW)
        call myprint('Nseq',Nseq)
        call myprint('Kseq',Kseq)
        call myprint('KNseq',mseq%KN)
        call myprint('KN_fcast1',m_fcast1%KN)
        call myprint('WW_fcast1',m_fcast1%WW)
        call myprint('KN_real1',m_real1%KN)
        call myprint('WW_real1',m_real1%WW)
        call myprint('KN_fcast50',m_fcast50%KN)
        call myprint('WW_fcast50',m_fcast50%WW)
        call myprint('KN_real50',m_real50%KN)
        call myprint('WW_real50',m_real50%WW)

        call myprint('R2_KN',R2_KN)

        !variables in mod_param
        call print_common_var()
        
        print*,'done printing'

    end subroutine print_all

! This computes the expectation of V(:,m) using the probability dsn dsn_m.
function E_over_m_5d_logical(d,dsn_m) result(Ed)
    use mod_param, only: nM
    implicit none
    logical, intent(in) :: d(:,:,:,:,:)
    real(8) :: dsn_m(1:) 
    real(8) :: Ed(size(d,1),size(d,2),size(d,3),size(d,4))
    integer :: dims(size(shape(d)))
    ! local
    real(8), allocatable :: V(:,:,:,:,:)

    dims = shape(d)

    allocate(V(dims(1),dims(2),dims(3),dims(4),dims(5)))

    where (d)
        V = 1
    elsewhere 
        V = 0
    endwhere
    
    Ed = reshape(E_over_m_core(V,size(V)/nM,dsn_m),dims(1:4))

end function 
function E_over_m_5d(V,dsn_m) result(EV)
    use mod_param, only: nM
    implicit none
    real(8) :: dsn_m(1:), V(1:,1:,1:,1:,1:), EV(size(V,1),size(V,2),size(V,3),size(V,4))
    integer :: dims(size(shape(V)))
    dims = shape(V)
    EV = reshape(E_over_m_core(V,size(V)/nM,dsn_m),dims(1:4))
end function 
function E_over_m_4d(V,dsn_m) result(EV)
    use mod_param, only: nM
    implicit none
    real(8) :: dsn_m(1:), V(1:,1:,1:,1:), EV(size(V,1),size(V,2),size(V,3))
    integer :: dims(size(shape(V)))
    dims = shape(V)
    EV = reshape(E_over_m_core(V,size(V)/nM,dsn_m),dims(1:3))
end function 
function E_over_m_3d(V,dsn_m) result(EV)
    use mod_param, only: nM
    implicit none
    real(8) :: dsn_m(1:), V(1:,1:,1:), EV(size(V,1),size(V,2))
    integer :: dims(size(shape(V)))
    dims = shape(V)
    EV = reshape(E_over_m_core(V,size(V)/nM,dsn_m),dims(1:2))
end function 
function E_over_m_2d(V,dsn_m) result(EV)
    use mod_param, only: nM
    implicit none
    real(8) :: dsn_m(1:), V(1:,1:), EV(size(V,1))
    integer :: dims(size(shape(V)))
    dims = shape(V)
    EV = reshape(E_over_m_core(V,size(V)/nM,dsn_m),dims(1:1))
end function 
function E_over_m_core(V,nV1,dsn_m) result(EV)
    use mod_param, only: nM
    implicit none
    integer, intent(in) :: nV1 
    real(8), intent(in) :: V(nV1,*)
    real(8), intent(in) :: dsn_m(nM)
    real(8) :: EV(nV1)
    EV = matmul(V(1:nV1,1:nM),dsn_m)
end function
! This is slightly different
function E_over_m_1d(V,dsn_m) result(EV)
    implicit none
    real(8) :: dsn_m(1:), V(1:), EV
    EV = dot_product(V,dsn_m)
end function 

end subroutine sub_BC

end module mod_main

