module mod_nonstochsimul
    use mod_param, only: nT,nA,nAmm,nE,Tend,nH
    use mod_stats, only: stat_type
    implicit none


contains

!Generate sequence of aggregate shocks
subroutine init_zseq(iZs,trans_z)
    use mod_matlab
    use mod_discreteAR1
    implicit none
    integer, dimension(:), intent(out) :: iZs
    real(8), dimension(:,:), intent(in) :: trans_z
    !local
    integer :: i

    call reset(getDefaultStream())

    iZs(1) = 1
    do i = 2,size(iZs)
        iZs(i) = drawAR1(trans_z,iZs(i-1),option=0)
    end do

end subroutine init_zseq


! Sets up the aggregate labor supply shifter
subroutine setup_N(eshift)
    use mod_matlab, only: hpfilter,std,hist
    use mod_param, only: trans_z, nE, nT, len_t0, nZ, trans_z, g_ind, b_ind
    use mod_shocks
    use mod_root1dim
    use mod_data, only: gE, init_dsn_e, invar_dsn_e_SS, trans_e
    implicit none
    real(8), dimension(2), intent(out) :: eshift
    !local
    real(8) :: maxN, minN
    real(8), dimension(nT) :: Nseq
    integer, dimension(nT) :: zseq_ind
    real(8) :: aggEShiftDelta, aggEg, aggEb

    !Initialize {z}
    call init_zseq(zseq_ind,trans_z)

    ! Find the aggE value
    print *, 'Searching for appropriate agg shift value'
    aggEShiftDelta = .0174d0
    call sub_brent(aggEShiftDelta,res,a_in = 0.d0,b_in=.05d0,printmod_in=1)

    ! Set eshift
    eshift = (/aggEg,aggEb/)

contains

    function res(aggEShiftDelta) 
        use mod_param, only: trans_z, eshiftStdevTarget
        use mod_matlab, only: nullspace, eye
        implicit none
        real(8), intent(in) :: aggEShiftDelta
        real(8) :: res
        !local
        real(8) :: stdN
        real(8) :: invar_dsn_z(2)
        
        invar_dsn_z = nullspace(eye(2) - transpose(trans_z))
        invar_dsn_z = invar_dsn_z/sum(invar_dsn_z)

        !Set the eshift(g) to 1+delta and find eshift(b) so that the mean is 1
        eshift(g_ind) = 1.d0 + aggEShiftDelta
        eshift(b_ind) = 1.d0/invar_dsn_z(b_ind)*(1.d0-eshift(g_ind)*invar_dsn_z(g_ind))
        
        !Calculate the exogenous sequence of labor supply
        call sub_nonstoch_simul_eonly(Nseq,zseq_ind,gE,trans_e,invar_dsn_e_SS,init_dsn_e)

        !Calculate the implied standard deviation of hp-filtered log N
        maxN = maxval(Nseq(len_t0+1:nT-1))
        minN = minval(Nseq(len_t0+1:nT-1))
        stdN = 100.d0*std(hpfilter(log(Nseq(len_t0+5:nT-5)),100.d0))
        aggEg = eshift(g_ind)
        aggEb = eshift(b_ind)

        res = stdN - 100d0*eshiftStdevTarget

    end function res

end subroutine setup_N


! sub_nonstoch_simul
! Nonstochastically simulate the distribution of households.
subroutine sub_nonstoch_simul(mseq,Kseq,Nseq,zseq_ind,K_SS,doComputeStats)
    use mod_param, only: nA,nE,Tend,nH,nM,nZ,g_ind,b_ind,probLive,grid_z,eshift,len_t0
    use mod_matlab
    use mod_law
    use mod_root1dim
    use mod_myprint
    use mod_stats
    use mod_convertAtoK
    use mod_data, only: v,c,apg,apb,q,d,gA,gE,trans_e,invar_dsn_SS,init_dsn_e
    implicit none
    real(8), intent(in) :: K_SS
    integer, dimension(nT), intent(in) :: zseq_ind
    real(8), dimension(nT), intent(out) :: Kseq, Nseq
    type(mom), dimension(nT), intent(out) :: mseq
    logical, intent(in) :: doComputeStats
    
    !local
    real(8), dimension(:,:,:,:), allocatable :: pdf, pdfp, pdfAvg
    real(8), dimension(:,:), allocatable :: statlist
    real(8), dimension(:,:,:,:,:), allocatable :: d_Work, papg_Work, papb_Work, qBond
    real(8), dimension(nE,nM) :: vNewborn
    real(8) :: Ktmp,Ntmp
    type(mom) :: mtmp
    integer :: iA,iE,iM,age,iH,iZ,T_ind
    integer, parameter :: pmod_sim = 100
    
    integer, parameter :: IRF_lb = 1857, IRF_ub = 2013
    real(8), dimension(IRF_lb:IRF_ub) :: IRF_K, IRF_N
    type(mom), dimension(IRF_lb:IRF_ub) :: IRF_m
    integer, dimension(IRF_lb:IRF_ub) :: IRF_z
    real(8), dimension(100,IRF_lb:IRF_ub) :: IRF_statlist

    if (any(isnan(apg))) STOP 'apg has nan'
    if (any(isnan(apb))) STOP 'apb has nan'

    allocate(qBond(nA,nE,Tend,nH,nM))
    allocate(pdf(nA,nE,Tend,nH),pdfp(nA,nE,Tend,nH),pdfAvg(nA,nE,Tend,nH))
    allocate(statlist(100,nT))

    if (abs(sum(invar_dsn_SS)-1.d0)>1d-6) then
        write(*,*) 'sub_nonstoch_simul: Invariant distribution does not sum to 1 but to ', sum(invar_dsn_SS)
        STOP
    end if

    !!! SETUP WORK ARRAYS THAT WILL BE USED FOR INTERPOLATION
    !!! I TAKE SPECIAL CARE TO MANAGE MEMORY B/C THESE ARE VERY LARGE ARRAYS
    
    ! Calculate work array for hp
    allocate(d_Work(nA,nE,Tend,nH,nM))
    where (d)
        d_Work = 1d0
    elsewhere
        d_Work = 0d0
    end where    
    
    ! Calculate work array for v
    vNewborn = v(nAmm+1,:,1,1,:)

    !Calculate the prob of repayment using the price schedules
    allocate(papg_Work(nA,nE,Tend,nH,nM),papb_Work(nA,nE,Tend,nH,nM))
    papg_Work = 0d0
    papb_Work = 0d0
    do iM = 1,nM
        do iH = 1,nH
            do age = 1,Tend-1 ! NOTE: this is only to Tend-1 because for age==Tend, they always save 0
                do iE = 1,nE
                    do iA = 1,nA
                        papg_Work(iA,iE,age,iH,iM) = pap_value(apg(iA,iE,age,iH,iM), probLive(age), q(:,iE,age,iH,iM,g_ind), gA)
                        papb_Work(iA,iE,age,iH,iM) = pap_value(apb(iA,iE,age,iH,iM), probLive(age), q(:,iE,age,iH,iM,b_ind), gA)
                    end do
                end do
            end do
        end do
    end do

    ! Compute the bond price
    qBond = 0d0
    do iZ = 1,nZ
        qBond = qBond + q(:,:,:,:,:,iZ)
    end do

    
    !Initialize the distribution and moments
    pdf = invar_dsn_SS
    
    pdfAvg = 0d0

    mseq%z_ind = zseq_ind
    mseq%z = grid_z(zseq_ind)    

    Kseq(1) = K_SS
    Nseq(1) = eshift(mseq(1)%z_ind)*sum(gE*sum(sum(pdf(:,:,:,:),4),1))
    mseq(1)%KN = Kseq(1)/Nseq(1)
    
    ! Simulate
    do T_ind = 1,nT

        ! Simulate one period
        if (T_ind<nT) then
            call sub_sim_one_period(pdfp,mseq(T_ind+1),Kseq(T_ind+1),Nseq(T_ind+1),statlist(:,T_ind),pdf,mseq(T_ind),&
                                    d_Work,qBond,papg_Work,papb_Work,zseq_ind(T_ind+1),doComputeStats)
        else
            call sub_sim_one_period(pdfp,mtmp,Ktmp,Ntmp,statlist(:,T_ind),pdf,mseq(T_ind),&
                                    d_Work,qBond,papg_Work,papb_Work,1,doComputeStats)
        end if

        ! Display some info
        if (T_ind==1 .or. mod(T_ind,pmod_sim)==0) &
            write(*,'(A,i5,3(A,f11.4))') 'T: ', T_ind,' K:',Kseq(T_ind),' W:',mseq(T_ind)%WW, ' N:',Nseq(T_ind)

        if (T_ind>len_t0) then
            pdfAvg = pdfAvg + pdf
        end if

        ! Update the distribution 
        pdf = pdfp

    end do


    !Write stats
    pdfAvg = pdfAvg/sum(pdfAvg) !Before this step pdfAvg is only the sum

    call myprint('seq_statlist',statlist)
    call myprint('seq_pdfAvg',pdfAvg)

!     call plot(statlist(39,:),'mean_c')

!     ! Also, compute the eq path for a sequence of shocks mimicking the great recession
! 
!     ! Get the shocks for the economy
!     IRF_z = [0,0,1,1,0,1,1,1,0,0,0,1,0,0,1,1,1,0,0,0,0,0,1,1,1,0,0,0,1,1,0,1,1,1,1,1,0,0,1,0,0,1,0,0,&
!              1,1,0,0,1,1,0,0,1,0,0,1,0,0,1,1,1,1,1,0,0,1,0,0,1,1,0,1,1,0,0,0,1,1,1,1,0,0,1,1,1,1,1,1,0,1,1,1,0,&
!              1,1,1,1,1,1,1,1,1,1,0,1,1,1,1,1,1,1,1,1,0,1,1,1,0,1,1,1,1,1,0,1,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,&
!              1,1,1,0,1,1,1,1,1,1,0,0,1,1,1,1]
!     IRF_z = merge(g_ind,b_ind,IRF_z==1) 
!     IRF_m%z_ind = IRF_z
! 
!     ! Set the initial state
!     pdf = invar_dsn_SS
! 
!     IRF_K(IRF_lb) = K_SS
!     IRF_N(IRF_lb) = eshift(IRF_z(IRF_lb))*sum(gE*sum(sum(pdf(:,:,:,:),4),1))
!     IRF_m(IRF_lb)%KN = IRF_K(IRF_lb)/IRF_N(IRF_lb)
! 
!     do T_ind = IRF_lb,IRF_ub
! 
!         ! Simulate one period
!         if (T_ind<IRF_ub) then
!             call sub_sim_one_period(pdfp,IRF_m(T_ind+1),IRF_K(T_ind+1),IRF_N(T_ind+1),IRF_statlist(:,T_ind),pdf,IRF_m(T_ind),&
!                                     d_Work,qBond,papg_Work,papb_Work,IRF_z(T_ind+1))
!         else
!             call sub_sim_one_period(pdfp,mtmp,Ktmp,Ntmp,IRF_statlist(:,T_ind),pdf,IRF_m(T_ind),&
!                                     d_Work,qBond,papg_Work,papb_Work,1)
!         end if
! 
!         ! Update
!         pdf = pdfp
! 
!     end do
! 
!     call myprint('IRF_statlist',IRF_statlist)
    
end subroutine sub_nonstoch_simul

subroutine dointerp(xi,x,indices,weights)
    implicit none
    real(8), intent(in) :: x(:,:,:,:,:),weights(:)
    integer, intent(in) :: indices(:)
    real(8), intent(out) :: xi(:,:,:,:)

    xi = weights(1)*x(:,:,:,:,indices(1)) + &
         weights(2)*x(:,:,:,:,indices(2)) + &
         weights(3)*x(:,:,:,:,indices(3)) + &
         weights(4)*x(:,:,:,:,indices(4))

end subroutine 

! Given a pdf, a current z, a future z, and policy functions, this routine simulates the economy forward one period.
subroutine sub_sim_one_period(pdfp,mp,Kp,Np,statlist,pdf,m,d_Work,qBond,papg_Work,papb_Work,iZp,doComputeStats)
    use mod_param, only: nT,len_t0,nH,nM,nZ,grid_z,trans_z,g_ind,b_ind,grid_WW,eshift,Tend,lambda,mom,alpha_K
    use mod_stats, only: stat_type, computeStats
    use mod_root1dim, only: sub_brent
    use mod_law, only: get_m_interp
    use mod_data, only: gA,gE,trans_e,init_dsn_e,v,c,apg,apb
    use mod_matlab
    implicit none
    real(8), intent(out) :: pdfp(:,:,:,:), Kp, Np
    real(8), intent(inout) :: statlist(100)
    type(mom), intent(out) :: mp
    real(8), intent(in) :: pdf(:,:,:,:)
    type(mom), intent(inout) :: m ! eq WW is computed by this routine
    integer, intent(in) :: iZp
    real(8), dimension(:,:,:,:,:), intent(in) :: d_Work, qBond, papg_Work, papb_Work
    logical, intent(in) :: doComputeStats
    ! local
    real(8), dimension(:,:,:,:), allocatable :: apInterp,cInterp,dInterp,vInterp,qBondInterp ! a,e,age,h
    real(8) :: pdf_marginal_e(nE,Tend), pdfpg_marginal_e(nE,Tend), pdfpb_marginal_e(nE,Tend)
    integer :: indices(4)
    real(8) :: weights(4), AApg_grid(nM), AApb_grid(nM)
    real(8) :: Aapg, Aapb, Npg, Npb
    real(8) :: WWeq, tmp
    integer :: iZ
    type(stat_type) :: stmp

    allocate(apInterp(nA,nE,Tend,nH),cInterp(nA,nE,Tend,nH),dInterp(nA,nE,Tend,nH),vInterp(nA,nE,Tend,nH),qBondInterp(nA,nE,Tend,nH)) 

    ! Get z_ind from m
    iZ = m%z_ind

    ! Ensure m%z agrees with m%z_ind
    m%z = grid_z(m%z_ind)
    
    ! Calculate Npg and Npb which are totally exogenous
    pdf_marginal_e = sum(sum(pdf,4),1)
    call sub_pdfp_eonly(pdfpg_marginal_e,pdf_marginal_e,trans_e(:,:,:,iZ,g_ind),init_dsn_e(:,g_ind)) ! This takes into account mortality risk
    call sub_pdfp_eonly(pdfpb_marginal_e,pdf_marginal_e,trans_e(:,:,:,iZ,b_ind),init_dsn_e(:,b_ind))
    Npg = eshift(g_ind)*sum(gE*pdfpg_marginal_e)
    Npb = eshift(b_ind)*sum(gE*pdfpb_marginal_e)

    ! Compute what AAp values are on the moments grid (these are interpolated inside the res routine)
    call mydgemv(AApg_grid,papg_Work,pdf,nA*nE*Tend*nH,nM,tA='T')
    call mydgemv(AApb_grid,papb_Work,pdf,nA*nE*Tend*nH,nM,tA='T')
    
    ! Call solver to find bond price that clears.  This implicitly updates the policy and value functions.
    call sub_brent(WWeq,res,a_in=minval(grid_WW),b_in=maxval(grid_WW),toler_in=1d-5,xtoler_in=1d-12,printmod_in=-1)

    !!! Once have eq val in hand, call internal function again to reset values, AApg, AApb, Kp, m%WW.
    tmp = res(WWeq)
         
    ! Using eq weighting, interpolate to recover asset and h' policy functions and also get apInd

    ! Get the interpolation weights
    call get_m_interp(indices,weights,m) ! Get interpolation weights and indices for the moment grid
    
    ! Interpolate the policies
    call dointerp(dInterp,d_Work,indices,weights)
    call dointerp(cInterp,c,indices,weights)
    call dointerp(vInterp,v,indices,weights)
    call dointerp(qBondInterp,qBond,indices,weights)
    if (iZp==g_ind) then
        call dointerp(apInterp,apg,indices,weights) !Note: this a' is actually conditional on repaying, not the standard definition               
    else
        call dointerp(apInterp,apb,indices,weights)
    end if

    ! Compute the next period distributions cdnal on the z' realization
    call sub_pdfp(pdfp,pdf,apInterp,dInterp,trans_e(:,:,:,iZ,iZp),init_dsn_e(:,iZp)) 

    ! Set next period values
    Np = merge(Npg,Npb,g_ind == iZp)
    mp%KN = Kp/Np
    mp%z_ind = iZp

    !Collect some statistics
    if (doComputeStats) then
        call computeStats(statlist,stmp,2,pdf,grid_z(m%z_ind),eshift(m%z_ind),vInterp,cInterp,apInterp,dInterp>.5d0,qBondInterp,gE) 
    end if

contains

    function res(WW_in) result(residual)
        use mod_matlab
        use mod_param, only: g,b
        use mod_convertAtoK
        implicit none
        real(8), intent(in) :: WW_in
        real(8) :: residual
        !local
        real(8) :: Kp_impliedby_AApg, Kp_impliedby_AApb !, accum

        ! Check that the moment sequence is initialize
        if (m%KN<0.d0 .or. m%z<0d0) STOP 'moment not init'

        ! Store the weighting guess in the moment
        m%WW = WW_in

        ! Get interpolation weights 
        call get_m_interp(indices,weights,m) ! Get interpolation weights and indices for the moment grid
        
        ! Interpolate the asset policy function and aggregate using today's pdf
        ! This is okay because if households default in one state and don't in another, we're taking a weighted 
        ! average of a'Repay and a'Default(=0).       
        AApg = dot_product(weights,AApg_grid(indices))
        AApb = dot_product(weights,AApb_grid(indices))

!         AApg = sum((weights(1)*papg_Work(:,:,:,:,indices(1)) + &
!                     weights(2)*papg_Work(:,:,:,:,indices(2)) + &
!                     weights(3)*papg_Work(:,:,:,:,indices(3)) + &
!                     weights(4)*papg_Work(:,:,:,:,indices(4)))    *   pdf)
! 
!         AApb = sum((weights(1)*papb_Work(:,:,:,:,indices(1)) + &
!                     weights(2)*papb_Work(:,:,:,:,indices(2)) + &
!                     weights(3)*papb_Work(:,:,:,:,indices(3)) + &
!                     weights(4)*papb_Work(:,:,:,:,indices(4)))    *   pdf)

        !Compute the implied capital next period
        Kp_impliedby_AApg = convertAtoK(AApg,Npg,g)
        Kp_impliedby_AApb = convertAtoK(AApb,Npb,b)
        Kp = .5d0*(Kp_impliedby_AApg + Kp_impliedby_AApb)

        ! Output the residual
        residual = .5d0*(Kp_impliedby_AApg - Kp_impliedby_AApb) 
        
    end function

end subroutine 

! function weightedsum(w,inds,x,y)
!     use mod_param, only: nA,nE,Tend,nH
!     implicit none
!     real(8), intent(in) :: w(4),x(nA*nE*Tend*nH,*),y(*)
!     integer, intent(in) :: inds(4)
!     real(8) :: weightedsum
!     ! local
!     integer :: i,n
!     real(8) :: accum
! 
!     n = nA*nE*Tend*nH
! 
!     accum = 0d0
! 
!     !$OMP PARALLEL DO SCHEDULE(STATIC) DEFAULT(SHARED) PRIVATE(i) REDUCTION(+:accum)
!     do i = 1,n
!         accum = accum + w(1)*x(i,inds(1))*y(i) + w(2)*x(i,inds(2))*y(i)+ w(3)*x(i,inds(3))*y(i)+ w(4)*x(i,inds(4))*y(i)
!     end do
!     !$OMP END PARALLEL DO
! 
!     weightedsum = accum
! 
! end function


function pap_value(ap,probLive,q,gA) result(pap)
    use mod_matlab, only: interp1q
    implicit none
    real(8), dimension(1:nA), intent(in) :: q,gA
    real(8), intent(in) :: ap,probLive
    real(8) :: pap
    ! local
    real(8) :: qval

    ! Recover the price
    qval = interp1q(gA,q,ap)

    ! Compute p (probLive*q/qriskfree) and multiply by a'
    pap = (probLive*qval/q(nA))*ap
    
end function

subroutine sub_nonstoch_simul_eonly(Nseq,zseq_ind,gE,trans_e,invar_dsn_e_SS,init_dsn_e)
    use mod_param, only: nE,Tend,nT,len_t0,nM,nZ,trans_z,g_ind,b_ind,eshift,lambda
    implicit none
    real(8), dimension(nT), intent(out) :: Nseq
    integer, dimension(nT), intent(in) :: zseq_ind
    real(8), dimension(nE,Tend), intent(in) :: invar_dsn_e_SS, gE
    real(8), dimension(nE,nZ), intent(in) :: init_dsn_e
    real(8), dimension(nE,nE,Tend,nZ,nZ), intent(in) :: trans_e
    !local
    real(8) :: pdf_marginal_e(nE,Tend), pdfp_marginal_e(nE,Tend)
    integer :: T_ind

    pdf_marginal_e = invar_dsn_e_SS
    Nseq(1) = eshift(zseq_ind(1))*sum(gE*pdf_marginal_e)

    ! Simulate the economy but only to obtain the exogenous sequence of labor supplies
    do T_ind = 1,nT-1
        if (abs(sum(pdf_marginal_e)-1.d0)>1d-12) then
            print*,'T_ind',T_ind,'sum(pdf)',sum(pdf_marginal_e)
            STOP 'pdf_marginal_e doesn''t sum to 1'
        end if

        call sub_pdfp_eonly(pdfp_marginal_e,pdf_marginal_e,trans_e(:,:,:,zseq_ind(T_ind),zseq_ind(T_ind+1)),init_dsn_e(:,zseq_ind(T_ind+1))) ! This takes into account mortality risk

        Nseq(T_ind+1) = eshift(zseq_ind(T_ind+1))*sum(gE*pdfp_marginal_e)
        pdf_marginal_e = pdfp_marginal_e
    end do 

end subroutine sub_nonstoch_simul_eonly

subroutine sub_pdfp_eonly(pdfp,pdf,pi_ee,init_dsn_e)
    use mod_param, only: probLive,nE,Tend
    use mod_matlab, only: mtimes
    implicit none
    real(8), intent(in) :: pdf(nE,Tend), init_dsn_e(nE), pi_ee(nE,nE,Tend)
    real(8), intent(out) :: pdfp(nE,Tend)
    !local
    integer :: age

    pdfp = 0d0

    !Transition the e's within a particular age's group
    do age = 1,Tend-1
        !Note the use of epl and epu in pdfp b/c I've already transitioned them by age
        pdfp(:,age+1) = probLive(age)*matmul(pdf(:,age),pi_ee(:,:,age)) !No transpose here because moving forward in time
    end do

    !Replace those who died with newborns who have 0 assets, good standing, and e according to init_dsn_e
    do age = 1,Tend
        pdfp(:,1) = pdfp(:,1) + ((1d0-probLive(age))*sum(pdf(:,age)))*init_dsn_e(:)
    end do

end subroutine sub_pdfp_eonly

subroutine sub_pdfp(pdfp,pdf,ap,d,pi_ee,init_dsn_e) !,api_lb,api_ub)
    use mod_matlab, only: mtimes,bsearch
    use mod_param, only: nA,nE,Tend,nH !,probLive,lambda
    implicit none
    real(8), dimension(nA,nE,Tend,nH), intent(out) :: pdfp
    real(8), dimension(nA,nE,Tend,nH), intent(in) :: pdf,d
    real(8), dimension(nA,nE,Tend,nH), intent(in) :: ap
    real(8), dimension(nE,nE,Tend), intent(in) :: pi_ee
    real(8), dimension(nE), intent(in) :: init_dsn_e
    !local
!     real(8), allocatable :: tmparray(:,:,:,:)
    real(8), allocatable, save :: work(:,:,:)
    integer :: age

    if (.not. allocated(work)) allocate(work(nA,nE,Tend))

    !Check that dsn looks okay
    if (abs(sum(pdf)- 1.d0)>1d-10) then
        write(*,*) 'WARNING: Invar dsn does not sum to 1. It sums to', sum(pdf)
    end if
    if (minval(pdf)<-1d-10) then
        write(*,*) 'WARNING: Invar dsn has some negative elements! ', minval(pdf)
    end if


    ! NOTE: I think this can be parallelized across age 
!     !$OMP PARALLEL DEFAULT(SHARED) PRIVATE(age,tmparray)
!     allocate(tmparray(nA,nE,Tend,nH))
!     tmparray = 0d0

!     !$OMP DO
    

    !$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(age)
    do age = 1,Tend
    
        
        ! sub_dsn_aux computes the dsn(:,:,age,:) using dsn(:,:,age-1,:) if age>1
        ! if (age>1) then
        !     tmparray(:,:,age-1,:) = pdf(:,:,age-1,:)
        ! end if
        ! call sub_dsn_aux(age,tmparray,d,ap,pi_ee,init_dsn_e)
        ! pdfp(:,:,age,:) = tmparray(:,:,age,:)
        call sub_dsn_aux(pdfp,pdf,age,d,ap,pi_ee,init_dsn_e,work(:,:,age))

!         if (any(isnan(pdfp(:,:,age,:)))) then
!             print*,'any(isnan(pdf))',any(isnan(pdf))
!             print*,'any(isnan(ap))',any(isnan(ap))
!             print*,'any(isnan(d))',any(isnan(d))
!             STOP 'found nan in pdfp'
!         end if

    end do
    !$OMP END PARALLEL DO

!     !$OMP END DO
!     !$OMP END PARALLEL


end subroutine

! Given an age and distribution dsn, computes the next period distribution over age.
! That is, given a distribution dsn, it uses the distribution dsn(.,age-1) to compute
! what dsn(.,age). For computational convenience, this info is written directly to dsn(.,age). 
! If age is 1, this code sets dsn(.,1) w/o referencing dsn(.,0). 
! Those who die are not accounted for at each iteration (i.e., this code does not write 
! to dsn(.,age=1) each time). However, they are accounted for in this initial distribution of age=1 households. 
! Consequently, if one wants to compute the steady state distribution, just call this code Tend times with
! ages 1 to Tend in order.
!
! NOTE: this is completely independent of the legal environment (for the ones we've considered). You can
! verify this by seeing that mod_param only uses basic demographics information.
!
! ALSO: dreal is used rather than d_SS because this code is called by sub_nonstochsimul where 
! d_SS is interpolated (and so not true or false but in [0,1]).
subroutine sub_dsn_aux(dsn_out,dsn_in,age_in,dreal,ap,trans_e,init_dsn_e,work)
    use mod_param, only: Tend, probLive, lambda, nA, nAmm, nE, nH
    use mod_matlab, only: cumprod, bsearch, mydgemm
    use mod_data, only: gA
    implicit none
    real(8), intent(inout) :: dsn_out(:,:,:,:)
    integer, intent(in) :: age_in
    real(8), intent(in) :: dsn_in(:,:,:,:)
    real(8), intent(in) :: ap(:,:,:,:), trans_e(:,:,:), init_dsn_e(:), dreal(:,:,:,:)
    real(8), intent(out) :: work(:,:)
    ! local
!     real(8), allocatable :: tmparray(:,:)
    real(8) :: mu, apval, aplo_m
    real(8) :: msrByAge(Tend)
    integer :: iA, iE, iH, aplo_i
    integer :: age

    ! Compute the measure of each age
    msrByAge = cumprod((/1d0,probLive(1:Tend-1)/))
    msrByAge = msrByAge/sum(msrByAge)


    ! Check for special case of age==1
    age = age_in
    if (age==1) then
        ! Initial dsn: everyone begins with zero assets, init_dsn_e, age 1, and good credit.
        ! The total measure is msrByAge(1).
        dsn_out(:,:,1,:) = 0d0
        dsn_out(nAmm+1,:,1,1) = msrByAge(1)*init_dsn_e 
        return
    end if

    ! So don't have to modify below code, set age to age_in - 1    
    age = age_in - 1

    dsn_out(:,:,age+1,:) = 0d0
!     !$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(iE,iH,iA,mu,apval,aplo_i,aplo_m)
    do iE = 1,nE
        do iH = 1,nH
            do iA = 1,nA
                
                if (iH==2 .and. iA<nAmm+1) cycle
                
                ! Get mass of households and immediately kill of those who don't survive
                mu = dsn_in(iA,iE,age,iH)*probLive(age)
                
                if (mu==0d0) cycle
                
                ! Get location of ap on the grid according to a'(a,e,age,h). This incorporates the default decision. 
                apval = ap(iA,iE,age,iH)
                call bsearch(aplo_i,apval,gA)
                aplo_m = (gA(aplo_i+1)-apval)/(gA(aplo_i+1)-gA(aplo_i))
                aplo_m = min(max(aplo_m,0d0),1d0)
                ! if (abs(aplo_m)<2d0*epsilon(0d0)) aplo_m = 0d0
                ! if (abs(1d0-aplo_m)<2d0*epsilon(0d0)) aplo_m = 1d0

                ! Distribute the mass to next period
                if (iH==1) then

                    ! If default, go to a'=0,h'=1 for sure
                    dsn_out(nAmm+1,iE,age+1,2) = dsn_out(nAmm+1,iE,age+1,2)     + mu*dreal(iA,iE,age,iH)
                    ! if don't default, go to h'=0 for sure
                    dsn_out(aplo_i  ,iE,age+1,1) = dsn_out(aplo_i  ,iE,age+1,1) + mu*aplo_m*(1d0-dreal(iA,iE,age,iH))
                    dsn_out(aplo_i+1,iE,age+1,1) = dsn_out(aplo_i+1,iE,age+1,1) + mu*(1d0-aplo_m)*(1d0-dreal(iA,iE,age,iH))

                else
                    ! if default, go to a'=0,h'=1 for sure
                    dsn_out(nAmm+1,iE,age+1,2) = dsn_out(nAmm+1,iE,age+1,2) + mu*dreal(iA,iE,age,iH)

                    ! if already in default and, so go to h'=0 with probability 1-lambda
                    dsn_out(aplo_i  ,iE,age+1,1) = dsn_out(aplo_i  ,iE,age+1,1) + mu*aplo_m*(1d0-lambda)*(1d0-dreal(iA,iE,age,iH))
                    dsn_out(aplo_i+1,iE,age+1,1) = dsn_out(aplo_i+1,iE,age+1,1) + mu*(1d0-aplo_m)*(1d0-lambda)*(1d0-dreal(iA,iE,age,iH))
                    dsn_out(aplo_i  ,iE,age+1,2) = dsn_out(aplo_i  ,iE,age+1,2) + mu*aplo_m*lambda*(1d0-dreal(iA,iE,age,iH))
                    dsn_out(aplo_i+1,iE,age+1,2) = dsn_out(aplo_i+1,iE,age+1,2) + mu*(1d0-aplo_m)*lambda*(1d0-dreal(iA,iE,age,iH))

                end if

            end do
        end do             
    end do
!     !$OMP END PARALLEL DO

    ! Add the e' transition
    do iH = 1,nH
        work = dsn_out(:,:,age+1,iH) 
        ! dsn_out(:,:,age+1,iH) = matmul(work,trans_e(:,:,age)) ! use matmul instead of blas because sub_dsn_aux is called within omp and don't want to descend into a threaded blas call 
        call mydgemm(dsn_out(:,:,age+1,iH),work,trans_e(:,:,age),nA,nE,nE,nE) ! but blas is so much faster it is probably better anyway
    end do

!     allocate(tmparray(nA,nE))
!     do iH = 1,nH
!         tmparray = dsn_out(:,:,age+1,iH) 
!         call mydgemm(dsn_out(:,:,age+1,iH),tmparray,trans_e(:,:,age),nA,nE,nE,nE)
!     end do

end subroutine sub_dsn_aux


end module mod_nonstochsimul

