module mod_price_fcns

implicit none

contains

!Calc_price_schedules.  
! Given probabilities of repayment and risk free rates, calculates price and expenditure schedules
!Note: this doesn't really belong in this module and should be but into a separate one.
!It weakly depends on the law module.
subroutine calc_price_schedules(q,qbar,d,age,w,gE,trans_e,trans_m_cdnal_zp,&
                                z_indices,gA)
    use mod_param, only: isUsuryLaw, usuryLimit, eshift, problive, chiGarnish, nA,nE,Tend,nM,nZ,nH,nAmm, lambda
    use mod_matlab, only: mydgemm
    implicit none
    integer, intent(in) :: age
    real(8), dimension(nA,nE,Tend,nH,nM,nZ), intent(inout) :: q
    logical, dimension(nA,nE,Tend,nH,nM), intent(in) :: d
    real(8), dimension(nE,nE,Tend,nZ,nZ),intent(in) :: trans_e
    integer, dimension(nM), intent(in) :: z_indices
    real(8), dimension(nM,nM,nZ), intent(in) :: trans_m_cdnal_zp
    real(8), dimension(nM,nZ), intent(in) :: qbar
    real(8), dimension(nA), intent(in) :: gA
    real(8), dimension(nE,Tend), intent(in) :: gE
    real(8), dimension(nM), intent(in) :: w
    !local
    real(8), dimension(:,:,:,:,:), allocatable :: p
    real(8), dimension(:,:,:,:), allocatable :: recovery_rate, tmpAEHM
    real(8) :: chiPaid
    integer :: iM, iEp, iAp, iMp, iZp, iHp

    allocate(recovery_rate(nA,nE,nH,nM),p(nA,nE,nH,nM,nZ),tmpAEHM(nA,nE,nH,nM))


    if (age>=Tend) STOP 'calc_price_schedules: can only applied to age<Tend'
    
    ! Given default decisions, compute the net recovery rate 1-d + d*chi*we/a. 
    ! The recovery rate depends on how much is garnished, summarized in the 
    ! chiGarnish function
    do iHp = 1,nH
        do iEp = 1,nE  
            do iMp = 1,nM
                chiPaid = chiGarnish(gE(iEp,age+1)*eshift(z_indices(iMp)),&
                                                      eshift(z_indices(iMp)))
                do iAp = 1,nA
                    if (iAp==nAmm+1) then 
                        recovery_rate(iAp,iEp,iHp,iMp) = recovery_rate(iAp-1,iEp,iHp,iMp) ! recoveryrate can be arbitrary, so, for pretty graphs, set to iAp-1
                        cycle
                    end if
                    if (d(iAp,iEp,age+1,iHp,iMp)) then
                        recovery_rate(iAp,iEp,iHp,iMp) = chiPaid*w(iMp)*gE(iEp,age+1)*eshift(z_indices(iMp))/(-gA(iAp))
                    else
                        recovery_rate(iAp,iEp,iHp,iMp) = 1d0
                    end if
                end do
            end do
        end do
    end do

    if (any(isnan(recovery_rate))) STOP 'recoveryrateisnan'

    ! The recovery rate is a function r(a',e',m') (Note that z' is implicit in m').
    ! We need to compute the expectation p(a',e,m;z') = E(a',e',m'|a',z',e,m) for each z'.  
    !!! To do this, first compute E(a',e',m'|a',z',e',m) = sum_m' r(a',e',m')*F(m'|z',m)
    do iZp = 1,nZ
        ! call dgemm('N','T',nA*nE,nM,nM,1d0,recovery_rate,nA*nE,trans_m_cdnal_zp(:,:,iZp),nM,0d0,tmpAEM,nA*nE) 
        call mydgemm(tmpAEHM,recovery_rate,trans_m_cdnal_zp(:,:,iZp),nA*nE*nH,nM,nM,nM,tB='T')
        !!! Second, compute p(a',e,m;z') by calculating E(a',e',m'|a',z',e,m) = sum_{e'} E(a',e',m'|a',z',e',m)*F(e'|e,z')
        do iHp = 1,nH
            do iM = 1,nM
                p(:,:,iHp,iM,iZp) = matmul(tmpAEHM(:,:,iHp,iM),transpose(trans_e(:,:,age,z_indices(iM),iZp)))
            end do
        end do
    end do
    p(:,:,2,:,:) = lambda*p(:,:,2,:,:) + (1d0-lambda)*p(:,:,1,:,:) ! now function of h rather than h'

    if (any(isnan(p))) STOP 'p is nan'
        
    !Use p to determine price schedule
    do iHp = 1,nH ! just use iHp but iH is more appropriate
        do iZp = 1,nZ
            do iM = 1,nM
                q(:,:,age,iHp,iM,iZp) = qbar(iM,iZp)*problive(age)*p(:,:,iHp,iM,iZp)
            end do
        end do
    end do

    if (any(isnan(q(:,:,age,:,:,:)))) STOP 'q(age) is nan'

    ! Impose the usury law if applicable
    if (isUsuryLaw) then
        do iZp = 1,nZ
            where (sum(q(:,:,age,:,:,:),5)<usuryLimit) 
                q(:,:,age,:,:,iZp) = 0d0
            endwhere
        end do
    end if

end subroutine calc_price_schedules

elemental function zKN_to_w(z,K,N) result(w)
    use mod_param, only: alpha_K
    implicit none
    real(8), intent(in) :: K,N,z
    real(8) :: w
    w = z*(1.d0-alpha_K)*K**(alpha_K)*N**(-alpha_K)
end function zKN_to_w

elemental function zKN_to_r(z,K,N) result(r)
    use mod_param, only: alpha_K
    implicit none
    real(8), intent(in) :: K,N,z
    real(8) :: r
    r = z*alpha_K*K**(alpha_K-1.d0)*N**(1.d0-alpha_K)
end function zKN_to_r

end module mod_price_fcns







    ! Compute the transition matrix for (e,m) to (e',m')
    
    ! Compute the expected recovery rates conditional on g or b shocks
!     pg = 0.d0
!     pb = 0.d0
! !     !$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(iEp,iMp,iM,z_ind,iE,iAp) REDUCTION(+:pg,pb)
!     do iEp = 1,nE  
!         do iMp=1,nM
!             do iM = 1,nM
!                 z_ind = z_indices(iM)
!                 !Calculate sum_{e',m'}(1-d(a',e';m'))Pr(e'|e,z,z')*Pr(m'|m,z')
!                 do iE = 1,nE
!                     do iAp = 1,nAmm
!                         pg(iAp,iE,iM) = pg(iAp,iE,iM) + trans_m_cdnal_zpg(iM,iMp)*trans_e(iE,iEp,age,z_ind,g_ind)*recovery_rate(iAp,iEp,iMp)
!                         pb(iAp,iE,iM) = pb(iAp,iE,iM) + trans_m_cdnal_zpb(iM,iMp)*trans_e(iE,iEp,age,z_ind,b_ind)*recovery_rate(iAp,iEp,iMp)
!                     end do
!                 end do
!             end do
!         end do        
!     end do
! !     !$OMP END PARALLEL DO
