! mod_shocks
! 
! Generates all the shock process info
! 
! A few definitions that are used throughout (except in mod_SS.f90):
! 1) init_dsn_e_SS(nE)                  ! (e) with Pr(e|age=1,ss)
! 2) init_dsn_e_g(nE)                   ! (e) with Pr(e|age=1,z=g)
! 3) init_dsn_e_b(nE)                   ! (e) with Pr(e|age=1,z=b) 
! 4) trans_e_SS(nE,nE,Tend)             ! (e,e',age) with Pr(e' | e,age)
! 5) trans_z(nZ,nZ)               ! (z,z') with Pr(z'|z)
! 6) trans_e(nE,nE,Tend,nZ,nZ)    ! (e,e',age,z,z') with Pr(e' | e,z,z',age)
! 7) trans_ez(nE,nE,nZ,nZ,Tend)   ! (e,e',z,z',age) with Pr(e',z' | e,z,age)
! 8) invar_dsn_e_SS(nE,Tend)            ! (e,age) with Pr(e,age) (in steady state)
!     
module mod_shocks
    use mod_param, only: nZ,Tend,npers,ntran,nCCNRaugment,nE,useSTYnotKO,useFlatEarn,useNoCEV,&
                         threshold, fracLastPeriod, fracGuaranteed, phiCCNRaugment, eloCCNRaugment,&
                         ehiCCNRaugment, pi11augment, pi21augment, pi1augment_init, stdevCoverpers, &
                         stdevCoverPerm, stdevCoverTran
    implicit none

    ! Defines a type used in calculating the shock process
    type state_type
        logical :: isLogProc ! Log process
        integer :: zi,ei ! Log Process State, z=persistent e=transitory
        logical :: isSuperRich ! Whether super rich or not
        integer :: ui ! Super rich process state
        integer :: xi ! Expenditure shock state
        real(8) :: expend ! Expenditure shock value
        real(8) :: persval,tranval ! Persistent shock value and transitory shock value (assuming linear in logs)
        real(8) :: richval ! Super rich shock logged
    end type

    ! Vector of state_type giving state associated with e
    ! Used by mov_value_fcns
!     type(state_type), dimension(nE) :: state
    type(state_type), dimension(:), allocatable :: state

    private :: sub_idshocks, sub_aggshocks, sub_jointshocks

    real(8) :: stdev_perm,stdev_tran
    real(8), dimension(1:Tend), target :: AR1_pers_prof,stdev_avg_pers_prof,stdev_g_pers_prof,stdev_b_pers_prof

    private :: stdev_perm,stdev_tran,stdev_avg_pers_prof,stdev_g_pers_prof,stdev_b_pers_prof

contains

! Computes the transition matrix for a particular grid using Tauchen's method
function tauchenTrans(grid,pers,stdev,mean) result(trans)
    use mod_matlab, only: normcdf
    implicit none
    real(8), dimension(1:), intent(in) :: grid
    real(8), intent(in) :: pers,stdev,mean
    real(8), dimension(1:size(grid),1:size(grid)) :: trans
    !local
    real(8), dimension(1:size(grid)-1) :: mz
    real(8) :: tmp
    integer :: n,zi,zpi

    n = size(grid)

    mz = (grid(1:n-1)+grid(2:n))/2.d0
    do zi = 1,n
        tmp = mean + pers*grid(zi)
        trans(zi,1) = normcdf(mz(1),tmp,stdev)
        trans(zi,n) = 1.d0-normcdf(mz(n-1),tmp,stdev)
        do zpi = 2,n-1
            trans(zi,zpi) = normcdf(mz(zpi),tmp,stdev)-normcdf(mz(zpi-1),tmp,stdev)
        end do
    end do

end function tauchenTrans

function tauchenIid(grid,stdev,mean) result(dsn)
    use mod_matlab, only: normcdf
    implicit none
    real(8), dimension(1:), intent(in) :: grid
    real(8), intent(in) :: stdev,mean
    real(8), dimension(1:size(grid)) :: dsn
    !local
    real(8), dimension(1:size(grid)-1) :: mz
    real(8) :: tmp
    integer :: n,zi

    n = size(grid)

    if (n>1) then 
        mz = (grid(1:n-1)+grid(2:n))/2.d0

        tmp = mean !Mean
        dsn(1) = normcdf(mz(1),tmp,stdev)
        dsn(n) = 1.d0-normcdf(mz(n-1),tmp,stdev)
        do zi = 2,n-1
            dsn(zi) = normcdf(mz(zi),tmp,stdev)-normcdf(mz(zi-1),tmp,stdev)
        end do
    else
        dsn = 1.d0
    end if

end function tauchenIid

! subroutine getShockInfo(stdev_perm,stdev_tran,AR1_pers_prof,stdev_avg_pers_prof,stdev_g_pers_prof,stdev_b_pers_prof)
subroutine getShockInfo()
    use mod_param
    use mod_matlab
    implicit none
!     real(8), intent(out) :: stdev_perm,stdev_tran
!     real(8), dimension(1:Tend), intent(out), target :: AR1_pers_prof,stdev_avg_pers_prof,stdev_g_pers_prof,stdev_b_pers_prof
    !local
    ! STY values
    real(8), parameter :: STY_AR1=.952d0,STY_stdev_g_pers=.125d0,STY_stdev_b_pers=.211d0,&
                          STY_stdev_avg_pers=.168d0,STY_perm = .378d0,STY_tran=.255d0
    ! KO values
    real(8), parameter :: KO_AR1_cubic(4) = (/.7596d0,.2039d0,-.0535d0,.0028d0/), &
                          KO_avg_pers_cubic(4) = (/.0518d0,-.0405d0,.0105d0,-.0002d0/), &
                          KO_pers_init = .0803d0, KO_perm=.0768d0, KO_tran=.0564d0
    ! Other
    real(8), dimension(:), pointer :: rho, sig
    integer :: age,h

    if (.not. useSTYnotKO) then

        ! KO
        do age = 1,Tend
            h = (T1age + age-1) - 24 + 1 ! To KO, h=1 means real age 24. (T1age+age-1) is the real age. Subtracting 24 and adding 1 means a real age of 24 correponds to h=1
            AR1_pers_prof(age) = KO_AR1_cubic(1) + KO_AR1_cubic(2)*(real(h,8)/10d0) &
                               + KO_AR1_cubic(3)*(real(h,8)/10d0)**2 + KO_AR1_cubic(4)*(real(h,8)/10d0)**3
            stdev_avg_pers_prof(age) = KO_avg_pers_cubic(1) + KO_avg_pers_cubic(2)*(real(h,8)/10d0) + &
                                       KO_avg_pers_cubic(3)*(real(h,8)/10d0)**2 + KO_avg_pers_cubic(4)*(real(h,8)/10d0)**3
        end do
        stdev_avg_pers_prof = sqrt(stdev_avg_pers_prof) ! Convert to standard deviations

        if (T1age/=20) STOP 'I need to fix getShockInfo'
        rho(20:84) => AR1_pers_prof
        sig(20:84) => stdev_avg_pers_prof
        ! Now have to use these cubic profiles to back out, assuming the profiles are right, the variance of (z_20) s.t. var(z_24) is KO_pers_init. 
        ! NOTE: I FOUND THIS IS IMPOSSIBLE, the variance actually has to be negative!
        ! Instead, what we can do is assume the variance of the innovations at age <=24 is constant, assume the persistent component is correct, and 
        ! compute the <=24 variance to make the unconditional variance of z at age 24 equal to the computed value
        sig(20:24) = sqrt( KO_pers_init/(1.d0 + rho(23)**2 + product(rho(22:23))**2 &
                   + product(rho(21:23))**2 + product(rho(20:23))**2 ))

        stdev_g_pers_prof = .744d0*stdev_avg_pers_prof
        stdev_b_pers_prof = 1.256d0*stdev_avg_pers_prof
        stdev_tran = sqrt(KO_tran)
        stdev_perm = sqrt(KO_perm)

        if (useNoCEV) then
            stdev_g_pers_prof = stdev_avg_pers_prof
            stdev_b_pers_prof = stdev_avg_pers_prof
        end if

    else

        ! STY
        stdev_perm = STY_perm
        stdev_tran = STY_tran
        stdev_avg_pers_prof = STY_stdev_avg_pers
        stdev_g_pers_prof = STY_stdev_g_pers
        stdev_b_pers_prof = STY_stdev_b_pers
        AR1_pers_prof = STY_AR1

    end if

end subroutine getShockInfo

! Initializes the earnings shocks in mod_data
subroutine sub_idshocks() !(gE,init_dsn_e_SS,init_dsn_e,invar_dsn_e_SS,trans_e_SS,trans_e)
    use mod_matlab
    use mod_param, only: nE, nZ, g_ind, b_ind, Tend, Tret, probLive, T1age, chi, nExpenditure
    use mod_myprint
    use mod_root1dim
    use mod_data
    implicit none
    ! local
    real(8), dimension(:,:,:), allocatable :: trans_e_b, trans_e_g 
    real(8) :: grid_pers(npers), grid_tran(ntran), grid_rich(nCCNRaugment)
    real(8) :: Fe(ntran)
    real(8), allocatable :: grid_expend(:), Fx(:)
    real(8), dimension(npers) :: FzzSS_0, Fzzg_0, Fzzb_0
    real(8), dimension(npers,npers,Tend) :: Fzz_by_age_SS,Fzz_by_age_g,Fzz_by_age_b
    real(8) :: m_rich(nCCNRaugment-1), Frich(nCCNRaugment), pi_11, pi_12, pi_21, pi_22 
    real(8), dimension(1:Tret) :: sharedProfNHS, sharedProfHS, sharedProfCOL, sharedProfAVG, workyears
    real(8), dimension(1:Tend) :: msrByAge
    real(8) :: popRetired, normalize
    integer :: zi, ei, ui, ii, iip, age, xi

    allocate(trans_e_b(nE,nE,Tend),trans_e_g(nE,nE,Tend))
    if (.not. allocated(state)) allocate(state(nE))

    ! Get the shock profiles (which may be constant)
!     call getShockInfo(stdev_perm,stdev_tran,AR1_pers_prof,stdev_avg_pers_prof,stdev_g_pers_prof,stdev_b_pers_prof)
    call getShockInfo()
    call myprint('stdev_perm',stdev_perm)
    call myprint('stdev_tran',stdev_tran)
    call myprint('AR1_pers_prof',AR1_pers_prof) ! do not delete, used by stochsimul
    call myprint('stdev_avg_pers_prof',stdev_avg_pers_prof)
    call myprint('stdev_g_pers_prof',stdev_g_pers_prof)
    call myprint('stdev_b_pers_prof',stdev_b_pers_prof)

    if (nE /= nExpenditure*(npers*ntran+nCCNRaugment)) STOP 'sub_idshocks: wrong nE!'

    if (nCCNRaugment/=0) then
        if (phiCCNRaugment == -100.d0) STOP 'sub_idshocks: not initialized'
        if (eloCCNRaugment == -100.d0) STOP 'sub_idshocks: not initialized'
        if (ehiCCNRaugment == -100.d0) STOP 'sub_idshocks: not initialized'
        if (pi11augment == -100.d0) STOP 'sub_idshocks: not initialized'
        if (pi21augment == -100.d0) STOP 'sub_idshocks: not initialized'
        if (pi1augment_init == -100.d0) STOP 'sub_idshocks: not initialized'
    else
        pi1augment_init = 0d0
        pi11augment = 0d0
        pi21augment = 0d0
    end if

    ! Setup state grid
    ii = 1
    do xi = 1,nExpenditure
        do ui = 1,nCCNRaugment
            state(ii)%isLogProc = .false.
            state(ii)%isSuperRich = .true.
            state(ii)%zi = -1
            state(ii)%ei = -1
            state(ii)%ui = ui
            state(ii)%xi = xi
            ii = ii+1
        end do
        do zi = 1,npers
            do ei = 1,ntran
                state(ii)%isLogProc = .true.
                state(ii)%isSuperRich = .false.
                state(ii)%zi = zi
                state(ii)%ei = ei
                state(ii)%ui = -1
                state(ii)%xi = xi
                ii = ii+1
            end do
        end do
    end do

    ! Setup the grids for the individual states
    grid_pers = merge(mean(stdev_avg_pers_prof)/sqrt(1.d0-mean(AR1_pers_prof)**2)*linspace(-stdevCoverPers,stdevCoverPers,npers),&
                        0.d0,npers>1)
    grid_tran = merge(stdev_tran*linspace(-stdevCoverTran,stdevCoverTran,ntran),&
                        0.d0,ntran>1)
    allocate(grid_expend(nExpenditure),Fx(nExpenditure))
    if (nExpenditure==1) then
        grid_expend = 0d0
        Fx = 1d0
    elseif (nExpenditure==3) then
        grid_expend(:) = [0d0,.762d0,2.373d0] ! See "CalibrationTargets.xlsx" for these values
        Fx(2:3) = [.0242640d0,.0015357d0] 
        Fx(1) = 1d0-sum(Fx(2:3))
    else
        STOP 'ERROR: don''t have this nExpenditure case setup'
    end if

    ! Compute the probabilities over initial states for newborns
    FzzSS_0 = tauchenIid(grid_pers,stdev_avg_pers_prof(1),0.d0) ! Iid draw 
    Fzzg_0 = tauchenIid(grid_pers,stdev_g_pers_prof(1),0.d0) 
    Fzzb_0 = tauchenIid(grid_pers,stdev_b_pers_prof(1),0.d0)
    Fe = tauchenIid(grid_tran,stdev_tran,0.d0)
    
    ! Setup the super rich grid
    if (nCCNRaugment>1) then
        ! log-spaced points
        grid_rich = logspace(log10(eloCCNRaugment),log10(ehiCCNRaugment),nCCNRaugment)                    
        m_rich = (grid_rich(1:nCCNRaugment-1)+grid_rich(2:nCCNRaugment))/2.d0
        Frich(1) = ((m_rich(1)-eloCCNRaugment)/(ehiCCNRaugment-eloCCNRaugment))**phiCCNRaugment
        Frich(2:nCCNRaugment-1) = ((m_rich(2:nCCNRaugment-1)-eloCCNRaugment)/(ehiCCNRaugment-eloCCNRaugment))**phiCCNRaugment - &
                    ((m_rich(1:nCCNRaugment-2)-eloCCNRaugment)/(ehiCCNRaugment-eloCCNRaugment))**phiCCNRaugment
        Frich(nCCNRaugment) = 1.d0-((m_rich(nCCNRaugment-1)-eloCCNRaugment)/(ehiCCNRaugment-eloCCNRaugment))**phiCCNRaugment            
    else
        STOP 'not prepared for nCCNRaugment==1'
    end if

    ! Include the levels
    state(:)%expend = grid_expend(state(:)%xi)
    where (state(:)%zi>0) 
        state(:)%persval = grid_pers(state(:)%zi)
    elsewhere
        state(:)%persval = -huge(0d0)
    endwhere
    where (state(:)%ei>0) 
        state(:)%tranval = grid_tran(state(:)%ei)
    elsewhere
        state(:)%tranval = -huge(0d0)
    endwhere
    where (state(:)%ui>0) 
        state(:)%richval = log(grid_rich(state(:)%ui)) ! note: logged
    elsewhere
        state(:)%richval = -huge(0d0)
    endwhere

    ! Compute the initial distribution over states
    do ii = 1,nE
        if (state(ii)%isSuperRich) then            

            ! Start with pi1 super rich
            init_dsn_e_SS(ii) = pi1augment_init * Fx(state(ii)%xi)*Frich(state(ii)%ui) ! The first term in brackets is "pi1"
            init_dsn_e(ii,:) = pi1augment_init * Fx(state(ii)%xi)*Frich(state(ii)%ui)
            
        elseif (state(ii)%isLogProc) then

            ! Start with 1d0-pi1 non-super rich
            init_dsn_e_SS(ii) = (1d0 - pi1augment_init) * Fx(state(ii)%xi)*Fe(state(ii)%ei)*FzzSS_0(state(ii)%zi)
            init_dsn_e(ii,g_ind) = (1d0 - pi1augment_init) * Fx(state(ii)%xi)*Fe(state(ii)%ei)*Fzzg_0(state(ii)%zi)
            init_dsn_e(ii,b_ind) = (1d0 - pi1augment_init) * Fx(state(ii)%xi)*Fe(state(ii)%ei)*Fzzb_0(state(ii)%zi)
        end if
    end do

    ! Compute the transitions for the persistent shock process
    do age = 1,Tend-1
        ! Fzz is conditional on your age today, where will you transit to tomorrow
        Fzz_by_age_SS(:,:,age) = tauchenTrans(grid_pers,AR1_pers_prof(age),stdev_avg_pers_prof(age+1),0.d0)
        Fzz_by_age_g(:,:,age) = tauchenTrans(grid_pers,AR1_pers_prof(age),stdev_g_pers_prof(age+1),0.d0)
        Fzz_by_age_b(:,:,age) = tauchenTrans(grid_pers,AR1_pers_prof(age),stdev_b_pers_prof(age+1),0.d0)
    end do    
    
    ! Compute the transition probabilities across states
    trans_e_SS = 0.d0
    trans_e_g = 0.d0
    trans_e_b = 0.d0
    do age = 1,Tend
        if (age==Tend) then
            ! Special case: transition is immaterial, but should put something in
            trans_e_SS(:,:,age) = 0.d0
            trans_e_g(:,:,age) = 0.d0
            trans_e_b(:,:,age) = 0.d0
        elseif (age>=Tret+1) then !retired
!             ! Remain where you are
!             trans_e_SS(:,:,age) = eye(nE)
!             trans_e_g(:,:,age) = eye(nE)
!             trans_e_b(:,:,age) = eye(nE)

            ! This is the retirment process. Need to let the x shock by iid and all the efficiency process values be deterministic
            do ii = 1,nE
                do iip = 1,nE
                    if ((state(ii)%isLogProc .eqv. state(iip)%isLogProc) .and. state(ii)%zi==state(iip)%zi .and. state(ii)%ei==state(iip)%ei .and. state(ii)%ui == state(iip)%ui) then
                        trans_e_SS(ii,iip,age) = Fx(state(iip)%xi)
                        trans_e_g(ii,iip,age) = Fx(state(iip)%xi)
                        trans_e_b(ii,iip,age) = Fx(state(iip)%xi)
                    end if
                end do
            end do

        else
            ! Reach here if and only if not retired
!             if (age==Tret) then !working but will retire next period
!                 ! If will retire period after next, then transit out of super rich process for sure
!                 pi_11 = 0.d0
!                 pi_12 = 1.d0
!                 pi_21 = 0.d0
!                 pi_22 = 1.d0
!             else
                ! If work this and next period, then have usual transition probabilities
                pi_11 = pi11augment
                pi_12 = 1.d0-pi11augment
                pi_21 = pi21augment
                pi_22 = 1.d0-pi21augment                
!             end if
            ! With transitions into and out of the log process/superrich process set, compute the transitions
            do ii = 1,nE
                do iip = 1,nE
                    if (state(ii)%isLogProc .and. state(iip)%isLogProc) then
                        trans_e_SS(ii,iip,age) = pi_22*Fx(state(iip)%xi)*Fe(state(iip)%ei)*Fzz_by_age_SS(state(ii)%zi,state(iip)%zi,age)
                        trans_e_g(ii,iip,age) = pi_22*Fx(state(iip)%xi)*Fe(state(iip)%ei)*Fzz_by_age_g(state(ii)%zi,state(iip)%zi,age)
                        trans_e_b(ii,iip,age) = pi_22*Fx(state(iip)%xi)*Fe(state(iip)%ei)*Fzz_by_age_b(state(ii)%zi,state(iip)%zi,age)
                    elseif (.not. state(ii)%isLogProc .and. state(iip)%isLogProc) then
                        ! In this case, they take a draw from the initial distribution for newborns.
                        trans_e_SS(ii,iip,age) = pi_12*Fx(state(iip)%xi)*Fe(state(iip)%ei)*FzzSS_0(state(iip)%zi)
                        trans_e_g(ii,iip,age) =  pi_12*Fx(state(iip)%xi)*Fe(state(iip)%ei)*Fzzg_0(state(iip)%zi)
                        trans_e_b(ii,iip,age) =  pi_12*Fx(state(iip)%xi)*Fe(state(iip)%ei)*Fzzb_0(state(iip)%zi)
                    elseif (state(ii)%isLogProc .and. .not. state(iip)%isLogProc) then
                        trans_e_SS(ii,iip,age) = pi_21*Fx(state(iip)%xi)*Frich(state(iip)%ui)
                        trans_e_g(ii,iip,age) =  pi_21*Fx(state(iip)%xi)*Frich(state(iip)%ui)
                        trans_e_b(ii,iip,age) =  pi_21*Fx(state(iip)%xi)*Frich(state(iip)%ui)
                    elseif (.not. state(ii)%isLogProc .and. .not. state(iip)%isLogProc) then
                        trans_e_SS(ii,iip,age) = pi_11*Fx(state(iip)%xi)*Frich(state(iip)%ui)
                        trans_e_g(ii,iip,age) =  pi_11*Fx(state(iip)%xi)*Frich(state(iip)%ui)
                        trans_e_b(ii,iip,age) =  pi_11*Fx(state(iip)%xi)*Frich(state(iip)%ui)
                    end if 
                end do
            end do
        end if
    end do
    
    trans_e(:,:,:,g_ind,g_ind) = trans_e_g(:,:,:)
    trans_e(:,:,:,b_ind,g_ind) = trans_e_g(:,:,:)
    trans_e(:,:,:,g_ind,b_ind) = trans_e_b(:,:,:)
    trans_e(:,:,:,b_ind,b_ind) = trans_e_b(:,:,:)

    ! Compute the steady state invariant distribution
    !!! Compute the invariant distribution over non age states conditional on every household living 
    invar_dsn_e_SS = 0.d0
    invar_dsn_e_SS(:,1) = init_dsn_e_SS

    do age = 1,Tend-1
        invar_dsn_e_SS(:,age+1) = matmul(invar_dsn_e_SS(:,age),trans_e_SS(:,:,age))
    end do

    if (any(abs(sum(invar_dsn_e_SS,1)-1.d0)>1d-12)) STOP 'Conditional distribution doesnt sum to 1'
    
    !!! Now weight it according to the measure of households in each generation
    msrByAge = cumprod((/1.d0,probLive(1:Tend-1)/))
    msrByAge = msrByAge/sum(msrByAge)

    do age = 1,Tend
        invar_dsn_e_SS(:,age) = msrByAge(age)*invar_dsn_e_SS(:,age)
    end do    

    ! Set up the deterministic income profile 
    !!! These estimates are from Hubbard for 12-15 years of educ for 1986
    !!! (THESE ARE IN LEVELS for ages 65 and under)
    do ii = 1,Tret
        workyears(ii) = T1age - 1 + real(ii,8) !Work from age T1age on
    end do

    if (.not. useFlatEarn) then
        sharedProfNHS = 10993.d0  -196.d0*workyears + 2167.d0*workyears**2/100.d0  -2655.d0*workyears**3/10000.d0 + 75.d0
        sharedProfHS = -11833.d0 + 1421.d0*workyears  -137.d0*workyears**2/100.d0  -2186.d0*workyears**3/10000.d0 + 620.d0
        sharedProfCOL = 72270.d0  -5579.d0*workyears + 19200.d0*workyears**2/100.d0  -18076.d0*workyears**3/10000.d0 + 2100.d0
        sharedProfAVG = .13d0*sharedProfNHS + .48d0*sharedProfHS + .39d0*sharedProfCOL  !educ attainment 2010 for ages >=30 years
        sharedProfAVG = sharedProfAVG/sharedProfAVG(1)
    else 
        sharedProfAVG = 1d0 ! flat
    end if
    call myprint('sharedProfAVG',sharedProfAVG)

    ! Setup earnings for everyone
    do age = 1,Tret
        do ii = 1,nE
            ! Working age households
            if (state(ii)%isLogProc) then
                gE(ii,age) = sharedProfAVG(age)*exp(grid_pers(state(ii)%zi) + grid_tran(state(ii)%ei))
            else
                gE(ii,age) = sharedProfAVG(age)*grid_rich(state(ii)%ui)
                ! No life cycle profile
                ! gE(ii,age) = grid_rich(state(ii)%ui) 
            end if    
        end do
    end do
    do age = Tret+1,Tend
        do ii = 1,nE
            ! In retirement assume they get .25* income in last period of life
            if (state(ii)%isLogProc) then
                gE(ii,age) = fracLastPeriod*sharedProfAVG(Tret)*exp(grid_pers(state(ii)%zi))
            else
                ! Give them the average efficiency in retirement
                gE(ii,age) = fracLastPeriod*sharedProfAVG(Tret)*&
                                (ehiCCNRaugment*phiCCNRaugment + eloCCNRaugment)/(1.d0+phiCCNRaugment) 
            end if
        end do
    end do
           
    !!! Add in the guaranteed fraction (this is a little algebra)
    popRetired = sum(invar_dsn_e_SS(:,Tret+1:Tend))
    normalize = sum(gE*invar_dsn_e_SS)/(1d0-fracGuaranteed*popRetired)
    gE(:,Tret+1:Tend) = gE(:,Tret+1:Tend) + fracGuaranteed*normalize

    gE = gE/sum(gE*invar_dsn_e_SS) ! Normalize so average N is 1

    ! If truncating efficiency, do that here
    if (threshold>0.d0) then
        print*,'**TRUNCATING EFFICIENCY DISTRIBUTION**'
        gE = max(gE,threshold)
        gE = gE/sum(gE*invar_dsn_e_SS)
    end if  

    ! Print out some additional variables
    call myprint('grid_pers',grid_pers)
    call myprint('grid_tran',grid_tran)
    call myprint('state_superrich',merge(1,0,state%isSuperRich))
    call myprint('state_pers',state%zi)
    call myprint('state_tran',state%ei)

    ! Check for errors
    if (abs(sum(init_dsn_e_SS)-1d0)>1d-12) STOP 'sub_idshocks: init_dsn_e_SS doesnt sum to 1'
    if (abs(sum(init_dsn_e(:,g_ind))-1d0)>1d-12) STOP 'sub_idshocks: init_dsn_e_g doesnt sum to 1'
    if (abs(sum(init_dsn_e(:,b_ind))-1d0)>1d-12) STOP 'sub_idshocks: init_dsn_e_b doesnt sum to 1'

    do age = 1,Tend-1
        if (any(abs(sum(trans_e_g(:,:,age),2)-1d0)>1d-12)) STOP 'sub_idshocks: error in constructing trans matrix trans_e_g'
        if (any(abs(sum(trans_e_b(:,:,age),2)-1d0)>1d-12)) then
            print*,'maxval(abs(sum(trans_e_b(:,:,age),2)-1.d0))',maxval(abs(sum(trans_e_b(:,:,age),2)-1.d0))
            STOP 'sub_idshocks: error in constructing trans matrix trans_e_b'
        end if
        if (any(abs(sum(trans_e_SS(:,:,age),2)-1.d0)>1d-12)) then
            print*,'sub_idshocks: error in constructing trans matrix trans_e_SS'
            print*,'maxval(abs(sum(trans_e_SS,2)))',maxval(abs(sum(trans_e_SS(:,:,age),2)))
            STOP 'sub_idshocks: error in constructing trans matrix trans_e_SS'
        end if 
    end do

    if (any(invar_dsn_e_SS<-1d-12)) then
        STOP 'sub_idshocks: negative values in invariant distribution'
    end if

    ! Check that N=1
    if (abs(sum(invar_dsn_e_SS*gE)-1.d0)>1d-9) then
        print*,'sub_idshocks: Normalization of N=1 failed'
        print*,'N-1 is',sum(invar_dsn_e_SS*gE)-1.d0
        STOP 
    end if

    ! Check that chi is not too high
    if (any(chi(gE,1d0)>.9d0)) then
        print*,'sub_idshocks: WARNING: chi is HIGH: higher than .9d0'
    end if

    
end subroutine sub_idshocks

subroutine sub_jointshocks() !(trans_ez, trans_e, trans_z)
    use mod_param, only: nE,nZ,Tend,trans_z
    use mod_data
    implicit none 
!     real(8), dimension(nZ,nZ), intent(in) :: trans_z !Pr(z -> z')
!     real(8), dimension(nE,nE,Tend,nZ,nZ), intent(in) :: trans_e !Pr(e,z,z' -> e' | Tend)
!     real(8), dimension(nE,nE,nZ,nZ,Tend), intent(out) :: trans_ez 
    ! local
    integer :: iE,iZ,iEp,iZp,age
    
    !Write joint transition matrix for (e,z)
    do age = 1,Tend
        do iE = 1,nE
            do iZ = 1,nZ
                do iEp = 1,nE
                    do iZp = 1,nZ
                        trans_ez(iE,iEp,iZ,iZp,age) = trans_e(iE,iEp,age,iZ,iZp)*trans_z(iZ,iZp)
                    end do
                end do
            end do
        end do 
    end do

end subroutine 

! Sets up the remaining aggregate shock information. Checks that the values
! in mod_param are valid.
subroutine sub_aggshocks()
    use mod_param, only: nZ, g_ind, b_ind, g, b, trans_gg, trans_bb, grid_z, trans_z
    use mod_matlab, only: eye
    implicit none

    if (grid_z(g_ind) /= g) STOP 'sub_aggshocks: wrong shocks! 1'
    if (grid_z(b_ind) /= b) STOP 'sub_aggshocks: wrong shocks! 2'
    if (trans_z(g_ind,g_ind) /= trans_gg) STOP 'sub_aggshocks: wrong shocks! 3'
    if (trans_z(g_ind,b_ind) /= 1.d0-trans_gg) STOP 'sub_aggshocks: wrong shocks! 4'
    if (trans_z(b_ind,b_ind) /= trans_bb) STOP 'sub_aggshocks: wrong shocks! 5'
    if (trans_z(b_ind,g_ind) /= 1.d0-trans_bb) STOP 'sub_aggshocks: wrong shocks!'

    if (maxval(abs(sum(trans_z,DIM=2)-1.d0))>1d-15) then 
        write(*,*) 'WARNING: Transition matrix trans_z rows do not sum to 1. Max diff is ', maxval(abs(sum(trans_z,DIM=2)-1.d0))
    end if

end subroutine sub_aggshocks

! Sets all the shocks in the model
subroutine sub_setshocks()
    implicit none

    call sub_idshocks()
    call sub_aggshocks()
    call sub_jointshocks()

end subroutine 

! Idiosyncratic shocks, but only those relevant to steady state calculations
! subroutine sub_idshocks_SS(gE, init_dsn_e_SS, invar_dsn_e_SS, trans_e_SS)
!     use mod_param, only: nE, nZ, Tend
!     use mod_myprint
!     implicit none
!     real(8), dimension(nE), intent(out) :: init_dsn_e_SS
!     real(8), dimension(nE,Tend), intent(out) :: gE, invar_dsn_e_SS
!     real(8), dimension(nE,nE,Tend), intent(out) :: trans_e_SS !Pr(e -> e')
!     !local
!     real(8), allocatable :: trans_e(:,:,:,:,:), init_dsn_e(:,:)
! 
!     allocate(trans_e(nE,nE,Tend,nZ,nZ),init_dsn_e(nE,nZ))
!    
!     call sub_idshocks(gE,init_dsn_e_SS,init_dsn_e,invar_dsn_e_SS,trans_e_SS,trans_e)
!     
! end subroutine sub_idshocks_SS

! Computes the median efficiency level for the steady state shocks
function fn_medianE_SS() result(medianE)
    use mod_param, only: nE, nZ, Tend
    use mod_myprint
    use mod_matlab, only: interp1, sort2, cumsum
    use mod_data
    implicit none
    real(8) :: medianE
    !local
!     real(8), allocatable :: init_dsn_e_SS(:), gE(:,:), invar_dsn_e_SS(:,:), trans_e_SS(:,:,:) 
!     real(8), allocatable :: trans_e(:,:,:,:,:), init_dsn_e(:,:)
    real(8), allocatable :: tmpE(:),tmpMsr(:)
    integer, allocatable :: tmpInds(:)

!     allocate(init_dsn_e_SS(nE),gE(nE,Tend),invar_dsn_e_SS(nE,Tend),trans_e_SS(nE,nE,Tend),&
!              trans_e(nE,nE,Tend,nZ,nZ),init_dsn_e(nE,nZ))
!    
!     call sub_idshocks(gE,init_dsn_e_SS,init_dsn_e,invar_dsn_e_SS,trans_e_SS,trans_e)

    allocate(tmpE(nE*Tend),tmpInds(nE*Tend),tmpMsr(nE*Tend))
    call sort2(tmpE,tmpInds,vec(gE))
    tmpMsr = vec(invar_dsn_e_SS)
    tmpMsr = tmpMsr(tmpInds) ! Now sorted according to E
    tmpMsr = cumsum(tmpMsr) 
    medianE = interp1(tmpMsr,tmpE,.5d0) ! Now find the middle value 

contains
    function vec(X)
        implicit none
        real(8) :: X(1:,1:)
        real(8) :: vec(1:size(X))
        vec = reshape(X,(/size(X)/))
    end function
end function

! subroutine sub_idshocks(gE, init_dsn_e_SS, init_dsn_e_g, init_dsn_e_b, trans_e)
!     use mod_param, only: nE, nZ, g_ind, b_ind, Tend
!     use mod_myprint
!     implicit none
!     real(8), dimension(nE), intent(out) :: init_dsn_e_SS, init_dsn_e_g, init_dsn_e_b
!     real(8), dimension(nE,Tend), intent(out) :: gE
!     real(8), dimension(nE,nE,Tend,nZ,nZ), intent(out) :: trans_e !Pr(e,z,z' -> e')
!     !local
!     real(8), allocatable :: trans_e_SS(:,:,:), invar_dsn_e_SS(:,:)
! 
!     allocate(trans_e_SS(nE,nE,Tend),invar_dsn_e_SS(nE,Tend))
!     
!     call sub_idshocks(gE,init_dsn_e_SS,init_dsn_e_g,init_dsn_e_b,invar_dsn_e_SS,trans_e_SS,trans_e)
! 
! !     call myread('gE',gE)
! !     call myread('init_dsn_e_SS',init_dsn_e_SS)
! !     call myread('init_dsn_e_g',init_dsn_e_g)
! !     call myread('init_dsn_e_b',init_dsn_e_b)
! !     call myread('trans_e',trans_e)
! !     STOP 'sub_idshocks: need to add a 5 dim read write or get rid of trans_e'
! 
! end subroutine sub_idshocks

end module mod_shocks
