module mod_stochsimul
    implicit none

    type hhstate
        ! state variables
        real(8) :: a
        integer :: iE,age,iH
        ! policies and values
        real(8) :: c,v,we
    end type

    ! To speed up simulation, many random numbers are drawn at once
    integer, private :: random_pos = 1
    integer, parameter, private :: random_size = 10000
    real(8), private :: random_cache(random_size)

contains

function get_rand() result(val)
    implicit none
    real(8) :: val

    if (random_pos==1) then
        call random_number(random_cache)
    end if

    val = random_cache(random_pos)

    random_pos = random_pos + 1

    if (random_pos>random_size) then
        random_pos = 1
    end if

end function 


subroutine disphhstate(hh)
    implicit none
    type(hhstate), intent(in) :: hh(:)
    ! local
    integer :: iT

    write(*,'(A10,3A4,2A10)') 'a','iE','age','iH','c','v'
    do iT = 1,size(hh)
        write(*,'(f10.4,3i4,2f10.4)') hh(iT)%a,hh(iT)%iE,hh(iT)%age,hh(iT)%iH,hh(iT)%c,hh(iT)%v
    end do

end subroutine 

! Simulates a panel of nI households for nT periods. When a household dies, it is replaced by a newborn.
! Assumes TFP follows the process in mseq unless mseq is not present, in which case it assumes steady state.
subroutine sub_panel(mseq,w_SS)
    use mod_myprint
    use mod_matlab
    use mod_param !, only: nA,nE,Tend,nH,nT,Tret,mom,b_ind,g_ind
    use mod_data
    use mod_law
    use mod_shocks, only: state,AR1_pers_prof
    implicit none
    type(mom), intent(in), optional :: mseq(:)
    real(8), intent(in), optional :: w_SS
    ! local
    integer :: nI = 50000, nBurn = len_t0
    type(hhstate), allocatable :: panel(:,:)
    real(8), allocatable :: reald(:,:,:,:)
    real(8) :: insureagg,insureagg_20,insureagg_80,insurepers,insuretran,insurerich,insureexpend, weights(4)
    real(8), dimension(:), allocatable :: insureagg_age,insureagg_20_age,insureagg_80_age
    real(8), dimension(:), allocatable :: insurepers_age,insuretran_age,insurerich_age,insureexpend_age
    real(8), dimension(:,:,:,:,:), allocatable :: dWork
    real(8), dimension(:,:,:,:), allocatable :: vInt,cInt,apInt
    integer :: iI,iT,iZ,iZp,indices(4),age
    character(3) :: suffix
    real(8), allocatable :: cumtransFee(:,:,:,:,:)
    real(8), allocatable :: cumtransFeeSS(:,:,:)

    allocate(cumtransFee(nE,nE,Tend,nZ,nZ))
    allocate(cumtransFeeSS(nE,nE,Tend))
    do iZp = 1,nZ
        do iZ = 1,nZ
            do age = 1,Tend
                cumtransFee(:,:,age,iZ,iZp) = cumsum(transpose(trans_e(:,:,age,iZ,iZp)))
                cumtransFeeSS(:,:,age) = cumsum(transpose(trans_e_SS(:,:,age)))
            end do
        end do
    end do

    ! Initialize the seed to a default state (conditional on a compiler!)
    call random_seed()

    allocate(dWork(nA,nE,Tend,nH,nM))
    allocate(vInt(nA,nE,Tend,nH))
    allocate(cInt(nA,nE,Tend,nH))
    allocate(apInt(nA,nE,Tend,nH))
        
    allocate(insureagg_age(Tend))
    allocate(insureagg_20_age(Tend))
    allocate(insureagg_80_age(Tend))
    allocate(insurepers_age(Tend))
    allocate(insuretran_age(Tend))
    allocate(insurerich_age(Tend))
    allocate(insureexpend_age(Tend))

    allocate(panel(nI,nT))
    if (present(mseq)) then
        if (size(mseq)/=nT) then
            print*,'WARNING: could not simulate panel as izseq has the wrong length'
            return
        end if
    end if

    suffix = merge('   ','_SS',present(mseq))

    allocate(reald(nA,nE,Tend,nH))

    ! Initialize all households to being newborn for ease (there is a large burn-in)
    ! WOULD BE BETTER TO START FROM SS DSN RATHER THAN NEWBORNS
    do iI = 1,nI
        call set_newborn(panel(iI,1),init_dsn_e_SS) 
    end do

    if (.not. present(mseq)) then

        reald = merge(1d0,0d0,d_SS)

        ! Simulate panel
        do iI = 1,nI
            do iT = 1,nT-1
                ! call sub_update(panel(iI,iT),panel(iI,iT+1),ap_SS,reald,trans_e_SS,init_dsn_e_SS)
                call sub_update(panel(iI,iT),panel(iI,iT+1),ap_SS,reald,cumtransFeeSS,init_dsn_e_SS)
                call sub_fill(panel(iI,iT+1),c_SS,v_SS,w_SS)
            end do
        end do

    else
        
        dWork = merge(1d0,0d0,d) 

        ! Simulate panel
        do iT = 1,nT-1

            iZ = mseq(iT)%z_ind
            iZp = mseq(iT+1)%z_ind

            ! Get interpolation weights
            call get_m_interp(indices,weights,mseq(iT)) ! Get interpolation weights and indices for the moment grid

            ! Interpolate at current moments
            call dointerp(reald,dWork,indices,weights)
            call dointerp(cInt,c,indices,weights)
            call dointerp(vInt,v,indices,weights)
            if (iZp==g_ind) then
                call dointerp(apInt,apg,indices,weights) !Note: this a' is actually conditional on repaying, not the standard definition               
            else
                call dointerp(apInt,apb,indices,weights)
            end if
            
            do iI = 1,nI
                ! call sub_update(panel(iI,iT),panel(iI,iT+1),apInt,reald,trans_e(:,:,:,iZ,iZp),init_dsn_e(:,iZp))
                call sub_update(panel(iI,iT),panel(iI,iT+1),apInt,reald,cumtransFee(:,:,:,iZ,iZp),init_dsn_e(:,iZp))
                call sub_fill(panel(iI,iT+1),cInt,vInt,mseq(iT+1)%z * (1d0-alpha_k)*(mseq(iT+1)%KN)**alpha_k * eshift(iZp))
            end do

        end do

        ! Compute the insurance coefficient against aggregate shocks
        call sub_agginsurecoeff(insureagg,insureagg_age,mseq,spread(.true.,1,nE),panel,nBurn)
        call sub_agginsurecoeff(insureagg_20,insureagg_20_age,mseq,state%isSuperRich,panel,nBurn)
        call sub_agginsurecoeff(insureagg_80,insureagg_80_age,mseq,state%isLogProc,panel,nBurn)
        write(*,*) 'Insure coeff agg: all ',insureagg,' top20 ',insureagg_20,' bottom80 ',insureagg_80
!         call plot(insureagg_age(2:Tret),'agg profile')
!         call plot(insureagg_20_age(2:Tret),'agg profile top 20')
!         call plot(insureagg_80_age(2:Tret),'agg profile bottom 80')
        call myprint('insureagg'//trim(suffix),insureagg)
        call myprint('insureagg_20'//trim(suffix),insureagg_20)
        call myprint('insureagg_80'//trim(suffix),insureagg_80)
        call myprint('insureagg_age'//trim(suffix),insureagg_age)
        call myprint('insureagg_20_age'//trim(suffix),insureagg_20_age)
        call myprint('insureagg_80_age'//trim(suffix),insureagg_80_age)

    end if

    ! For each idiosyncratic shock, compute the insurance coefficient
    call sub_idinsurecoeff(insurepers,insurepers_age,state%persval,state%isLogProc,AR1_pers_prof,panel,nBurn,.false.)
    call sub_idinsurecoeff(insuretran,insuretran_age,state%tranval,state%isLogProc,0d0*AR1_pers_prof,panel,nBurn,.false.)
    call sub_idinsurecoeff(insurerich,insurerich_age,state%richval,state%isSuperRich,0d0*AR1_pers_prof,panel,nBurn,.false.)
    call sub_idinsurecoeff(insureexpend,insureexpend_age,state%expend,state%isLogProc,0d0*AR1_pers_prof,panel,nBurn,.true.) 

    write(*,*) 'Insurance coeffs: pers ',insurepers,' tran ',insuretran,' rich ',insurerich, 'expend ',insureexpend
    
!     call plot(insurepers_age(2:Tret),'pers profile')
!     call plot(insuretran_age(2:Tret),'tran profile')
!     call plot(insurerich_age(2:Tret),'rich profile')
!     call plot(insureexpend_age(2:Tret),'expend profile')

    call myprint('insurepers'//trim(suffix),insurepers)
    call myprint('insuretran'//trim(suffix),insuretran)
    call myprint('insurerich'//trim(suffix),insurerich)
    call myprint('insureexpend'//trim(suffix),insureexpend)
    call myprint('insurepers_age'//trim(suffix),insurepers_age)
    call myprint('insuretran_age'//trim(suffix),insuretran_age)
    call myprint('insurerich_age'//trim(suffix),insurerich_age)
    call myprint('insureexpend_age'//trim(suffix),insureexpend_age)

!     do iI = 1,nI
!         call disphhstate(panel(iI,:))
!     end do

end subroutine 

subroutine dointerp(xi,x,indices,weights)
    implicit none
    real(8), intent(in) :: x(:,:,:,:,:),weights(:)
    integer, intent(in) :: indices(:)
    real(8), intent(out) :: xi(:,:,:,:)

    xi = weights(1)*x(:,:,:,:,indices(1)) + &
         weights(2)*x(:,:,:,:,indices(2)) + &
         weights(3)*x(:,:,:,:,indices(3)) + &
         weights(4)*x(:,:,:,:,indices(4))

end subroutine 

subroutine sub_agginsurecoeff(insurecoeff,insurecoeff_age,mseq,useiE,panel,nBurn)
    use mod_matlab
    use mod_param, only: Tret,Tend,mom,grid_z
    implicit none
    type(mom), intent(in) :: mseq(:)
    logical, intent(in) :: useiE(:) 
    type(hhstate), intent(in), target :: panel(:,:)
    integer, intent(in) :: nBurn
    real(8), intent(out) :: insurecoeff,insurecoeff_age(:)
    ! local
    type(hhstate) :: hh,hhprior
    integer :: nI,nT,iI,iT,iX,ninsample
    logical :: insample
    real(8) :: Ec,Es,Ecs,Ess,COVcs,VARs,tmp,tmp2
    real(8), allocatable, dimension(:) :: Ec_age,Es_age,Ecs_age,Ess_age,COVcs_age,VARs_age
    integer, allocatable, dimension(:) :: ninsample_age

    nI = size(panel,1)
    nT = size(panel,2)

    allocate(ninsample_age(Tend))
    allocate(Ec_age(Tend))
    allocate(Es_age(Tend))
    allocate(Ecs_age(Tend))
    allocate(Ess_age(Tend))
    allocate(COVcs_age(Tend))
    allocate(VARs_age(Tend))
    
    ninsample = 0
    Ec = 0d0
    Es = 0d0
    Ecs = 0d0
    Ess = 0d0
    ninsample_age = 0
    Ec_age = 0d0
    Es_age = 0d0
    Ecs_age = 0d0
    Ess_age = 0d0
    do iT = nBurn+1,nT ! discard 1,nBurn
        do iI = 1,nI

            hh = panel(iI,iT)
            hhprior = panel(iI,iT-1)

            ! Check sample criterion 
            insample = hh%age>=2 .and. hh%age<=Tret ! no shocks in retirement, so degenerate variance
            insample = useiE(hh%iE) .and. insample 
            insample = useiE(hhprior%iE) .and. insample 

            ! If in sample, need to compute covariance between log c growth and shock value. 
            ! Use formula Cov(x,y) = E[x*y] - E[x]*E[y].
            ! Also need variance of shock, use Var(x) = E[x*x] - E[x]*E[x].
            if (insample) then
                ninsample = ninsample + 1
                tmp = log(hh%c/hhprior%c)
                ! tmp2 = shockvals(hh%iE) - pers_prof(hh%age-1)*shockvals(hhprior%iE)
                ! tmp2 = log(grid_z(mseq(iT)%z_ind)) - log(grid_z(mseq(iT-1)%z_ind)) !
                ! tmp2 = log(grid_z(mseq(iT)%z_ind)) - 1d0/3d0*log(grid_z(mseq(iT-1)%z_ind)) ! rouwenhorst discretization implies the Markov chain with 2/3 persistence is 1/3. So, can treat it like rho=1/3.
                tmp2 = log(grid_z(mseq(iT)%z_ind)) ! saying the z shock is iid
                Ec = Ec + tmp
                Es = Es + tmp2
                Ecs = Ecs + tmp*tmp2
                Ess = Ess + tmp2**2
                ! Also do by age
                ninsample_age(hh%age) = ninsample_age(hh%age) + 1
                Ec_age(hh%age) = Ec_age(hh%age) + tmp
                Es_age(hh%age) = Es_age(hh%age) + tmp2
                Ecs_age(hh%age) = Ecs_age(hh%age) + tmp*tmp2
                Ess_age(hh%age) = Ess_age(hh%age) + tmp2**2
            end if
        end do
    end do

    Ec = Ec/ninsample
    Es = Es/ninsample
    Ecs = Ecs/ninsample
    Ess = Ess/ninsample
    COVcs = Ecs - Ec*Es
    VARs = Ess - Es**2
    insurecoeff = 1d0 - COVcs/VARs

    where (ninsample_age>0)
        Ec_age = Ec_age/ninsample_age
        Es_age = Es_age/ninsample_age
        Ecs_age = Ecs_age/ninsample_age
        Ess_age = Ess_age/ninsample_age
    elsewhere
        Ec_age = 0d0
        Es_age = 0d0
        Ecs_age = 0d0
        Ess_age = 0d0
    endwhere
    COVcs_age = Ecs_age - Ec_age*Es_age
    VARs_age = Ess_age - Es_age**2
    where (VARs_age>0d0) 
        insurecoeff_age = 1d0 - COVcs_age/VARs_age
    elsewhere
        insurecoeff_age = 0d0
    endwhere

    write(*,'(7(A,e20.6))') 'Ec',Ec,' Es',Es,' Ecs',Ecs,' Ess',Ess,' COVcs',COVcs,' VARs',VARs,' insurecoeff',insurecoeff 


end subroutine


subroutine sub_idinsurecoeff(insurecoeff,insurecoeff_age,shockvals,useiE,pers_prof,panel,nBurn,isExpend)
    use mod_matlab
    use mod_param, only: Tret,Tend
    implicit none
    real(8), intent(in) :: shockvals(:), pers_prof(:)
    integer, intent(in) :: nBurn
    logical, intent(in) :: useiE(:)
    type(hhstate), intent(in), target :: panel(:,:)
    real(8), intent(out) :: insurecoeff,insurecoeff_age(:)
    logical, intent(in) :: isExpend 
    ! local
    type(hhstate) :: hh,hhprior
    integer :: nI,nT,iI,iT,iX,ninsample
    logical :: insample
    real(8) :: Ec,Es,Ecs,Ess,COVcs,VARs,tmp,tmp2,Ecc
    real(8), allocatable, dimension(:) :: Ec_age,Es_age,Ecs_age,Ess_age,COVcs_age,VARs_age
    integer, allocatable, dimension(:) :: ninsample_age

    nI = size(panel,1)
    nT = size(panel,2)

    allocate(ninsample_age(Tend))
    allocate(Ec_age(Tend))
    allocate(Es_age(Tend))
    allocate(Ecs_age(Tend))
    allocate(Ess_age(Tend))
    allocate(COVcs_age(Tend))
    allocate(VARs_age(Tend))
    
    ninsample = 0
    Ec = 0d0
    Es = 0d0
    Ecs = 0d0
    Ess = 0d0
    Ecc = 0d0
    ninsample_age = 0
    Ec_age = 0d0
    Es_age = 0d0
    Ecs_age = 0d0
    Ess_age = 0d0
    do iT = nBurn+1,nT ! discard 1,nBurn
        do iI = 1,nI

            hh = panel(iI,iT)
            hhprior = panel(iI,iT-1)

            ! Check sample criterion 
            insample = hh%age>=2 .and. hh%age<=Tret ! no shocks in retirement, so degenerate variance
            insample = useiE(hh%iE) .and. insample ! satisfies sample criterion this period
            insample = useiE(hhprior%iE) .and. insample ! satisfies sample criterion last period

            ! If in sample, need to compute covariance between log c growth and shock value. 
            ! Use formula Cov(x,y) = E[x*y] - E[x]*E[y].
            ! Also need variance of shock, use Var(x) = E[x*x] - E[x]*E[x].
            if (insample) then
                ninsample = ninsample + 1
                tmp = log(hh%c/hhprior%c)
                tmp2 = shockvals(hh%iE) - pers_prof(hh%age-1)*shockvals(hhprior%iE)
                if (isExpend) then
                    tmp2 = -shockvals(hh%ie)/hh%we ! expenditure shock, use percent change in pre-shock earnings caused by x_t,  -x_t/(w_t-1 e_t-1)
                    ! tmp2 = 1.3513d0*(-shockvals(hh%ie)) - .02972d0 ! using z = a*(-x) + b, choose a,b so that Ez = 0 and z at the small shock is 1. This leads to a = 1.3513 and b = -.02972.
                end if
                Ec = Ec + tmp
                Es = Es + tmp2
                Ecs = Ecs + tmp*tmp2
                Ess = Ess + tmp2**2
                Ecc = Ecc + tmp**2
                ! Also do by age
                ninsample_age(hh%age) = ninsample_age(hh%age) + 1
                Ec_age(hh%age) = Ec_age(hh%age) + tmp
                Es_age(hh%age) = Es_age(hh%age) + tmp2
                Ecs_age(hh%age) = Ecs_age(hh%age) + tmp*tmp2
                Ess_age(hh%age) = Ess_age(hh%age) + tmp2**2
            end if
        end do
    end do

    Ec = Ec/ninsample
    Es = Es/ninsample
    Ecs = Ecs/ninsample
    Ess = Ess/ninsample
    Ecc = Ecc/ninsample
    COVcs = Ecs - Ec*Es
    VARs = Ess - Es**2
    insurecoeff = 1d0 - COVcs/VARs

    where (ninsample_age>0)
        Ec_age = Ec_age/ninsample_age
        Es_age = Es_age/ninsample_age
        Ecs_age = Ecs_age/ninsample_age
        Ess_age = Ess_age/ninsample_age
    elsewhere
        Ec_age = 0d0
        Es_age = 0d0
        Ecs_age = 0d0
        Ess_age = 0d0
    endwhere
    COVcs_age = Ecs_age - Ec_age*Es_age
    VARs_age = Ess_age - Es_age**2
    where (VARs_age>0d0) 
        insurecoeff_age = 1d0 - COVcs_age/VARs_age
    elsewhere
        insurecoeff_age = 0d0
    endwhere

    write(*,'(10(A,e20.6))') 'Ec',Ec,' Es',Es,' Ecs',Ecs,' Ess',Ess,' COVcs',COVcs,' VARs',VARs,' stdc',sqrt(Ecc-Ec**2),' stds',sqrt(VARs),' CORRcs',COVcs/sqrt(Ecc-Ec**2)/sqrt(VARs),' insurecoeff',insurecoeff 


end subroutine

! Fills out the hh state adding policies and values
subroutine sub_fill(hh,c,v,w)
    use mod_matlab
    use mod_data, only: gA,gE
    implicit none
    type(hhstate), intent(inout) :: hh
    real(8), dimension(:,:,:,:), intent(in) :: c,v
    real(8), intent(in) :: w
    ! local
    integer :: ilo
    real(8) :: mlo

    call interp1qlocweight(ilo,mlo,hh%a,gA)
    hh%c = mlo*c(ilo,hh%iE,hh%age,hh%iH) + (1d0-mlo)*c(ilo+1,hh%iE,hh%age,hh%iH)
    hh%v = mlo*v(ilo,hh%iE,hh%age,hh%iH) + (1d0-mlo)*v(ilo+1,hh%iE,hh%age,hh%iH)
    hh%we = w*gE(hh%iE,hh%age)

end subroutine 

subroutine set_newborn(hh,inite)
    use mod_param, only: nE
    implicit none
    type(hhstate), intent(out) :: hh
    real(8), intent(in) :: inite(:)
    ! local
    real(8) :: tmp,rand01
    integer :: iE

    ! Die and be born again
    hh%age = 1
    hh%a = 0d0
    ! call random_number(rand01)
    rand01 = get_rand()
    tmp = 0d0
    do iE = 1,nE-1
        tmp = tmp + inite(iE)
        if (rand01<tmp) exit
    end do
    hh%iE = iE
    hh%iH = 1

end subroutine 

subroutine sub_update(hh,hhp,ap,dreal,cumtransFee,inite)
    use mod_matlab
    use mod_param
    use mod_data, only: gA
    use mod_discreteAR1, only: drawAR1core
    implicit none
    type(hhstate), intent(in) :: hh
    type(hhstate), intent(out) :: hhp
    real(8), intent(in) :: ap(:,:,:,:),cumtransFee(:,:,:),inite(:),dreal(:,:,:,:)
    ! local
    integer :: iEp
    real(8) :: rand01,tmp

    ! Determine if die or not
    ! call random_number(rand01)
    rand01 = get_rand()
    if (rand01<probLive(hh%age)) then
        ! Live
        hhp%age = hh%age + 1

        ! Get e'
        ! call random_number(rand01)
        ! tmp = 0d0
        ! do iEp = 1,nE-1
        !     tmp = tmp + Fee(hh%iE,iEp,hh%age)
        !     if (rand01<tmp) exit ! if never exits, then h is nE on exit
        ! end do
        ! hhp%iE = iEp
        rand01 = get_rand()
        hhp%iE = drawAR1core(cumtransFee(:,:,hh%age),hh%iE,rand01)

        ! Get a' if not defaulting
        hhp%a = interp1q(gA,ap(:,hh%iE,hh%age,hh%iH),hh%a)

        ! Get h'
        !!! Determine if default
        ! call random_number(rand01)
        rand01 = get_rand()
        if (rand01<interp1q(gA,dreal(:,hh%iE,hh%age,hh%iH),hh%a)) then
            ! default
            hhp%a = 0d0
            hhp%iH = 2
        else
            ! don't default
            if (hh%iH==1) then
                !!! if in good standing, remain
                hhp%iH = 1
            else 
                !!! if in bad standing, probabilistically leave
                ! call random_number(rand01)
                rand01 = get_rand()
                if (rand01<lambda) then
                    ! Remain in autarky
                    hhp%iH=2
                else
                    ! Leave
                    hhp%iH=1
                end if
            end if
        end if
    else

        ! Die and be born again
        call set_newborn(hhp,inite)

    end if

end subroutine 

end module
