! This is a module for computing the transition between steady states.
module mod_transition
    use mod_stats, only: stat_type
    implicit none

    integer, parameter :: nT = 50

    type trans_type
        real(8), dimension(:,:,:,:), allocatable :: v,ap,c,q,dsn
        logical, dimension(:,:,:,:), allocatable :: d
        real(8) :: K,Kimply,N,w,r,qbar
        real(8), dimension(:), allocatable :: statlist
        type (stat_type) :: s
    end type

    type(trans_type), dimension(nT) :: t ! transition

contains
    
    ! Allocates the data for the transition
    subroutine allocate_transition()
        use mod_param, nT_mod_param => nT
        implicit none
        integer :: i

        ! Compute the required memory requirement
        print*,'Memory requirement for transition (GB): ',5*real(nT,8)*real(nA*nE*Tend*nH,8)*8/real(1028**3,8)

        print*,'nT',nT

        ! Allocate the space
        do i = 1,nT

            ! Allocate
            allocate(t(i)%v(nA,nE,Tend,nH))
            allocate(t(i)%ap(nA,nE,Tend,nH))
            allocate(t(i)%c(nA,nE,Tend,nH))
            allocate(t(i)%d(nA,nE,Tend,nH))
            allocate(t(i)%q(nA,nE,Tend,nH))
            allocate(t(i)%dsn(nA,nE,Tend,nH))
            allocate(t(i)%statlist(100))

            ! Zero out so that it immediately shows the data as reserved
            t(i)%v = 0d0
            t(i)%ap = 0d0
            t(i)%c = 0d0
            t(i)%q = 0d0
            t(i)%dsn = 0d0
            t(i)%d = .false.
            t(i)%statlist = 0d0

        end do

    end subroutine

    ! Copy's the steady data into the specified time period. Either t=1 or t=nT
    ! If opt = 1, set as initial state 
    ! If opt = 2, set as terminal state
    subroutine copyin_SS_to_transition(opt)
        use mod_data
        implicit none
        integer, intent(in) :: opt
        integer :: i

        if (opt>2) STOP 'ERROR: copyin_SS_values: opt must be 1 or 2'

        i = merge(1,nT,opt==1)
        
        t(i)%v = v_SS
        t(i)%ap = ap_SS
        t(i)%c = c_SS
        t(i)%d = d_SS
        t(i)%q = q_SS
        t(i)%dsn = invar_dsn_SS

    end subroutine
    
    subroutine sub_transition(caseold,casenew,switchdate)
        use mod_matlab
        use mod_param, nT_mod_param => nT
        use mod_SS
        use mod_data, only: gE,gA,trans_e_SS,init_dsn_e_SS,invar_dsn_e_SS
        use mod_nonstochsimul, only: sub_dsn_aux
        use mod_stats
        implicit none
        integer, intent(in) :: caseold,casenew,switchdate
        real(8), allocatable :: tmparray(:,:,:,:),dreal(:,:,:,:),work(:,:)
        real(8) :: Kold,Knew,sup !,Kcheck1,Kcheck2
        integer :: i,age,iter
        real(8), parameter :: epsilon_K = 1d-4, delta_K = 1d-3

        allocate(tmparray(nA,nE,Tend,nH))
        allocate(dreal(nA,nE,Tend,nH))
        allocate(work(nA,nE))

        ! Rather than redefining variables It borrows the data located in mod_SS_data. 
        ! Select variables are gA,gE,trans_e,...  see the use mod_SS_data statement above
        ! They should be allocated upon entry to this routine with grids and shocks already set

        ! computeKp always works (regardless of garnishment or not) in steady state
        Kold = computeKp(t(1)%dsn,t(1)%q,t(1)%ap)
        Knew = computeKp(t(nT)%dsn,t(nT)%q,t(nT)%ap)

        ! The exogenous labor supply is constant, so set it and forget it
        t%N = sum(invar_dsn_e_SS*gE)

        t(1)%K = Kold
        t(1)%w = (1d0-alpha_k)*(t(1)%K/t(1)%N)**alpha_k

        ! Guess on 2:nT values of K
        t(1:nT)%K = linspace(Kold,Knew,nT) 

        do iter = 1,200

            ! Compute factor prices
            t%w = (1d0-alpha_k)*(t%K/t%N)**alpha_k
            t%r = alpha_k*(t%K/t%N)**(alpha_k-1d0)
            t%qbar = 1d0/(1d0+t%r-delta)

            ! Using backward induction, compute the price schedule, value functions, and policy functions
            call set_case(casenew,.false.)
            write(*,'(A)',advance='no') 'VFI '
            do i = nT-1,1,-1
                
                write(*,'(A)',advance='no') '.'

                ! for i=1,switchdate, the old law is in effect
                ! if switchdate=0, the old law is never in effect (which is what would usually be assumed)
                if (i==switchdate) call set_case(caseold,.false.)

                ! NOTE: LEAVE age IN ASCENDING ORDER!!!
                do age = 1,Tend

                    ! The trick here is subValue only writes to "age". It reads in v(age+1) and d(age+1)
                    ! to determine what the continuation utility and price schedule should be. So, we need
                    ! to temporarily modify v(age+1) and d(age+1) to reflect that they are also associated
                    ! with next period prices. The most computationally efficient way is to work in ascending
                    ! order from age 1 to Tend. For each age, replace v(age+1) and d(age+1) with the 
                    ! v(age+1) and d(age+1) associated with next period. Then, since we're solving for age
                    ! in ascending order, the age+1 values for this period correctly overwrite the ones
                    ! associated with next period.

                    if (age<Tend) then
                        t(i)%v(:,:,age+1,:) = t(i+1)%v(:,:,age+1,:)
                        t(i)%d(:,:,age+1,:) = t(i+1)%d(:,:,age+1,:)
                    end if
                    call subValue(age,t(i)%v,t(i)%c,t(i)%ap,t(i)%d,t(i)%q,t(i)%qbar,t(i)%w,gA,gE,trans_e_SS) ! note: c is a temp

                end do

!                 if (mod(i,5)==1) call plot(gA,t(i)%v(:,:,1,1))

            end do
            write(*,'(A)') ' done'

            ! Using forward induction, compute the implied sequence of aggregate capital.
            ! Note: this part is completely independent of the legal environment.
            write(*,'(A)',advance='no') 'DSN '
            do i = 2,nT

                write(*,'(A)',advance='no') '.'

                tmparray = 0d0
                do age = 1,Tend

                    if (age>1) then
                        tmparray(:,:,age-1,:) = t(i-1)%dsn(:,:,age-1,:)
                    end if

                    ! need to convert d (logical) to dreal (real(8))
                    dreal = merge(1d0,0d0,t(i-1)%d)

                    ! sub_dsn_aux computes the dsn(:,:,age,:) using dsn(:,:,age-1,:) if age>1
                    ! call sub_dsn_aux(age,tmparray,dreal,t(i-1)%ap,trans_e_SS,init_dsn_e_SS)
                    call sub_dsn_aux(tmparray,tmparray,age,dreal,t(i-1)%ap,trans_e_SS,init_dsn_e_SS,work)

                    ! Copy from the temporary array out
                    t(i)%dsn(:,:,age,:) = tmparray(:,:,age,:)

                end do

                ! Check that the distribution sums to one
                if (abs(sum(t(i)%dsn)-1d0)>1d-10) STOP 'ERROR: sub_transition: sum(t(i)%dsn doesn''t sum to 1'

            end do
            write(*,'(A)') ' done'

            ! Compute the sequence of capital stocks associated with the aggregate capital
            !!! NOTE: both of the measures below should only be used to compute K(t) for t>1
            !!! The first measure cannot be used because computeKp computes the capital stock 
            !!! for the next period.
            !!! The second measure cannot be used because computeK is set to use the method 
            !!! associated with the new policy---it does not apply to the first period. 
            do i = 1,nT-1
                t(i+1)%Kimply = computeKp(t(i)%dsn,t(i)%q,t(i)%ap)
            end do
!             do i = 2,nT
!                 t(i)%Kimply2 = computeK(t(i)%dsn,t(i)%d,gA,gE,t(i)%w,nA,nAmm,nE,Tend,nH)
!             end do

            do i = 2,nT
                write(*,'(A,i5,2(f16.6))') 't,K,Kimply',i,t(i)%K,t(i)%Kimply ! ,t(i)%Kimply2
            end do 
            write(*,'(A,(f16.6))') 'New SS K:',Knew

            ! Check the supnorm 
            sup = maxval(abs(t(2:nT)%K - t(2:nT)%Kimply))

            print*,'iter',iter,'sup',sup

            if (sup<epsilon_K) then
                print*,'sub_transition: converge successfully',sup,epsilon_K
                exit
            else
                t(2:nT)%K = .8d0*t(2:nT)%K + .2d0*t(2:nT)%Kimply 
            end if

            ! Collect statistics from the transition
            call set_case(caseold,.false.)
            write(*,'(A)',advance='no') 'STATS '
            do i = 1,nT
                write(*,'(A)',advance='no') '.'

                ! for i=1,switchdate, the old law is in effect
                ! if switchdate=0, the old law is never in effect (which is what would usually be assumed)
                if (i==switchdate+1) call set_case(casenew,.false.)

                call computeStats(t(i)%statlist,t(i)%s,2,t(i)%dsn,1d0,1d0,t(i)%v,t(i)%c,t(i)%ap,t(i)%d,t(i)%q,gE)
            end do
            write(*,'(A)') ' done'

            ! Save time series of select stats
!             call plot(t%s%pop_def,'pop def')
!             call plot(t%s%pop_in_debt,'pop in debt')
!             call plot(t%s%total_debt,'total debt')
!             call plot(t%s%K,'K')
!             call plot(t%s%avgIntPaid,'avgIntPaid')
!             call plot(t%s%frac_def_below_median,'frac def below median')

        end do

        ! Check whether the transition period is long enough
        if (sup<epsilon_K) then
            sup = abs(Knew - t(nT)%Kimply)
            if (sup<delta_K) then
                print*,'sub_transition: the transition length is sufficient:',sup,delta_K
            else
                print*,'sub_transition: the transition period is not long enough:',sup,delta_K
            end if
        else
            print*,'sub_transition: failed to converge',sup,epsilon_K
        end if

        ! Collect statistics from the transition
        call set_case(caseold,.false.)
        do i = 1,nT
            ! for i=1:switchdate, the old law is in effect
            ! if switchdate=0, the old law is never in effect (which is what would usually be assumed)
            if (i==switchdate+1) call set_case(casenew,.false.)

            call computeStats(t(i)%statlist,t(i)%s,2,t(i)%dsn,1d0,1d0,t(i)%v,t(i)%c,t(i)%ap,t(i)%d,t(i)%q,gE)
        end do

        ! Save time series of select stats
!         call plot(t%s%pop_def,'pop def')
!         call plot(t%s%pop_in_debt,'pop in debt')
!         call plot(t%s%total_debt,'total debt')
!         call plot(t%s%avgIntPaid,'avgIntPaid')
!         call plot(t%s%frac_def_below_median,'frac def below median')
!         call plot(t%s%K,'K')

        call save_statlist()
    
    end subroutine

    subroutine save_statlist()
        use mod_myprint
        implicit none
        real(8), allocatable :: transition_statlist(:,:)
        ! local
        integer :: iT

        allocate(transition_statlist(nT,100))
        do iT = 1,nT
            transition_statlist(iT,:) = t(iT)%statlist
        end do
        call myprint('transition_statlist',transition_statlist)

    end subroutine save_statlist

end module
