module mod_value_fcns
    implicit none

contains

    ! subVFI
    ! If maxiter = 1, computes one update of the value function
    ! If maxiter > 1, computes up to maxiter updates of the value function stopping if convergence is achieved
    ! Calculates an update for the value function
    subroutine subVFI(v,c,apg,apb,d,q,qbar,w,gA,gE,trans_e,trans_ez,trans_m_cdnal_zp,z_indices)
!         use mod_value_fcns_priv
        use mod_param, only: discount, death, lambda, chi, chiBnk, defBnkAllowed, &
                                  g_ind, b_ind, isResrictedParticipation, Tret, &
                                  eshift, isRestrictedPortfolio, probLive, csnEquiv, trans_z, defZcontingent, defZcontingent_AllowedZind, &
                                  nA,nAmm,nE,Tend,nH,nZ,nM,nExpenditure
        use mod_shocks, only: state
        use mod_compute_vR, only: compute_vR_BC
        ! use mod_debug
        use mod_matlab, only: mtimes, colon, plot, interp1, tic, toc, disp, mydgemm, diff
        use mod_price_fcns
        implicit none

        ! Updating value function
        !
        ! in: guess on v(a,e,h)
        ! out: update on v(a,e,h) with policy functions a'(a,e,h), and possibly stochastic transitions h'(a,e,h), and e'(e)
        !

        real(8), dimension(nA,nE,Tend,nH,nM), intent(out) :: v,c,apg,apb
        logical, dimension(nA,nE,Tend,nH,nM), intent(out) :: d

        integer, dimension(nM), intent(in) :: z_indices
        real(8), dimension(nM,nM,nZ), intent(in) :: trans_m_cdnal_zp
        real(8), dimension(nA,nE,Tend,nH,nM,nZ), intent(out) :: q
        real(8), dimension(nM), intent(in) :: w
        real(8), dimension(nM,nZ), intent(in) :: qbar
        real(8), dimension(nE,nE,Tend,nZ,nZ),  intent(in) :: trans_e
        real(8), dimension(nE,nE,nZ,nZ,Tend),  intent(in) :: trans_ez
        real(8), intent(in) :: gA(nA), gE(nE,Tend)

        !local
        real(8), dimension(:,:,:,:), allocatable :: vR,cR
        real(8), dimension(:,:,:,:,:), allocatable :: apR
        real(8), dimension(:,:,:), allocatable :: endow

        real(8), dimension(:,:,:,:,:), allocatable :: CV,qap
        real(8), dimension(:,:,:,:), allocatable :: vNextPeriod,tmp_AEHM
        real(8), dimension(nE,nM) :: vDefBnk,cDefBnk 
        integer, dimension(nE) :: option
        integer :: iE, iA, ap_eq0_ind, iH, iM, iZp

        logical :: limitChoices(nE,Tend)
        integer :: age

        allocate(vR(nA,nE,nH,nM),cR(nA,nE,nH,nM))
        allocate(apR(nA,nE,nH,nM,nZ))
        allocate(endow(nE,nH,nM))

        allocate(CV(nA,nE,nH,nM,nZ),qap(nA,nE,nH,nM,nZ),vNextPeriod(nA,nE,nH,nM),tmp_AEHM(nA,nE,nH,nM))

        ! Setup the list of restricted participants
        if (isResrictedParticipation) then
            ! Restrict the non-super rich
            do iE = 1,nE
                limitChoices(iE,:) = .not. state(iE)%isSuperRich 
            end do
        else
            limitChoices = .false.
        end if


        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!        
        do age = Tend,1,-1

            if (age==Tend) then

                !!! Compute the ``price schedule'' faced in the last period of life (assume hhs default for sure next period)
                q(1:nAmm,:,age,:,:,:) = -.01234d0
                q(nAmm+1:nA,:,age,:,:,:) = 1d0
                CV = 0d0

            else

                ! Compute the price schedules for a given age given lots of information.
                call calc_price_schedules(q,qbar,d,age,w,gE,trans_e,trans_m_cdnal_zp,z_indices,gA)

                ! Compute the continuation value taking into account next periods value functions
                do iZp = 1,nZ

                    !!! First compute just the expectation taken over m conditional on z'=g
                    vNextPeriod = v(:,:,age+1,:,:)            
                    call mydgemm(tmp_AEHM,vNextPeriod,trans_m_cdnal_zp(:,:,iZp),nA*nE*nH,nM,nM,nM,tB='T')

                    !!! Now compute the expectation taken over e' conditional on z'=g
                    do iH = 1,nH
                        do iM = 1,nM

                            !Calculate expectation cdal on (e,z) by using E CV(a'|e,z,m) = sum_{e',z'} (CV(a'|e',z',m)*Pr(e',z'|e,z))
                            call mydgemm(CV(:,:,iH,iM,iZp),tmp_AEHM(:,:,iH,iM),trans_ez(:,:,z_indices(iM),iZp,age),nA,nE,nE,nE,tB='T')

                        end do
                    end do
                end do
                CV = probLive(age)*discount*CV  ! Note mortality risk and beta hav been included here 

            end if

            ! Compute the "expenditure" schedule q(a')*a'
            do iZp = 1,nZ
                do iM = 1,nM
                    do iH = 1,nH
                        do iE = 1,nE
                            qap(:,iE,iH,iM,iZp) = q(:,iE,age,iH,iM,iZp)*gA
                        end do
                    end do
                end do
            end do

            if (any(isnan(q(:,:,age,:,:,:)))) STOP 'found q is nan'
            if (any(isnan(qap))) STOP 'found qap nan'

            ! Set the portfolio options according to the parameter values
            do iE = 1,nE
                if (limitChoices(iE,age)) then
                    option(iE) = 1
                else
                    if (isRestrictedPortfolio) then
                        option(iE) = 2
                    else
                        option(iE) = 0
                    end if
                end if
            end do

            ! Compute value of default into bank
            do iM = 1,nM
                cDefBnk(:,iM) = w(iM)* gE(:,age)*eshift(z_indices(iM)) * (1d0-chi(gE(:,age)*eshift(z_indices(iM)),eshift(z_indices(iM))) )
            end do
            do iM = 1,nM
                do iE = 1,nE
                    vDefBnk(iE,iM) = -csnEquiv(age)/cDefBnk(iE,iM) + sum(CV(nAmm+1,iE,2,iM,:))
                end do
            end do

            ! Take CV, which is really contingent on h', and make contingent on h
            CV(:,:,2,:,:) = lambda*CV(:,:,2,:,:) + (1d0-lambda)*CV(:,:,1,:,:)
!             where (abs(CV(:,:,2,:,:)-CV(:,:,1,:,:))<10d0*epsilon(0d0)*abs(CV(:,:,1,:,:))) CV(:,:,2,:,:) = CV(:,:,1,:,:) ! cv(h=0) should always be >= cv(h=1) in theory. due to numercal precision, this may not be the case computationally. but we can force it via this check.

            ! Solve for the value of repaying and value of being in bankruptcy
            do iM = 1,nM
                endow(:,1,iM) = w(iM)*gE(:,age)*eshift(z_indices(iM)) - state(:)%expend
                endow(:,2,iM) = w(iM)*gE(:,age)*eshift(z_indices(iM))*(1d0-chiBnk(gE(:,age)*eshift(z_indices(iM)),eshift(z_indices(iM)))) - state(:)%expend
            end do

            call compute_vR_BC(vR,cR,apR,CV,qap,endow,gA,csnEquiv(age),option)

            if (any(isnan(apR(:,:,1,:,:)))) STOP 'apR(h=1) has nan'
            if (any(isnan(apR(nAmm+1:nA,:,2,:,:)))) STOP 'apR(h=2,a+) has nan'
            if (any(isnan(apR(:,:,2,:,:)))) STOP 'apR(h=2) has nan'
            if (any(isnan(apR))) STOP 'apR has nan'

            ! Check some theoretical results
            if (nExpenditure==1) then
                if (any(vR(nAmm+1:nA,:,1,:)<=.1d0*death)) print*,'WARNING: vRepay(a>=0) is really negative' 
                do iM = 1,nM
                    if (any(diff(vR(:,:,1,iM))<-1d-10)) print*,'WARNING: vRepay not increasing in a'
                    if (any(diff(vR(nAmm+1:nA,:,2,iM))<-1d-10)) print*,'WARNING: vBnk not increasing in a'
!                     if (any(vR(nAmm+1:nA,:,1,iM)<spread(vDefBnk(:,iM),1,nA-nAmm))) then
!                         print*,'maxval(spread(vDefBnk,1,nA-nAmm)-vRepay(nAmm+1:nA,:))',maxval(spread(vDefBnk(:,iM),1,nA-nAmm)-vR(nAmm+1:nA,:,1,iM))
!                         print*,'WARNING: vRepay(a>=0) is less than vDefBnk'
!                     end if
                end do
 !                if (any(vR(nAmm+1:nA,:,1,:)<=vR(nAmm+1:nA,:,2,:)-1d-8*abs(vR(nAmm+1:nA,:,2,:)))) then
 !                    print*,'maxval(vBnk-vRepay(nAmm+1:nA,:))',maxval(vR(nAmm+1:nA,:,2,:)-vR(nAmm+1:nA,:,1,:))
 !                    print*,'WARNING: vRepay(a>=0) is worse than vBnk(a>=0)' 
 !                end if
            end if
            
            ! Calculate the values of good and bad standing (good is upper envelope of not def and def)
            !Initialize credit flag policy (set to zero the probability of going anywhere)

            !!! NOTE: 
            !!! Since we know that a'=0 upon default, I've decided to say a'(a,e,age,h=0,m) is actually a'R
            !!! This avoids having to pass a separate a'R policy function around. 

            ! Determine whether to default or not and update value of good standing
            do iH = 1,nH
                do iM = 1,nM
                    do iE = 1,nE
                        do iA = 1,nA

                            ! Do not change: See the note above
                            apg(iA,iE,age,iH,iM) = apR(iA,iE,iH,iM,g_ind)
                            apb(iA,iE,age,iH,iM) = apR(iA,iE,iH,iM,b_ind)

                            ! Get the default decision
                            if (defBnkAllowed) then
                                if (.not. defZcontingent) then
                                    ! Normal case: default if you want to
                                    d(iA,iE,age,iH,iM) = vDefBnk(iE,iM)-abs(vDefBnk(iE,iM))*2d0*epsilon(0d0)>vR(iA,iE,iH,iM)
                                else
                                    ! Z contingent: default if you want to but only if allowed in that zind
                                    if (z_indices(iM)==defZcontingent_AllowedZind) then
                                        d(iA,iE,age,iH,iM) = vDefBnk(iE,iM)-abs(vDefBnk(iE,iM))*2d0*epsilon(0d0)>vR(iA,iE,iH,iM)
                                    else 
                                        ! If no default allowed, and there are expenditure shocks, permit default only when x>0
                                        d(iA,iE,age,iH,iM) = vDefBnk(iE,iM)-abs(vDefBnk(iE,iM))*2d0*epsilon(0d0)>vR(iA,iE,iH,iM) .and. &
                                                             state(iE)%expend>1d-10
                                    end if
                                end if
                            else
                                ! If no default allowed, and there are expenditure shocks, permit default only when x>0
                                d(iA,iE,age,iH,iM) = vDefBnk(iE,iM)-abs(vDefBnk(iE,iM))*2d0*epsilon(0d0)>vR(iA,iE,iH,iM) .and. &
                                                     state(iE)%expend>1d-10
                            end if

                            ! Using default decision, recover value and policy functions
                            if (d(iA,iE,age,iH,iM)) then
                                v(iA,iE,age,iH,iM) = vDefBnk(iE,iM)
                                c(iA,iE,age,iH,iM) = cDefBnk(iE,iM)
                            else
                                v(iA,iE,age,iH,iM) = vR(iA,iE,iH,iM)
                                c(iA,iE,age,iH,iM) = cR(iA,iE,iH,iM)
                            end if

                        end do
                    end do           
                end do

            end do

            ! Display some progress
            if (age<Tend .and. mod(age,20)==0) then 
                print*, 'subVFI: age ',age, 'supnorm', maxval(abs(v(:,:,age,:,:)-v(:,:,age+1,:,:)))
                ! write(*,'(A)',advance='no') 'VFItime: '; call toc()
            end if

        end do

!         call plot(gA,v(:,:,1,1,1),'v')
!         call plot(gA,c(:,:,1,1,1),'c')
!         call plot(gA,apg(:,:,1,1,1),'apg')
!         call plot(gA,apb(:,:,1,1,1),'apb')

    end subroutine subVFI

end module mod_value_fcns
