% bicubicSplineCoeff.m
% 
% Constructs the coefficients for a bicubic spline interpolation.
% The underlying function is y(x1,x2) with x1 having n1 pts and x2 n2.
% The algorithm is essentially that in Numerical recipes, where 
% first n2 the d^2y(x1,x2)/dx1^2 values are tabulated for each x2.
% Subsequently, when bicubicSplineEval.m is called, these are used
% to quickly construct one additional 1-dimensional spline and then
% evaluate it.
%
%   Method  | Spline Type
%   ========================
%   1       | not-a-knot (y''' constant across 1,2 interval and n-2,n-1 interval)
%   2       | natural (y''=0 at endpoints)
%   3       | secant hermite (y' at endpoints given by secant value)
%
% Note: the routine could be modified to also use a hermite spline (with 
% the user specifying derivatives; however, this would be a pain.
function [abcd_table] = bicubicSplineCoeff(x1,x2,y,method)

% Check inputs
if (~isvector(x1) || ~isvector(x2))
    error('x1 and x2 must both be vectors');
end
if (any(diff(x1)<=0) || any(diff(x2)<=0)) 
    error('x1 and x2 must be monotonically increasing');
end
if (size(y)~=[numel(x1),numel(x2)])
    error('y must be n1 by n2');
end

% Ensure column vectors
x1 = x1(:);
x2 = x2(:);

% Get lengths
n1 = length(x1);
n2 = length(x2);

% Loop over the columns and construct spline values for each
abcd_table = NaN(n1-1,4,n2);
for i2 = 1:n2
    abcd_table(:,:,i2) = cubicSplineCoeff(x1,y(:,i2),method);
end 

end
