% [abcd ypp] = cubicSplineCoeff(x,y,method,varargin)
% 
% Constructs the coefficients of a cubic spline y = a + bx + cx^2 + dx^3
% Also returns the second derivatives of the spline at each x. 
% The method argument selects a spline type according to the values
%
%   Method  | Spline Type
%   ========================
%   1       | not-a-knot (y''' constant across 1,2 interval and n-2,n-1 interval)
%   2       | natural (y''=0 at endpoints)
%   3       | secant hermite (y' at endpoints given by secant value)
%   4       | hermite (y' at endpoints specified by user)
%
% If method==4 selecting the hermite spline, a length 2 vector must be
% given providing the desired derivatives at x_1 and x_n
%
% Btw, the second derivatives of the spline at any x in [x_i,x_i+1] are 
% y''_i*theta + y''_i+1*(1-theta) where theta = (x_i+1-x)/(x_i+1-x_i).
% Hence to check concavity/convexity, enough to chech y''_i<0 or >0 for all
% i.
%
function [abcd ypp] = cubicSplineCoeff(x,y,method,varargin)

% Check inputs
if (~iscolumn(y) || ~iscolumn(x))
    error('x and y must both be column vectors');
end
if (any(diff(x)<=0)) 
    error('x must be monotonically increasing');
end

% Set the number of points n (intervals are n-1)
n = numel(y);

% Get the spline type
isNatural=0; % assumes y''(x1)=0=y''(xn)
isNotAKnot=0; % assumes y''' is constant across second and penultimate knots
isSecantHermite=0; % assumes y'(x1)=(y2-y1)/(x2-x1) and y'(xn) is similar
isHermite=0; % imposes y'(x1) and y'(xn) given by the user
switch method
    case 1,
        isNotAKnot=1;
    case 2,
        isNatural=1;
    case 3,
        isSecantHermite=1;
        
        yp1 = (y(2)-y(1))/(x(2)-x(1));
        ypN = (y(n)-y(n-1))/(x(n)-x(n-1));

    case 4,
        
        isHermite=1;
        
        if (~isempty(varargin)) 
            tmp = varargin{1};
            if (isvector(tmp) && numel(tmp)==2) 
                yp1 = tmp(1);
                ypN = tmp(2);
            else    
                error('the end point derivatives must be supplied as a length 2 vector')
            end
        else
            error('if selecting hermite spline, the end point derivatives must be supplied')
        end
    otherwise,    
        error('method must be one of {1,2,3,4}')
end

% Allocate space
ypp = NaN(n,1); % Unknown second derivatives, to be solved for .

% Compute the second derivatives of at each value of x
del = diff(x);
m = diff(y)./del; 
if (isNatural)
    
    % Condition is second derivatives at end points is set to 0
    
    ypp(1) = 0;
    ypp(n) = 0;
    
    d = (del(2:n-1)+del(1:n-2))/3; % (x(3:n)-x(1:n-2))/3; % The main diagonal values
    dp1 = del(2:n-2)/6; % diff(x(2:n-1))/6; % One above the main diagonal
    dm1 = dp1; 
        
%     A = diag(d)+diag(dp1,1)+diag(dm1,-1);
%     A = sparse(A);

    A = spdiags([[dm1;0],d,[0;dp1]],-1:1,n-2,n-2);
    
    ypp(2:n-1) = A\diff(m);
        
elseif (isHermite || isSecantHermite)
    
    % Conditions is derivatives at end points are set to yp1 and ypN
    
    d = [2;(del(2:n-1)+del(1:n-2))/3;2]; % (x(3:n)-x(1:n-2))/3; % The main diagonal values
    dp1 = [1;del(2:n-1)/6]; % % One above the main diagonal
    dm1 = [del(1:n-2)/6;1]; 
    
%     A = diag(d)+diag(dp1,1)+diag(dm1,-1);
%     A = sparse(A);
    
    A = spdiags([[dm1;0],d,[0;dp1]],-1:1,n,n);
    
    b = [-6/del(1)*(yp1-m(1));diff(m);6/del(n-1)*(ypN-m(n-1))];
    
    ypp(1:n) = A\b;
    
elseif (isNotAKnot) 
    
    % Conditions is third derivatives are the same in the 1 and 2nd
    % intervals, and the n-2 and n-1 th intervals.
    
    d = [1/del(1);(del(2:n-1)+del(1:n-2))/3;1/del(n-1)];
    dp1 = [-1/del(1)-1/del(2);del(2:n-1)/6]; % % One above the main diagonal
    dm1 = [del(1:n-2)/6;-1/del(n-1)-1/del(n-2)]; 
    
%     A = diag(d)+diag(dp1,1)+diag(dm1,-1);
%     A(1,3) = 1/del(2);
%     A(n,n-2) = 1/del(n-2);
%     A = sparse(A);

    A = spdiags([[dm1;0],d,[0;dp1]],-1:1,n,n);
    A(1,3) = 1/del(2);
    A(n,n-2) = 1/del(n-2);
    
    b = [0;diff(m);0];
    ypp(1:n) = A\b;
    
end


% Put the solution into standard form y=a_i+b_i(x-xi)+ c_i(x-xi)^2 + d_i (x-xi)^3
% A simple way to do this is to match second derivatives with the
% non standard form. This gives c_i and d_i. Then use these with 
% the interpolation conditions to back out b_i and a_i. 
d = diff(ypp)./(6*del);
c = ypp(1:n-1)/2;
b = m - c.*del - d.*del.^2;
a = y(1:n-1);

abcd = [a,b,c,d];


end 
