% [stats] = discreteStats(pdf,grid)
% Given a mdf (mass density function) over a particular grid, calculates
% several summary statistics and puts them in a structure stats
% 
% Note this treats the distribution truly as being discrete so some
% statistics, like median, may be weird without interpolation.
% 
% Statistics            Description
%  mean
%  var
%  stdev
%  median
%  meanmedian
%  gini
%  shareTop1
%  shareTop5
%  shareQuint5
%  shareQuint4
%  shareQuint3
%  shareQuint2
%  shareQuint1
%  level100             Level of grid held by 100% guy (max of grid)
%  level99              Level of grid held by 99% guy (above this is in top1%)
%  level95           
%  level80         
%  level60         
%  level40         
%  level20         
%  level5
%  level1
%  level0         
function [stats] = discreteStats(pdf,grid)

    if abs(sum(pdf(:))-1)>1d-12 
       fprintf(1,'warning: pdf doesn''t sum to 1 it is %f away\n',1-sum(pdf(:)))
    end
    
    if (numel(pdf)~=numel(grid)) 
        error('pdf and grid must have the same size')
    end
        
    grid = grid(:);
    pdf = pdf(:);
    n = length(grid);
    
    if isnan(pdf) 
        warning('pdf has nan')
    end
    if isnan(grid) 
        warning('grid has nan')
    end
    
    [grid I] = sort(grid);
    pdf = pdf(I);
    
    stats.mean = pdf'*grid;
    if isnan(stats.mean)
        warning('mean is nan')
    end
    stats.var = sum(pdf.*grid.^2) - stats.mean.^2;   
    stats.stdev = sqrt(stats.var);
        
    cdf = cumsum(pdf);
    stats.median = grid(find(cdf>=.5,1));

    stats.meanmedian = stats.mean/stats.median;

    % Compute the gini and Lorenz curve according to the discrete formula
    S = NaN(n+1,1);
    S(1) = 0;
    S(2:n+1) = cumsum(pdf.*grid);
    stats.gini = 1-sum(pdf.*(S(1:n)+S(2:n+1)))/S(n+1);
    stats.Lg = [0;cumsum(pdf)];
    stats.Lv = S/S(n+1);
%     plot(stats.Lg,stats.Lg,stats.Lg,stats.Lv)
    
    % Compute shares of wealth held by various groups
    i0 = find(stats.Lg>=1-1d-12,1);
    i1 = find(stats.Lg>=.99,1);
    i5 = find(stats.Lg>=.95,1);
    i20 = find(stats.Lg>=.8,1);
    i40 = find(stats.Lg>=.6,1);
    i60 = find(stats.Lg>=.4,1);
    i80 = find(stats.Lg>=.2,1);
    i100 = find(stats.Lg>=.0,1);
    
    stats.shareTop1 = stats.Lv(i0) - stats.Lv(i1);
    stats.shareTop5 = stats.Lv(i0) - stats.Lv(i5);
    stats.shareQuint5 = stats.Lv(i0) - stats.Lv(i20);
    stats.shareQuint4 = stats.Lv(i20) - stats.Lv(i40);
    stats.shareQuint3 = stats.Lv(i40) - stats.Lv(i60);
    stats.shareQuint2 = stats.Lv(i60) - stats.Lv(i80);
    stats.shareQuint1 = stats.Lv(i80) - stats.Lv(i100);

    stats.level100 = grid(n);
    stats.level99 = grid(find(cdf>=.99,1));
    stats.level95 = grid(find(cdf>=.95,1));
    stats.level80 = grid(find(cdf>=.8,1));
    stats.level60 = grid(find(cdf>=.6,1));
    stats.level40 = grid(find(cdf>=.4,1));
    stats.level20 = grid(find(cdf>=.2,1));
    stats.level5 = grid(find(cdf>=.05,1));
    stats.level1 = grid(find(cdf>=.01,1));
    stats.level0 = grid(1);
    

%     return
%     [stats.gini Li eq Lg]=sub_gini(pdf,grid);
%     
%     [tmp i1] = min(abs(Lg-.99));
%     [tmp i2] = min(abs(Lg-.98));
%     [tmp i5] = min(abs(Lg-.95));
%     [tmp i20] = min(abs(Lg-.80));
%     [tmp i40] = min(abs(Lg-.60));
%     [tmp i60] = min(abs(Lg-.40));
%     [tmp i80] = min(abs(Lg-.20));
%     [tmp i100] = min(abs(Lg-.00));
% 
%     stats.shareTop1 = 1-Li(i1);
%     stats.shareTop5 = 1 - Li(i5);
%     stats.shareQuint5 = 1 - Li(i20);
%     stats.shareQuint4 = Li(i20)-Li(i40);
%     stats.shareQuint3 = Li(i40)-Li(i60);
%     stats.shareQuint2 = Li(i60)-Li(i80);
%     stats.shareQuint1 = Li(i80)-Li(i100);

    
end