module mod_interp

implicit none

interface interp1
    module procedure interp1_mat,interp1_vec,interp1_pt
end interface interp1

contains

!MATLAB STYLE INTERPOLATION ROUTINES

!INTERP1 1-D interpolation (table lookup)
!    YI = INTERP1(X,Y,XI) interpolates to find YI, the values of the
!    underlying function Y at the points in the array XI. X must be a
!    vector of length N.
function interp1_mat(x,y,xi,option) result(yi)
    use mod_types
    implicit none
    real(dp), dimension(1:), intent(in) :: x, xi
    real(dp), dimension(1:,1:), intent(in) :: y
    real(dp), dimension(1:size(xi,1),1:size(y,2)) :: yi
    integer, intent(in), optional :: option
    !local
    integer :: col_ind    
    !Check inputs
    if (size(x)/=size(y,1)) STOP 'interp1: size x /= size y'    
    !Interp
    do col_ind = 1,size(y,2)
        yi(:,col_ind) = interp1_vec(x,y(:,col_ind),xi,option)
    end do
end function interp1_mat

function interp1_vec(x,y,xi,option) result(yi)
    use mod_types
    implicit none
    real(dp), dimension(1:), intent(in) :: x, y, xi
    real(dp), dimension(1:size(xi)) :: yi
    integer, intent(in), optional :: option
    !local
    integer :: xi_ind
    
    !Check inputs
    if (size(x)/=size(y)) STOP 'interp1: size x /= size y'

    do xi_ind = 1,size(xi)
        if (present(option)) then
            yi(xi_ind) = interp1_pt(x,y,xi(xi_ind),option)
        else
            yi(xi_ind) = interp1_pt(x,y,xi(xi_ind))
        end if
    end do

end function interp1_vec

function interp1_pt(x,y,xi,option) result(yi)
    use mod_types
    implicit none
    real(dp), dimension(1:), intent(in) :: x,y
    real(dp) :: xi, yi
    integer, intent(in), optional :: option
    !local
    integer :: len_x, xi_ilo, xi_ihi, xi_ind, x_ind
    real(dp) :: xi_mlo, xi_mhi, mu_x 
    integer :: flag
    real(dp) :: sum

    len_x = size(x)
    
    !Check inputs
    if (size(y)/=size(x)) STOP 'interp1: y does not have #x elts'

    !Get the closest grid points and their weights
    call interp_get_interval(xi_ilo,xi_ihi,xi_mlo,xi_mhi,xi,x,len_x,flag)

    if (present(option)) then                
        if (option==1) then
            !Don't allow extrapolation
            xi_mhi = max(min(xi_mhi,1.d0),0.d0)
            xi_mlo = max(min(xi_mlo,1.d0),0.d0)
        end if
    end if

    !Interpolation weight is just mu(x)
    sum = 0.d0
    do xi_ind = 1,2
        x_ind = merge(xi_ilo,xi_ihi,xi_ind==1)
        mu_x = merge(xi_mlo,xi_mhi,xi_ind==1)       

        sum = sum + mu_x*y(x_ind)
    end do
    yi = sum

end function interp1_pt

!INTERP2 2-D interpolation (table lookup).
!    ZI = INTERP2(X,Y,Z,XI,YI) interpolates to find ZI, the values of the
!    underlying 2-D function Z at the points in vectors XI and YI.
!    Vectors X and Y specify the points at which the data Z is given.
! 
!    XI is a vector.  Similarly, YI can be a column vector and it 
!    specifies a matrix with constant rows. 
!
!Option specifies the following
! 0: (default) allow extrapolation
! 1: only interpolation
!
function interp2(x,y,z,xi,yi,option) result(zi)
    use mod_types
    implicit none
    real(dp), dimension(1:), intent(in) :: x, y, xi, yi
    real(dp), dimension(1:,1:), intent(in) :: z 
    real(dp), dimension(1:size(xi),1:size(yi)) :: zi
    integer, intent(in), optional :: option
    !local
    integer :: xi_ind, yi_ind
    
    !Check inputs
    if (size(z,1)/=size(x)) STOP 'interp2: z does not have #x rows'
    if (size(z,2)/=size(y)) STOP 'interp2: z does not have #y columns'

    !Interpolate for each point
    do xi_ind = 1,size(xi)
        do yi_ind = 1,size(yi)
            if (present(option)) then
                zi(xi_ind,yi_ind) = interp2_pt(x,y,z,xi(xi_ind),yi(yi_ind),option)
            else
                zi(xi_ind,yi_ind) = interp2_pt(x,y,z,xi(xi_ind),yi(yi_ind))
            end if
        end do
    end do

end function interp2

function interp2_pt(x,y,z,xi,yi,option) result(zi)
    use mod_types
    implicit none
    real(dp), dimension(1:), intent(in) :: x, y
    real(dp), dimension(1:,1:), intent(in) :: z 
    real(dp) :: xi, yi
    real(dp) :: zi
    integer, intent(in), optional :: option
    !local
    integer :: len_x, len_y, xi_ilo, xi_ihi, yi_ilo, yi_ihi, xi_ind, yi_ind, x_ind, y_ind
    real(dp) :: xi_mlo, xi_mhi, mu_x, yi_mlo, yi_mhi, mu_y
    integer, dimension(1:2) :: flags
    integer :: flag
    real(dp) :: sum

    len_x = size(x)
    len_y = size(y)
    
    !Check inputs
    if (size(z,1)/=size(x)) STOP 'interp2: z does not have #x rows'
    if (size(z,2)/=size(y)) STOP 'interp2: z does not have #y columns'
   
    !Get the closest grid points for each moment on their respective grids
    call interp_get_interval(xi_ilo,xi_ihi,xi_mlo,xi_mhi,xi,x,len_x,flag); flags(1) = flag
    call interp_get_interval(yi_ilo,yi_ihi,yi_mlo,yi_mhi,yi,y,len_y,flag); flags(2) = flag

    if (present(option)) then                
        if (option==1) then
            !Don't allow extrapolation
            xi_mhi = max(min(xi_mhi,1.d0),0.d0)
            xi_mlo = max(min(xi_mlo,1.d0),0.d0)
            yi_mhi = max(min(yi_mhi,1.d0),0.d0)
            yi_mlo = max(min(yi_mlo,1.d0),0.d0)
        end if
    end if

    !Joint interpolation weight is just mu(x,y) = mu(x)*mu(y)
    sum = 0.d0
    do xi_ind = 1,2
        x_ind = merge(xi_ilo,xi_ihi,xi_ind==1)
        mu_x = merge(xi_mlo,xi_mhi,xi_ind==1)       
        do yi_ind = 1,2
            y_ind = merge(yi_ilo,yi_ihi,yi_ind==1)
            mu_y = merge(yi_mlo,yi_mhi,yi_ind==1)

            sum = sum + (mu_x*mu_y)*z(x_ind,y_ind)                        
        end do
    end do
    zi = sum

end function interp2_pt


!Matlab style wrapper for my interp subroutine
!INTERP3 3-D interpolation (table lookup).
!    VI = INTERP3(X1,X2,X3,V,X1I,X2I) interpolates to find VI, the values of the
!    underlying 3-D function V at the points in vectors X1I and Y2I.
! 
!    X1I and X2I are vectors.  
!
!Option specifies the following
! 0: (default) allow extrapolation
! 1: only interpolation
!
function interp3(x1,x2,x3,v,x1i,x2i,x3i,option) result(vi)
    use mod_types
    implicit none
    real(dp), dimension(1:), intent(in) :: x1, x2, x3, x1i, x2i, x3i
    real(dp), dimension(1:,1:,1:), intent(in) :: v 
    real(dp), dimension(1:size(x1i),1:size(x2i),1:size(x3i)) :: vi
    integer, intent(in), optional :: option
    !local
    integer :: x1i_ind, x2i_ind, x3i_ind
    
    !Check inputs
    if (size(v,1)/=size(x1)) STOP 'interp3: v does not have #x1 rows'
    if (size(v,2)/=size(x2)) STOP 'interp3: v does not have #x2 columns'
    if (size(v,3)/=size(x3)) STOP 'interp3: v does not have #x3 columns'

    !Interpolate for each point
    do x1i_ind = 1,size(x1i)
        do x2i_ind = 1,size(x2i)
            do x3i_ind = 1,size(x3i)
                if (present(option)) then
                    vi(x1i_ind,x2i_ind,x3i_ind) = interp3_pt(x1,x2,x3,v,x1i(x1i_ind),x2i(x2i_ind),x3i(x3i_ind),option)
                else
                    vi(x1i_ind,x2i_ind,x3i_ind) = interp3_pt(x1,x2,x3,v,x1i(x1i_ind),x2i(x2i_ind),x3i(x3i_ind))
                end if
            end do
        end do
    end do

end function interp3

function interp3_pt(x1,x2,x3,v,x1i,x2i,x3i,option) result(vi)
    use mod_types
    implicit none
    real(dp), dimension(1:), intent(in) :: x1, x2, x3
    real(dp), dimension(1:,1:,1:), intent(in) :: v 
    real(dp) :: x1i, x2i, x3i
    real(dp) :: vi
    integer, intent(in), optional :: option
    !local
    integer :: len_x1, len_x2, len_x3, &
               x1i_ilo, x1i_ihi, x2i_ilo, x2i_ihi, x3i_ilo, x3i_ihi, &
               x1i_ind, x2i_ind, x3i_ind, &
               x1_ind, x2_ind, x3_ind
    real(dp) :: x1i_mlo, x1i_mhi, mu_x1, &
                x2i_mlo, x2i_mhi, mu_x2, &
                x3i_mlo, x3i_mhi, mu_x3
    integer, dimension(1:3) :: flags
    integer :: flag
    real(dp) :: sum

    len_x1 = size(x1)
    len_x2 = size(x2)
    len_x3 = size(x3)
    
    !Check inputs
    if (size(v,1)/=size(x1)) STOP 'interp3: z does not have #x1 rows'
    if (size(v,2)/=size(x2)) STOP 'interp3: z does not have #x2 columns'
    if (size(v,3)/=size(x3)) STOP 'interp3: z does not have #x3 columns'
   
    !Get the closest grid points for each moment on their respective grids
    call interp_get_interval(x1i_ilo,x1i_ihi,x1i_mlo,x1i_mhi,x1i,x1,len_x1,flag); flags(1) = flag
    call interp_get_interval(x2i_ilo,x2i_ihi,x2i_mlo,x2i_mhi,x2i,x2,len_x2,flag); flags(2) = flag
    call interp_get_interval(x3i_ilo,x3i_ihi,x3i_mlo,x3i_mhi,x3i,x3,len_x3,flag); flags(3) = flag

    if (present(option)) then                
        if (option==1) then
            !Don't allow extrapolation
            x1i_mhi = max(min(x1i_mhi,1.d0),0.d0)
            x1i_mlo = max(min(x1i_mlo,1.d0),0.d0)
            x2i_mhi = max(min(x2i_mhi,1.d0),0.d0)
            x2i_mlo = max(min(x2i_mlo,1.d0),0.d0)
            x3i_mhi = max(min(x3i_mhi,1.d0),0.d0)
            x3i_mlo = max(min(x3i_mlo,1.d0),0.d0)
        end if
    end if

    !Joint interpolation weight is just mu(x1,x2) = mu(x1)*mu(x2)
    sum = 0.d0
    do x1i_ind = 1,2
        x1_ind = merge(x1i_ilo,x1i_ihi,x1i_ind==1)
        mu_x1 = merge(x1i_mlo,x1i_mhi,x1i_ind==1)       
        do x2i_ind = 1,2
            x2_ind = merge(x2i_ilo,x2i_ihi,x2i_ind==1)
            mu_x2 = merge(x2i_mlo,x2i_mhi,x2i_ind==1)
            do x3i_ind = 1,2
                x3_ind = merge(x3i_ilo,x3i_ihi,x3i_ind==1)
                mu_x3 = merge(x3i_mlo,x3i_mhi,x3i_ind==1)

                sum = sum + (mu_x1*mu_x2*mu_x3)*v(x1_ind,x2_ind,x3_ind)                        
            end do
        end do
    end do
    vi = sum

end function interp3_pt

!Matlab style wrapper for my interp subroutine
!INTERP4 4-D interpolation (table lookup).
!    VI = INTERP3(X1,X2,X3,X4,V,X1I,X2I,X3I,X4I) interpolates to find VI, the values of the
!    underlying function V at the points in vectors X1I and Y2I.
! 
!    Note X1,X2,X3,X4 need not be sorted, but V must be structured V(a,b,c,d) = V(X1(a),X2(b),...)
!    X1I,X2I,X3I,X4I are vectors.  
!
!Option specifies the following
! 0: (default) allow extrapolation
! 1: only interpolation
!
function interp4(x1,x2,x3,x4,v,x1i,x2i,x3i,x4i,option) result(vi)
    use mod_types
    implicit none
    real(dp), dimension(1:), intent(in) :: x1, x2, x3, x4, x1i, x2i, x3i, x4i
    real(dp), dimension(1:,1:,1:,1:), intent(in) :: v 
    real(dp), dimension(1:size(x1i),1:size(x2i),1:size(x3i),1:size(x4i)) :: vi
    integer, intent(in), optional :: option
    !local
    integer :: x1i_ind, x2i_ind, x3i_ind, x4i_ind
    
    !Check inputs
    if (size(v,1)/=size(x1)) STOP 'interp4: v does not have #x1 rows'
    if (size(v,2)/=size(x2)) STOP 'interp4: v does not have #x2 columns'
    if (size(v,3)/=size(x3)) STOP 'interp4: v does not have #x3 columns'
    if (size(v,4)/=size(x4)) STOP 'interp4: v does not have #x4 columns'

    !Interpolate for each point
    do x1i_ind = 1,size(x1i)
        do x2i_ind = 1,size(x2i)
            do x3i_ind = 1,size(x3i)
                do x4i_ind = 1,size(x4i)
                    if (present(option)) then
                        vi(x1i_ind,x2i_ind,x3i_ind,x4i_ind) = &
                            interp4_pt(x1,x2,x3,x4,v,x1i(x1i_ind),x2i(x2i_ind),x3i(x3i_ind),x4i(x4i_ind),option)
                    else
                        vi(x1i_ind,x2i_ind,x3i_ind,x4i_ind) = &
                            interp4_pt(x1,x2,x3,x4,v,x1i(x1i_ind),x2i(x2i_ind),x3i(x3i_ind),x4i(x4i_ind))
                    end if
                end do
            end do
        end do
    end do

end function interp4

function interp4_pt(x1,x2,x3,x4,v,x1i,x2i,x3i,x4i,option) result(vi)
    use mod_types
    implicit none
    real(dp), dimension(1:), intent(in) :: x1, x2, x3, x4
    real(dp), dimension(1:,1:,1:,1:), intent(in) :: v 
    real(dp) :: x1i, x2i, x3i, x4i
    real(dp) :: vi
    integer, intent(in), optional :: option
    !local
    integer :: len_x1, x1i_ilo, x1i_ihi, x1i_ind, x1_ind, &
               len_x2, x2i_ilo, x2i_ihi, x2i_ind, x2_ind, &
               len_x3, x3i_ilo, x3i_ihi, x3i_ind, x3_ind, &
               len_x4, x4i_ilo, x4i_ihi, x4i_ind, x4_ind               
    real(dp) :: x1i_mlo, x1i_mhi, mu_x1, &
                x2i_mlo, x2i_mhi, mu_x2, &
                x3i_mlo, x3i_mhi, mu_x3, &
                x4i_mlo, x4i_mhi, mu_x4
    integer, dimension(1:4) :: flags
    integer :: flag
    real(dp) :: sum

    len_x1 = size(x1)
    len_x2 = size(x2)
    len_x3 = size(x3)
    len_x4 = size(x4)
    
    !Check inputs
    if (size(v,1)/=size(x1)) STOP 'interp4: z does not have #x1 rows'
    if (size(v,2)/=size(x2)) STOP 'interp4: z does not have #x2 columns'
    if (size(v,3)/=size(x3)) STOP 'interp4: z does not have #x3 columns'
    if (size(v,4)/=size(x4)) STOP 'interp4: z does not have #x4 columns'
   
    !Get the closest grid points for each moment on their respective grids
    call interp_get_interval(x1i_ilo,x1i_ihi,x1i_mlo,x1i_mhi,x1i,x1,len_x1,flag); flags(1) = flag
    call interp_get_interval(x2i_ilo,x2i_ihi,x2i_mlo,x2i_mhi,x2i,x2,len_x2,flag); flags(2) = flag
    call interp_get_interval(x3i_ilo,x3i_ihi,x3i_mlo,x3i_mhi,x3i,x3,len_x3,flag); flags(3) = flag
    call interp_get_interval(x4i_ilo,x4i_ihi,x4i_mlo,x4i_mhi,x4i,x4,len_x4,flag); flags(4) = flag
    
    if (present(option)) then                
        if (option==1) then
            !Don't allow extrapolation
            x1i_mhi = max(min(x1i_mhi,1.d0),0.d0)
            x1i_mlo = max(min(x1i_mlo,1.d0),0.d0)
            x2i_mhi = max(min(x2i_mhi,1.d0),0.d0)
            x2i_mlo = max(min(x2i_mlo,1.d0),0.d0)
            x3i_mhi = max(min(x3i_mhi,1.d0),0.d0)
            x3i_mlo = max(min(x3i_mlo,1.d0),0.d0)
            x4i_mhi = max(min(x4i_mhi,1.d0),0.d0)
            x4i_mlo = max(min(x4i_mlo,1.d0),0.d0)
        end if
    end if

    !Joint interpolation weight is just mu(x1,x2,x3,x4) = mu(x1)*mu(x2)*mu(x3)*mu(x4)
    sum = 0.d0
    do x1i_ind = 1,2
        x1_ind = merge(x1i_ilo,x1i_ihi,x1i_ind==1)
        mu_x1 = merge(x1i_mlo,x1i_mhi,x1i_ind==1)       
        do x2i_ind = 1,2
            x2_ind = merge(x2i_ilo,x2i_ihi,x2i_ind==1)
            mu_x2 = merge(x2i_mlo,x2i_mhi,x2i_ind==1)
            do x3i_ind = 1,2
                x3_ind = merge(x3i_ilo,x3i_ihi,x3i_ind==1)
                mu_x3 = merge(x3i_mlo,x3i_mhi,x3i_ind==1)
                do x4i_ind = 1,2
                    x4_ind = merge(x4i_ilo,x4i_ihi,x4i_ind==1)
                    mu_x4 = merge(x4i_mlo,x4i_mhi,x4i_ind==1)

                    sum = sum + (mu_x1*mu_x2*mu_x3*mu_x4)*v(x1_ind,x2_ind,x3_ind,x4_ind)
                end do
            end do
        end do
    end do
    vi = sum

end function interp4_pt






function interp2_mesh(x,y,z,xi,yi) result(zi)
    use mod_types
    implicit none
    real(dp), dimension(1:,1:), intent(in) :: x, y, z, xi, yi
    real(dp), dimension(1:size(xi,1),1:size(xi,2)) :: zi
    !local
    integer :: len1, len2, len1i, len2i, x_ind, y_ind, yi_ind
    real(dp), dimension(1:size(x,1),1:size(xi,2)) :: z_interped_in_y
    real(dp), dimension(1:size(x,1)) :: xvec
    real(dp), dimension(1:size(x,2)) :: yvec

    len1 = size(x,1)
    len2 = size(x,2)
    len1i = size(xi,1)
    len2i = size(xi,2)
    
    !Check inputs
    if (size(x,1)/=size(z,1) .OR. size(y,1)/=size(z,1)) STOP 'interp2: size1 x /= size1 z or size1 y /= size1 z'
    if (size(x,2)/=size(z,2) .OR. size(y,2)/=size(z,2)) STOP 'interp2: size2 x /= size2 z or size2 y /= size2 z'
    if (size(xi,1)/=size(yi,1)) STOP 'interp2: size1 xi /= size1 yi'
    if (size(xi,2)/=size(yi,2)) STOP 'interp2: size2 xi /= size2 yi'
    do x_ind = 1,len1
        if (any(y(1,:)/=y(x_ind,:))) STOP 'interp2: y is not meshed!'    
    end do    
    do y_ind = 1,len2
        if (any(x(:,1)/=x(:,y_ind))) STOP 'interp2: x is not meshed!'    
    end do

    xvec = x(:,1)
    yvec = y(1,:)
    
    !Perform interp
    do x_ind = 1,len1
        !Fix x, interpolate each z(x,.) from y grid onto yi grid
        z_interped_in_y(x_ind,:)= interp1(yvec,z(x_ind,:),yi(x_ind,:)) !result is maxtrix len1 x len2i
    end do
    do yi_ind=1,len2i
        !Fix yi, interpolate each z(:,yi) from x grid onto xi grid
        zi(:,yi_ind) = interp1(xvec,z_interped_in_y(:,yi_ind),xi(:,yi_ind))
    end do    

end function interp2_mesh


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Given a function f defined (grid_x1,grid_x2) and a pt (x1,x2), 
! give the interpolated value (will extrapolate if asked)
pure function interp_linear(f,xhat,grid_x,len_x) result(fhat)
use mod_types
implicit none
real(dp) :: fhat
integer, intent(in) :: len_x
real(dp), dimension(1:len_x), intent(in) :: f
real(dp), intent(in) :: xhat
real(dp), dimension(1:len_x), intent(in) :: grid_x

!local
integer :: ilo, ihi
integer, dimension(1:2) :: i_vec
real(dp) :: mlo, mhi
real(dp), dimension(1:2) :: m_vec
real(dp) :: Ql, Qh

    i_vec = interp_get_interval_fn(xhat,grid_x,len_x)
    m_vec = interp_get_mass_fn(i_vec,xhat,grid_x,len_x)    
    
    ilo = i_vec(1)
    ihi = i_vec(2)
    mlo = m_vec(1)
    mhi = m_vec(2)
    
    Ql = f(ilo)
    Qh = f(ihi)

    !Interpolate
    fhat= interp_use_interval(Ql,Qh,mlo,mhi)

end function interp_linear

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Given a function f defined (grid_x1,grid_x2) and a pt (x1,x2), 
! give the interpolated value
pure subroutine interp_bilinear(fhat,f,x1,x2,grid_x1,grid_x2,len_x1,len_x2)

use mod_types
implicit none

real(dp), intent(out) :: fhat
integer, intent(in) :: len_x1, len_x2
real(dp), dimension(1:len_x1,1:len_x2), intent(in) :: f
real(dp), intent(in) :: x1,x2
real(dp), dimension(1:len_x1), intent(in) :: grid_x1
real(dp), dimension(1:len_x2), intent(in) :: grid_x2

!local
integer :: ilo1, ilo2, ihi1, ihi2
real(dp) :: mlo1, mlo2, mhi1, mhi2
real(dp) :: Qll, Qhl, Qlh, Qhh

    call interp_get_interval(ilo1,ihi1,mlo1,mhi1,x1,grid_x1,len_x1)
    call interp_get_interval(ilo2,ihi2,mlo2,mhi2,x2,grid_x2,len_x2)    
    
    Qll = f(ilo1,ilo2)
    Qhl = f(ihi1,ilo2)
    Qlh = f(ilo1,ihi2) 
    Qhh = f(ihi1,ihi2) 

    !Interpolate in first direction then in second direction
    fhat= &
        interp_use_interval(&
            interp_use_interval(Qll,Qhl,mlo1,mhi1),& !interpolated in first direction for second=lo
            interp_use_interval(Qlh,Qhh,mlo1,mhi1),& !interpolated in first direction for second=hi
                mlo2,mhi2)

end subroutine interp_bilinear

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Given a grid and a pt, finds the closest points below and above the point on the grid.  
! Returns the index of these two points, (ilo, ihi), and the mass that should be 
! placed on each (mlo,mhi) such that to interpolate any value, f(ihi)*mhi + f(ilo)*mlo  
! is the linearly interpolated value
!
! If pt is within a certain tolerance, ilo and ihi are set equal to each other, 
! and mhi=mlo=.5
!
! flag meaning
! 0: exited with no problem
! 1: extrapolated down
! 2: extrapolated up
! 3: only one point on grid
! 4: case not found

! Returns mass m_vec = [m_lo,m_hi] s.t. f(x) = dot(m_vec,grid_x(i_vec))
pure function interp_get_mass_fn(i_vec,x,grid_x,len_x) result(m_vec)
    use mod_types
    implicit none
    real(dp), dimension(1:2) :: m_vec
    integer, dimension(1:2), intent(in) :: i_vec
    integer, intent(in) :: len_x
    real(dp), dimension(1:len_x), intent(in) :: grid_x
    real(dp), intent(in) :: x    
    !local
    real(dp) :: mlo,mhi,dist
    integer :: ilo,ihi
    
    ilo = i_vec(1)
    ihi = i_vec(2)
    
    mlo = 1d14
    mhi = 1d14

    !Check if only one point on grid
    if (len_x==1) then
        mlo = 1.d0
        mhi = 0.d0
    elseif (ilo/=-1 .AND. ihi /=-1) then

        dist = abs(grid_x(ilo)-grid_x(ihi))
        if(dist<=1d-40) then 
            mhi=0.d0
            mlo=1.d0
        else
            mhi=(x-grid_x(ilo))/dist
            mlo=1.d0-mhi
        end if

    end if
    
    m_vec = (/mlo,mhi/)

end function interp_get_mass_fn
! Returns interval i_vec = [i_lo,i_hi] s.t. x is in grid_x(i_vec)
pure function interp_get_interval_fn(x,grid_x,len_x) result(i_vec)
    use mod_types
    implicit none
    integer, dimension(1:2) :: i_vec
    integer, intent(in) :: len_x
    real(dp), dimension(1:len_x), intent(in) :: grid_x
    real(dp), intent(in) :: x    
    !local    
    integer :: ilo, ihi
    integer, dimension(1:1) :: loc_dummy
    logical, dimension(1:len_x) :: check
    logical :: case1,case2,case3,case4

    ilo = -1
    ihi = -1

    case1 = len_x==1
    case2 = any(grid_x<=x) .AND. any(grid_x>=x)
    case3 = all(grid_x>x)
    case4 = all(grid_x<x) 
    
    if (case1) then
        !Only one point on grid
        ilo = 1 
        ihi = 1
    elseif (case2) then 
        !Interpolation
        check = grid_x<=x
        loc_dummy = minloc(abs(grid_x-x),mask=check)
        ilo = loc_dummy(1)

        check = grid_x>=x
        loc_dummy = minloc(abs(grid_x-x),mask=check)
        ihi = loc_dummy(1)   

    elseif (case3) then
        !Extrapolation downward
        check = grid_x>x
        loc_dummy = minloc(abs(grid_x-x),mask=check)
        ilo = loc_dummy(1)

        check(ilo) = .FALSE.
        loc_dummy = minloc(abs(grid_x-x),mask=check)
        ihi = loc_dummy(1)    

    elseif (all(grid_x<x)) then
        !Extrapolation upward
        check = grid_x<x
        loc_dummy = minloc(abs(grid_x-x),mask=check)
        ihi = loc_dummy(1)

        check(ihi) = .FALSE.
        loc_dummy = minloc(abs(grid_x-x),mask=check)
        ilo = loc_dummy(1)    
    end if    

    i_vec = (/ilo,ihi/)
    
end function interp_get_interval_fn

pure subroutine interp_get_interval(ilo,ihi,mlo,mhi,x,grid_x,len_x,flag)
use mod_types
implicit none

integer, intent(in) :: len_x
real(dp), dimension(1:len_x), intent(in) :: grid_x
real(dp), intent(in) :: x
integer, intent(out) :: ilo, ihi
real(dp), intent(out) :: mlo, mhi
integer, intent(out), optional :: flag

integer, dimension(1:1) :: loc_dummy
logical, dimension(1:len_x) :: check
real(dp) :: dist

ilo = -1
ihi = -1
mlo = 1.d14
mhi = 1.d14

!Check if only one point on grid
if (len_x==1) then
    if (present(flag)) flag = 3
    ilo = 1 
    ihi = 1
    mlo = 1.d0
    mhi = 0.d0
else

    !Is this interpolation?
    if (any(grid_x<=x) .AND. any(grid_x>=x)) then 

        if (present(flag)) flag = 0

        check = grid_x<=x
        loc_dummy = minloc(abs(grid_x-x),mask=check)
        ilo = loc_dummy(1)

        check = grid_x>=x
        loc_dummy = minloc(abs(grid_x-x),mask=check)
        ihi = loc_dummy(1)   

    !Is this extrapolation downward?
    elseif (all(grid_x>x)) then

        if (present(flag)) flag = 1

        check = grid_x>x
        loc_dummy = minloc(abs(grid_x-x),mask=check)
        ilo = loc_dummy(1)

        check(ilo) = .FALSE.
        loc_dummy = minloc(abs(grid_x-x),mask=check)
        ihi = loc_dummy(1)    

    !Is this extrapolation upward
    elseif (all(grid_x<x)) then

        if (present(flag)) flag = 2

        check = grid_x<x
        loc_dummy = minloc(abs(grid_x-x),mask=check)
        ihi = loc_dummy(1)

        check(ihi) = .FALSE.
        loc_dummy = minloc(abs(grid_x-x),mask=check)
        ilo = loc_dummy(1)    
       
    else
        
        if (present(flag)) flag = 4

    end if

    !Get associated weights
    if (ilo/=-1 .AND. ihi /=-1) then

        dist = abs(grid_x(ilo)-grid_x(ihi))
        if(dist<=1d-40) then 
            mhi=0.d0
            mlo=1.d0
        else
            mhi=(x-grid_x(ilo))/dist
            mlo=1.d0-mhi
        end if

    end if

end if    


end subroutine interp_get_interval

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Take the interval [ilo,ihi] and use mass (mlo,mhi) to 
! linearly interpolate.  f = flo*mlo + fhi*mhi
elemental function interp_use_interval(flo,fhi,mlo,mhi) result(f)
    use mod_types
    implicit none

    real(dp) :: f
    real(dp), intent(in) :: flo, fhi, mlo, mhi

    f = flo*mlo + fhi*mhi

end function




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Given xj on monotonic increasing grid_x and xhat in [xj-1,xj] cup [xj, xj+1]
! with f defined on grid_x, finds fhat = f(xhat)  

! flag meaning
! 0: exited with no problem
! 1: wanted to extrapolate down, but this program doesn't extrapolate
! 2: wanted to extrapolate up, but this program doesn't extrapolate
! 3: case not found
pure subroutine interplocal(fhat,grid_x,xj_ind,xhat,f,len_x,flag) 
use mod_types
implicit none

real(dp), intent(out) :: fhat
integer, intent(in) :: len_x
real(dp), dimension(1:len_x), intent(in) :: grid_x,f
real(dp), intent(in) :: xhat
integer, intent(in) :: xj_ind
integer, intent(out), optional :: flag

real(dp) :: xj

xj = grid_x(xj_ind) 

if (abs(xhat-xj)<1d-12) then

    fhat = f(xj_ind)
    flag = 0

elseif (xj_ind>1 .AND. xj_ind<len_x) then
    
    !Interp up
    if (xhat>xj) then 
        fhat = f(xj_ind) + (f(xj_ind+1)-f(xj_ind))/(grid_x(xj_ind+1)-xj)*(xhat-xj)
    !Interp down
    elseif (xhat<xj) then
        fhat = f(xj_ind) + (f(xj_ind-1)-f(xj_ind))/(grid_x(xj_ind-1)-xj)*(xhat-xj)
    endif
    
    flag = 0
    
elseif (xj_ind==1 .AND. xj_ind<len_x) then

    if (xhat>xj) then
        fhat = f(xj_ind) + (f(xj_ind+1)-f(xj_ind))/(grid_x(xj_ind+1)-xj)*(xhat-xj)
        flag = 0
    elseif (xhat<xj) then
        !write(stdout,*) 'You want me to extrapolate down, but I wont' 
        fhat = f(xj_ind)
        flag = 1
    end if 


elseif (xj_ind>1 .AND. xj_ind==len_x) then

    if (xhat>xj) then 
        !write(stdout,*) 'You want me to extrapolate up, but I wont'
        fhat = f(xj_ind)
        flag = 2
    elseif (xhat<xj) then
        fhat = f(xj_ind) + (f(xj_ind-1)-f(xj_ind))/(grid_x(xj_ind-1)-xj)*(xhat-xj)
        flag = 0
    endif    
else
    !Case note found
    flag = 3
end if

end subroutine interplocal

end module mod_interp


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! pp - Piecewise polynomials
! This module attempts to mimic the Matlab routines ppmak and 
! ppval, but combines them in the following way.  
!
! An example of ppform is the following:
! P(x) = x**2-3x+1 if x in [1,2) or 3x-5 if x in [2,3]
! Note the intervals are assumed to be [,) except the last.
! Then rewrite P as
! (x-1)**2-(x-1)-1 and 3(x-2)+1
! Then breaks will be [1,2,3]
! Coefs = [1,-1,-1;0,3,1]
! pp = ppmak(breaks,coefs)
! y = ppval(pp,x)
! 
! Instead of returning structure pp, just substitute:
! y = ppval(breaks,coefs,x)
! Remember that coefs should be in descending order!
!
module mod_pp
implicit none
contains

pure function ppval_fn(breaks,coefs,xpt) result(ypt)
    use mod_types
    implicit none
    real(dp), dimension(1:), intent(in) :: breaks
    real(dp), dimension(1:,1:), intent(in) :: coefs
    real(dp), intent(in) :: xpt
    real(dp) :: ypt
    !local 
    real(dp), dimension(1:1) :: yvec
    yvec = ppval(breaks,coefs,(/xpt/))
    ypt = yvec(1)
end function ppval_fn

pure function ppval(breaks,coefs,grid_x) result(y)
    use mod_types
    implicit none
    real(dp), dimension(1:), intent(in) :: breaks
    real(dp), dimension(1:,1:), intent(in) :: coefs
    real(dp), dimension(1:), intent(in) :: grid_x
    real(dp), dimension(1:size(grid_x)) :: y
    !local
    integer :: x_ind,a_ind,order
    real(dp) :: a,sum,x
      
    !Cycle of values of x
    do x_ind = 1,size(grid_x)

        x = grid_x(x_ind)
        
        !Find interval in breaks [a,b) s.t. x
        call get_interval(a,a_ind,x)

        !Evaluate x in that interval y(x_ind) = coefs(3)*(x-a)**2 + coefs(2)*(x-a)**1 + coeffs(1)
        !sum = coefs(a_ind,size(coefs,2)) !0th order term, just a coefficient        
        sum = 0.d0
        do order = 0,size(coefs,2)-1
            sum = sum + coefs(a_ind,size(coefs,2)-order)*(x-a)**order !orderth order term
        end do
        y(x_ind) = sum
        
    end do 
    
contains

    !Get interval in breaks of form [a,b) (except the last interval [a,b]) with x in it
    pure subroutine get_interval(a,a_ind,x)
        implicit none
        real(dp), intent(out) :: a
        integer, intent(out) :: a_ind
        real(dp), intent(in) :: x
        !local
        integer :: breaks_ind
        
        if (x<breaks(1)) then
            a = breaks(1)
            a_ind = 1
            return
        elseif (x>=breaks(size(breaks))) then
            a = breaks(size(breaks))
            a_ind = size(breaks)
            return
        else        
            do breaks_ind = 1,size(breaks)-1
                if (x>=breaks(breaks_ind) .AND. x<breaks(breaks_ind+1)) then
                    a = breaks(breaks_ind)
                    a_ind = breaks_ind                
                    return
                end if
            end do
        end if
        
!        a = breaks(size(breaks))
!        a_ind = size(breaks)
        
!        if (x>a+1d-15) then 
!            write(stdout,*) 'Interpolation point', x    
!            write(stdout,*) 'Max of grid', a
!            STOP 'Asking me to extrapolate in ppval'
!        end if
        
    end subroutine get_interval
    
end function ppval
end module mod_pp


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Schumaker shape preserving interpolation
!
! Adapted from Judd (Numerical Methods) and Leonardo Rezende
!
! copyright 2001 Leonardo Rezende.
! This program can be freely copied, used and modified, as long as this
! copyright notice is kept in place. Please inform the author of any
! useful modifications at lrezende@stanford.edu.
module mod_schumaker
implicit none
contains

function schumaker_interp(x,f,xnew,fp_in) result(fnew)
use mod_types
use mod_pp
implicit none
real(dp), dimension(1:), intent(in) :: x
real(dp), dimension(1:), intent(in) :: f
real(dp), dimension(1:), intent(in) :: xnew
real(dp), dimension(1:), intent(in), optional :: fp_in
real(dp), dimension(1:size(xnew)) :: fnew

!local
real(dp), dimension(1:size(x)) :: fp
integer :: interval_ind, breaks_ind, len_x
real(dp) :: casi, A1, B1, C1, A2, B2, C2

real(dp), dimension(1:2*(size(x)-1)+1) :: breaks_max !2*(size(x)-1)+1 is the maximum number of breaks (2 for each interval + end pt)
real(dp), dimension(1:2*(size(x)-1)+1,1:3)  :: coefs_max

real(dp), dimension(:), allocatable :: breaks
real(dp), dimension(:,:), allocatable :: coefs

!Check inputs
if (size(f)/=size(x)) STOP 'using schumaker_interp with incompatible f,x pair'
if (present(fp_in)) then 
    if (size(fp_in)/=size(x)) STOP 'using schumaker_interp with incompatible f_in,x pair'
end if

!Get break and coeficient information
if (.NOT. present(fp_in)) then
    call schumaker_ppmak(breaks,coefs,x,f)
else
    call schumaker_ppmak(breaks,coefs,x,fp_in)
end if

!evaluate the polynomial using pp
fnew = ppval(breaks,coefs,xnew)

end function schumaker_interp

pure subroutine schumaker_ppmak(breaks,coefs,x,f,fp_in)
use mod_types
use mod_pp
implicit none
real(dp), dimension(1:), intent(in) :: x
real(dp), dimension(1:size(x)), intent(in) :: f
real(dp), dimension(1:size(x)), intent(in), optional :: fp_in

real(dp), dimension(:), allocatable, intent(out) :: breaks 
real(dp), dimension(:,:), allocatable, intent(out) :: coefs 

!local
real(dp), dimension(1:size(x)) :: fp
integer :: interval_ind, breaks_ind, len_x
real(dp) :: casi, A1, B1, C1, A2, B2, C2

real(dp), dimension(1:2*(size(x)-1)+1) :: breaks_max !2*(size(x)-1)+1 is the maximum number of breaks (2 for each interval + end pt)
real(dp), dimension(1:2*(size(x)-1)+1,1:3)  :: coefs_max

len_x = size(x)

!compute slopes if not present
if (.NOT. present(fp_in)) then
    call schumaker_slopes(fp,f,x,len_x)
else
    fp = fp_in
end if

!find break and coef information
breaks_ind = 1
do interval_ind = 1,len_x-1

    !check for knot in [x(i),x(i+1))
    casi = schumaker_casi(f(interval_ind),f(interval_ind+1),fp(interval_ind),fp(interval_ind+1),x(interval_ind),x(interval_ind+1))

    !get polynomial coefficients breaks
    call schumaker_poly(f(interval_ind),f(interval_ind+1),fp(interval_ind),fp(interval_ind+1),x(interval_ind),x(interval_ind+1),&
                        casi,A1,B1,C1,A2,B2,C2)

    !store lefthand side of [x(i),x(i+1)) in breaks and coeffs
    breaks_max(breaks_ind) = x(interval_ind)
    coefs_max(breaks_ind,1:3) = (/C1,B1,A1/) !stored in descending order
    breaks_ind = breaks_ind+1

    !Store a break and additional coefficients if placed a knot
    if (casi<x(interval_ind+1)) then
        breaks_max(breaks_ind) = casi
        coefs_max(breaks_ind,1:3) = (/C2,B2,A2/)
        breaks_ind = breaks_ind+1
    end if    

end do
breaks_max(breaks_ind) = x(len_x)
coefs_max(breaks_ind,1:3) = (/0.d0,0.d0,f(len_x)/) !There may be a problem with this

!Output the breaks and coefficients
allocate(breaks(1:breaks_ind),coefs(1:breaks_ind,1:3))
breaks = breaks_max(1:breaks_ind)
coefs = coefs_max(1:breaks_ind,1:3)

end subroutine schumaker_ppmak


pure function schumaker_casi(f1,f2,fp1,fp2,x1,x2) result(casi)
use mod_types
implicit none
real(dp), intent(in) :: f1,f2,fp1,fp2,x1,x2
real(dp) :: casi
!f is f(x)
!f is derivative of f(x)

!local
real(dp) :: d

if ((fp1+fp2)*(x2-x1) == 2.d0*(f2-f1)) then
    casi = x2
    return
else
    d = (f2-f1)/(x2-x1)
    if ((fp1-d)*(fp2-d)>=0.d0) then
        casi = .5d0*(x1+x2)
    elseif (abs(fp2-d)<abs(fp1-d)) then
        !casibar = x1 + 2*(x2-x1)*(fp2-d)/(fp2-fp1)
        !casi = .5d0*(x1+casibar)
        ! => casi = x1 + (x2-x1)*(fp2-d)/(fp2-fp1)
        casi = x1 + (x2-x1)*(fp2-d)/(fp2-fp1)
    else
        !casibar = x2 + 2*(x2-x1)*(fp1-d)/(fp2-fp1)
        !casi = .5d0*(x2+casibar)
        ! => casi = x2 + (x2-x1)*(fp1-d)/(fp2-fp1)        
        casi = x2 + (x2-x1)*(fp1-d)/(fp2-fp1)
    end if
end if

end function schumaker_casi

pure subroutine schumaker_poly(f1,f2,fp1,fp2,x1,x2,casi,A1,B1,C1,A2,B2,C2)
use mod_types
implicit none
real(dp), intent(in) :: f1,f2,fp1,fp2,x1,x2,casi
real(dp), intent(out) :: A1,B1,C1,A2,B2,C2
!local
real(dp) :: fpbar,a,b

a = casi-x1 !(alpha)
b = x2-casi !(beta)
fpbar = (2.d0*(f2-f1)-(a*fp1+b*fp2))/(x2-x1)

A1 = f1
B1 = fp1
C1 = (fpbar-fp1)/(2.d0*a)

A2 = A1+a*B1+(a**2)*C1
B2 = fpbar
!C2 = (fp2-fpbar)/(2.d0*b)
C2 = merge((fp2-fpbar)/(2.d0*b),(fp2-fpbar)/(2.d0*sqrt(huge(0.d0))),b/=0.d0)

end subroutine schumaker_poly

pure subroutine schumaker_slopes(fp,f,x,n)

use mod_types
implicit none
integer, intent(in) :: n
real(dp), dimension(1:n), intent(in) :: f,x
real(dp), dimension(1:n), intent(out) :: fp

!local
real(dp), dimension(1:n-1) :: L,d
real(dp), dimension(1:n-2) :: sb

L = sqrt((x(2:n)-x(1:n-1))**2 + (f(2:n)-f(1:n-1))**2) 
d = (f(2:n)-f(1:n-1))/(x(2:n)-x(1:n-1))
where (d(1:n-2)*d(2:n-1)>0.d0)
    sb = (L(1:n-2)*d(1:n-2)+L(2:n-1)*d(2:n-1))/(L(1:n-2)+L(2:n-1))
elsewhere
    sb = 0.d0
endwhere    
fp(1) = (3.d0*d(1)-sb(1))/2.d0
fp(2:n-1) = sb
fp(n) = (3.d0*d(n-1)-sb(n-2))/2.d0

end subroutine schumaker_slopes

end module mod_schumaker




!program test_interp2
!
!use mod_interp
!implicit none
!
!real(8), dimension(1:3,1:2) :: x,y,z
!real(8), dimension(1:4,1:4) :: xi,yi,zi
!integer :: xi_ind, yi_ind
! 
!real(8), dimension(1:2) :: x1vec,x2vec,x3vec
!real(8), dimension(1:2,1:2,1:2) :: v
!real(8) :: xi1,xi2,xi3
!integer :: i1,i2,i3
!
!x(:,1) = (/0.d0,.5d0,1.d0/)
!x(:,2) = x(:,1)
!
!y(1,:) = (/-1.d0, 4.d0/)
!y(2,:) = y(1,:)
!y(3,:) = y(1,:)
!
!!Mesh grid function z = x+y
!z = x+y
!
!xi(:,1) = (/-.5d0,1.d0,.25d0,1.25d0/)
!xi(:,2) = xi(:,1)
!xi(:,3) = xi(:,1)
!xi(:,4) = xi(:,1)
!
!yi = transpose(xi)
!
!zi = interp2_mesh(x,y,z,xi,yi)
!
!write(stdout,*) zi(1,:)
!write(stdout,*) zi(2,:)
!write(stdout,*) zi(3,:)
!write(stdout,*) zi(4,:)
!
!write(stdout,*) 'CHECK:'
!do xi_ind = 1,4
!do yi_ind = 1,4
!write(stdout,*) zi(xi_ind,yi_ind) - interp2_pt(x(:,1),y(1,:),z,xi(xi_ind,yi_ind),yi(xi_ind,yi_ind))
!end do
!end do
!
!
!!interp 3d
!x1vec=(/0.d0, 1.d0/)
!x2vec=(/-100.d0, 50.d0/)
!x3vec=(/-2.d0, -5.d0/)
!do i1=1,2
!    do i2=1,2
!        do i3=1,2
!            v(i1,i2,i3) = x1vec(i1) + 3*x2vec(i2) + x3vec(i3)**2
!        end do
!    end do
!end do       
!xi1 = .75d0
!xi2 = -99.d0
!xi3 = -5.d0
!
!write(stdout,*) interp3_pt(x1vec,x2vec,x3vec,v,xi1,xi2,xi3)
!
!end program
