module mod_root1dim

implicit none
contains

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! sub_bisect
!
! Uses bisection to find a root f(x) = 0 in a bracket [a,b]
! f should be defined as in the below interface block as just f(x)
!
! Optional arguments are the evaluation of f at a AND or b.  Alternatively,
! just the correct sign of fa AND fb can be passed.
!
subroutine sub_bisect(x_out,f,a_in,b_in,toler_in,maxiter_in,fa_in,fb_in,xtoler_in,printmod_in)
use mod_types
implicit none
real(dp), intent(out) :: x_out
interface
    function f(x)
    use mod_types
    implicit none
    real(dp), intent(in) :: x
    real(dp) :: f
    end function f
end interface
real(dp), intent(in) :: a_in,b_in
real(dp), intent(in), optional :: toler_in, fa_in,fb_in, xtoler_in
integer, intent(in), optional :: maxiter_in, printmod_in

!local
real(dp) :: a,b,x,fa,fb,fx,res,toler,xtoler
logical :: fapos, faneg, case1,case2,case3,case4,case5
integer :: maxiter,printmod, iter

a = a_in
b = b_in

!default values
printmod = 1
xtoler = 1.d-9
toler = 1.d-6
maxiter = 10000
if (present(printmod_in)) printmod = printmod_in
if (present(xtoler_in)) xtoler = xtoler_in
if (present(toler_in)) toler = toler_in
if (present(maxiter_in)) maxiter = maxiter_in

if (present(fa_in)) then 
    fa = fa_in
else
    fa = f(a)
end if
if (present(fb_in)) then 
    fb = fb_in
else
    fb = f(b)
end if

if ((fa>0.d0 .AND. fb>0.d0) .OR. (fa<0.d0 .AND. fb<0.d0)) then
    !if (printmod/=-1) write(stdout,*) 'Error: zero not bracketed', fa, fb
    write(stdout,*) 'sub_bisect: zero not bracketed:', fa, fb
    x_out = merge(a,b,abs(fa)<abs(fb))
    return
end if

!Main routine
fapos = fa>=0.d0
faneg = fa<=0.d0
res = 1d10
if (printmod/=-1) write(stdout,"('|~~~~~~~~~~~~~~Beginning Bisection Algorithm~~~~~~~~~~~~~~~~~~|')")
if (printmod/=-1) write(stdout,"('| iter,            a,            b,            x,           fx|')")
iter = 1
do while(res>toler)

    x = .5d0*a + .5d0*b
    if (abs(x-a)<=xtoler) then 
        write(stdout,*) 'Warning: xtoler exceeded'
        exit
    end if
    fx = f(x)
    
    if (printmod/=-1) write(stdout,"('|',i5,4e14.5,'|')") iter,a,b,x,fx

    case1 = fx<0.d0 .AND. faneg
    case2 = fx>0.d0 .AND. faneg
    case3 = fx<0.d0 .AND. fapos
    case4 = fx>0.d0 .AND. fapos
    case5 = fx==0.d0

    if (case1) then
        fa = fx
        a = x
    elseif (case2) then
        fb = fx
        b = x
    elseif (case3) then
        fb = fx
        b = x
    elseif (case4) then
        fa = fx
        a = x
    elseif (case5) then
        exit
    end if
     
    res = abs(fx)
    if (res<=toler) then 
        if (printmod/=-1) write(stdout,"('|',i5,4e14.5,'|')") 0, 0.d0, 0.d0, x, fx
        if (printmod/=-1) write(stdout,"('|~~~~~~~~~~~~~~Converged Successfully~~~~~~~~~~~~~~~~~~~~~~~~~|')")
    end if

    iter = iter+1
end do

x_out = x
if (printmod/=-1) write(stdout,"('|~~~~~~~~~~~~~~End Bisection Algorithm~~~~~~~~~~~~~~~~~~~~~~~~|')")

end subroutine sub_bisect


!Adapted from NR Fortran 77
subroutine sub_brent(x_out,f,a_in,b_in,toler_in,maxiter_in,fa_in,fb_in,xtoler_in,printmod_in,c_in,fc_in)
    use mod_types
    implicit none    
    real(dp), intent(out) :: x_out
    interface
        function f(x)
        use mod_types
        implicit none
        real(dp), intent(in) :: x
        real(dp) :: f
        end function f
    end interface
    real(dp), intent(in) :: a_in,b_in
    real(dp), intent(in), optional :: c_in,toler_in,fa_in,fb_in,fc_in,xtoler_in
    integer, intent(in), optional :: maxiter_in, printmod_in
    !local
    real(dp), parameter :: eps = epsilon(0.d0) !1.d-14 !3.d-8 !Machine floating-point precision. 1d-15??
    integer :: maxiter
    real(dp) :: a,b,c,d,e,fa,fb,fc,p,q,r,s,tol1,xm,zbrent
    real(dp) :: toler,xtoler 
    integer :: iter,printmod

    !set defaults    
    maxiter = 1000
    toler = 1.d-4
    xtoler = 1d-7
    printmod = 1
    
    !override defaults if present
    if (present(toler_in)) toler = toler_in
    if (present(xtoler_in)) xtoler = xtoler_in
    if (present(maxiter_in)) maxiter = maxiter_in    
    if (present(printmod_in)) printmod = printmod_in
    
    if (printmod/=-1) write(stdout,"('|~~~~~~~~~~~~~~Beginning Brent''s Method Algorithm~~~~~~~~~~~~~|')")
    if (printmod/=-1) write(stdout,"('|~~~~~~~~~~~~~~Calculating value at endpoints...~~~~~~~~~~~~~~|')")

    a=a_in
    b=b_in
    
    fa=f(a)
    fb=f(b)

    if (printmod/=-1) write(stdout,"('|~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~...Finished~~~~~~~~~~~~~~~|')")

    if((fa>0.d0 .AND. fb>0.d0).or.(fa<0.d0 .AND. fb<0.d0)) then
        write(stdout,*) 'sub_brent: root is not bracketed, returing best endpoint'
        x_out = merge(a,b,abs(fa)<abs(fb))         
        return
    end if

    c = b
    fc = fb

!    if (present(c_in)) then
!        c = c_in
!        if (present(fc_in)) then
!            fc = fc_in
!        else
!            fc = f(c)
!        end if    
!    end if
    
    !main solver
    if (printmod/=-1) write(stdout,"('|~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~|')")
    if (printmod/=-1) write(stdout,"('| iter,            a,            b,            c,           fb|')")
    do iter=1,maxiter
        if((fb>0.d0 .AND. fc>0.d0).or.(fb<0.d0 .AND. fc<0.d0)) then
            c=a !Rename a, b, c AND adjust bounding interval d.
            fc=fa
            d=b-a
            e=d
        endif
        if(abs(fc)<abs(fb)) then
            a=b
            b=c
            c=a
            fa=fb
            fb=fc
            fc=fa
        endif
        
        if (printmod/=-1) write(stdout,"('|',i5,4e14.5,'|')") iter,a,b,c,fb
        
        tol1=2.d0*EPS*abs(b)+0.5d0*toler !Convergence check.
        xm=.5d0*(c-b)
        if(abs(xm)<=tol1 .or. fb==0.d0)then
            zbrent=b
            exit !return
        endif
        if(abs(e)>=tol1 .AND. abs(fa)>abs(fb)) then
            s=fb/fa !Attempt inverse quadratic interpolation.
            if(a==c) then
                p=2.d0*xm*s
                q=1.d0-s
            else
                q=fa/fc
                r=fb/fc
                p=s*(2.d0*xm*q*(q-r)-(b-a)*(r-1.d0))
                q=(q-1.d0)*(r-1.d0)*(s-1.d0)
            endif
            if(p>0.d0) q=-q !Check whether in bounds.
            p=abs(p)
            if(2.d0*p < min(3.d0*xm*q-abs(tol1*q),abs(e*q))) then
                e=d !Accept interpolation.
                d=p/q
            else
                d=xm !Interpolation failed, use bisection.
                e=d
            endif
        else !Bounds decreasing too slowly, use bisection.
            d=xm
            e=d
        endif
        a=b !Move last best guess to a.
        fa=fb
        if(abs(d) > tol1) then !Evaluate new trial root.
            b=b+d
        else
            b=b+sign(tol1,xm)
        endif
        fb=f(b)
    enddo

    if (printmod/=-1) write(stdout,"('|',i5,4e14.5,'|')") 0,0.d0, b, 0.d0, fb
    if (printmod/=-1 .AND. iter>=maxiter) write(stdout,*) 'sub_brent: Max No of Iterations Exceeded'
    if (printmod/=-1 .AND. iter<maxiter .AND. abs(fb)<toler) &
                      write(stdout,"('|~~~~~~~~~~~~~~Converged Successfully~~~~~~~~~~~~~~~~~~~~~~~~~|')")
    if (printmod/=-1) write(stdout,"('|~~~~~~~~~~~~~~End Brent''s Method~~~~~~~~~~~~~~~~~~~~~~~~~~~~~|')")
    
    zbrent=b   
    x_out = zbrent
    
end subroutine sub_brent

end module mod_root1dim


!Sample Output for f = (x+3.d0)*(x-1.d0)**8 + .1d0

!|~~~~~~~~~~~~~~Beginning Brent's Method Algorithm~~~~~~~~~~~~~|
!| iter,            a,            b,            c,           fb|
!|    1  -0.40000E+01   0.13333E+01  -0.40000E+01   0.10066E+00|
!|    2   0.13333E+01   0.13333E+01  -0.40000E+01   0.10066E+00|
!|    3   0.13333E+01  -0.13333E+01  -0.40000E+01   0.14645E+04|
!|    4  -0.13333E+01  -0.26667E+01  -0.40000E+01   0.10891E+05|
!|    5  -0.33333E+01  -0.26667E+01  -0.33333E+01   0.10891E+05|
!|    6  -0.26667E+01  -0.28054E+01  -0.33333E+01   0.85576E+04|
!|    7  -0.28054E+01  -0.30694E+01  -0.28054E+01  -0.52162E+04|
!|    8  -0.30694E+01  -0.29694E+01  -0.30694E+01   0.18860E+04|
!|    9  -0.29694E+01  -0.29959E+01  -0.30694E+01   0.26361E+03|
!|   10  -0.29959E+01  -0.30001E+01  -0.29959E+01  -0.33224E+01|
!|   11  -0.30001E+01  -0.30000E+01  -0.30001E+01   0.27065E-01|
!|   12  -0.30000E+01  -0.30000E+01  -0.30001E+01   0.27437E-05|
!|   13  -0.30000E+01  -0.30000E+01  -0.30001E+01   0.26193E-11|
!|   14  -0.30000E+01  -0.30000E+01  -0.30000E+01   0.26193E-11|
!|    0   0.00000E+00  -0.30000E+01   0.00000E+00   0.26193E-11|
!|~~~~~~~~~~~~~~Converged Successfully~~~~~~~~~~~~~~~~~~~~~~~~~|
!|~~~~~~~~~~~~~~End Brent's Method~~~~~~~~~~~~~~~~~~~~~~~~~~~~~|
!  
!|~~~~~~~~~~~~~~Beginning Bisection Algorithm~~~~~~~~~~~~~~~~~~|
!| iter,            a,            b,            x,           fx|
!|    1  -0.40000E+01   0.13333E+01  -0.13333E+01   0.14645E+04|
!|    2  -0.40000E+01  -0.13333E+01  -0.26667E+01   0.10891E+05|
!|    3  -0.40000E+01  -0.26667E+01  -0.33333E+01  -0.41443E+05|
!|    4  -0.33333E+01  -0.26667E+01  -0.30000E+01   0.10000E+00|
!|    5  -0.33333E+01  -0.30000E+01  -0.31667E+01  -0.15141E+05|
!|    6  -0.31667E+01  -0.30000E+01  -0.30833E+01  -0.64407E+04|
!|    7  -0.30833E+01  -0.30000E+01  -0.30417E+01  -0.29666E+04|
!|    8  -0.30417E+01  -0.30000E+01  -0.30208E+01  -0.14232E+04|
!|    9  -0.30208E+01  -0.30000E+01  -0.30104E+01  -0.69692E+03|
!|   10  -0.30104E+01  -0.30000E+01  -0.30052E+01  -0.34481E+03|
!|   11  -0.30052E+01  -0.30000E+01  -0.30026E+01  -0.17146E+03|
!|   12  -0.30026E+01  -0.30000E+01  -0.30013E+01  -0.85456E+02|
!|   13  -0.30013E+01  -0.30000E+01  -0.30007E+01  -0.42622E+02|
!|   14  -0.30007E+01  -0.30000E+01  -0.30003E+01  -0.21247E+02|
!|   15  -0.30003E+01  -0.30000E+01  -0.30002E+01  -0.10570E+02|
!|   16  -0.30002E+01  -0.30000E+01  -0.30001E+01  -0.52342E+01|
!|   17  -0.30001E+01  -0.30000E+01  -0.30000E+01  -0.25669E+01|
!|   18  -0.30000E+01  -0.30000E+01  -0.30000E+01  -0.12334E+01|
!|   19  -0.30000E+01  -0.30000E+01  -0.30000E+01  -0.56668E+00|
!|   20  -0.30000E+01  -0.30000E+01  -0.30000E+01  -0.23334E+00|
!|   21  -0.30000E+01  -0.30000E+01  -0.30000E+01  -0.66668E-01|
!|   22  -0.30000E+01  -0.30000E+01  -0.30000E+01   0.16666E-01|
!|   23  -0.30000E+01  -0.30000E+01  -0.30000E+01  -0.25000E-01|
!|   24  -0.30000E+01  -0.30000E+01  -0.30000E+01  -0.41670E-02|
!|   25  -0.30000E+01  -0.30000E+01  -0.30000E+01   0.62497E-02|
!|   26  -0.30000E+01  -0.30000E+01  -0.30000E+01   0.10414E-02|
!|   27  -0.30000E+01  -0.30000E+01  -0.30000E+01  -0.15628E-02|
!|   28  -0.30000E+01  -0.30000E+01  -0.30000E+01  -0.26072E-03|
!|   29  -0.30000E+01  -0.30000E+01  -0.30000E+01   0.39032E-03|
!|   30  -0.30000E+01  -0.30000E+01  -0.30000E+01   0.64799E-04|
!|   31  -0.30000E+01  -0.30000E+01  -0.30000E+01  -0.97962E-04|
!|   32  -0.30000E+01  -0.30000E+01  -0.30000E+01  -0.16581E-04|
! Warning: xtoler exceeded
!|~~~~~~~~~~~~~~End Bisection Algorithm~~~~~~~~~~~~~~~~~~~~~~~~|