% [x w] = myquad(d,mean,varcov,method,n)
%
% Provides quadrature nodes (x) and weights (w) that can be used to 
% integrate a function f of dim d with respect to a multivariate normal
% distribution N(mean,varcov) using monomial rules taken from Stroud. If d
% equals 1, there are also several non-monomial rules available. For
% a single dimension, i.e. d=1, and a non-monimal rule, n controls the 
% number of quadrature points. 
%
% There are several methods available
% 
% Method name      Description
% 'deg3B'          A degree 3 rule using spread out equally weighted points 
% 'deg5'           A degree 5 rule using 2d^2+1 pts at various distances
% 'gauss-hermite'  A nonmonomial rule with degree n (specified by the user).
%                  
function [x w] = myquad(d,mean,varcov,method,n)

% If mean is a scalar, adjust it to a vector
if isscalar(mean)
    mean = repmat(mean,[d 1]);
end

% Construct the cholesky decomposition of varcov
Omega = chol(varcov)';
absdetOmega = abs(det(Omega));

% Get the weights for standard form
if strcmp(method,'deg3B')
    [x w] = int_deg3ruleB(d);
elseif strcmp(method,'deg5')
    [x w] = int_deg5rule(d);
elseif strcmp(method,'gauss-hermite')
    [x w] = gaussHermite(n);
%     y = sqrt(2)*sqrt(varcov)*x
%     ww = w*pi^(-.5)
else
    error('Method not found.  Remember it''s case sensitive.')
end

% Convert to non-standard form
% x = sqrt(2)*Omega*x' + repmat(mean,[1,size(x,1)]);
x = sqrt(2)*x*Omega' + repmat(mean,[1,size(x,1)])';
w = w*(2^(d/2))*absdetOmega*1/sqrt(abs(det(varcov)))*(2*pi)^(-d/2);

% Do an error check to make sure that that E(sum(x)) = sum(mean); and
% that if f(x) = 1, that E(f(x)) = 1; and that E(sum(x^2)) = sum(diag(varcov))
if abs(w'*sum(x,2) - sum(mean))>1d-10
    error('Error in quadrature formula: E(sum(x))~=sum(mean)')
elseif abs(sum(w)-1)>1d-10
    error('Error in quadrature formula: E(1)~=1')
elseif abs(w'*sum(x.^2,2) - sum(diag(varcov)))>1d-10
    error('Error in quadrature formula: E(sum(x^2))~=sum(variances)')
end
    
% 
% 'should be 1:'
% sum(w)
% 
% disp({'should be ',sum(diag(varcov)),':'})
% w'*sum(x.^2,2)

end 

% Provides quadrature weights for the following integral in standard form 
% \int^\infty_{-\infty} f(x) e^{-x^2} dx.
function [x w] = gaussHermite(n)
    if (n==2)
        x1 = .7071067811;        
        w1 = .8862269254;        
        x = [x1;-x1];
        w = [w1;w1];
    elseif (n==10)
        x1 = .3436159118837738d+1;
        x2 = .2532731674232790d+1;
        x3 = .1756683649299882d+1;
        x4 = .1036610829789514d+1;
        x5 = .342901327223705;
        w1 = .7640432855233d-5;
        w2 = .1343645746781d-2;
        w3 = .3387439445548d-1;
        w4 = .2401386110823;
        w5 = .6108626337353;
        
        x = [x1;-x1;x2;-x2;x3;-x3;x4;-x4;x5;-x5];
        w = [w1;w1;w2;w2;w3;w3;w4;w4;w5;w5];
    elseif (n==17)
        x1 = 0.0000000000000;
        x2 = 0.5316330013427;
        x3 = 1.0676487257435;
        x4 = 1.6129243142212;
        x5 = 2.1735028266666;
        x6 = 2.7577629157039;
        x7 = 3.3789320911415;
        x8 = 4.0619466758755;
        x9 = 4.8713451936744;
        
        w1 = 0.5309179376249;
        w2 = 0.4018264694704;
        w3 = 0.1726482976701;
        w4 = 0.4092003414976d-1;
        w5 = 0.5067349957628d-2;
        w6 = 0.2986432866978d-3;
        w7 = 0.7112289140021d-5;
        w8 = 0.4977078981631d-7;
        w9 = 0.4580578930799d-10;
        
        x = [x1;x2;-x2;x3;-x3;x4;-x4;x5;-x5;-x6;x6;-x7;x7;-x8;x8;-x9;x9];
        w = [w1;w2;w2;w3;w3;w4;w4;w5;w5;w6;w6;w7;w7;w8;w8;w9;w9];
    end
end

function [x w] = int_deg3ruleB(d)
    r = sqrt(d/2);    
    x = [diag(repmat(r,[d 1]));diag(repmat(-r,[d 1]))];    

    V = pi^(d/2);
%     E = V/(2*d)*sum(f);
    w = repmat(V/(2*d),[2*d 1]);
end

%Integrate \int_{R^d} f(x) e^{-\sum^d_{i=1} x^2_i dx}
%using a degree 5 rule as in Judd p.275.
function [x w] = int_deg5rule(dint)

    x = NaN(2*dint.^2+1,dint); %Values of x that need to be evaluated
    
    %Variable definitions using monomial approach due to Stroud (1971) and found in Judd
    %integer, parameter :: dint = 4 %underscore J is for Judd

    N = 2*dint.^2+1; %Number of points used in formula
    d = dint;
    r = sqrt(1+.5d0*d);
    s = sqrt(.5d0+.25d0*d);
    
    %x = 0            
    x(1,:) = 0;
    % x= r*ei
    x(2:2+d-1,:) = diag(repmat(r,[d 1]));
    % x= -r*ei
    x(2+d:2+d+d-1,:) = diag(repmat(-r,[d 1]));

    % x = +-s*ei +-s*ej
    ind = 2+d+d;
    for i = 1:d-1
        for j = i+1:d
            tmpPP = zeros([1 d]);
            tmpPP(i) = s;
            tmpPP(j) = s;
            tmpPM = zeros([1 d]);
            tmpPM(i) = s;
            tmpPM(j) = -s;
            tmpMP = zeros([1 d]);
            tmpMP(i) = -s;
            tmpMP(j) = s;
            tmpMM = zeros([1 d]);
            tmpMM(i) = -s;
            tmpMM(j) = -s;
            x(ind:ind+3,:) = [tmpPP;tmpPM;tmpMP;tmpMM];
            ind = ind+4;
        end
    end

    N = 2*dint.^2+1;
    d = dint;
    V = (pi).^(d/2.d0);
    A = 2.d0/(d + 2.d0)*V;
    B = (4.d0-d)/(2.d0*(d+2.d0).^2)*V;
    C = V/(d+2.d0).^2;

    w = NaN(N,1);
    w(1) = A;
    w(2:2+d-1) = B;
    w(2+d:2+d+d-1) = B;
    w(2+d+d:N) = C;
%     E = (A*fx(1)+ B*sum(fx(2:2+d-1)+fx(2+d:2+d+d-1)) + C*sum(fx(2+d+d:size(fx,1))));
    %E = (A*fx(1)+ B*sum(fx(2:5)+fx(6:9)) + C*sum(fx(10:33)));

    
end