% [grad hess f0] = numGradHess(f,x0,varargin)
% 
% Computes the gradiant and hessian numerically. The hessian is only 
% computed if desired. If the hessian is computed, then f(x0) is also
% returned.
% 
% The step size is set according to a rule of thumb and changes from 
% the first to second derivative calculations. However, it can be over-
% ridden by varargin{1}.
% 
function [grad hess f0] = numGradHess(f,x0,varargin)

    % Check that x0 is vector
    if (~isvector(x0)) 
        error('x0 must be a vector')
    end

    h = max(min(abs(x0)),1)*(eps(1.d0)).^(1/3);

    yesHess = nargout>1;
    
    if (size(varargin)>=1)
        if ~isempty(varargin{1})
            h = varargin{1};
        end
    end


    n = numel(x0);
    
    x0 = x0(:);
    
    % Compute the gradient via central differencing    
    f0phi = NaN(n,1);
    f0mhi = NaN(n,1);
    for i = 1:n
        e = zeros(n,1);
        e(i) = h;
        f0phi(i) = f(x0 + e);
        f0mhi(i) = f(x0 - e);
    end

    grad = (f0phi - f0mhi)./(2*h);
    
    % If desired, compute the hessian
    
    h = max(min(abs(x0)),1)*(eps(1.d0)).^(1/4);
    if (yesHess) 
        
        % Compute and store the value at x0 (which is used multiple times)
        f0 = f(x0);
                
        % Compute the cross partials d^2f/dxidxj
        hess = zeros(n,n);
        for i = 1:n
            for j = i:n
                ei = zeros(n,1);
                ej = zeros(n,1);
                ei(i) = h;
                ej(j) = h;
                if (i~=j) 
                    hess(i,j) = (f(x0+ei+ej)+f(x0-ei-ej)-f(x0+ei-ej)-f(x0-ei+ej))./(4*h^2);
                else
                    hess(i,j) = (f(x0+2*ei)+f(x0-2*ei)-2*f0)./(4*h^2);
                end
            end
        end
        
    end
    
    for i = 1:n-1
        hess = hess + diag(diag(hess,i),-i);
    end
    
end

function [] = testNumGradHess()

x0 = ones(2,1); 


f = @(x) x(1).^2 + x(2).^2 - 6*x(1)*x(2);
[g H]=numGradHess(f,x0)

    
a = 2;
f = @(x) (1-x(1)).^2 + a*(x(1)-x(2).^2).^2; 
gf = @(x) [-2*x(1)*(1-x(1)) + a*2*(x(1)-x(2).^2) ; a*2*(x(1)-x(2).^2)*(-x(2))*2];
% H11f = @(x) (-2 + a*2); 
% H22f = @(x)  -4*a*( x(1)-3*x(2).^2 );
% H12f = @(x) 2*a*(-2*x(2));
% Hf = @(x) [H11f(x),H12f(x);H12f(x),H22f(x)];

[g H]=numGradHess(f,x0)

g - gf(x0)
H - Hf(x0)

end