function [x s] = smolyakapprox_step1(d,mu,a,b)
%smolyakapprox_step1 Smolyak Grid Construction 
%   [x s] = smolyakapprox_step1(d,mu,a,b) constructs the Smolyak
%   grid points for a hypercube in R^d defined by bounds a,b with level of 
%   approximation mu.  It outputs two elements, x and s.  x is
%   the collocation points that your function should be evaluated.
%   s is a structure containing information used by
%   smolyakapprox_step2 and smolyakapprox_step3.
% 
%   Grey Gordon 2011.  
%   "But God chose the foolish things of the world to shame the wise; 
%    God chose the weak things of the world to shame the strong." -1Cor12:47

%   This program is free software: you can redistribute it and/or modify
%   it under the terms of the GNU General Public License as published by
%   the Free Software Foundation, either version 3 of the License, or
%   (at your option) any later version.
% 
%   This program is distributed in the hope that it will be useful,
%   but WITHOUT ANY WARRANTY; without even the implied warranty of
%   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%   GNU General Public License for more details.
% 
%   You should have received a copy of the GNU General Public License
%   along with this program.  If not, see <http://www.gnu.org/licenses/>.

%   Revision History
%   Date: 12/09/10
%   Modified: 2/26/11
%   Modified: 3/29/11 added switch to allow for Fortran compatibility.  
%   Modified: 6/2/11 to remove the switch

    q = max(d,mu+1);        
    
    s.q = q;
    s.d = d;
    s.mu = mu;
    s.a = a;
    s.b = b;
    s.M_mup1 = smolyakM(mu+1); %This constant is used in step3
    
    if (length(a)~=d), error('length of a is longer than dim d'), end
    if (length(b)~=d), error('length of b is longer than dim d'), end
    if (any(b<=a)), error('b is less than or equal to a'), end
    
    %Construct necessary coefficients using fx values.
    % First, enumerate all necessary theta
    % -For all i satisfying q<=ibar<=d+mu, 
    % -need { l | l(1) is in 1...m(i1), l(2) is in 1...m(i1)}
    
    %Enumerate all such i, then all such m(i), then all such l
    % Determine all i that fit the criterion q<=|i|<=d+mu
%     tmp=[];
%     for ibar = q:d+mu
%         enum = 1 + smolyakEnumerate(d,ibar-d);
%         tmp = [enum;tmp];
%     end
%     s.i = unique(tmp,'rows');
    
    tmp = [];
    for ibar = q:d+mu
%         enum = 1 + spgetseq(ibar-d,d);
        enum = smolyakEnumerate2(d,ibar);
        tmp = [enum;tmp];
    end
    s.i = sortrows(tmp); % here just to ensure compatibility with old solutions
        
    
    % # of i combinations
    leni = size(s.i,1);
    s.leni = leni;
        
    % store |i|  
    s.ibar = sum(s.i,2);
    
    % compute k = 2^(i-1)+1
    k = smolyakM(s.i);
    
    % Lb(i) and Ub(i) give the starting and ending range for 
    % all the cartesian products of G(m(i1))x...xG(m(id))
    s.Lb = NaN(s.leni,1);
    s.Ub = NaN(s.leni,1);    
    tmpInt = 0;
    for iind = 1:leni        
        s.Lb(iind) = tmpInt + 1;
        tmpInt = tmpInt + prod(k(iind,:));
        s.Ub(iind) = tmpInt;
    end

    % Preallocate
    z = NaN(s.Ub(leni),d); % now longer saved in the s structure
    s.j = NaN(s.Ub(leni),d);
    
    % Compute the cartesian products of the points, both their indices
    % and their values
    %%% NOTE: instead of the cartesian product, I should just have a
    %%% routine that enumerate all possible combinations of indices. Then
    %%% get the corresponding numeric values from the knots.
    %%%
    %%% NEW WAY:    
    %%%%% Get indices
    for iind = 1:leni
        s.j(s.Lb(iind):s.Ub(iind),:) = cartProdInt(k(iind,:));
    end
    %%%%%% With indices, compute corresponding knots. 
    for i = 1:max(k(:)) % max number of extrema anywhere 
        Gtmp{i} = smolyakG(i); % extrema for all possible cases
    end    
    for iind = 1:leni
        for dind = 1:d
            z(s.Lb(iind):s.Ub(iind),dind) = Gtmp{k(iind,dind)}(s.j(s.Lb(iind):s.Ub(iind),dind));
        end
    end

    % Get the smolyak grid points (b/c of their nested nature, only return unique values)
    [x s.redo s.undo] = unique(z,'rows');

    % Convert the unit cube to the desired cube by linear transform
    for dind = 1:d
        x(:,dind) = (x(:,dind)+1.d0)*(b(dind)-a(dind))/2.d0 + a(dind);
    end    
        
    % Allocate space 
%     s.l = s.j; % l and j have the same form but are used differently
%%% I NOW CHANGED ALL THE REFERENCES TO S.L TO S.J
    
    % Evalute the Chebyshev polynomials at the extrema
    % NOTE: WE CAN DERIVE THE ANALYTICAL FORMULA WHICH, IN KRUEGER'S 
    % NOTATION IS JUST T_l(xsi^n_j) = (-1)^(l-1)*cos((l-1)*(j-1)/(n-1)*pi)
%     maxErrorFnd = 0;
    s.T = NaN(s.Ub(leni),max(s.Ub-s.Lb)+1);
%     s.T = sparse(s.Ub(leni),max(s.Ub-s.Lb)+1);
    for iind = 1:leni
        for lind = s.Lb(iind):s.Ub(iind)
            for jind = s.Lb(iind):s.Ub(iind)
                tmpProd = 1.d0;
                for dind = 1:d

                    % OLD WAY:
                    %%% Evaluates the Chebyshev polynomial of order
                    %%% s.l(lind,dind)-1 at s.z(jind,dind).   
                    %%% s.z() is the Cartezian product of the extrema of 
                    %
                    % NEW WAY:
                    % Computes T(z) where T has order (in the usual sense)
                    % l-1 and z is the jth extrema (out of k=m(i)).
                    if (k(iind,dind)>1) 
                        if (mod(s.j(lind,dind),2))
                            exact = cos(pi*(s.j(lind,dind)-1)*(s.j(jind,dind)-1)/(k(iind,dind)-1));
                        else
                            exact = -cos(pi*(s.j(lind,dind)-1)*(s.j(jind,dind)-1)/(k(iind,dind)-1));
                        end
                        tmpProd = tmpProd*exact;
                    end
                end
                s.T(jind,lind-s.Lb(iind)+1) = tmpProd;
            end
        end       
    end        

    s.cjprod = NaN(length(s.undo),1);
    for iind = 1:leni
        for lind = s.Lb(iind):s.Ub(iind)
            s.cjprod(lind) = 1.d0;
            for dind = 1:d
                % Dimensions with only one dimensions are "dropped", that
                % is they're coefficient in the tensor product is just 1
                if (k(iind,dind)>1) 
                    % The product is 2^(l==1 or l==k)
                    if ((s.j(lind,dind)==1) || (s.j(lind,dind)==k(iind,dind))) 
                        s.cjprod(lind) = s.cjprod(lind)*2.d0;
                    end 
                end 
            end 
        end 
    end 

    % cjprod is used differently but has the same values as clprod 
%     s.cjprod = s.clprod;

    % for each i, there is a unique "const"
    s.const = NaN(leni,1);
    for iind = 1:leni
        tmpProd = 1.d0;
        tmpSum = 0.d0;
        for dind = 1:d
            if (k(iind,dind)>1) 
                tmpSum = tmpSum + 1.d0;
                tmpProd = tmpProd*(k(iind,dind)-1);
            end
        end
        s.const(iind) = (2.d0^tmpSum)/tmpProd;
    end
    
    % Store the "choose" constant values
    s.constVec = NaN(d+mu+1-q,1);
    for ibar = q:d+mu
        s.constVec(ibar-q+1) = (-1.d0)^(d+mu-ibar)*choose(d-1,d+mu-ibar);
    end

    % For each dind, find the location of the all nonzero j(:,dind)>1 elements
    % This saves a large amount of "if" statements in step3
    
    % Non-adaptive smolyak treats all dimensions equally, so can easily get
    % count of dimensions
%     s.j_gt_1 = NaN(sum(s.j(:,1)>1),d); 
%     for dind = 1:d
%         s.j_gt_1(:,dind) = find(s.j(:,dind)>1);
%     end
%     for dind = 1:d
%         s.j_eq_2(:,dind) = find(s.j(:,dind)==2); 
%         s.j_eq_3(:,dind) = find(s.j(:,dind)==3); 
%         s.j_eq_4(:,dind) = find(s.j(:,dind)==4); 
%         s.j_eq_5(:,dind) = find(s.j(:,dind)==5); 
%         s.j_gt_5(:,dind) = find(s.j(:,dind)>5); 
%     end
%     s.j_gt_1 = s.j>1;
    
end

% Given an integer i, deliver the function m(i) = 2^(i-1) + 1
function o = smolyakM(i)        
    o = NaN(size(i));
    if sum(i<1)>0
        error('i must be >= 1')
    else
        o(i==1) = 1;
        o(i>1) =  2.^(i(i>1)-1) + 1;
    end   
end

% Given an integer n, construct the set of n extrema of the Chebyshev 
% polynomials zeta_j = -cos(pi*(j-1)/(n-1)), j = 1,n
function [set] = smolyakG(n)
    if (n<1) 
        error('smolyakG: n must be >=1')
    elseif (n==1)
        set = 0;
    else       
        j = (1:n)';
        set = -cos(pi*(j-1)/(n-1));
        set(abs(set)<1d-12) = 0;
        set(abs(set-1.d0)<1d-12) = 1;
        set(abs(set+1.d0)<1d-12) = -1;
    end
end