
function feval = smolyakapprox_step3(s,x)
%smolyakapprox_step3 Smolyak Polynomial Evaluation
%   [feval] = smolyakapprox_step3(s,x)
%   Given pol which defines an approximation to f, compute the approx
%   value of f at the pts "x".  For performance, it also returns a struct
%   called work that can be used on subsequent calls to improve performance.
% 
%   Grey Gordon 2011.  
%   "But God chose the foolish things of the world to shame the wise; 
%    God chose the weak things of the world to shame the strong." -1Cor12:47

%   This program is free software: you can redistribute it and/or modify
%   it under the terms of the GNU General Public License as published by
%   the Free Software Foundation, either version 3 of the License, or
%   (at your option) any later version.
% 
%   This program is distributed in the hope that it will be useful,
%   but WITHOUT ANY WARRANTY; without even the implied warranty of
%   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%   GNU General Public License for more details.
% 
%   You should have received a copy of the GNU General Public License
%   along with this program.  If not, see <http://www.gnu.org/licenses/>.

    d = s.d;

    if (d~=size(x,2))
        disp({'d is ',d,' but size(x,2) is',size(x,2)})
        error('smolyakapprox_step3: x is of wrong dimension')
    end
    
    % The smolyak approximation can be written as 
    % \sum_{q\le |i| \le d+mu} (-1)^{d+mu-|i|}choose(d-1,d+\mu-|i|)p^i(x)
    % p^i(x) itself is composed of a summation across l-values (in the notation of MKK)
    % of theta_l*T_l1(x_1)*T_l2(x_2)*...*T_ld(x_d). So, get the product across T and save
    % as Tprod for each l.
    % So, the summation can be written as 
    % \sum_{q\le |i| \le d+mu} [(-1)^{d+mu-|i|}choose(d-1,d+\mu-|i|)theta_l ]*
    %                          [T_l1(x_1)*T_l2(x_2)*...*T_ld(x_d)]
    % The first bracketed object has already been computed and saved as constTimesTheta(iapprox,l)
    % The second has not been computed since it depends on the (transformed) x values
    
    T = ones(size(x,1),s.M_mup1);    
    Tprod = ones(size(x,1),size(s.j,1));
    for dind = 1:d
        
        % Convert the x points to z via linear transformation
        z = 2.d0*(x(:,dind)-s.a(dind))/(s.b(dind)-s.a(dind)) - 1.d0;
        
        % Evaluate the cheby values beforehand for all orders
        T(:,2) = z;        
        % q<=ibar<=d+mu.  So ibar=d+mu is largest.  
        %Largest any order Cheby can be is the largest any component of m(i) can be.  
        %This occurs when i = 1 except for one component. 
        % => icomp = d+mu-(d-1) = mu+1 => i = mu+1 => m(i) = m(mu+1);
        % NOTE: I precompute this now as s.M_mup1
        twoz = 2*z;
        for oind = 3:s.M_mup1
            T(:,oind) = twoz.*T(:,oind-1) - T(:,oind-2);
        end
        
        %For each possible value of l, compute the product across i of
        %T(:,li) where li is a component of l    
        % Note: since s.j==s.l, I dropped s.l and now use s.j
        tmp = s.j(:,dind)>1; % A cheby polynomial of degree 0 is always 1
        Tprod(:,tmp) = Tprod(:,tmp).*T(:,s.j(tmp,dind));

%         Tprod(:,s.j_eq_2(:,dind)) = Tprod(:,s.j_eq_2(:,dind)).*T(:,repmat(2,[size(s.j_eq_2,1),1]));
%         Tprod(:,s.j_eq_3(:,dind)) = Tprod(:,s.j_eq_3(:,dind)).*T(:,repmat(3,[size(s.j_eq_3,1),1]));
%         if (s.M_mup1>=4)
%             Tprod(:,s.j_eq_4(:,dind)) = Tprod(:,s.j_eq_4(:,dind)).*T(:,repmat(4,[size(s.j_eq_4,1),1]));
%         end
%         if (s.M_mup1>=5)
%             Tprod(:,s.j_eq_5(:,dind)) = Tprod(:,s.j_eq_5(:,dind)).*T(:,repmat(5,[size(s.j_eq_5,1),1]));
%         end
%         % The below only occurs for mu>=3 I believe
%         if (s.M_mup1>5)
%            Tprod(:,s.j_gt_5(:,dind)) = Tprod(:,s.j_gt_5(:,dind)).*T(:,s.j(s.j_gt_5(:,dind),dind));            
%         end

    end
    
    feval = Tprod*s.constTimesTheta';
    
end
